/*global cc*/
/*global require*/
/*global lumosity*/

/**
*  InteractionType is an enum representing all the different possible types of an Interaction.
*  All Interactions are created with a single type.
*  
*  @name InteractionType
*  @class
*  @readonly
*  @enum {number}
*/
var InteractionType =
{
    /** Unknown interaction type, if you see this then there is an error. */
    Unknown:0,
    
    /** A Point interaction is an interaction based upon a location, like a touch or a mouse event. */
    Point:1,
    
    /** A MultiplePoints interaction is an interaction based upon multiple locations, like multi-touch. */
    MultiplePoints:2,
    
    /** A Button interaction is an interaction based upon a digital press of a button, like the keyboard. */
    Button:3,

    /** Describes an interaction where the user swiped the touch interface using one point of contact. */
    Swipe:4
};

/**
*  InteractionPhase is an enum representing all the different possible phases of an Interaction.
*  All Interactions are created with a single phase.
*  
*  @name InteractionPhase
*  @class
*  @readonly
*  @enum {number}
*/
var InteractionPhase =
{
    /** Unknown interaction phase, if you see this then there is an error. */
    Unknown:0,
    
    /** The phase at which an interaction first begins. Is called just once per interaction. */
    Began:1,
    
    /** The phase at which an interaction changes in any way. This might mean adjusting mouse position, or releasing a touch. */
    Changed:2,
    
    /** The phase at which an interaction ends. Is called just once per interaction. */
    Ended:3,
    
    //** If an interaction is somehow cancelled, like from OS events, this phase will be represented. */
    Cancelled:4
};

/**
*  InteractionSource is an enum representing all the different possible ways that an Interaction
*  can be triggered. This will usually depend on device, and generally you only need this enum
*  for the purpose of creating metadata or recording analytics.
*  
*  @name InteractionSource
*  @class
*  @readonly
*  @enum {number}
*/
var InteractionSource =
{
    /** Unknown source for an interaction, if you see this then there is an error. */
    Unknown: 0,
    
    /** The touch screen was used to cause this interaction. */
    TouchScreen: 1,
    
    /** The mouse was used to cause this interaction. */
    Mouse: 2,
    
    /** The keyboard was used to cause this interaction. */
    Keyboard: 3,
    
    /** A button on the device was used to cause this interaction. */
    DeviceButton : 4
};

var Interaction = cc.Class.extend(
{
    /** @lends Interaction.prototype */
    
    type : InteractionType.Unknown,
    phase: InteractionPhase.Unknown,
    data : {},
    
    /**
    *  An Interaction is an abstraction of any input event. This allows us to
    *  potentially allow touch and mouse and any other interaction without needing
    *  to code separate functions for each.
    *  
    *  Proper usage is to create an Interaction with a valid InteractionType,
    *  then in your Game or GameView respond to that InteractionType within the
    *  interaction_received( interaction ) function.
    *  
    *  Note that this constructor may modify the contents of data to make it "complete".
    *
    *  @constructs
    *  @param {InteractionType} type The InteractionType of this Interaction. Can be retrieved later with getType().
    *  @param {InteractionPhase} phase The InteractionPhase of this Interaction. Can be retrieved later with getPhase().
    *  @param {Object} data The data object for this interaction. Should include members useful to this type.
    */
    ctor:function( type, phase, data )
    {
        if ( type )
        {
            this.type = type;
        }

        if ( phase )
        {
           this.phase = phase;
        }

        if ( data )
        {
            this.data = data;

            this._doConvenienceDataPopulation();
        } 
    },
    
    /**
    *  Returns the InteractionType of this Interaction.
    *  
    *  @returns {InteractionType} The InteractionType.
    */
    getType: function()
    {
        return this.type;
    },
    
    /**
    *  Returns the InteractionPhase of this Interaction.
    *  
    *  @returns {InteractionPhase} The InteractionPhase.
    */
    getPhase: function()
    {
        return this.phase;
    },
    
    /**
    *  Returns the data object for this Interaction. Contains details of how the Interaction was performed.
    *  You can use the static members of the Interaction class to get individual non-custom members.
    *  
    *  @returns {Object} The data object of this Interaction.
    */
    getData: function()
    {
        return this.data;
    },
    
    /**
    *  Convenience function, simply uses the position key to return position directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {cc.Point} this.getData()[ Interaction.getPositionKey() ]
    */
    getPosition:function()
    {
        return this.data[ Interaction.getPositionKey() ];
    },
    
    /**
    *  Convenience function, simply uses the index key to return index directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {number} this.getData()[ Interaction.getIndexKey() ]
    */
    getIndex:function()
    {
        return this.data[ Interaction.getIndexKey() ];
    },
    
    /**
    *  Convenience function, simply uses the positions key to return positions directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {Array} this.getData()[ Interaction.getPositionsKey() ]
    */
    getPositions:function()
    {
        return this.data[ Interaction.getPositionsKey() ];
    },

    /**
    *  Convenience function, simply uses the indices key to return indices directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {Array} this.getData()[ Interaction.getIndicesKey() ]
    */
    getIndices:function()
    {
        return this.data[ Interaction.getIndicesKey() ];
    },

    /**
    *  Convenience function, simply uses the buttonName key to return buttonName directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {string} this.getData()[ Interaction.getButtonNameKey() ]
    */
    getButtonName:function()
    {
        return this.data[ Interaction.getButtonNameKey() ];
    },
    
    /**
    *  Convenience function, simply uses the rawButtonName key to return rawButtonName directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {string} this.getData()[ Interaction.getRawButtonNameKey() ]
    */
    getRawButtonName:function()
    {
        return this.data[ Interaction.getRawButtonNameKey() ];
    },
    
    /**
     *  Convenience function, uses the buttonName key and the common "keyboard." prefix
     *  to decide whether or not this button press is that specific keyboard interaction.
     *
     *  @param {number} keyCode The keycode we want to check for. Use cc.KEY.code, like cc.KEY.up, cc.KEY.a, etc.
     *  @param {boolean} [useRawName=false] If true, we'll check the raw button name instead of the actual name.
     *  @returns {boolean} True if that specific keyCode is for this interaction.
     */
    isKeyboardButton:function( keyCode, useRawName )
    {
        return ( useRawName ? this.getRawButtonName() : this.getButtonName() ) === ( Interaction.keyboardPrefix + keyCode );
    },
    
    /**
    *  Convenience function, simply uses the swipeDirection key to return swipeDirection directly.
    *  May return undefined depending on the data provided to this Interaction.
    *  
    *  @returns {string} this.getData()[ Interaction.getSwipeDirectionKey() ]
    */
    getSwipeDirection:function()
    {
        return this.data[ Interaction.getSwipeDirectionKey() ];
    },
    
    /**
     *  Convenience function, simply uses the source key to return the source directly.
     *  This will always have a value, because by default all interactions are mouse or touchscreen.
     *
     *  @returns {InteractionSource} this.getData()[ Interaction.getSourceKey() ]
     */
    getSource: function()
    {
        return this.data[ Interaction.getSourceKey() ];
    },

    /////////////////Private//////////////////
    
    /**
    *  A private function that basically ensures that you will have a complete data object,
    *  even if you pass in only partial data. Would be annoying otherwise to ensure
    *  complete data every single time you want to create an Interaction.
    *  
    *  @private
    */
    _doConvenienceDataPopulation:function()
    {
        //automatically populate indices if we don't have any but we do have positions
        if ( this.data[ Interaction.getPositionsKey() ] && !this.data[ Interaction.getIndicesKey() ] )
        {
            var indices = [];
            var positionIndex;
            for ( positionIndex = 0; positionIndex < this.data[ Interaction.getPositionsKey() ].length; positionIndex++ )
            {
                indices[ positionIndex ] = positionIndex;
            }
        }

        //automatically populate a single index if we don't have one but do have a position
        if ( this.data[ Interaction.getPositionKey() ] && !this.data[ Interaction.getIndexKey() ] )
        {
            this.data[ Interaction.getIndexKey() ] = 0;
        }

        //automatically turn the positions into an array if it turns out it isn't one but has data
        if ( this.data[ Interaction.getPositionsKey() ] && !( this.data[ Interaction.getPositionsKey() ] instanceof Array ) )
        {
            this.data[ Interaction.getPositionsKey() ] = [ this.data[ Interaction.getPositionsKey() ] ];
        }

        //automatically turn the indices into an array if it turns out it isn't one but has data
        if ( this.data[ Interaction.getIndicesKey() ] && !( this.data[ Interaction.getIndicesKey() ] instanceof Array ) )
        {
            this.data[ Interaction.getIndicesKey() ] = [ this.data[ Interaction.getIndicesKey() ] ];
        }

        //automatically add a positions array if we have just a single position
        if ( this.data[ Interaction.getPositionKey() ] && !this.data[ Interaction.getPositionsKey() ] )
        {
           this.data[ Interaction.getPositionsKey() ] = [ this.data[ Interaction.getPositionKey() ] ];
        }

        //automatically add a indices array if we have just a single index
        if ( this.data[ Interaction.getIndexKey() ] && !this.data[ Interaction.getIndicesKey() ] )
        {
          this.data[ Interaction.getIndicesKey() ] = [ this.data[ Interaction.getIndexKey() ] ];
        }

        //automatically add a single position if we have a positions array
        if ( this.data[ Interaction.getPositionsKey() ] && !this.data[ Interaction.getPositionKey() ] )
        {
           this.data[ Interaction.getPositionKey() ] = this.data[ Interaction.getPositionsKey() ][ 0 ];
        }

        //automatically add a single index if we have an indices array
        if ( this.data[ Interaction.getIndicesKey() ] && !this.data[ Interaction.getIndexKey() ] )
        {
           this.data[ Interaction.getIndexKey() ] = this.data[ Interaction.getIndicesKey() ][ 0 ];
        }
        
        //be default, the source of all interactions is touch or mouse, depending on device type
        if ( !this.data[ Interaction.getSourceKey() ] )
        {
            this.data[ Interaction.getSourceKey() ] = lumosity.device.id === "html5" ? InteractionSource.Mouse : InteractionSource.TouchScreen; //TODO - this is inferior, what about mobile web?
        }
    }
});

/**
*  Use to get the key in the data object that corresponds with position.
*  Position will be the location of a touch or a mouse interaction on the screen.
*  @returns {string} The key for position.
*/
Interaction.getPositionKey = function()
{
    return "position";
};

/**
*  Use to get the key in the data object that corresponds with index.
*  Index will be the index of this touch, or of the button on the mouse that was pressed.
*  @returns {string} The key for index.
*/
Interaction.getIndexKey = function()
{
    return "index";
};

/**
*  Use to get the key in the data object that corresponds with an array of positions.
*  Positions will be the locations of all touches or a mouse interactions on the screen.
*  @returns {string} The key for positions array.
*/
Interaction.getPositionsKey = function()
{
    return "positions";
};

/**
*  Use to get the key in the data object that corresponds with an array of indices.
*  Indices will be the indices of all touches, or of the buttons on the mouse that were pressed.
*  @returns {string} The key for indices array.
*/
Interaction.getIndicesKey = function()
{
    return "indices";
};

/**
*  Use to get the key in the data object that corresponds with the name of the button pressed.
*  The button pressed name will usually be custom per what the situation calls for.
*  @returns {string} The key for button name.
*/
Interaction.getButtonNameKey = function()
{
    return "buttonName";
};

/**
 *  It's possible to use aliases and whatnot to change the buttonName value. That's all well and good
 *  unless you need to know the real source of the button press, mainly for metadata. This will
 *  return the for reals button name.
 */
Interaction.getRawButtonNameKey = function()
{
    return "rawButtonName";
};

/**
*  This value is appended as the prefix to all button presses that come from the keyboard.
*  It should only be used internally, or for automation.
*/
Interaction.keyboardPrefix = "keyboard.";

/**
*  Use to get the key in the data object that corresponds with the direction of a swipe.
*  The value of the swipe will be an integer, as defined at the top of SwipeInteractionRecognizer
*  in the SwipeGestureDirection enum.
*  @returns {string} The key for button name.
*/
Interaction.getSwipeDirectionKey = function()
{
    return "swipeDirection";
};

/**
 *  The source represents where the Interaction came from, using an InteractionSource.
 *  Will usually be touch or mouse, but on other systems it may be keyboard or deviceButton.
 */
Interaction.getSourceKey = function()
{
    return "source";
};