/////////////////////////////////////////////////////////////////////
// This fragment shader should allow the user to replace up to two //
// colors (also takes into account gradients).                     //
/////////////////////////////////////////////////////////////////////
varying mediump vec2 v_texCoord;
uniform sampler2D CC_Texture0;

//////////////////////////////////////////////////////////////
//NOTE: {x: Color to be replaced, y: Color to replace with} //
//////////////////////////////////////////////////////////////
uniform vec2 swapR1;
uniform vec2 swapG1;
uniform vec2 swapB1;

uniform vec2 swapR2;
uniform vec2 swapG2;
uniform vec2 swapB2;

const vec4 errorColor = vec4( 0, 0, 0, 0 );

/////////////////
// PROTOTYPES  //
/////////////////
bool colorWithinRange( vec4 currentColor, vec4 targetColor, float threshold );
bool epsilonCompare( vec4 currentColor, vec4 targetColor, float epsilon );
vec4 exchangeColors( vec4 keyColor, vec4 swapColor, vec4 sourceColor );

void main()
{
    vec4 sourceColor = texture2D(CC_Texture0, v_texCoord);
    vec4 finalColor = sourceColor;

    //
    //NOTE: We have to organize code this way because GLSL ES 2.0 is only
    //      guranteed to support a constant expression for indexing  an array (in the fragment shader).
    //      
    //      This was experienced on a Nexus device, but seems to be an issue with Qualcomm's drivers in general.
    //      
    //      Please see post #10 in this thread:
    //      
    //      https://developer.qualcomm.com/comment/6284#comment-6284
    //      
    if( sourceColor.a == 1.0 )
    {
        finalColor = exchangeColors( vec4( swapR1.x, swapG1.x, swapB1.x, 1.0 ),
                                     vec4( swapR1.y, swapG1.y, swapB1.y, 1.0 ),
                                     sourceColor );

        if( finalColor == errorColor )
        {
            finalColor = exchangeColors( vec4( swapR2.x, swapG2.x, swapB2.x, 1.0 ),
                                         vec4( swapR2.y, swapG2.y, swapB2.y, 1.0 ),
                                         sourceColor );
        }
        if( finalColor == errorColor )
        {
            finalColor = sourceColor;
        }
    }
    gl_FragColor = finalColor;
}

///////////////////////////////////////
// Performs the swapping algorithm.  //
///////////////////////////////////////
vec4 exchangeColors( vec4 keyColor, vec4 swapColor, vec4 sourceColor )
{
    const float epsilon = 0.0001;
    const float threshold = 0.5;
    if( epsilonCompare( sourceColor, keyColor, epsilon ) )
    {
        return swapColor;
    }
    else if( colorWithinRange( sourceColor, keyColor, threshold ) ) 
    {
        return vec4( swapColor.r + ( swapColor.r * abs( sourceColor.r - keyColor.r ) ),
                       swapColor.g + ( swapColor.g * abs( sourceColor.g - keyColor.g ) ),
                       swapColor.b + ( swapColor.b * abs( sourceColor.b - keyColor.b ) ), 1.0 );
    }
    return errorColor;
}

//////////////////////////////////////////////////////////////////////
// Checks that currentColor is within +/- threshold of targetColor. //
//////////////////////////////////////////////////////////////////////
bool colorWithinRange( vec4 currentColor, vec4 targetColor, float threshold )
{
    vec4 colorMinus = vec4( targetColor.r - threshold, targetColor.g - threshold, targetColor.b - threshold, 1.0 );
    vec4 colorPlus = vec4( targetColor.r + threshold, targetColor.g + threshold, targetColor.b + threshold, 1.0 );

    return( (( currentColor.r > colorMinus.r ) && ( currentColor.r < colorPlus.r )) &&
            (( currentColor.g > colorMinus.g ) && ( currentColor.g < colorPlus.g )) &&
            (( currentColor.b > colorMinus.b ) && ( currentColor.b < colorPlus.b )) );
}

////////////////////////////////////////////////////////
// Checks that two vectors are approximately "equal". //
////////////////////////////////////////////////////////
bool epsilonCompare( vec4 currentColor, vec4 targetColor, float epsilon )
{
    return( ( abs( currentColor.r - targetColor.r ) < epsilon ) &&
            ( abs( currentColor.g - targetColor.g ) < epsilon ) &&
            ( abs( currentColor.b - targetColor.b ) < epsilon ) );
}
