/*global cc*/
/*global require*/
/*global ll*/

var TiledMap;
( function()
{
    "use strict";

    TiledMap = cc.Node.extend(
    {
        /** @lends TiledMap.prototype */
        
        /**
         * Returns a tile coordinate for position.
         * 
         * @param  {cc.p} position The position.
         * 
         * @return {cc.p} The tile's coordinates.
         */
        tileCoordForPosition: function( position )
        {
            var tileCoord = cc.PointZero();
            var tileSize = this.getTileSize();
            var mapSize = this.getMapSize();

            tileCoord.x = Math.floor( position.x / tileSize.width );
            tileCoord.y = Math.floor( ( ( mapSize.height * tileSize.height ) - position.y ) / tileSize.height );
            return tileCoord;
        },

        isoTileCoordForPosition: function( position )
        {
            var tileSize = this.getTileSize();
            var mapSize = this.getMapSize();

            var tw = tileSize.width;
            var th = tileSize.height;
            var mw = mapSize.width;
            var mh = mapSize.height;

            var posX = mh + position.x/tw - mw/2 - position.y/th;
            var posY = mh - position.x/tw + mw/2 - position.y/th;

            return cc.p( Math.floor( posX ), Math.floor( posY ) );
        },

        /**
         * Returns a tile rect for tile coordinates.
         * 
         * @param  {cc.p} tileCoords The tile's coordinates.
         * 
         * @return {cc.rect}         The tile's rect.
         */
        tileRectFromTileCoords: function( tileCoords )
        {
            var tileRect = cc.RectZero();
            var tileSize = this.getTileSize();
            var mapSize = this.getMapSize();

            tileRect.x = tileCoords.x * tileSize.width;
            tileRect.y = ( mapSize.height * tileSize.height ) - ( ( tileCoords.y + 1 ) * tileSize.height );
            tileRect.width = tileSize.width;
            tileRect.height = tileSize.height;
            return tileRect;
        },

        /**
         * Returns whether or not the tile coordinates is within
         * the layer's tile space.
         * 
         * @param  {string} layerName The name of the layer.
         * @param  {cc.p} tileCoord The tiles coordinates.
         * 
         * @return {Boolean} Whether the tile coordinates is within the layer's tile space.
         */
        doesContainTileCoords: function( layerName, tileCoords )
        {
            var doesContainTileCoords = false;
            var layer = null;

            layer = this.getLayer( layerName );

            if( tileCoords.x < layer.getLayerSize().width && tileCoords.y < layer.getLayerSize().height &&
                tileCoords.x >= 0 && tileCoords.y >= 0 )
            {
                doesContainTileCoords = true;
            }

            return doesContainTileCoords;
        },

        /**
         * Returns a tile's properties for position.
         * 
         * @param  {string} layerName The name of the layer.
         * @param  {cc.p} position  The position.
         * 
         * @return {cc.p} The tile's properties.
         */
        tilePropertiesForPosition: function( layerName, position )
        {
            var tileCoord = cc.PointZero();
            var tileProperties = null;

            tileCoord = this.tileCoordForPosition( position );
            this.tilePropertiesForTileCoord( layerName, tileCoord );
            return tileProperties;
        },

        /**
         * Returns a tile's properties for tile coordinates.
         * 
         * @param  {string} layerName The name of the layer.
         * @param  {cc.p} tileCoord The tiles coordinates.
         * 
         * @return {cc.p} The tile's properties.
         */
        tilePropertiesForTileCoord: function( layerName, tileCoord )
        {
            var tileGID = 0;
            var layer = null;
            var tileProperties = null;

            layer = this.getLayer( layerName );
            if( !layer )
            {
                ll.warning( "TiledMap", "tilePropertiesForPosition", "layer not found." );
                return null;
            }

            tileGID = layer.getTileGIDAt( tileCoord );
            if( !tileGID )
            {
                ll.warning( "TiledMap", "tilePropertiesForPosition", "gid not found." );
                return null;
            }

            tileProperties = this.propertiesForGID( tileGID );
            return tileProperties;
        },

        /**
         * Changes the tile at the given position with a new tile based on gid.
         * 
         * @param {string} layerName The name of the layer.
         * @param {cc.p} position The position.
         * @param {int} gid The id of the tile.
         */
        setTilePositionWithGID: function( layerName, position, gid )
        {
            var tileCoord = cc.PointZero();

            tileCoord = this.tileCoordForPosition( position );
            this.setTileCoordWithGID( layerName, tileCoord, gid );
        },

        /**
         * Changes the tile at the given tile coordinates with a new tile based on gid.
         * 
         * @param {string} layerName The name of the layer.
         * @param {cc.p} tileCoord The tiles coordinates.
         * @param {int} gid The id of the tile.
         */
        setTileCoordWithGID: function( layerName, tileCoord, gid )
        {
            var layer = null;

            layer = this.getLayer( layerName );
            if( !layer )
            {
                ll.warning( "TiledMap", "setTileWithGID", "layer not found." );
                return;
            }

            layer.setTileGID( gid, tileCoord );
        }
    } );
}() );

TiledMap.create = function( tmxFile ) 
{
    "use strict";
    var tiledMap = new TiledMap();
    var newTiledMap = cc.TMXTiledMap.create( tmxFile );

    newTiledMap.tileCoordForPosition = tiledMap.tileCoordForPosition;
    newTiledMap.isoTileCoordForPosition = tiledMap.isoTileCoordForPosition;
    newTiledMap.tileRectFromTileCoords = tiledMap.tileRectFromTileCoords;
    newTiledMap.doesContainTileCoords = tiledMap.doesContainTileCoords;
    newTiledMap.tilePropertiesForPosition = tiledMap.tilePropertiesForPosition;
    newTiledMap.tilePropertiesForTileCoord = tiledMap.tilePropertiesForTileCoord;
    newTiledMap.setTilePositionWithGID = tiledMap.setTilePositionWithGID;
    newTiledMap.setTileCoordWithGID = tiledMap.setTileCoordWithGID;

    return newTiledMap;
};
