/*jslint node: true */
/*global cc*/
"use strict";

require("src/View/CCBReaderHelper.js");
require("src/View/LumosityLabel.js");
require("src/Utility/LumosityUtils.js");

/**
 * Adapt scene loaded from CocosBuilder into Lumosity's custom scaling of Cocos2D.
 */
var CCBReaderAdapter = {};

/**
 * CocosBuilder defaults to resolution 320x480.  
 * Substitute another value like 640x960 if you are using that design width.
 * Test case:  2014-11 CocosBuilder expects 320x480.  2015-02-05 Wouter expects 640x960.
 */
CCBReaderAdapter.designWidth = 320;
CCBReaderAdapter.designHeight = 480;

/**
 * Rescales images from iPhone 3 coordinate space to Lumosity coordinate space and compensates for pixel rounding error.
 * Boiler shows different scale than iOS simulator.
 * iPad2: width 768
 * iPad Air: width 1536
 * @param   screenWidth     Default director win size (which is set by setDesignResolutionSize).
 * @param   design          Default 400 for a 320x480 iPhone 3.
 * @return  Content scale factor.
 *
 * GAME-2785 If width of the screen is less than or equal to 400 pixels, then scale content by 400 instead of by 640.  Test case:  2015-02-23 LGL34C "Optimus Fuel", with small screen resolution (320x480).  Look at layout.  [~rivey] expects images are proportional to small resolution.  Got images are even smaller.  
 */
CCBReaderAdapter.calculateContentScale = function(design, screenWidth)
{
    design = design || 400.0;
    screenWidth = screenWidth || cc.Director.getInstance().getWinSize().width;
    var contentScaleFactor = 1;
    var screen;
    if(screenWidth <= 400)
    {
        var iPhone3 = 400.0;
        screen = iPhone3;
    }
    else if(screenWidth <= 640)
    {
        var iPhone4 = 640.0;
        screen = iPhone4;
    }
    else if (screenWidth <= 800)
    {
        var iPad2 = 800.0;
        screen = iPad2;
    }
    else if (screenWidth <= 1200)
    {
        var androidTablet = 1200.0;
        screen = androidTablet;
    }
    else 
    {
        var iPadAir = 1600.0;
        screen = iPadAir;
    }
    contentScaleFactor = screen / design;
    // cc.log("CCBReaderAdapter.calculateContentScale: width " + screenWidth + " scale " + contentScaleFactor);
    return contentScaleFactor;
};

/**
 * Scale content according to design resolution.
 * Set before images are loaded.
 * This is more robust than scaling child nodes,
 * because keyframes and text had needed special cases.
 * Example @see Trouble Brewing/shared/src/launcher.js
 *
 * @param   designBucket    {cc.Size} to Override default design width and height.
 *
 * I gave up on setDesignResolutionSize for our base game.
 * This made blurry images and text.
 * Maybe there is something incompatible with design resolution, asset loading and font crispness.
 */
CCBReaderAdapter.setContentScaleFactor = function(designBucket)
{
    var scale = CCBReaderAdapter.calculateContentScale(designBucket);
    cc.Director.getInstance().setContentScaleFactor(scale);
};

CCBReaderAdapter.requires = function(all) {
    if (all) {
        // html:
        require('src/View/CCBReader/CCNodeLoader.js');
        require('src/View/CCBReader/CCSpriteLoader.js');
        require('src/View/CCBReader/CCControlLoader.js');
        require('src/View/CCBReader/CCNodeLoaderLibrary.js');
        require('src/View/CCBReader/CCBAnimationManager.js');
        require('src/View/CCBReader/CCBSequence.js');
        require('src/View/CCBReader/CCBRelativePositioning.js');
        require('src/View/CCBReader/CCBKeyframe.js');
        require('src/View/CCBReader/CCBValue.js');
        // mobile:
        require('src/View/GUI/CCControlExtension/CCControl.js');
        require('src/View/GUI/CCControlExtension/CCControlButton.js');
        require('src/View/GUI/CCControlExtension/CCControlUtils.js');
        require('src/View/GUI/CCControlExtension/CCInvocation.js');
        require('src/View/GUI/CCControlExtension/CCScale9Sprite.js');
        require('src/View/CCBReader/CCBReaderUtil.js');
        require('src/View/CCBReader/CCBReader.js');
    }
};

/**
 * PlayingKoi does not need this on mobile because Cocos2d-x JavaScript bindings implement CCBReader.
 * It would need to require CCBReader in an HTML5 platform, such as Boiler.
 * 
 * @param   all {Boolean}   On cocos2d-html5 without loadExtensions, include dependencies.
 */
CCBReaderAdapter.requireEach = function(all)
{
    require("src/View/CCBReaderHelper.js");
    CCBReaderAdapter.requires(all);
};

/**
 * Load sprite sheets before loading any .ccbi file that uses a sprite on a sheet.
 * Expects spritesheets follow this naming convention:
 *      spritesheet_0.plist, spritesheet_1.plist, ...
 *
 * @param   count   By default, try to load up to 3 sprite sheets.
 *                  As more graphics are added, number of sprite sheets may vary.
 *                  The number of spritesheets also varies per resolution.
 * Test case:  2015-01-06 Xcode: Expect to start.  Crash.  Read log:
 *     Cocos2d: Get data from file(timer1Enabled.png) failed!
 *     TypeError: texture is null
 */
CCBReaderAdapter.loadSpriteSheets = function(count)
{
    count = undefined === count ? 3 : count;
    var cache = cc.SpriteFrameCache.getInstance();
    for (var i = 0; i < count; i++) {
        try {
            cache.addSpriteFrames( "spritesheet_" + i + ".plist", "spritesheet_" + i + ".png" );
        }
        catch (err) {
            if (0 == i) {
                throw err;
            }
            // ll.warning("Failed to load second sprite sheet used only by high resolution. " + err);
        }
    }
    CCBReaderAdapter.requireEach("html5" == lumosity.device.id);
};

/**
 * iPhone 3:    320, 480
 * iPhone 4:    640, 960
 * iPhone 5:    640, 1136
 * iPad:        768, 1028
 * iPad Retina: 1536, 2048
 */
CCBReaderAdapter.getScaleByAspect = function()
{
    var design = new cc.size(CCBReaderAdapter.designWidth, CCBReaderAdapter.designHeight);
    var screen = cc.Director.getInstance().getWinSize();
    return CCBReaderAdapter.getScaleByDimensions(design.width, design.height, 
                                                   screen.width, screen.height);
};

CCBReaderAdapter.mapLeftAndRight = function(node)
{
    var screen = cc.Director.getInstance().getWinSize();
    var left = node.convertToNodeSpace(cc.p(0, 0));
    left = Math.round(left.x);
    var right = node.convertToNodeSpace(cc.p(screen.width, 0));
    right = Math.round(right.x);
    return [left, right];
};

/**
 * @return scale {Number}   To fit by minimum aspect of width or height.
 */
CCBReaderAdapter.getScaleByDimensions = function(designWidth, designHeight, 
                                                   screenWidth, screenHeight)
{
    var designAspect = designHeight / designWidth;
    var screenAspect = screenHeight / screenWidth;
    var designPerScreen = designAspect / screenAspect;
    var screenPerDesign;
    if (screenAspect < designAspect) {
        screenPerDesign = screenHeight / designHeight;
    }
    else {
        screenPerDesign = screenWidth / designWidth;
    }
    var scale = screenPerDesign;
    return scale;
};

/**
 * GAME-2272 GAME-2019 Substract parent's world coordinates from world coordinates of root.  The parent may be shifted off-screen, such as when starting the tutorial.  Test case:  2015-01-05 In Xcode load iPad.  Set Debug Mode.  Tap Trouble Brewing.  Tap "D".  Tap "Test layout".  Take a screenshot.  Inspect left and right edge that they are no more than 4 pixels different.  2015-01-21 From level select, How to Play.  Expect to read instructions.  Got screen is off to the left.  
 * GAME-2159 Convert arbitrary design center to world space.  Was hard-coded 160, 240.  Test case:  2015-02-24 Select How to Play mode.  Select Train of Thought.  Expect text and callout fits on screen.  Got too high and to the left.
 */
CCBReaderAdapter.mapCenterOnScreen = function(root, screen, design)
{
    design = design || cc.size(CCBReaderAdapter.designWidth, CCBReaderAdapter.designHeight);
    var world = root.convertToWorldSpace(cc.p(0.5 * design.width, 0.5 * design.height));
    var parent = root.getParent().convertToWorldSpace(cc.p(0, 0));
    var child = {x: world.x - parent.x, y: world.y - parent.y};
    // cc.log("CCBReaderAdapter.mapCenterOnScreen: " + world.x + ", " + world.y);
    var x = screen.width * 0.5 - child.x;
    var y = screen.height * 0.5 - child.y;
    var center = {x: x, y: y};
    return center;
};

/**
 * Scale from center.
 * Image scale factor is related to content scale factor.
 */
CCBReaderAdapter.fit = function(root)
{
    var screen = cc.Director.getInstance().getWinSize();
    var scale = CCBReaderAdapter.getScaleByAspect();
    root.setScale(scale);
    var centered = CCBReaderAdapter.mapCenterOnScreen(root, screen);
    // cc.log("CCBReaderAdapter.fit: center " + centered.x + ", " + centered.y );
    root.setPosition(centered.x, centered.y);
};

/**
 * GAME-2451 When loading an animation separately, also invert scale of text, so that is higher font size, rather than a magnified scale (which its ancestors are based on aspect ratio of the device).  Test case:  2015-01-13 Serve.  Observe score.  [~sclark] expects font of score on cup is crisp.  Got pixelated.
 */
CCBReaderAdapter.create = function(ccbFileName)
{
    var node = cc.BuilderReaderHelper.load(ccbFileName);
    return node;
};

/**
 * @param   nodeName    Key node name in cache for next time.
 * @return  Adopted from cache or created.
 */
CCBReaderAdapter.adoptOrCreate = function(nodeName, ccbFileBaseName, nextParent)
{
    var previousParent = nextParent;
    var key = nodeName + "_" + ccbFileBaseName;
    nextParent = cc.BuilderReaderHelper.ccbCacheAdopt(key, nextParent);
    if (nextParent)
    {
    }
    else
    {
        nextParent = CCBReaderAdapter.create(ccbFileBaseName);
        cc.BuilderReaderHelper.ccbCache[key] = nextParent;
    }
    if (previousParent && nextParent != previousParent)
    {
        cc.BuilderReaderHelper.adopt(previousParent, nextParent);
    }
    if (nodeName != nextParent.getName())
    {
        nextParent.setName(nodeName);
    }
    return nextParent;
};

/**
 * @param   ccbFileName Expects valid variable name (without "." or " " or "-" or leading number.  Assigns this name to parent as a member variable containing the new child.
 */
CCBReaderAdapter.addChild = function(parent, ccbFileName, currentScene, owner)
{
    if(currentScene)
    {
        parent.removeChild(currentScene);
    }
    currentScene = cc.BuilderReaderHelper.load(ccbFileName, owner);
    parent.addChild(currentScene);
    parent[ccbFileName] = currentScene;
    CCBReaderAdapter.fit(currentScene);
    return currentScene;
};


// TestCCBReaderAdaptor


/**
 * Currently shows in Boiler, but not in iOS simulator iPhone 4s.
 * Properties appear as if it were visible.
 */
CCBReaderAdapter.describe = function(currentScene) {
    var message = "CCBReaderAdapter.describe:";
    try {
        var child = currentScene.getChildren()[0];
        var count = child.getChildrenCount();
        var grandchild = child.getChildren()[0];
        if (!grandchild) {
            grandchild = child;
            child = currentScene;
            currentScene = currentScene.getParent();
        }
        var director = cc.Director.getInstance();
        var design = director.getWinSize();
        message += ""
                + " dww " + design.width
                + " dwh " + design.height
                + " csf " + director.getContentScaleFactor();
        message += ""
                + " gname " + currentScene.getName()
                + " gpx " + currentScene.getPositionX()
                + " gpy " + currentScene.getPositionY()
                + " gpax " + currentScene.getAnchorPoint().x
                + " gpay " + currentScene.getAnchorPoint().y;
        message += " count " + count
                + " name " + child.getName()
                + " px " + child.getPositionX()
                + " py " + child.getPositionY()
                + " pax " + child.getAnchorPoint().x
                + " pay " + child.getAnchorPoint().y;
        message += ""
                + " x " + grandchild.getPositionX()
                + " y " + grandchild.getPositionY()
                + " z " + grandchild.getZOrder()
                + " sx " + grandchild.getScaleX()
                + " sy " + grandchild.getScaleY()
                + " w " + grandchild.getContentSize().width
                + " h " + grandchild.getContentSize().height
                + " a " + grandchild.getOpacity()
                + " visible " + grandchild.isVisible()
                + " t " + texture;
        var texture = grandchild.getTexture();
        if (texture) {
            message += " tw " + texture.getContentSizeInPixels().width
                     + " th " + texture.getContentSizeInPixels().height
        }
    }
    catch (err) {
        cc.log(err);
    }
    cc.log(message);
};

CCBReaderAdapter.testLayout = function(stage, ccbFileName)
{
    if ( CCBReaderAdapter.toggleTestLayout() ) {
        return;
    }
    ccbFileName = ccbFileName || "testRoot";
    CCBReaderAdapter.testRoot = CCBReaderAdapter.addChild(stage, ccbFileName,
        CCBReaderAdapter.testRoot);
    CCBReaderAdapter.describe(CCBReaderAdapter.testRoot);
    cc.log("CCBReaderAdapter.testRoot: Expect to see bars fitting at corners, diagonals, and sides.");
};

/**
 * Toggle visible.  Test case:  2015-01-20 Test layouts repeatedly.  Expect see and disappear.  Got crash.  Cocos2d-x v2.2.2 seems to crash when reparenting animated nodes or after some duration.
 * @return  If exists already.
 */
CCBReaderAdapter.toggleTestLayout = function()
{
    var isAdded = false;
    if (CCBReaderAdapter.testRoot) {
        isAdded = true;
        var isVisible = CCBReaderAdapter.testRoot.isVisible();
        CCBReaderAdapter.testRoot.setVisible(!isVisible);
    }
    return isAdded;
};
