/*global require*/
/*global cc*/
/*global localize*/
/*global LumosityButtonState*/
/*global LumosityButton*/
/*global LayoutLoader*/
/*global LumosityUtils*/
/*global lg*/

require( "src/View/LumosityButton.js" );
require( "src/View/LumosityButtonLayer.js" );
require( "src/Animation/LayoutLoader.js" );
require( "src/Utility/LumosityUtils.js" );

var LumosityDebugMenu;
( function()
{
    "use strict";

    LumosityDebugMenu = LumosityButtonLayer.extend(
    {
        /** @lends LumosityDebugMenu.prototype */

        BG_COLOR: cc.c3b( 242, 104, 46 ),

        BUTTON_COLOR: cc.WHITE,

        NAV_BUTTON_FONT_SIZE: 44,

        DEBUG_BUTTON_FONT_SIZE: 28,

        _debugButtons: [],

        _delegate: undefined,

        /**
         *  LumosityDebugMenu is the main class for the debug menu view!
         *
         *  @constructs
         */
        ctor: function()
        {
            this._super();

            var screenSize = cc.Director.getInstance().getWinSize();
            this.setContentSize( screenSize );
            this.setAnchorPoint( cc.p( 0, 0 ) );
            this.setPosition( cc.p( 0, 0 ) );
            this.nodeName = "LumosityDebugMenu";

            this._createViews();

            this.open( false );
        },

        /**
         * Adds a debug button to the menu.
         *
         * @param {Object} buttonOptions User's options specified for button, including label, callback, etc.
         * @returns {LumosityButton}
         */
        addDebugButton:function( buttonOptions )
        {
            var labelText = buttonOptions.labelText;
            var callback = buttonOptions.callback;
            var target = buttonOptions.target;
            var closeMenuOnCallback = buttonOptions.closeMenuOnCallback;

            var button = this._createButton( this.debugButtonSize, labelText, this.DEBUG_BUTTON_FONT_SIZE );
            if( button )
            {
                this._menuNode.addChild( button, 10 );
                this.addButton( button );

                var buttonIndex = this._debugButtons.length;
                this._debugButtons.push( button );

                var buttonPos = this._getDebugButtonPos( buttonIndex );
                button.setPosition( buttonPos );

                var releasedCallback = this._debugButtonReleased.bind( this, callback, target, closeMenuOnCallback );
                button.setReleasedCallback( releasedCallback );
                button.setTarget( this );

                button.setEnabled( false );
            }
            return button;
        },

        /**
         * Adds a debug label to the menu.
         *
         * @param {Object} labelText Text that the lable should display. (default is "!*!*!*!*!")
         * @param {Object} fontSize Size of the font. (default is 36)
         * @returns {cc.LabelTTF}
         */
        addDebugLabel: function( labelText, fontSize )
        {
            labelText = labelText || "!*!*!*!*!";
            fontSize = fontSize || 36;
            var label = cc.LabelTTF.create( labelText, "fonts/MuseoSans-700.ttf", fontSize );
            label.setColor( this.BUTTON_COLOR );
            label.setAnchorPoint( 0.5, 0.5 );
            label.setHorizontalAlignment( cc.TEXT_ALIGNMENT_CENTER );
            label.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
            this._menuNode.addChild( label );
            return label;
        },

        /**
         * Creates view elements for pause menu, including open/close buttons, background, and debug buttons.
         *
         * @private
         */
        _createViews:function()
        {
            // create debug menu ui and scale it appropriately
            var screenSize = cc.Director.getInstance().getWinSize();
            var scaleFactor = Math.min( screenSize.width / 640, screenSize.height / 960 );

            var bgMargin = 32;
            this.bgSize = cc.size( 640 - bgMargin * 2, 960 - bgMargin * 2 );
            this.navButtonSize = cc.size( this.bgSize.width * 0.15, this.bgSize.width * 0.15 );
            this.buttonMargin = cc.size( this.bgSize.width * 0.05, this.bgSize.width * 0.15 );
            this.debugButtonSize = cc.size( this.bgSize.width * 0.45, this.bgSize.height * 0.085 );

            // create menu node to hold background and debug buttons
            this._menuNode = cc.Node.create();
            var menuPos = cc.p( screenSize.width / 2, screenSize.height / 2 );
            this._menuNode.setPosition( menuPos );
            this._menuNode.setScale( scaleFactor );
            this.addChild( this._menuNode );

            // create rectangular background
            this.bg = LumosityUtils.createRectangularSprite( this.bgSize, this.BG_COLOR );
            this._menuNode.addChild( this.bg );

            // create open button that sits in the lower right of game screen and opens debug menu
            this.openButton = this._createButton( this.navButtonSize, "D", this.NAV_BUTTON_FONT_SIZE );
            var openButtonPos = cc.p( screenSize.width - this.navButtonSize.width * scaleFactor, 0 );
            this.openButton.setPosition( openButtonPos );
            this.openButton.setScale( scaleFactor );
            this.openButton.setTarget( this );
            this.openButton.setReleasedCallback( this._onNavButtonReleased.bind( this, true ) );
            this.addChild( this.openButton, 10 );
            this.addButton( this.openButton );

            // create close button that sits in upper right of debug menu and closes it
            this.closeButton = this._createButton( this.navButtonSize, "X", this.NAV_BUTTON_FONT_SIZE, this._onCloseButtonRelease, this );
            var closeButtonPos = cc.p( this.bgSize.width / 2 - this.navButtonSize.width, this.bgSize.height / 2 - this.navButtonSize.height );
            this.closeButton.setPosition( closeButtonPos );
            this.closeButton.setTarget( this );
            this.closeButton.setReleasedCallback( this._onNavButtonReleased.bind( this, false ) );
            this._menuNode.addChild( this.closeButton, 10 );
            this.addButton( this.closeButton );

            // create title so that people understand whose house they're in. debug menu's house.
            this.title = cc.LabelTTF.create( "DEBUG", "fonts/MuseoSans-700.ttf", 36 );
            this.title.setColor( this.BUTTON_COLOR );
            this.title.setAnchorPoint( 0.5, 0.5 );
            this.title.setPosition( cc.p( 0, this.bgSize.height / 2 - this.bgSize.height * 0.05 ) );
            this.title.setHorizontalAlignment( cc.TEXT_ALIGNMENT_CENTER );
            this.title.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
            this._menuNode.addChild( this.title );
        },

        /**
         * Creates a button in the debug menu.
         *
         * @param {cc.size} buttonSize Design size of button.
         * @param {string} labelText Text for button's label.
         * @param {number} fontSize Text's font size.
         * @returns {LumosityButton}
         * @private
         */
        _createButton:function( buttonSize, labelText, fontSize )
        {
            var buttonSprite = LumosityUtils.createRectangularSprite( buttonSize, this.BUTTON_COLOR );
            var button = new LumosityButton();
            button.setTitleForState( labelText, LumosityButtonState.ButtonStateNormal );
            button.setTitleColorForState( this.BG_COLOR, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( buttonSprite, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( buttonSprite, LumosityButtonState.ButtonStateSelected );
            button.titleLabel.setInitialFontSize( fontSize );
            button.titleLabel.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            button.titleLabel.setPosition( cc.p( buttonSize.width / 2, buttonSize.height / 2 ) );
            button.titleLabel.setHorizontalAlignment( cc.TEXT_ALIGNMENT_CENTER );
            return button;
        },

        /**
         * Handles debug button presses.
         *
         * @param {function} callback Callback invoked when button is pressed.
         * @param {Object} target Object upon which callback is invoked.
         * @param {boolean} closeMenuOnCallback Whether to close the debug menu when the button is pressed.
         * @private
         */
        _debugButtonReleased:function( callback, target, closeMenuOnCallback )
        {
            if( callback )
            {
                if( target )
                {
                    callback.call( target );
                }
                else
                {
                    callback();
                }
            }
            if( closeMenuOnCallback )
            {
                this.open( false );
            }
        },

        /**
         * Handles open/close button presses.
         *
         * @param {boolean} open Whether the nav button was open (or close)
         * @private
         */
        _onNavButtonReleased:function( open )
        {
            this.open( open );
        },

        /**
         * Gets a debug button's position, based on button's index.
         *
         * @param {number} index Button's index in menu.
         * @returns {cc.p} Position of button in question.
         * @private
         */
        _getDebugButtonPos:function( index )
        {
            var x = this.bgSize.width * -0.5 + this.buttonMargin.width;
            var y = this.bgSize.height * 0.5 - this.buttonMargin.height - this.debugButtonSize.height - index * this.debugButtonSize.height * 1.2;
            return cc.p( x, y );
        },

        /**
         * Opens or closes the debug menu.
         *
         * @param {boolean} open Whether to open (or close) the menu
         */
        open:function( open )
        {
            this._menuNode.setVisible( open );
            this.openButton.setVisible( !open );
            this.closeButton.setEnabled( open );
            this.openButton.setEnabled( !open );
            //
            // enable/disable individual buttons
            //
            if( this._debugButtons )
            {
                for( var i = 0; i < this._debugButtons.length; i++ )
                {
                    var button = this._debugButtons[i];
                    if( button )
                    {
                        button.setEnabled( open );
                    }
                }
            }
            //
            if( this._delegate )
            {
                if( open && this._delegate.debugMenuOpen )
                {
                    this._delegate.debugMenuOpen();
                }
                else if( !open && this._delegate.debugMenuClose )
                {
                    this._delegate.debugMenuClose();
                }
            }
        },

        setDelegate: function ( delegate )
        {
            this._delegate = delegate;
        }

    } );
}() );
