/*global cc*/
/*global require*/

require( "src/Utility/LumosityUtils.js" );
/*global LumosityUtils*/

require( "src/Animation/EaseWrapper.js" );
/*global EaseWrapper*/

var PropertyAnimator;
(function()
{
    "use strict";
    
    PropertyAnimator = cc.Class.extend(
    {
        /** @lends PropertyAnimator.prototype */        
        
        /**
         * A PropertyAnimator is capable of animating a single property of a node by appending it onto a
         * sequence. It should be subclassed for each individual property.
         *
         * @constructs
         * @param {string} animatedActionClassName The class name to create for the animated action. "ScaleTo", "MoveTo", "FadeTo", etc
         */
        ctor: function( animatedActionClassName )
        {
            this._animatedActionClassName = animatedActionClassName;
        },
        
        /**
         * Used to set a property on a node, but not in an animated way.
         *
         * @param {cc.Sequence} prevSequence If a sequence already exists to animate this property, it will be appended to.
         * @param {cc.Node} node The node that will be animated.
         * @param {boolean} immediate If true, then the setFunction will be called immediately. Otherwise, it's added to the sequence.
         */
        addSetProperty:function( prevSequence, node, value, immediate )
        {
            value = this._convertDesignValueToGameValue( value, node );
            
            if ( immediate )
            {
                this._setFunction( node, value );
            }
            else
            {
                var action = cc.CallFunc.create( this._setFunction.bind( this, node, value ), this );
                return this._addAction( prevSequence, action );
            }
        
            return prevSequence;
        },
    
        /**
         * Used to animate a property on a node. This will create a Cocos action from its current value
         * to the endValue
         */
        addAnimateProperty:function( prevSequence, node, endValue, duration, easeType, timeStarted )
        {
            endValue = this._convertDesignValueToGameValue( endValue, node );
            
            var action = this._createAnimationAction( duration, endValue );
            action = EaseWrapper.wrapActionWithEase( action, easeType );
            return this._addAction( prevSequence, action );
        },
        
        /**
         * Many values coming from design data must be converted to a different value in order to work
         * in-engine. If that's the case with a property, it must override this function to return
         * something else.
         *
         * @param {*} designValue The value that was in the animation data.
         * @param {cc.Node} [node] You may need the node to compute your value, so this is passed in.
         * @returns {*} A game value that can be used by Cocos.
         */
        _convertDesignValueToGameValue: function( designValue, node )
        {
            return designValue;
        },
        
        /**
         * Normally, uses the animatedActionClassName ( ctor parameter ) to create an action, and
         * passes the value in as its own parameter. However, some actions need more parameters,
         * so for this case you can override this.
         *
         * @param {*} duration The duration that the action will be applied for.
         * @param {*} endValue The value to create the action with.
         */
        _createAnimationAction: function( duration, endValue )
        {
            return cc[ this._animatedActionClassName ].create( duration, endValue );
        },
        
        /**
         * This should be overridden by subclasses. Use this to apply a setProperty value.
         *
         * @param {cc.Node} node The node that will have the property applied.
         * @param {*} value The value that will be applied.
         */
        _setFunction: function( node, value )
        {
            return undefined;
        },
        
        //adds an action to a sequence, or simply returns the action if there is no sequence
        _addAction:function( prevSequence, action )
        {
            if ( !prevSequence )
            {
                return action;
            }
            return cc.Sequence.create( prevSequence, action );
        }
    });
}() );