/*global cc*/
/*global LumosityUtils*/
/*global Interaction*/
/*global InteractionType*/
/*global InteractionPhase*/

/**
*  An enumeration that describes swipe directions.
*/
var SwipeGestureDirection = 
{
    Up:    0,
    Right: 1,
    Down:  2,
    Left:  3
};

var SwipeInteractionRecognizer;
( function()
{
    "use strict";

    /**
    *  SwipeInteractionRecognizer performs the necessary functions to detect different types of
    *  gestures.  
    */
    SwipeInteractionRecognizer = cc.Class.extend(
    {
        /** Keeps track of the most currently recorded direction of the user's swipe. */
        _swipeDirection: null,

        /** Represents the first touch position when detecting a swipe motion. */
        _swipeInitialPosition: null,

        /** Models the smallest distance between points that can be interpreted as a swipe. */
        _swipeMinimumDistance: 50,

        /** The touch id of the first valid touch object. */
        _touchID: null,

        /** An array of unique touch id's. Helps with detecting special gestures on iOS. */
        _uniqueTouches: [],

        /**
        *  Checks the input for a valid swipe interaction. It has to deal with two different phases:
        *  onTouchBegin() and onTouchEnd().
        *
        *  @param {cc.Touch|Array} touches Either an array of touches or a single touch.
        *  @param {InteractionPhase} phase The phase of the touch you are reflecting.
        *  @param {Object} interactionResponder The object that we send interaction events to.
        */
        checkSwipeInteraction:function( touches, phase, interactionResponder )
        {
            LumosityUtils.assert( touches, "Expects a valid touches object." );
            LumosityUtils.assert( phase > InteractionPhase.Unknown && phase <= InteractionPhase.Cancelled, "Expects InteractionPhase between (InteractionPhase.Unknown, InteractionPhase.Cancelled]." );
            LumosityUtils.assert( interactionResponder, "Expects a valid interactionResponder object." );

            if( phase === InteractionPhase.Began )
            {
                this._beginSwipeDetection( touches );
            }
            else if( phase === InteractionPhase.Ended )
            {
                this._endSwipeDetection( touches, interactionResponder );
            }
        },

        /**
        *  Records the initial swipe position (where swiping begins), and is the start
        *  of detecting a swipe interaction.
        *
        *  @param {cc.Touch|Array} touches Either an array of touches or a single touch.
        */
        _beginSwipeDetection:function( touches )
        {
            LumosityUtils.assert( touches, "Expects a valid touches object." );

            var i = 0;
            for( i = 0; i < touches.length; i++ )
            {
                if( touches[i].getId )
                {
                    var touchID = touches[i].getId();
                    if( this._uniqueTouches.indexOf( touchID ) === -1 )
                    {
                        this._uniqueTouches.push( touchID );
                    }
                }
            }

            if( this._touchID === null )
            {
                this._swipeInitialPosition = touches[ 0 ].getLocation();
                this._touchID = touches[ 0 ].getId();
            }
        },

        /**
        *  This is where we kick off the process to determine if the interaction performed
        *  qualifies as a swipe gesture.  
        *
        *  @param {cc.Touch|Array} touches Either an array of touches or a single touch.
        *  @param {Object} interactionResponder The object that we send interaction events to.
        */
        _endSwipeDetection:function( touches, interactionResponder )
        {
            LumosityUtils.assert( touches, "Expects a valid touches object." );
            LumosityUtils.assert( interactionResponder, "Expects an interactionResponder." );
            LumosityUtils.assert( typeof interactionResponder.interaction_received === "function", "Expects that interactionResponder has a method named interaction_received." );

            var uniqueTouchesLength = this._uniqueTouches.length;

            if ( touches.length <= 0 )
            {
                return;
            }

            if( uniqueTouchesLength >= 4 )
            {
                this._uniqueTouches = [];
                this._touchID = null;
                this._swipeInitialPosition = null;
                return;
            }
            else
            {
                this._uniqueTouches = this._uniqueTouches.filter( function(x) { return touches.indexOf(x) < 0; } );
            }

            var incomingTouchID = touches[ 0 ].getId();
            var finalPosition = touches[ 0 ].getLocation();

            if( incomingTouchID === this._touchID )
            {
                this._touchID = null;
                this._checkForValidSwipeGesture( finalPosition, interactionResponder, this._distanceBetweenPoints( this._swipeInitialPosition, finalPosition ) );
                this._swipeInitialPosition = null;
            }
        },

        /**
        *  Determines if interaction qualifies as a swipe. Checks the distance between the
        *  intial touch position and the ending touch position is within some acceptable range.
        *  Creates an Interaction object and invokes the interactionResponder::interaction_received
        *  method if a swipe was correctly identified.
        *
        *  @param {cc.Point} finalPosition Describes the last point of touch that's part of the swipe gesture.
        *  @param {Object} interactionResponder The object that we send interaction events to.
        *  @param {number} distance The length of the swipe.
        */
        _checkForValidSwipeGesture:function( finalPosition, interactionResponder, distance )
        {
            var swipeInteractionObject = null;
            if( distance >= this._swipeMinimumDistance )
            {
                swipeInteractionObject = this.createSwipeInteractionObject( this._swipeInitialPosition, finalPosition );
                interactionResponder.interaction_record( swipeInteractionObject );
                interactionResponder.interaction_received( swipeInteractionObject );
            }
        },

        /**
        *  Creates an Interaction object with the characteristics of a swipe.
        *
        *  @param {cc.Point} position Ending coordinate of the swipe gesture.
        *  returns {Object} An Interaction object of type Swipe.
        */
        createSwipeInteractionObject:function( initialPosition, finalPosition )
        {
            var interactionData = {};
            interactionData[ Interaction.getPositionsKey() ] = [ initialPosition, finalPosition ];
            interactionData[ Interaction.getSwipeDirectionKey() ] = this._getSwipeDirection( initialPosition, finalPosition );
            return new Interaction( InteractionType.Swipe, InteractionPhase.Ended, interactionData );
        },

        /**
        *  Calculates the distance between two points.
        *
        *  @param {cc.Point} point1 First, initial point.
        *  @param {cc.Point} point2 Second point.
        *  returns {number} The distance between two points.
        */
        _distanceBetweenPoints:function( point1, point2 )
        {
            var deltaX = point2.x - point1.x,
                deltaY = point2.y - point1.y;
            return Math.sqrt( ( deltaX * deltaX ) + ( deltaY * deltaY ) );
        },

        /**
        *  Calculates the direction of the user's swipe interaction [Up, Left].
        *  We basically test to see if the user's swipe motion tends to any particular direction, and return
        *  the direction that was greater. So, if a user swipes up and to the left slightly, the up direction would
        *  be returned.
        *
        *  @param {cc.Point} point1 First, initial point.
        *  @param {cc.Point} point2 Second point.
        *  returns {SwipeGestureDirection} An enumeration type that maps to the set {Up, Right,
        *                                     Down, Left}.
        */
        _getSwipeDirection:function( point1, point2 )
        {
            LumosityUtils.assert( point1 && point2, "Expects valid point objects." );

            var horizontalSwipeDirection = null,
                horizontalPointDelta = null,
                verticalSwipeDirection = null,
                verticalPointDelta = null;

            if( point2.x - point1.x >= this._swipeMinimumDistance )
            {
                //RIGHT
                horizontalPointDelta = point2.x - point1.x;
                horizontalSwipeDirection = SwipeGestureDirection.Right;
            }
            else
            {
                //LEFT
                horizontalPointDelta = point1.x - point2.x;
                horizontalSwipeDirection = SwipeGestureDirection.Left;
            }

            if( point1.y - point2.y >= this._swipeMinimumDistance )
            {
                //DOWN
                verticalPointDelta = point1.y - point2.y;
                verticalSwipeDirection = SwipeGestureDirection.Down;
            }
            else
            {
                //UP
                verticalPointDelta = point2.y - point1.y;
                verticalSwipeDirection = SwipeGestureDirection.Up;
            }
            return ( verticalPointDelta > horizontalPointDelta ) ? verticalSwipeDirection : horizontalSwipeDirection;
        }
    });
}());
