/*jslint node: true */
/*global cc*/
"use strict";

/**

 Features
 --------

 * Set parent variable to child names, like Flash instance names.

 * Load image from sprite sheet or file.

 * Run animation on root node of CCB file.
   Cocos2d-x js v2.2.2 C++ reader:
   To access animations, that reader expects to set jsControlled true.
   Otherwise that reader assigns animation manager to a node user object 
   but does not expose the method to get the user object.
   JS controlled nodes disable auto play animation, so this JS loader auto plays.
   js_bindings_ccbreader.cpp

 * Splice child CCB with clipping node.


 Not supported
 -------------

 * Cocos2d-x js v2.2.2 does not support:
 
    * Custom class node loader in JavaScript.

    * Reparenting over time or with animations.

 */

cc.BuilderReaderHelper = cc.BuilderReaderHelper || {};

/**
 * This helper depends on the following patches to cc.Node and cc.BuilderAnimationManager.
 */
cc.BuilderReaderHelper.monkeyPatch = function()
{
    /**
     * Cocos2d-x v3 uses setName and getName.  
     */
    if (!cc.Node.prototype.setName) {
        cc.Node.prototype.setName = function(name) {
            this._name = name;
        };
        cc.Node.prototype.getName = function() {
            return this._name;
        };
    }
};

/**
 * Set property "name".
 * Do not set tag.  Version 2.2.2 supports integer tag.  Cocos2d-x crashes if passing a string.
 * Version 3 supports string name.  Would be more compatible to getName/setName.
 */
cc.BuilderReaderHelper.setName = function(node, name)
{
    if (node.setName) {
        node.setName(name);
    }
    else {
        node.name = name;
    }
};

/**
 * Set parent variable name of node, if valid and not already taken.
 * Flash Professional sets parent member variable to instance name.
 * @param   isOverride  If true, then override previous value.
 */
cc.BuilderReaderHelper.trySetParentVariable = function(parent, name, node, isOverride)
{
    if (!parent || !name) {
    }
    else if (undefined === parent[name] || isOverride) {
        try {
            parent[name] = node;
        }
        catch (err) {
            cc.log('cc.BuilderReaderHelper.trySetParentVariable: Cannot set "' 
                + parent + '" variable with a name of "' + name + '" to node "' + node + '".');
        }
    }
    else if (node !== parent[name]) {
        cc.log('cc.BuilderReaderHelper.trySetParentVariable: "' 
            + parent + '" already has variable "' + name + '" of value "' + parent[name] + '".');
    }
};

/**
 * @param   grandparent {cc.Node}   Set member variables.  Useful to wire {cc.ScrollView}, from CocosBuilder, which inserts another node in-between.  
 *
 * @param   memberNames {Array}     Granchild names that will become member variables of the grandparent.
 *
 * @return  {cc.Node} Single child of grandparent.
 *
 * GAME-4509 Test case:  2015-10-16 Highway Hazards.  Press pause.  Expect background and scenery are cropped.  Pause screen slides in.  Got background behind pause screen.
 */
cc.BuilderReaderHelper.setGrandchildVariables = function(grandparent, memberNames)
{
    var child = grandparent.getChildren()[0];
    for (var n = 0; n < memberNames.length; n++) {
        var name = memberNames[n];
        var node = child[name];
        cc.BuilderReaderHelper.trySetParentVariable(grandparent, name, node);
    }
    return child;
};

/**
 * Expects children already have names and node variables.
 */
cc.BuilderReaderHelper.setup = function(node)
{
    var helper = cc.BuilderReaderHelper;
    helper.mayScaleCapInsets(node, helper.capInsetsScale, false);
    cc.BuilderReaderHelper.spliceClippingNode(node);
};

/**
 * @param   owner   If set, set variables on owner.  Default of root.
 */
cc.BuilderReaderHelper.trySetChildrenVariables = function(root, owner)
{
    var children = root.getChildren();
    owner = owner || root;
    for (var c = 0; c < children.length; c++) {
        var child = children[c];
        if (child.getName) {
            var name = child.getName();
            cc.BuilderReaderHelper.trySetParentVariable(owner, name, child);
        }
    }
}

/**
 * Recursively set parent member variables to child.
 */
cc.BuilderReaderHelper.setupTree = function(root)
{
    cc.BuilderReaderHelper.setup(root);
    var children = root.getChildren();
    for (var c = 0; c < children.length; c++) {
        var child = children[c];
        if (child.getName) {
            var name = child.getName();
            cc.BuilderReaderHelper.trySetParentVariable(root, name, child);
        }
        cc.BuilderReaderHelper.setupTree(child);
    }
};

/**
 * Assign node name and parent member variable to each child.
 */
cc.BuilderReaderHelper.load = function(ccbFileName, owner)
{
    cc.BuilderReaderHelper.monkeyPatch();
    var root = cc.BuilderReaderHelper._load(ccbFileName, owner, null);
    cc.BuilderReaderHelper.setupTree(root);
    return root;
};

//
// CocosBuilder definitions
//

cc.BuilderReaderHelper = cc.BuilderReaderHelper || {};

var _ccbGlobalContext = this;

/**
 * HTML 5 has no cc._Reader.create or reader.load;
 * Cocos2d-x js makes these extensions already.
 */
cc.BuilderReaderHelper.extendHtml5 = function()
{
    if (!cc._Reader) 
    {
        cc._Reader = {};
    }
    if (!cc._Reader.create) 
    {
        cc._Reader.create = function()
        {
            return new cc.BuilderReader(
                cc.NodeLoaderLibrary.newDefaultCCNodeLoaderLibrary());
        };
    }
    var prototype = cc.BuilderReader.prototype;
    if (prototype && !prototype.load) {
        prototype.load = prototype.readNodeGraphFromFile;
    }
};

cc.BuilderReaderHelper.nodesWithAnimationManagers = [];

/**
 * Load file.
 * Name nodes from owner member variable in JS controlled CCB.
 * Expects each control button has an owner variable.
 * @param   owner   If none, then a stub is created to enable naming nodes.
 *
 * Adapted from Cocos2d-x v2.2.2 jsb_cocosbuilder.js
 */
cc.BuilderReaderHelper._load = function(file, owner, parentSize)
{
    owner = owner || {};
    cc.BuilderReaderHelper.extendHtml5();
    // Load the node graph using the correct function
    var reader = cc._Reader.create();
    // Cocos2d-js assigns extension, but HTML5 does not.
    if((file.length < 5)||(file.toLowerCase().lastIndexOf(".ccbi") != file.length - 5))
        file = file + ".ccbi"; 

    var node;

    if (parentSize)
    {
        node = reader.load(file, null, parentSize);
    }
    else
    {
        node = reader.load(file);
    }

    // Assign owner callbacks & member variables
    if (owner)
    {
        // Callbacks
        var ownerCallbackNames = reader.getOwnerCallbackNames();
        var ownerCallbackNodes = reader.getOwnerCallbackNodes();

        for (var i = 0; i < ownerCallbackNames.length; i++)
        {
            var callbackName = ownerCallbackNames[i];
            var callbackNode = ownerCallbackNodes[i];

           if (owner[callbackName] === undefined)
            {
                cc.log("Warning: " + "owner." + callbackName + " is undefined.");
            }
            else
            {
                if(callbackNode instanceof cc.ControlButton)
                {
                    var ownerCallbackControlEvents = reader.getOwnerCallbackControlEvents();
                    callbackNode.addTargetWithActionForControlEvents(owner, owner[callbackName], ownerCallbackControlEvents[i]);
                }
                else
                {
                    callbackNode.setCallback(owner[callbackName], owner);
                }
            }
        }

        // Variables
        var ownerOutletNames = reader.getOwnerOutletNames();
        var ownerOutletNodes = reader.getOwnerOutletNodes();

        for (var i = 0; i < ownerOutletNames.length; i++)
        {
            var outletName = ownerOutletNames[i];
            var outletNode = ownerOutletNodes[i];
            owner[outletName] = outletNode;
            outletNode.setName(outletName);
            // cc.log("load: ownerOutlet " + outletNode + " name " + outletNode.getName());
        }
    }

    var nodesWithAnimationManagers = reader.getNodesWithAnimationManagers();
    var animationManagersForNodes = reader.getAnimationManagersForNodes();
    var length = nodesWithAnimationManagers ? nodesWithAnimationManagers.length : 0;

    // Attach animation managers to nodes and assign root node callbacks and member variables
    for (var i = 0; i < length; i++)
    {
        var innerNode = nodesWithAnimationManagers[i];
        var animationManager = animationManagersForNodes[i];

        innerNode.animationManager = animationManager;

        var documentControllerName = animationManager.getDocumentControllerName();
        if (!documentControllerName) continue;

        // Create a document controller
        var controller = new _ccbGlobalContext[documentControllerName]();
        controller.controllerName = documentControllerName;

        innerNode.controller = controller;
        controller.rootNode = innerNode;

        // Callbacks
        var documentCallbackNames = animationManager.getDocumentCallbackNames();
        var documentCallbackNodes = animationManager.getDocumentCallbackNodes();

        for (var j = 0; j < documentCallbackNames.length; j++)
        {
            var callbackName = documentCallbackNames[j];
            var callbackNode = documentCallbackNodes[j];

            if (controller[callbackName] === undefined)
            {
                cc.log("Warning: " + documentControllerName + "." + callbackName + " is undefined.");
            }
            else
            {
                if(callbackNode instanceof cc.ControlButton)
                {
                    var documentCallbackControlEvents = animationManager.getDocumentCallbackControlEvents();
                    callbackNode.addTargetWithActionForControlEvents(controller, controller[callbackName], documentCallbackControlEvents[j]); 
                }
                else
                {
                    callbackNode.setCallback(controller[callbackName], controller);
                }
            }
        }
        cc.BuilderReaderHelper.nodesWithAnimationManagers = 
            cc.BuilderReaderHelper.nodesWithAnimationManagers.concat(
                reader.getNodesWithAnimationManagers() );


        // Variables
        var documentOutletNames = animationManager.getDocumentOutletNames();
        var documentOutletNodes = animationManager.getDocumentOutletNodes();

        for (var j = 0; j < documentOutletNames.length; j++)
        {
            var outletName = documentOutletNames[j];
            var outletNode = documentOutletNodes[j];

            controller[outletName] = outletNode;
        }

        if (typeof(controller.onDidLoadFromCCB) == "function")
        {
            controller.onDidLoadFromCCB();
        }

        // Setup timeline callbacks
        var keyframeCallbacks = animationManager.getKeyframeCallbacks();
        for (var j = 0; j < keyframeCallbacks.length; j++)
        {
            var callbackSplit = keyframeCallbacks[j].split(":");
            var callbackType = callbackSplit[0];
            var callbackName = callbackSplit[1];
            
            if (callbackType == 1) // Document callback
            {
                var callfunc = cc.CallFunc.create(controller[callbackName], controller);
                animationManager.setCallFunc(callfunc, keyframeCallbacks[j]);
            }
            else if (callbackType == 2 && owner) // Owner callback
            {
                var callfunc = cc.CallFunc.create(owner[callbackName], owner);
                animationManager.setCallFunc(callfunc, keyframeCallbacks[j]);
            }
        }
        
        // Start animation
        var autoPlaySeqId = animationManager.getAutoPlaySequenceId();
        if (autoPlaySeqId != -1)
        {
            animationManager.runAnimationsForSequenceIdTweenDuration(autoPlaySeqId, 0);
        }
    }

    return node;
};


// Replace:

/**
 * Besides usual node properties, also copy animations and CCB file base name.
 */
cc.BuilderReaderHelper.clone = function(previousNode, nextNode, animationManager)
{
    if (previousNode == nextNode) {
        return;
    }
    nextNode.setPosition(previousNode.getPosition());
    nextNode._positionType = previousNode._positionType;
    nextNode.setRotation(previousNode.getRotation());
    nextNode.setScaleX(previousNode.getScaleX());
    nextNode.setScaleY(previousNode.getScaleY());
    if (previousNode.getName) {
        var name = previousNode.getName();
        cc.BuilderReaderHelper.setName(nextNode, name);
    }
    var visible = previousNode.isVisible() !== false;
    if (!visible) {
        cc.log("cc.BuilderReaderHelper.clone: hiding nextNode " + nextNode);
    }
    nextNode.setVisible(visible);
    var nodes = cc.BuilderReaderHelper.nodesWithAnimationManagers;
    if (!animationManager) {
        var index = nodes.indexOf(previousNode);
        if (0 <= index) {
            animationManager = nodes[index].animationManager;
        }
    }
    if (animationManager) {
        animationManager.moveAnimationsFromNode(previousNode, nextNode);
        if (nodes.indexOf(nextNode) <= -1) {
            nodes.push(nextNode);
        }
    }
};


cc.BuilderReaderHelper.adoptVariable = function(previousParent, nextParent, node)
{
    var name = node.getName();
    if (name) {
        cc.BuilderReaderHelper.trySetParentVariable(nextParent, name, node, true);
        if (node == previousParent[name]) {
            previousParent[name] = undefined;
        }
    }
};

/**
 * Adopt children.  Adopt parent variable.  Test case:  2015-01-10 Restart with different parents.  Expect to see children.
 */
cc.BuilderReaderHelper.adopt = function(previousParent, nextParent)
{
    if (previousParent == nextParent)
    {
        return;
    }
    var children = previousParent.getChildren();
    while (1 <= children.length)
    {
        var child = children[0];
        child.retain();
        child.removeFromParent(false);
        nextParent.addChild(child);
        child.release();
        cc.BuilderReaderHelper.adoptVariable(previousParent, nextParent, child);
    }
    var animationManager = previousParent.animationManager;
    if (animationManager) {
        animationManager.moveAnimationsFromNode(previousParent, nextParent);
    }
}


cc.BuilderReaderHelper.ccbCache = cc.BuilderReaderHelper.ccbCache || {};

/**
 * Replace cache with new parent.
 * @param   key         Cache key.  Format:  "name_ccbFileBaseName".  
 * @param   nextParent  If undefined, then create a node.
 *
 * To configure to cache next time it is loaded, set value to true.
 *     cache[nodeName] = true;
 * To save by node name.  Initialize keys of node names with value of node.
 *     cache[nodeName] = node;
 */
cc.BuilderReaderHelper.ccbCacheAdopt = function(key, nextParent)
{
    var previousParent = cc.BuilderReaderHelper.ccbCache[key];
    if (previousParent instanceof cc.Node) {
        nextParent = nextParent || cc.Node.create();
        cc.BuilderReaderHelper.adopt(previousParent, nextParent);
        if (nextParent != previousParent) {
            cc.BuilderReaderHelper.ccbCache[key] = nextParent;
        }
        return nextParent;
    }
}


/**
 * XXX Not supported in cocos2d-x v2.2.2 JS when used over multiple frames with animated nodes.
 *
 * Replace member variable of parent.
 * Clone previous child.
 * Even after adding the child, do not release.  Test case:  2015-01-11 Cocos2d-x v2.2.2 JS.  Replace a node.  Wait 5 seconds.  Expect game continues.  Got crash with "Invalid Native Object".  Cocos2d-html5 has no problem.  Test case:  2015-01-11 Cocos2d-x v2.2.2 JS.  Restart.  Loading.  Game hangs.  Cococs2d-html5 has no problem.
 */
cc.BuilderReaderHelper.replace = function(parent, name, nextChild, previousChild)
{
    previousChild = previousChild || parent[name];
    var z = parent.getChildrenCount();
    if (null != previousChild && previousChild != nextChild) {
        if (previousChild.getParent()) {
            cc.BuilderReaderHelper.adopt(previousChild, nextChild);
            cc.BuilderReaderHelper.clone(previousChild, nextChild);
            z = previousChild.getZOrder();
            // XXX previousChild.retain();
            previousChild.removeFromParent();
            // previousChild.release();
        }
        var previousParent = nextChild.getParent();
        var isMoving = previousParent && parent != previousParent;
        if (isMoving) {
            nextChild.retain();
            nextChild.removeFromParent();
        }
        if (name) {
            parent[name] = nextChild;
        }
        if (!nextChild.getParent()) {
            parent.addChild(nextChild, z);
            if (isMoving) {
                nextChild.release();
            }
        }
    }
};

/**
 * If parent is a cc.ClippingNode and child is named "stencil", then set stencil.
 * Adapted from example: http://www.waitingfy.com/archives/1093
 *
 * Clipping UI convention, like Flash Professional "mask".
 *     Flash Professional
 *         mask layer
 *             masked layer
 *             masked layer
 *     Cocos2d
 *         ClippingNode (class)
 *             masked
 *             stencil (name of a cc.Sprite or cc.Node)
 *             masked
 * HTML5 without WebGL does not trim the pixels.  It shows rectangle rotated to the stencil.
 * Cocos-html5 v2 renders clipping node with part of the area unclipped.
 *
 * GAME-2019 Set stencil opacity to 1.  To ensure stencil area renders the rest of the clipping node children, do not set stencil invisible.  Test case:  2015-01-21 Trouble Brewing level select.  Wouter expects to see steam rise.  Got no visible animation.
 */
cc.BuilderReaderHelper.clipByStencilName = function(clippingNode, nodeNamedStencil) {
    if (clippingNode instanceof cc.ClippingNode) {
        if ("stencil" == nodeNamedStencil.getName()) {
            clippingNode.setStencil(nodeNamedStencil);
            clippingNode.setAlphaThreshold(0.000001);
            var stencilSize = nodeNamedStencil.getContentSize();
            clippingNode.setContentSize(stencilSize.width, stencilSize.height);
            nodeNamedStencil.setOpacity(1);
            if (clippingNode._cangodhelpme) {
                clippingNode._cangodhelpme(true);
            }
        }
    }
}

/**
 * cocos2d-x js v2.2.2 Reparent.  App hangs.
 * Replace node in scene graph with clipping node.
 * Expects node has a parent and name.
 * Adopt all properties of node, such as children, parent, and animation.
 * Structure:  node.clippingNode.root.
 */
cc.BuilderReaderHelper.replaceWithClippingNode = function(node, ccbFileName) {
    node.removeAllChildren();
    var content = cc.BuilderReaderHelper.load(ccbFileName);
    cc.BuilderReaderHelper.setupTree(content);
    cc.BuilderReaderHelper.trySetChildrenVariables(content, node);
    var clippingNode = cc.ClippingNode.create();
    content.setName("root");
    clippingNode.addChild(content);
    cc.BuilderReaderHelper.trySetChildrenVariables(clippingNode);
    cc.BuilderReaderHelper.clipByStencilName(clippingNode, node.stencil);
    clippingNode.setName("clippingNode");
    node.addChild(clippingNode);
    cc.BuilderReaderHelper.trySetChildrenVariables(node);
    var animationManager = content.animationManager;
    if (animationManager) 
    {
        // Start animation
        var autoPlaySeqId = animationManager.getAutoPlaySequenceId();
        if (autoPlaySeqId != -1)
        {
            animationManager.runAnimationsForSequenceIdTweenDuration(autoPlaySeqId, 0);
        }
    }
}

/**
 * Rename parent variable if found.
 */
cc.BuilderReaderHelper.rename = function(node, previousName, name)
{
    node.setName(name);
    var parent = node.getParent();
    if (parent)
    {
        if (node == parent[previousName])
        {
            if (name)
            {
                parent[name] = node;
            }
            delete parent[previousName];
        }
    }
};

/**
 * Create clipping node of child CCB. 
 *         content_class_ClippingNode_CupDoubleClippingNode
 * 2014-11 Wouter expects to animate clipping node of coffee filling.
 * 
 * Details:
 *
 * Duplicate CCB.
 * Stub previous CCB.  Otherwise animation manager crashes.
 * Create a clipping node from a CCBI.
 *     From owner variable.  Example:
 *         content_class_ClippingNode_CupDoubleClippingNode
 *         content_class_ClippingNode_CupScrollButtonClippingNode
 *         content_class_ClippingNode_CupSingleClippingNode
 *         highlight_class_ClippingNode_LevelButtonClippingNode
 * Remove children of parent.
 * Trim name at "_class_ClippingNode".
 * Parent adds clipping node.
 * Wire loaded children to parent of clipping node.
 * Setup clipping node.
 *     Find first CCB content child named "stencil".
 * Move parent variable to trimmed name.
 * Animate parent and clipping node child at same time.
 * Wire stencil.
 *
 * The above hierarchy works around these unfortunate situations with cocos2d-x js v2.2.2:
 *      Reader does not support custom class JS node loader.
 *      Engine does not support reparenting animated nodes.
 */
cc.BuilderReaderHelper.spliceClippingNode = function(node)
{
    var name = node.getName();
    var isClippingNode = name && 0 <= name.indexOf("_class_ClippingNode");
    if (isClippingNode)
    {
        var previousName = name;
        var index = name.indexOf("_class_ClippingNode");
        name = name.substr(0, index);
        cc.BuilderReaderHelper.rename(node, previousName, name);
        var words = previousName.split("_");
        var ccbFileName = words[words.length - 1];
        cc.BuilderReaderHelper.replaceWithClippingNode(node, ccbFileName);
    }
};

// Button:

/**
 * Workaround: For each target, create a unique function.
 * If readd method to same control, ignore.
 *
 * GOTCHA: In Cocos2d-x, adding the same function and event to a different target silently fails.
 * "if (it->second->_jsFunc == jsFunc && arg2 == it->second->_type)"
 * https://github.com/cocos2d/cocos2d-x/blame/v2/scripting/javascript/bindings/jsb_cocos2dx_extension_manual.cpp#L687
 *
 * Test case:
 * cocos2d-html5 v2.2.2 accepts cc.Control.addTargetWithActionForControlEvents
 * However, cocos2d-x jsb this code had no apparent effect on second and third button.
 * The first button responds.
 * http://www.cocos2d-x.org/reference/html5-js/V2.2.3/index.html
 *
 * Bug reported here:
 * https://github.com/cocos2d/cocos2d-x/issues/9480
 */
cc.BuilderReaderHelper.addTargetWithActionForControlEvents = function(control, methodOwner, method, event)
{
    if (null == method) {
        throw new Error("Calling will expect a method.");
    }
    var uniqueFunction = function(){
        return method.call(methodOwner, control, event);
    };
    if (!control.listens) {
        control.listens = [];
    }
    if (0 <= control.listens.indexOf(method)) {
        // cc.log("cc.BuilderReaderHelper.addTargetWithActionForControlEvents: target has this method.");
        return;
    }
    control.listens.push(method);
    return control.addTargetWithActionForControlEvents(methodOwner, uniqueFunction, event); 
};

cc.BuilderReaderHelper.capInsetsScale = cc.rect(0, 0, 1, 1);

/**
 * Scale cap insets of ControlButton Scale9Sprite.  
 * To disable overwriting default cap insets, set this to null.
 * At (0, 0, 1, 1) cap insets disable Scale9Sprite.
 * Set preferred size to original sprite size.
 * Refers to numerical control states. 
 * Test case:  Cocos2d-js v2 does not accept string states.
 *
 * @param   enableZoomOnTouch   If defined, set zoom on touch down.
 */
cc.BuilderReaderHelper.mayScaleCapInsets = function(node, scale, enableZoomOnTouch) {
    function hasArea(rect) {
        return rect && (rect.x || rect.y || rect.width || rect.height);
    }
    if (node instanceof cc.ControlButton) {
        if (undefined !== enableZoomOnTouch) {
            node.setZoomOnTouchDown(enableZoomOnTouch);
        }
        if (hasArea(scale)) {
            var states = [cc.CONTROL_STATE_NORMAL, cc.CONTROL_STATE_HIGHLIGHTED, 
                          cc.CONTROL_STATE_DISABLED, cc.CONTROL_STATE_SELECTED ];
            for (var s = 0; s < states.length; s++) {
                var state = states[s];
                var scale9Sprite = node.getBackgroundSpriteForState(state);
                if (scale9Sprite) {
                    var size = scale9Sprite.getOriginalSize();
                    var width = size.width;
                    var height = size.height;
                    var insets = cc.rect(scale.x * width, scale.y * height, 
                        scale.width * width, scale.height * height);
                    scale9Sprite.setCapInsets(insets);
                    var size = cc.size(width, height);
                    node.setPreferredSize(size);
                }
            }
        }
    }
}



// Animation:


/**
 * All descendents of a CCB root simultaneously play the named sequence.
 * @param   ccbRootNode     LearnCocos2D expects to call animation on CCB root node.  (stackoverflow.com 2012)
 * @param   sequenceName    If the node does not have this name, game crashes.
 * @param   doNotRestart    If currently running and not yet finished, do not restart.
 * @return  If started playing now.
 *
 * Would be nicer to check if animation exists instead of crashing when it does not.
 * Cocos2d-x v2.2.2 JavaScript bindings only expose "getSequences", not "getSequenceId".
 * Unfortunately, Cocos2d-x v2.2.2 crashes when "getSequences" is called,
 * because bindings do not support a Sequence.
 * (garfield_ho and dumganhar, July 2013, discuss.coco2d-x.org)
 * Cocos2d: Assert failed: the type isn't suppored.
 * Assertion failed: (false), function ccarray_to_jsval, file /Users/ethan/workspace/Locomotive/DoubleWide/cocos2d-x/scripting/javascript/bindings/ScriptingCore.cpp, line 1426.
 *
 */
cc.BuilderReaderHelper.runAnimations = function(ccbRootNode, sequenceName, doNotRestart)
{
    var isStarting = false;
    var animationManager = ccbRootNode.animationManager;
    if (animationManager)
    {
        var willStart = !doNotRestart 
            || sequenceName != animationManager.getRunningSequenceName(); 
        if (willStart)
        {
            animationManager.runAnimations(sequenceName);
            // cc.log("runAnimations: " + sequenceName);
            isStarting = true;
        }
    }
    return isStarting;
}

/**
 * GAME-3705 On restart, nullify current sequence and current state on the CocosBuilder root node.  Test case: 2015-06-22 Select level.  Immediately, before HUD appears, pause.  Tap How to Play or Restart.  Sarah Clark expects HUD does not appear.
 */
cc.BuilderReaderHelper.onRestart = function(rootNode)
{
    if (rootNode) {
        rootNode.currentSequence = null;
        rootNode.currentState = null;
    }
}

/**
 * @param   root        CCB root node with animation manager.
 * @param   delegate    When root animation finishes and not looped, call delegate.completedAnimationSequenceNamed
 *                      That method name is a Cocos2d-x convention.
 *                      If delegate is not defined, then do not call.
 * Unfortunately I didn't find a way for cocos2d-x js v2.2.2 to call completedAnimationSequenceNamed
 * @return  Completed sequeunce name.  undefined if not completed.  This is more flexible than a delegate.  Example:  @see Editors Choice
 *
 * Test case:  2015-02-06 Trouble Brewing:  Ethan expects to call function at end of scene and tutor animation.  Paddle Fighter:  Konstantin expects to call function on end of animation.
 * Test case:  2015-06-04 Editors Choice:  Ethan expects to softwire cross-platform update.
 */
cc.BuilderReaderHelper.update = function(root, delegate)
{
    var completedSequenceName;
    if (root)
    {
        var previous = root.currentSequence;
        var next = root.animationManager.getRunningSequenceName();
        if (previous != next)
        {
            if (previous)
            {
                completedSequenceName = previous;
                if (delegate) {
                    delegate.completedAnimationSequenceNamed(previous);
                }
            }
            root.currentSequence = next;
        }
    }
    return completedSequenceName;
}


/**
 * Call this function to wire callbacks to CocosBuilder animation events before playing those animations.
 *
 * @param   ccbRootNode     {cc.Node} Loaded from CCB that contains the animation manager.
 * @param   selectorTarget  {object} The owner of the method.
 * @param   selectorName    {string} The name of the method.
 *
 * @author Mike Edwards
 */
cc.BuilderReaderHelper.setCallFunc = function( ccbRootNode, selectorTarget, selectorName )
{
    var finishedAction = cc.CallFunc.create( selectorTarget[ selectorName ], selectorTarget );
    var animationManager = ccbRootNode.animationManager;
    if( animationManager )
    {
        animationManager.setCallFunc( finishedAction, "1:" + selectorName );
    }
}

/**
 * GAME-3280 When getting a node's state, if no animation is running, then return the node's member variable named "currentState", which was defined by function "setState".  Otherwise, when an animation finishes running and yet the last frame of the animation is still displaying, the state would be returned as null.  Test case: 2015-06-19 Xcode simulator and Nexus 9:  Tutor.  Progress Page appears.  Correct.  Get next trial wrong.  Get next trial correct.  Bev Chung expects tutor text updates.  Got tutor text remains.  On Flash version, this symptom does not occur.
 */
cc.BuilderReaderHelper.getState = function(rootNode)
{
    var state;
    if (rootNode)
    {
        if (rootNode.animationManager) {
            state = rootNode.animationManager.getRunningSequenceName();
        }
        if (null == state) {
            state = rootNode.currentState;
        }
    }
    return state;
}


/**
 * runAnimations sequence name if not already at that currentState.
 * @return if did go.
 */
cc.BuilderReaderHelper.setState = function(rootNode, sequenceName)
{
    var isChanging = false;
    if (sequenceName != rootNode.currentState) {
        rootNode.currentState = sequenceName;
        var isRunning = rootNode.animationManager 
            && sequenceName == rootNode.animationManager.getRunningSequenceName();
        if (!isRunning) {
            isChanging = cc.BuilderReaderHelper.runAnimations(rootNode, sequenceName);
        }
    }
    return isChanging;
}


/**
 * Recursively pause or resume, or whatever on every node with that has that method name.
 *
 * @param   methodArgument  If defined, then pass this single argument into the method of the node. 
 *                          Useful to set opacity.  Example from Highway Hazards:
 *                          cc.BuilderReaderHelper.visit(sprite, "setOpacity", Math.round(alpha * 255));
 *
 * GAME-1885 Visit all descendent nodes, not just descendents with animation managers.  2015-01-20 Begin filling a cup.  Observe height of coffee.  Pause.  Wait 5 seconds or more.  Resume.  [~mlow] expects coffee continues filling from previously observed height.  Coffee reaches fill line.  Matt expects the serve button appears.  Got animation had continued during pause, although serve button did appear at the right time.  
 * If no node, then log and return.  Test case:  2015-02-02 Trouble Brewing.  Click Play.  Immediately click pause before level loaded or immediately press back arrow on Android.  Android subscriber expects to exit app without error.  John Tighe expects no error on Crashlytics.    Got crash report.
 * https://www.crashlytics.com/lumos-labs-inc/android/apps/com.lumoslabs.lumosity/issues/54c6cf8265f8dfea15b43e36
 *
 */
cc.BuilderReaderHelper.visit = function(rootNode, methodName, methodArgument)
{
    if (undefined == rootNode)
    {
        cc.log("cc.BuilderReaderHelper.visit: root does not exist for method " + methodName);
        return;
    }
    if (rootNode[methodName])
    {
        // cc.log("visit: " + rootNode.getName() + " method " + methodName);
        if (undefined === methodArgument) {
            rootNode[methodName]();
        }
        else {
            // cc.log("visit: method " + methodName + " methodArgument " + methodArgument);
            rootNode[methodName](methodArgument);
        }
    }
    var children = rootNode.getChildren();
    for (var c = 0; c < children.length; c++) 
    {
        var child = children[c];
        cc.BuilderReaderHelper.visit(child, methodName, methodArgument);
    }
}

cc.BuilderReaderHelper.pause = function(rootNode)
{
    cc.BuilderReaderHelper.visit(rootNode, "pauseSchedulerAndActions");
}

cc.BuilderReaderHelper.resume = function(rootNode)
{
    cc.BuilderReaderHelper.visit(rootNode, "resumeSchedulerAndActions");
}
