var cc = cc;
var require = require;

require( "src/Utility/MultiDimensionArray.js" );
var MultiDimensionArray = MultiDimensionArray;

require( "src/Game/Maze/MazeConstants.js" );
var MazeConstants = MazeConstants;
var MazeUtility = MazeUtility;
var MazeDirections = MazeDirections;

require( "src/Utility/LumosityUtils.js" );

/**
 * A* is an incredibly commonly used AI approach for pathfinding. However, it is
 * very rarely used for maze-like instances because of the randomness of maze. Backtracking
 * will always work, but it does so at a fairly random rate. A* will also always work, but
 * every move taken is specificially addressed for odds of success. In most cases, A*
 * should perform less checks than the backtrack approach, although this is not always true.
 *
 * @author Eli Delventhal
 * @name AStarAI
 * @class
 */
var AStarAI =
{
    /** @lends AStarAI */
    
    /** The maximum number of times we will search a maze before giving up. */
    MAX_SEARCH_COUNT: 1000,
    
    /**
     * Solve the maze with this AI. Returns an array that is in stack order, which represents
     * the moves to take to solve the path. To play back the moves, pop off the back of the array
     * and move to that position (each object is an absolute position in the maze).
     * Once the stack is empty, you will have reached your target.
     *
     * @param {Maze} maze The maze to solve.
     * @param {number} startX The x position to start at.
     * @param {number} startY The y position to start at.
     * @param {number} goalX The x position to end at.
     * @param {number} goalY The y position to end at.
     * @return {Array} A stack (FILO) of moves to solve the maze. Each move is an object with {x,y}.
     */
    solveMaze: function( maze, startX, startY, goalX, goalY )
    {
        //early out if we have invalid positions passed in
        if ( startX < 0 || startY < 0 || startX >= maze.getWidth() || startY >= maze.getHeight() ||
             goalX  < 0 || goalY  < 0 || goalX  >= maze.getWidth() || goalY  >= maze.getHeight() )
        {
            return [];
        }
        
        //Create arrays to keep track of the various information we need.
        var open = [];
        var closed = [];//{};
        
        //A 2D array of room info objects, stores additional temporary info about each room
        var info = new MultiDimensionArray( maze.getWidth(), maze.getHeight() );
        
        //Start A* at the initial position, and mark that location cost of 0.
        info.set( AStarAI._createRoomInfo( 0, null ), startX, startY );
        open.push( maze.getRoom( startX, startY ) );
        var closestX = startX;
        var closestY = startY;
        
        var foundGoal = false;
        var searchCount = 0;
        var startTime = new Date().getTime();
        
        //Keep searching as long as we haven't tried every possibility.
        while ( open.length > 0 && searchCount < AStarAI.MAX_SEARCH_COUNT )
        {
            searchCount++;
            
            //Start with the first Room in the open list and remove it.
            var r = open.shift();
            
            //Set what the closest room is.
            if ( Math.abs( r.x - goalX ) + Math.abs( r.y - goalY ) < Math.abs( closestX - goalX ) + Math.abs( closestY - goalY ) )
            {
                closestX = r.x;
                closestY = r.y;
            }
            
            //If we're at the goal room, that means we're good to go, so stop searching.
            if ( r === maze.getRoom( goalX, goalY ) )
            {
                foundGoal = true;
                break;
            }
            
            //Close this Room, because it's already being evaluated.
            closed[ r.x * maze.getWidth() + r.y ] = r;
            
            var direction;
            for ( direction = 0; direction < MazeDirections.Count; direction++ )
            {
                //See which of the adjacent rooms is the best option.
                if ( r.isOpen( direction ) )
                {
                    //Set the direction to match the looped value.
                    var delta = MazeConstants.positionDeltas[ direction ];
                    
                    //The cost of each move is always 1, because you can only move 1 space at a time.
                    var nextCost = info.get( r.x, r.y ).cost + 1;
                    
                    //Store the next room.
                    var r2 = maze.getRoom( r.x + delta.x, r.y + delta.y );
                    if ( !info.get( r.x + delta.x, r.y + delta.y ) )
                    {
                        info.set( AStarAI._createRoomInfo(nextCost,r), r.x + delta.x, r.y + delta.y );
                    }
                    
                    //Now check to see if this room has a cheaper cost than the adjacent one.
                    //If it does, we don't want to bother with the adjacent room.
                    if ( nextCost < info.get( r.x + delta.x, r.y + delta.y ).cost )
                    {
                        open.remove( r2 );
                        closed[ r2.x * maze.getWidth() + r2.y ] = r2;
                    }
                    //If the adjacent room is neither open nor closed, add it to the open list.
                    else if ( !open.contains( r2 ) && !closed[ r2.x * maze.getWidth() + r2.y ] )
                    {
                        this._addInOrder( open, info, r2, goalX, goalY );
                    }
                }
            }
        }
        
        //We're done. Use the parent info to find the correct path back.
        cc.log("  Solved maze: " + foundGoal + "; Took " + searchCount + " tries, took " + ( new Date().getTime() - startTime ) + "ms." );
        var path = [];
        var x = foundGoal ? goalX : closestX;
        var y = foundGoal ? goalY : closestY;
        searchCount = 0;
        while ( x !== startX || y !== startY )
        {
            if ( searchCount >= AStarAI.MAX_SEARCH_COUNT )
            {
                cc.log("AStarAI: Couldn't find a path, returning an empty one!");
                return [];
            }
            
            path.push( { x:x, y:y } );
            
            if ( !info.get( x, y ) )
            {
                cc.log("AStarAI: Couldn't find a path, returning an empty one!");
                return [];
            }
            else
            {
                var p = info.get( x, y ).parent;
                x = p.x;
                y = p.y;
            }
            
            searchCount++;
        }
        
        return path;
    },
    
    /**
     * Add a room into the open list and order it by its distance from the goal.
     * @param {Array} open The open list.
     * @param {MultiDimensionArray} info The info for this room.
     * @param {Room} r The room.
     * @param {number} goalX The goal X.
     * @param {number} goalY The goal Y.
     */
    _addInOrder: function( open, info, r, goalX, goalY )
    {
        //Loop through the open list to add in the correct location.
        var openIndex;
        for ( openIndex = 0; openIndex < open.length; openIndex++ )
        {
            var r2 = open[ openIndex ];
            if ( Math.abs( goalX - r.x ) + Math.abs( goalY - r.y ) < Math.abs( goalX - r2.x ) + Math.abs( goalY - r2.y ) )
            {
                open.splice( openIndex, 0, r );
                return;
            }
        }
        
        //If we got here, that means we should add at the end.
        open.push( r );
    },
    
    /**
     * Stores various extra information about each Room.
     * This information is necessary for A* to function.
     *
     * @param {number} cost The cost in moves to reach the Room.
     * @param {Room} room The room that this info is for.
     */
    _createRoomInfo:function( cost, room )
    {
        return { cost: cost, parent: room };
    }
};
