/**
 * Using this file:
 * General:
 *      When the component is loaded it will automatically play the file in
 *    settings.ambientSound or settings.stageSelectAmbientSound
 *
 * fadeMusicOut/In:             fades background music volume out or in
 * fadeEffectsOut/In:           fades all sound effect volume out or in
 * crossfade:                   fades current bkgrd music out and new music in afterwards
 * preloadMusic:                preloads music file (so there is no lag in game)
 * preloadEffects:              preloads effects file (so there is no lag in game)
 * stopSound:                   stops all sound
 *
 */


/*global cc*/
/*global audio*/
/*global require*/
/*global GameComponent*/

require( "src/Game/Component/GameComponent.js" );

var GameComponentAudioMixer;
( function()
{
    "use strict";

    var SoundTypes = 
    {
        musicSound:  0,
        soundEffect: 1,
        _typesNames: [ "Music", "Effects" ],

        getSoundType: function( typeIndex )
        {
            var soundType = "";
            if( typeIndex >= 0 && typeIndex <= this._typesNames.length - 1 )
            {
                soundType = this._typesNames[ typeIndex ];
            }
            return soundType;
        }
    };

    /**
     * Controls how sound effects and music (ambient tracks) fade in and out according to
     * particular events (onBegin, onPause, onEndGameplay, etc.) NOTE: in order to use an
     * ambient sound (like Ebb and Flow or Rain Drops) simply add a key called AmbientSound
     * and the path to your sound in your specific games' settings file.
     */
    GameComponentAudioMixer = GameComponent.extend(
    {
        _settings:         null,
        _isFadingAmbient: null,
        _isFadingSoundEffects: null,
        _stopAllSound:     null,
        _updateMusic:      null,
        _updateSounds:     null,
        _resetFromFade:    null,
        crossFadeAudio:    null,
        loadedSounds:      null,

        ctor:function( controller )
        {
            this._super( controller );
            this._settings = controller.settings;
            this.resumeSound();
            this.loadedSounds = [];
        },

        getType:function()
        {
            return "GameComponentAudioMixer";
        },

        onBegin: function()
        {
            this.resumeSound();
            // Play Stage Select ambient sound if you have one in your settings file.
            if( this._settings.stageSelectAmbientSound )
            {
                this._playStageSelectAmbientSound();
            }
            else
            {
                this._playAmbientSound();
            }
        },

        _playStageSelectAmbientSound: function()
        {
            this._playMusic( this._settings.stageSelectAmbientSound );
        },

        _playAmbientSound: function()
        {
            this._playMusic( this._settings.ambientSound );
        },

        _playMusic: function( music )
        {
            if( music && this._stopAllSound === false )
            {
                audio.setMusicVolume( 1 );
                //
                //NOTE: The playMusic() function has issues across android devices. Sometimes results in
                //      what seems like an infinite loop. Looping playEffect seems to work fine. 
                //
                audio.playMusic( music, true );
            }
        },

        onPause: function()
        {
            this.pauseSound();
        },

        onEndGameplay: function()
        {
            this.pauseSound();
            this._isFadingSoundEffects = false;
            this._updateSounds     = false;
            audio.setEffectsVolume( 1 );
        },

        pauseSound: function()
        {
            this._isFadingAmbient = true;
            this._isFadingSoundEffects = true;
            this._stopAllSound     = false;
            this._updateMusic      = true;
            this._updateSounds     = true;
            //
            //
            //
            this._resetFromFade = true;
        },

        onResume: function()
        {
            this.resumeSound();
        },

        resumeSound: function()
        {
            this._isFadingAmbient = false;
            this._isFadingSoundEffects = false;
            this._stopAllSound     = false;
            this._updateMusic      = true;
            this._updateSounds     = true;
        },

        fadeMusicIn: function ()
        {
            this._isFadingAmbient = false;
            this._updateMusic = true;
        },
        fadeMusicOut: function ()
        {
            this._isFadingAmbient = true;
            this._updateMusic = true;
        },
        fadeEffectsIn: function ()
        {
            this._isFadingSoundEffects = false;
            this._updateSounds = true;
        },
        fadeEffectsOut: function ()
        {
            this._isFadingSoundEffects = true;
            this._updateSounds = true;
        },
        crossFade: function ( newMusic )
        {
            this.fadeMusicOut();
            this.crossFadeAudio = newMusic;
        },
        preloadMusic: function ( fileName )
        {
            audio.preloadMusic( fileName );
            this.loadedSounds.push( fileName );
        },
        preloadEffects: function ( fileName )
        {
            audio.preloadEffects ( fileName );
            this.loadedSounds.push( fileName );

        },

        onUpdate: function( delta )
        {
            this.updateSound( delta );
        },
     
        updateSound: function( delta )
        {
            var volumeDelta = delta;

            if( this._updateMusic === true )
            {
                this._updateAmbientSound( volumeDelta );
            }
            
            //if crossfading & old audio is silenced, fade new music in
            if (this.crossFadeAudio && audio.getMusicVolume() === 0)
            {
                audio.playMusic(this.crossFadeAudio, true);
                this.fadeMusicIn();
                this.crossFadeAudio = null;
            }

            if( this._updateSounds === true )
            {
                this._updateSoundEffects( volumeDelta );
            }
        },

        _updateAmbientSound: function( volumeDelta )
        {
            if( this._settings.ambientSound && this._stopAllSound === false )
            {
                this._updateMusic = this._fadeSound( this._isFadingAmbient,
                                                    SoundTypes.getSoundType( SoundTypes.musicSound ),
                                                    volumeDelta );
            }
        },

        _updateSoundEffects: function( volumeDelta )
        {
            this._updateSounds = this._fadeSound( this._isFadingSoundEffects,
                                                SoundTypes.getSoundType( SoundTypes.soundEffect ),
                                                volumeDelta );
            if( this._updateSounds === false && this._resetFromFade === true )
            {
                audio.setEffectsVolume( 1.0 );
                this._resetFromFade = false;
            }
        },

        _fadeSound: function( fadeSound, soundType, volumeDelta )
        {
            var getterFunctionName = "get"+soundType+"Volume";
            var setterFunctionName = "set"+soundType+"Volume";
            var continueUpdating   = true;
            if( fadeSound === true )
            {

                continueUpdating = this._fadeSoundOut( getterFunctionName, setterFunctionName, soundType, volumeDelta );
            }
            else
            {
                continueUpdating = this._fadeSoundIn( getterFunctionName, setterFunctionName, soundType, volumeDelta );
            }
            return continueUpdating;
        },

        _fadeSoundOut: function( getterFunctionName, setterFunctionName, soundType, volumeDelta )
        {
            var currentVolume = audio[ getterFunctionName ].call( audio );
            if( currentVolume === 0 )
            {
                if( this._stopAllSound === true )
                {
                    audio[ "stop"+soundType ].call( audio );
                }
                else if( this._stopAllSound === false )
                {
                    audio[ "pause"+soundType ].call( audio );
                }
            }
            audio[ setterFunctionName ].call( audio, currentVolume - volumeDelta );
            return Boolean( currentVolume > 0 );
        },

        _fadeSoundIn: function( getterFunctionName, setterFunctionName, soundType, volumeDelta )
        {
            var currentVolume = audio[ getterFunctionName ].call( audio );
            if( audio.isMusicPlaying() === false )
            {
                audio[ "resume"+soundType ].call( audio );
            }
            audio[ setterFunctionName ].call( audio, currentVolume + volumeDelta );
            return Boolean( currentVolume < 1 );
        },

        onFinish: function()
        {
            this.stopSound();
        },

        stopSound: function()
        {
            audio.stopMusic( true );
            audio.stopAllEffects();
        },

        stopEffects: function()
        {
            audio.stopAllEffects();
            if( this._settings.ambientSound )
            {
                audio.playEffect( this._settings.ambientSound, true );
            }
        }

    } );
}() );
