/*global require*/
/*global cc*/
/*global GameComponentScore*/
/*global GameWidgetHud*/

require( "src/Game/Component/GameComponentScore.js" );
require( "src/Game/Widget/GameWidgetHud.js" );

var GameComponentScoreMultiplier;
( function()
{
    "use strict";

    GameComponentScoreMultiplier = GameComponentScore.extend(
    {
        _multiplier:0,
        _meterValue:0,

        /** @lends GameComponentScoreMultiplier.prototype */
        
        /**
         * A GameComponentScoreMultiplier is a subclass of GameComponentScore that can be used to set up games with
         * score multiplier. Provides functionality to manage the game score. 
         *
         * !IMPORTANT!
         *
         * This class expects that a game's settings.json file includes definitions for the following:
         *     minMultiplier
         *     maxMultiplier
         *     minMeterValue
         *     maxMeterValue
         *     meterIncreasePerCorrect
         *     meterDecreasePerWrong
         *     
         * @constructs
         */
        ctor:function( controller )
        {
            this._super( controller );

            var model = this._controller.model;
            var settings = this._controller.settings;

            this._multiplier = settings.minMultiplier;
            this._meterValue = settings.minMeterValue;

            // Initialize the model
            model.multiplier = this._multiplier;
            model.meterValue = this._meterValue;
            model.prevMultiplier = this._multiplier - 1;
        },

        /**
        * Returns Component type as a string.
        *
        * @returns {string}
        */
        getType:function()
        {
            return "GameComponentScoreMultiplier";
        },

        /**
        *  Increases the multiplier progress.
        */
        increaseMultiplierProgress:function()
        {
            var settings = this._controller.settings;
            var newMultiplier = this._multiplier;
            var newMeterValue = this._meterValue;
            var newMeterValueDelta = 0;
            var newScore = this._score;

            if( newMultiplier < settings.maxMultiplier )
            {
                newMeterValue += settings.meterIncreasePerCorrect;
            }
           
            while ( newMeterValue > settings.maxMeterValue )
            {
                if( newMultiplier < settings.maxMultiplier )
                {
                    newMultiplier+=1;
                }

                newMeterValueDelta = newMeterValue - ( settings.maxMeterValue + 1 );
                newMeterValue = settings.minMeterValue + newMeterValueDelta;
            }

            newScore += settings.pointsPerCorrect * this._multiplier;
            this._setMultiplier( newMultiplier );

            this._setMeterValue( newMeterValue );
            this.setScore( newScore );

            // cc.log("GameComponentScoreMultiplier::increaseMultiplierProgress _meterValue:" + this._meterValue);
            // cc.log("GameComponentScoreMultiplier::increaseMultiplierProgress _multiplier:" + this._multiplier);
        },

        /**
        *  Decreases the multiplier progress.
        */
        decreaseMultiplierProgress:function()
        {
            var settings = this._controller.settings;

            if ( this._meterValue <= settings.minMeterValue )
            {
                this._multiplier = Math.max( settings.minMultiplier, this._multiplier - 1 );
            }
            this._meterValue = settings.minMeterValue;
        },
        
        /**
         * Gets the value you use for "progress" metadata.
         */
        getProgressMetadata:function()
        {
            return this._meterValue - 1;
        },

        /**
         * Original code for decreasing multiplier progress. May want to keep just in case?
         */
        // decreaseMultiplierProgress:function()
        // {
        //     var settings = this._controller.settings;
        //     var newMultiplier = this._multiplier;
        //     var newMeterValue = this._meterValue;
        //     var newMeterValueDelta = 0;
        //     var newScore = this._score;

        //     if( newMultiplier > settings.minMultiplier ||
        //         newMeterValue > settings.minMeterValue )
        //     {
        //         newMeterValue -= settings.meterDecreasePerWrong;
        //     }

        //     while( newMeterValue < settings.minMeterValue )
        //     {
        //         if( newMultiplier > settings.minMultiplier )
        //         {
        //             newMultiplier-=1;
        //             this._setMultiplier( newMultiplier );
        //         }

        //         newMeterValueDelta = newMeterValue - (settings.minMeterValue - 1);
        //         newMeterValue = settings.maxMeterValue + newMeterValueDelta;
        //     }

        //     newScore -= settings.pointsPerIncorrect * this._multiplier;

        //     this._setMeterValue( newMeterValue );
        //     this.setScore( newScore );

        //     // cc.log("GameComponentScoreMultiplier::decreaseMultiplierProgress _meterValue:" + this._meterValue);
        //     // cc.log("GameComponentScoreMultiplier::decreaseMultiplierProgress _multiplier:" + this._multiplier);
        // },

        _setMultiplier:function( multiplier )
        {
            var settings = this._controller.settings;
            var model = this._controller.model;

            model.prevMultiplier = this._multiplier;
            this._multiplier = Math.min( settings.maxMultiplier, Math.max( multiplier, settings.minMultiplier ) );
            model.multiplier = this._multiplier;
        },

        _setMeterValue:function( meterValue )
        {
            var settings = this._controller.settings;
            var model = this._controller.model;

            model.prevMeterValue = this._meterValue;
            this._meterValue = Math.min( settings.maxMeterValue, Math.max( meterValue, settings.minMeterValue ) );
            model.meterValue = this._meterValue;
        },

        onEndGameplay:function()
        {  
            var settings = this._controller.settings;
            var newScore = this._score;
            var bonusScore = settings.lastTrialBonus;

            if ( !settings.fixedLastTrialBonus )
            {
                bonusScore *= this._multiplier;
            }

            newScore += bonusScore;

            this._setBonus(bonusScore);
            this.setScore( newScore );
        }

    } );
}() );