/*global cc*/
/*global require*/
/*global ll*/
/*global GameWidget*/
/*global LumosityLabel*/
/*global LumosityUtils*/
/*global localize*/



var GameWidgetHudElementType =
{
    HudElementSingleLabel:              "HudElementSingleLabel",
    HudElementLabelValue:               "HudElementLabelValue",
    HudElementLabelTimeValue:           "HudElementLabelTimeValue",
    HudElementLabelSomeOfManyValue:     "HudElementLabelSomeOfManyValue"
};



var GameWidgetHud;
(function()
{
    "use strict";

    require( "src/Game/Widget/GameWidget.js" );
    require( "src/Utility/LumosityUtils.js" );
    require( "src/View/LumosityLabel.js" );

    GameWidgetHud = GameWidget.extend(
    {
        /** @lends GameWidgetHud.prototype */

        /** Game settings */
        _settings: null,

        /** Node that contains all other hud elements. */
        _hudContainer: null,


        /** Hud components */
        _components: [],
        _componentsMap: {},


        /** The full scale of this popup widget. Can be increased to change the maximum size. */
        _fullScale: 1.0,
        /** Defaults, will be read from the flash file. */
        _font: "fonts/MuseoSans-700.ttf",
        _fontSize: 26,
        _fontColor: cc.WHITE, //cc.c3b( 4, 69, 81 )
        _spacerWidth: 3,
        _marginX: 20,
        _marginY: 20,
        _anchorX: 30,
        _hudBackgroundColor: cc.RED,
        _hudHighlightColor: cc.GREEN,
        _hudTotalWidth: 640,
        _hudTotalHeight: 60,
        _trialHudSizeFactor: 0.85,

        /**
         * Returns Widget type as a string.
         *
         * @returns {string}
         */
        getType: function()
        {
            return "GameWidgetHud";
        },


        /**
         *  A GameWidgetHud is a representation of the HUD on the screen. It
         *  can currently display 2 separate fields, out of trial #, time,
         *  and score.
         *
         *  @constructs
         *  @param {boolean} showTrials Whether or not this HUD should show trials.
         *  @param [boolean] showTime Whether or not this HUD should show time.
         *  @param {boolean} showScore Whether or not this HUD should show score.
         *  @param {boolean} onlyShowPositiveTime Whether to show zero when time is negative.
         */
        ctor: function( settings, delegate, preload )
        {
            this._super();

            this._settings = settings;
            this._delegate = delegate;

            this._fullScale = this.getScale();

            this._buttons = [];

            this._showScore = settings.showScore;
            this._showTime = settings.showTime;
            this._showTrials = settings.showTrials;
            this._onlyShowPositiveTime = settings.onlyShowPositiveTime;

            this._fontColor = cc.YELLOW;
            this._hudBackgroundColor = cc.RED;
            this._hudHighlightColor = cc.GREEN;

            // Set initial content size, just needs to be larger than 0 x 0.
            var contentSize = cc.size( this._hudTotalWidth, this._hudTotalHeight );
            this.setContentSize( contentSize );

            // Set anchor point
            this.setAnchorPoint( cc.p( 0.5, 0.5 ) );

            //by default, preload all our images on creation
            if ( ( preload === undefined || preload ) && settings !== undefined )
            {
                this.preload( settings );
            }

            // Create sub views
            this._createViews();

            // Position views
            this._positionComponents ();
        },

        onPause: function()
        {
            this._super();
            
            this.pauseAnimations();
        },

        onResume: function()
        {
            this._super();

            this.resumeAnimations();
        },

        onRestart: function()
        {
            this._super();

            this.stopAnimations();
        },



        /**
         *  Animates the huds text objects with a given sequence.
         *
         *  @param {cc.Sequence} Animation sequence.
         */
        setTextAnimation: function( sequence )
        {
            var component;
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                component = this._components [ i ];
                if ( component )
                {
                    if ( component.label )
                    {
                        component.label.runAction( sequence.clone() );
                    }

                    if ( component.value )
                    {
                        component.value.runAction( sequence.clone() );
                    }
                }
            }
        },

        /**
        * Pauses all current running animations.
        */
        pauseAnimations: function()
        {
            var component;
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                component = this._components [ i ];
                if ( component )
                {
                    if ( component.label )
                    {
                        component.label.pauseSchedulerAndActions();
                    }

                    if ( component.value )
                    {
                        component.value.pauseSchedulerAndActions();
                    }
                }
            }
        },

        /**
        * Resumes all previous running animations.
        */
        resumeAnimations: function()
        {
            var component;
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                component = this._components [ i ];
                if ( component )
                {
                    if ( component.label )
                    {
                        component.label.resumeSchedulerAndActions();
                    }

                    if ( component.value )
                    {
                        component.value.resumeSchedulerAndActions();
                    }
                }
            }
        },

        /**
        * Stops all previous running animations.
        */
        stopAnimations: function()
        {
            var component;
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                component = this._components [ i ];
                if ( component )
                {
                    if ( component.label )
                    {
                        component.label.stopAllActions();
                    }

                    if ( component.value )
                    {
                        component.value.stopAllActions();
                    }
                }
            }
        },

        /**
         * Preloads images
         *
         * @param settings Game settings that provides
         */
        preload:function( settings )
        {
        },

        /**
         * Creates views.
         *
         * @private
         */
        _createViews: function()
        {
            var screenSize = cc.Director.getInstance().getWinSize();
            var widgetContentSize = this.getContentSize();

            // Create hud container node
            this._hudContainer = cc.Node.create();
            this._hudContainer.setContentSize( cc.size( this._hudTotalWidth, this._hudTotalHeight ) );
            this._hudContainer.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this._hudContainer.setPosition( cc.p( widgetContentSize.width / 2, widgetContentSize.height / 2 ) );
            this.addChild( this._hudContainer );

            if ( this._showScore )
            {
                this._createScoreViews();
            }

            if ( this._showTime )
            {
                this._createTimeViews();
            }

            if ( this._showTrials )
            {
                this._createTrialViews();
            }

            if ( LumosityUtils.isFitTestModeOn() )
            {
                this._createFitTestViews();
            }
        },


        /**
         * Creates generic hud element.
         *
         * @param componentKey
         * @param componentType
         * @param labelString
         * @param valueString
         * @returns {{}}
         * @private
         */
        _createHudElement: function ( componentKey, componentType, labelString, valueString )
        {
            var component = {};
            component.type = componentType;

            var hudContainerSize = this._hudContainer.getContentSize();

            // Create Label
            if ( labelString && labelString.length > 0 )
            {
                component.label = LumosityLabel.create( labelString, this._font, this._fontSize );
                component.label.setColor( this._fontColor );
                component.label.setAnchorPoint( cc.p( 0, 0.5 ) );
                component.label.setHorizontalAlignment( cc.TEXT_ALIGNMENT_LEFT );
                component.label.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
                component.label.nodeName = "HUD." + componentKey + ".Label";
            }

            // Create Value
            if ( valueString && valueString.length > 0 )
            {
                component.value = LumosityLabel.create( valueString, this._font, this._fontSize );
                component.value.setColor( this._fontColor );
                component.value.setAnchorPoint( cc.p( 1, 0.5 ) );
                component.value.setHorizontalAlignment( cc.TEXT_ALIGNMENT_RIGHT );
                component.value.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
                component.value.nodeName = "HUD." + componentKey + ".Value";
            }

            // Figure out size of the background image
            var componentWidth = this._marginX;
            if ( component.label )
            {
                var componentLabelWidth = component.label.getContentSize().width;

                // reduce spacing between label and value if "TRIAL" hud is used
                if ( componentType == GameWidgetHudElementType.HudElementLabelSomeOfManyValue )
                {
                    componentLabelWidth *= this._trialHudSizeFactor;
                }

                componentWidth += componentLabelWidth + this._marginX;
            }
            if ( component.value )
            {
                componentWidth += component.value.getContentSize().width + this._marginX;
            }
            var componentSize = cc.size( componentWidth, this._hudTotalHeight );

            // Create score container
            component.container = cc.Node.create();
            component.container.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            component.container.setContentSize( componentSize );
            component.container.nodeName = "HUD." + componentKey + ".Container";

            // Create Background image
            component.backgroundSprite = LumosityUtils.createRectangularSprite( componentSize, this._hudBackgroundColor );
            component.backgroundSprite.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            component.backgroundSprite.nodeName = "HUD." + componentKey + ".Background";

            // Add Background Image
            component.container.addChild( component.backgroundSprite );
            var backgroundPosition = cc.p( componentSize.width / 2 , componentSize.height / 2 );
            component.backgroundSprite.setPosition( backgroundPosition );

            // Add Label
            if ( component.label )
            {
                component.container.addChild( component.label );
                var labelPosition = cc.p( this._marginX, this._hudTotalHeight / 2 );
                component.label.setPosition( labelPosition );
            }

            // Add Value
            if ( component.value )
            {
                component.container.addChild( component.value );
                var valuePosition = cc.p( componentSize.width - this._marginX, this._hudTotalHeight / 2 );
                component.value.setPosition( valuePosition );
            }

            // Add Container
            this._hudContainer.addChild( component.container );
            var containerPosition = cc.p( hudContainerSize.width - ( this._anchorX + componentSize.width / 2 ), hudContainerSize.height / 2 );
            component.container.setPosition( containerPosition );

            return component;
        },


        /**
         * Resizes component based on its children's size.
         *
         * @param component
         * @private
         */
        _sizeComponent: function ( component )
        {
            if ( component )
            {
                var componentWidth = 2 * this._marginX;

                // Figure out size of the background image
                if ( component.label )
                {
                    var componentLabelWidth = component.label.getContentSize().width * component.label.getScaleX();

                    // reduce spacing between label and value if "TRIAL" hud is used
                    if ( component.type == GameWidgetHudElementType.HudElementLabelSomeOfManyValue )
                    {
                        componentLabelWidth *= this._trialHudSizeFactor;
                    }

                    componentWidth += componentLabelWidth + this._marginX;
                }
                if ( component.value )
                {
                    componentWidth += ( component.value.getContentSize().width * component.value.getScaleX() ) + this._marginX;
                }

                var componentSize = cc.size( componentWidth, this._hudTotalHeight );

                // Set container new size
                component.container.setContentSize ( componentSize );

                // Resize background sprite
                lu.resizeNode ( component.backgroundSprite, componentSize );

                // Position Background Image
                var backgroundPosition = cc.p( componentSize.width / 2 , componentSize.height / 2 );
                component.backgroundSprite.setPosition( backgroundPosition );

                // Position Label
                if ( component.label )
                {
                    var labelPosition = cc.p( this._marginX, this._hudTotalHeight / 2 );
                    component.label.setPosition( labelPosition );
                }

                // Position Value
                if ( component.value )
                {
                    var valuePosition = cc.p( componentSize.width - this._marginX, this._hudTotalHeight / 2 );
                    component.value.setPosition( valuePosition );
                }
            }
        },

        /**
         * Positions components in order they appear in the list.
         * Invisible components are ignored.
         *
         * @private
         */
        _positionComponents: function ()
        {
            var hudContainerSize = this._hudContainer.getContentSize();
            var positionY = this._hudContainer.getContentSize().height / 2;
            var positionX = hudContainerSize.width - this._anchorX;
            var component;
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                component = this._components [ i ];
                if ( component && component.container && component.container.isVisible() )
                {
                    positionX -= component.container.getContentSize().width / 2;

                    component.container.setPosition( cc.p ( positionX, positionY ) );

                    positionX -= ( component.container.getContentSize().width / 2 + this._spacerWidth );
                }
            }
        },



        /**
         * Creates SCORE component & views.
         *
         * @private
         */
        _createScoreViews: function()
        {
            var componentKey = "Score";
            var labelString = localize ( "HUD_ScoreText", "SCORE" );
            var component = this._createHudElement( componentKey, GameWidgetHudElementType.HudElementLabelValue, labelString, "88888" );
            this._components.push ( component );
            this._componentsMap [ componentKey ] = component;

            // Setup just like the component's value.
            var gainedPoints = LumosityLabel.create( "+88888", this._font, this._fontSize );
            gainedPoints.setColor( cc.c3b( 255, 255, 255 ) );
            gainedPoints.setAnchorPoint( cc.p( 1, 0.5 ) );
            gainedPoints.setHorizontalAlignment( cc.TEXT_ALIGNMENT_RIGHT );
            gainedPoints.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
            component.container.addChild( gainedPoints );
            component.gainedPoints = gainedPoints;

            // Start invisible.
            gainedPoints.setOpacity( 0 );
        },

        /**
         * Creates TIME component & views.
         *
         * @private
         */
        _createTimeViews: function()
        {
            var componentKey = "Time";
            var labelString = localize ( "HUD_TimeText", "TIME" );
            var component = this._createHudElement( componentKey, GameWidgetHudElementType.HudElementLabelTimeValue, labelString, "12:12" );
            this._components.push ( component );
            this._componentsMap [ componentKey ] = component;
        },


        /**
         * Creates TRIALS component & views.
         *
         * @private
         */
        _createTrialViews: function()
        {
            var componentKey = "Trials";
            var labelString = localize ( "HUD_TrialText", "TRIALS" );
            var component = this._createHudElement( componentKey, GameWidgetHudElementType.HudElementLabelSomeOfManyValue, labelString, "10 of 10" );
            this._components.push ( component );
            this._componentsMap [ componentKey ] = component;
        },


        /**
         * Creates Fit Test component & views.
         *
         * @private
         */
        _createFitTestViews: function ()
        {
            var componentKey = "Tutorial";
            var labelString = localize ( "HUD_TutorialText", "TUTORIAL" );
            var component = this._createHudElement( componentKey, GameWidgetHudElementType.HudElementSingleLabel, labelString );
            this._components.push ( component );
            this._componentsMap [ componentKey ] = component;
        },


        /**
         * Hides all visible components.
         *
         * @private
         */
        _hideAll: function()
        {
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                this._components[ i ].container.setVisible ( false );
            }
        },


        /**
         * Internal method for setting background colors
         *
         * @param backgroundSprite
         * @param color
         * @private
         */
        _setBackgroundColor: function ( backgroundSprite, color )
        {
            if ( backgroundSprite && color )
            {
                backgroundSprite.setColor( color );

                if ( color.a )
                {
                    backgroundSprite.setOpacity( color.a );
                }
            }
        },

        /**
         * Sets background color and updates each component.
         *
         * @param {cc.Color|number} colorOrR Either the full color object, or just the red component of the color, 0->255.
         * @param {number} [g] The green component of the color, 0->255.
         * @param {number} [b] The blue component of the color, 0->255.
         * @param {number} [a] The alpha component of the color, 0->255.
         */
        setBackgroundColor: function( colorOrR, g, b, a )
        {
            if ( colorOrR.r )
            {
                this._hudBackgroundColor = colorOrR;
            }
            else if ( colorOrR.Red )
            {
                this._hudBackgroundColor = cc.c4b( colorOrR.Red, colorOrR.Green, colorOrR.Blue, colorOrR.Alpha );
            }
            else
            {
                this._hudBackgroundColor = cc.c4b( colorOrR, g, b, a );
            }

            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                this._setBackgroundColor ( this._components[ i ].backgroundSprite, this._hudBackgroundColor );
            }
        },

        /**
         * Sets background highlight color.
         *
         * @param {cc.Color|number} colorOrR Either the full color object, or just the red component of the color, 0->255.
         * @param {number} [g] The green component of the color, 0->255.
         * @param {number} [b] The blue component of the color, 0->255.
         * @param {number} [a] The alpha component of the color, 0->255.
         */
        setHighlightColor: function( colorOrR, g, b, a )
        {
            if ( colorOrR.r )
            {
                this._hudHighlightColor = colorOrR;
            }
            else if ( colorOrR.Red )
            {
                this._hudHighlightColor = cc.c4b( colorOrR.Red, colorOrR.Green, colorOrR.Blue, colorOrR.Alpha );
            }
            else
            {
                this._hudHighlightColor = cc.c4b( colorOrR, g, b, a );
            }
        },

        /**
         * Sets text color and updates each component.
         *
         * @param {cc.Color|number} colorOrR Either the full color object, or just the red component of the color, 0->255.
         * @param {number} [g] The green component of the color, 0->255.
         * @param {number} [b] The blue component of the color, 0->255.
         * @param {number} [a] The alpha component of the color, 0->255.
         */
        setTextColor: function( colorOrR, g, b, a )
        {
            if ( colorOrR.r !== undefined && colorOrR.r !== null )
            {
                this._fontColor = colorOrR;
            }
            else if ( colorOrR.Red )
            {
                this._fontColor = cc.c4b( colorOrR.Red, colorOrR.Green, colorOrR.Blue, colorOrR.Alpha );
            }
            else
            {
                this._fontColor = cc.c4b( colorOrR, g, b, a );
            }

            var component;
            var i;
            for ( i = 0; i < this._components.length; i++ )
            {
                component = this._components [ i ];
                if ( component )
                {
                    if ( component.label )
                    {
                        component.label.setColor( this._fontColor );
                        if ( this._fontColor.a )
                        {
                            component.label.setOpacity( this._fontColor.a );
                        }
                    }

                    if ( component.value )
                    {
                        component.value.setColor( this._fontColor );
                        if ( this._fontColor.a )
                        {
                            component.value.setOpacity( this._fontColor.a );
                        }
                    }
                }
            }
        },


        _getScoreComponent: function ( create )
        {
            var scoreComponent = this._componentsMap [ "Score" ];
            if ( !scoreComponent && create )
            {
                this._createScoreViews();
                scoreComponent = this._componentsMap [ "Score" ];
            }

            return scoreComponent;
        },


        _getTimeComponent: function ( create )
        {
            var timeComponent = this._componentsMap [ "Time" ];
            if ( !timeComponent && create )
            {
                this._createTimeViews();
                timeComponent = this._componentsMap [ "Time" ];
            }

            return timeComponent;
        },

        _getTrialsComponent: function ( create )
        {
            var trialsComponent = this._componentsMap [ "Trials" ];
            if ( !trialsComponent && create )
            {
                this._createTrialViews();
                trialsComponent = this._componentsMap [ "Trials" ];
            }

            return trialsComponent;
        },

        _getTutorialComponent: function ( create )
        {
            var component = this._componentsMap [ "Tutorial" ];
            if ( !component && create )
            {
                this._createFitTestViews();
                component = this._componentsMap [ "Tutorial" ];
            }

            return component;
        },



        showScore: function ( )
        {
            var scoreComponent = this._getScoreComponent( true );
            if ( scoreComponent )
            {
                scoreComponent.container.setVisible ( true );
            }
        },

        hideScore: function()
        {
            var scoreComponent = this._getScoreComponent( false );
            if ( scoreComponent )
            {
                scoreComponent.container.setVisible ( false );
            }
        },

        /**
         *  Set the current score. Will show the score formatted as an integer.
         *  Does nothing if there is no score label in this HUD.
         *
         *  @param {number} score The score to show.
         */
        setScore: function( score, resize )
        {
            var component = this._getScoreComponent( false );
            if ( component && component.value )
            {
                component.value.setString( score );

                if ( resize )
                {
                    this._sizeComponent ( component );
                    this._positionComponents();
                }
            }
        },

        setScoreLabel: function( scoreLabelText )
        {
            var scoreComponent = this._getScoreComponent( false );
            if ( scoreComponent && scoreComponent.label )
            {
                scoreComponent.label.setString( scoreLabelText );

                this._sizeComponent ( scoreComponent );
                this._positionComponents();
            }
        },


        setScoreHighlighted: function( highlight )
        {
            var scoreComponent = this._getScoreComponent( false );
            if ( scoreComponent && scoreComponent.backgroundSprite )
            {
                var newColor;
                if ( highlight )
                {
                    newColor = this._hudHighlightColor;
                }
                else
                {
                    newColor = this._hudBackgroundColor;
                }

                scoreComponent.backgroundSprite.setColor( newColor );
                if ( newColor.a )
                {
                    scoreComponent.backgroundSprite.setOpacity( newColor.a );
                }
            }
        },

        setScoreAnimated: function( gainedPoints, score )
        {
            var scoreComponent = this._getScoreComponent( false );

            if ( scoreComponent )
            {
                // Update the display string.
                scoreComponent.gainedPoints.setString( "+" + gainedPoints );

                // Move from below the score value position to the actual score value position.
                var scoreValuePos = scoreComponent.value.getPosition();
                var gainedValuePos = scoreComponent.value.getPosition();
                gainedValuePos.y -= this._hudTotalHeight;

                // Sequence to appear, hold position, then move towards score value.
                var initialPosition = cc.MoveTo.create( 0.0, gainedValuePos );
                var waitToMove = cc.DelayTime.create( 0.766666667 );
                var move = cc.MoveTo.create( 0.133333333, scoreValuePos );

                // Final action is to update the actual score value.
                var hud = this;
                var functor = { setScore: function() { hud.setScore( score ); } };
                var finishedAction = cc.CallFunc.create( functor.setScore, functor );

                var moveSequence = cc.Sequence.create( initialPosition, waitToMove, move, finishedAction );
                scoreComponent.gainedPoints.runAction( moveSequence );

                // Sequence to fade in, hold, then fade out.
                var initialOpacity = cc.FadeTo.create( 0.0, 0 );
                var fadeIn = cc.FadeTo.create( 0.166666667, 255 );
                var waitToFade = cc.DelayTime.create( 0.6 );
                var fadeOut = cc.FadeTo.create( 0.133333333, 0 );

                var fadeSequence = cc.Sequence.create( initialOpacity, fadeIn, waitToFade, fadeOut );
                scoreComponent.gainedPoints.runAction( fadeSequence );
            }
        },

        setTimeHighlighted: function( highlight )
        {
            var timeComponent = this._getTimeComponent( false );
            if ( timeComponent && timeComponent.backgroundSprite )
            {
                var newColor;
                if ( highlight )
                {
                    newColor = this._hudHighlightColor;
                }
                else
                {
                    newColor = this._hudBackgroundColor;
                }

                timeComponent.backgroundSprite.setColor( newColor );
                if ( newColor.a )
                {
                    timeComponent.backgroundSprite.setOpacity( newColor.a );
                }
            }
        },


        /**
         * Shows TRIALS component.
         */
        showTrials: function()
        {
            var component = this._getTrialsComponent( true );
            if ( component )
            {
                component.container.setVisible ( true );
                this._positionComponents();
            }
        },

        /**
         * Hides TRIALS component.
         */
        hideTrials: function()
        {
            var component = this._getTrialsComponent( false );
            if ( component )
            {
                component.container.setVisible ( false );
                this._positionComponents();
            }
        },

        /**
         * Sets current trial and max trials in the TRIALS component.
         *
         * @param trialNumber
         * @param maxTrials
         */
        setTrial:function( trialNumber, maxTrials, resize )
        {
            var component = this._getTrialsComponent( false );
            if ( component && component.value )
            {
                var string = trialNumber + " of " + maxTrials;
                var localizedString = localize( "HUD_TrialValueText", string ).format( trialNumber, maxTrials );

                component.value.setString( localizedString );

                if ( resize )
                {
                    this._sizeComponent ( component );
                    this._positionComponents();
                }
            }
        },

        setTrialsLabel: function( trialsLabelText )
        {
            var component = this._getTrialsComponent( false );
            if ( component && component.label )
            {
                component.label.setString ( trialsLabelText );

                this._sizeComponent ( component );
                this._positionComponents();
            }
        },


        /**
         * If time is negative, then just display zero (default: false)
         *
         * @param {boolean} flag - whether to only show positive numbers or to take the absolute
         *                         value for presentation
         */
        setOnlyShowPositiveTime: function( flag )
        {
            this._onlyShowPositiveTime = flag;
        },


        /**
         *  Set the time remaining. Will show the time formatted as 0:00,
         *  where the minutes place will always be shown, and the tens place
         *  will always be shown before the ones place. You should pass the
         *  time in as seconds, and let this widget format it.
         *  Does nothing if there is no time label in this HUD.
         *
         *  @param {number} timeInSeconds The time to show, in seconds. Do not pass in a formatted string.
         */
        setTimeRemaining: function( timeInSeconds, resize )
        {
            
            if ( this._onlyShowPositiveTime && timeInSeconds < 0 )
            {
                timeInSeconds = 0;
            }

            var component = this._getTimeComponent ( false );
            if ( component && component.value )
            {
                var minutes = Math.floor( timeInSeconds / 60 );
                var seconds = timeInSeconds - minutes * 60;
                var tens = Math.floor( seconds / 10 );
                seconds = seconds - tens * 10;

                component.value.setString( minutes + ":" + tens + seconds );

                if ( resize )
                {
                    this._sizeComponent ( component );
                    this._positionComponents();
                }
            }
        },


        /**
         * Sets Time Label to provided string and resizes the HUD to fit in the text.
         *
         * @param timeText
         */
        setTimeLabel: function( timeLabelText )
        {
            var component = this._getTimeComponent ( false );
            if ( component && component.label )
            {
                component.label.setString ( timeLabelText );

                this._sizeComponent ( component );
                this._positionComponents();
            }
        },

        /**
         * Shows TIME component.
         */
        showTime: function()
        {
            var component = this._getTimeComponent( true );
            if ( component )
            {
                component.container.setVisible ( true );
                this._positionComponents();
            }
        },

        /**
         * Hides TIME component.
         */
        hideTime: function()
        {
            var component = this._getTimeComponent( false );
            if ( component )
            {
                component.container.setVisible ( false );
                this._positionComponents();
            }
        },


        /**
         * Shows Fit Test components
         */
        showFitTest: function ()
        {
            var component = this._getTutorialComponent( true );
            if ( component )
            {
                component.container.setVisible ( true );
                this._positionComponents();
            }
        },


        /**
         * Enables Fit Test mode, which shows Fit Test views and hides all others.
         *
         * @param enable
         */
        enableFitTestMode: function ( enable )
        {
            ll.verbose("ENABLING FIT TEST MODE: " + enable );
            if ( enable )
            {
                this._hideAll();

                this.showFitTest();
            }
            else
            {
                this._hideAll();

                if ( this._settings.showScore )
                {
                    this.showScore();
                }

                if ( this._settings.showTime )
                {
                    this.showTime();
                }

                if ( this._settings.showTrials )
                {
                    this.showTrials();
                }
            }
        },


        /**
         * Adds custom component.
         *
         * @param componentKey - Key that identifies the component.
         * @param componentType - Component type, although it doesn't do anything currently.
         * @param labelString - String to be set, should be localized.
         * @param valueString - String to be set, should be localized.
         */
        addCustomComponent: function ( componentKey, componentType, labelString, valueString )
        {
            // Component was already added.
            if( this._componentsMap [ componentKey ] )
            {
                ll.warning( "Trying to add a hud component that already exists: " + componentKey );
                return;
            }

            var component = this._createHudElement( componentKey, componentType, labelString, valueString );
            this._components.push ( component );
            this._componentsMap [ componentKey ] = component;

            if ( component.backgroundSprite )
            {
                this._setBackgroundColor ( component.backgroundSprite, this._hudBackgroundColor );
            }
        },

        /**
         * Sets label value for a custom component.
         *
         * @param componentKey - Key that identifies the component.
         * @param labelString - String to be set, should be localized.
         * @param resize - If true, component will be resized.
         */
        setCustomComponentLabel: function ( componentKey, labelString, resize )
        {
            var component = this._componentsMap [ componentKey ];
            if ( component && component.label )
            {
                component.label.setString ( labelString );

                if ( resize )
                {
                    this._sizeComponent ( component );
                    this._positionComponents();
                }
            }
        },

        /**
         * Sets value for a custom component.
         *
         * @param componentKey - Key that identifies the component.
         * @param valueString - String to be set, should be localized.
         * @param resize - If true, component will be resized.
         */
        setCustomComponentValue: function ( componentKey, valueString, resize )
        {
            var component = this._componentsMap [ componentKey ];
            if ( component && component.value )
            {
                component.value.setString ( valueString );

                if ( resize )
                {
                    this._sizeComponent ( component );
                    this._positionComponents();
                }
            }
        },


        /**
         * Sets order for the specified component.
         *
         * @param componentKey - Key that identifies the component.
         * @param orderIndex - Index where component should be, other components will be shifted.
         */
        setComponentOrder: function ( componentKey, orderIndex )
        {
            var component = this._componentsMap [ componentKey ];
            if ( component )
            {
                this._components.remove( component );
                this._components.splice( orderIndex, 0, component );

                this._positionComponents();
            }
        }
    });
}() );
