/*global cc*/
/*global require*/
/*global lumosity*/

require( "src/Input/Interaction.js" );
/*global Interaction*/
/*global InteractionPhase*/
/*global InteractionType*/

require( "src/Input/SwipeInteractionRecognizer.js" );
/*global SwipeGestureDirection*/

require( "src/Utility/LumosityUtils.js" );
/*global lu*/

/**
 * InteractionAutomator can be used to remotely fire Interactions using simple json objects.
 * This is useful for automated testing, etc.
 *
 * api:
 * {
 *    "type" : "point" || "multiPoint" || "button" || "swipe",
 *    "phase" : "began" || "changed" || "ended" || "canceled",
 *    "data" : { ... }
 * }
 *
 * The "data" object should be populated with relevant data using the keys at the bottom
 * of Interaction.js. Some examples:
 * { "position" : { "x" : 10, "y" : 75 } }
 * { "buttonName" : "yes" }
 * { "swipeDirection" : "up" || "down" || "left" || "right" }
 *  
 * @name InteractionAutomator
 * @class
 * @readonly
 */
var InteractionAutomator;
( function()
{
    "use strict";

    InteractionAutomator = cc.Class.extend(
    {
        _game: null,
        report: null,
        timer: 0,

        /** @lends InteractionAutomator */
        
        _typeMap:
        {
            "point" : InteractionType.Point,
            "multiPoint" : InteractionType.MultiplePoints,
            "button" : InteractionType.Button,
            "swipe" : InteractionType.Swipe
        },
        
        _reverseTypeMap: {}, //dynamically populated
        
        _phaseMap:
        {
            "began" : InteractionPhase.Began,
            "changed" : InteractionPhase.Changed,
            "ended" : InteractionPhase.Ended,
            "canceled" : InteractionPhase.Canceled
        },
        
        _reversePhaseMap: {}, //dynamically populated
        
        _swipeDirectionMap:
        {
            "up" : SwipeGestureDirection.Up,
            "down" : SwipeGestureDirection.Down,
            "left" : SwipeGestureDirection.Left,
            "right" : SwipeGestureDirection.Right
        },
        
        _reverseSwipeDirectionMap: {}, //dynamically populated

        _filteredViewPropertyList:
        [

        ],

        ctor: function( game )
        {
            this._game = game;
            this.report = 
            {
                tests: [],
                status: "pass"
            };
            this._populateReverseMaps();
        },
    
        /**
         * Creates an interaction from a data blob that matches the API.
         *
         * @param {Object} apiData An object that should be in the format of the api, or weird Interactions will be returned.
         * @returns {Interaction} An Interaction object that matches the passed in data.
         */
        createInteraction: function( apiData )
        {
            var type = this._typeMap[ apiData.type ];
            if ( type === undefined )
            {
                type = InteractionType.Unknown;
            }
            
            var phase = this._phaseMap[ apiData.phase ];
            if ( phase === undefined )
            {
                phase = InteractionType.Unknown;
            }
            
            var interactionData = lu.duplicateObject( apiData.data, true );
            if ( interactionData.swipeDirection !== undefined )
            {
                interactionData.swipeDirection = this._swipeDirectionMap[ interactionData.swipeDirection ];
            }
            
            return new Interaction( type, phase, interactionData );
        },

        /**
         * Compares the current state of the game with a recorded state.
         *
         * @param {object} gameState The recorded game state.
         * @param {string} assertMessage The assert message upon missmatch.
         *
         * @return {boolean} Whether or not the comparison failed or succeeded. 
         */
        _compareGameState: function( gameState, tests )
        {
            var testInfo = null;
            var didSucceed = true;
            if( !gameState )
            {
                return;
            }

            // Compare the game model.
            testInfo = { test: "Model", assertMessage: "" };
            if( !this._compareModel( gameState.model, testInfo ) )
            {
                testInfo.status = "fail";
                tests.push( testInfo );
                didSucceed = false;
            }

            // Compare view.
            testInfo = { test: "View", assertMessage: "" };
            if( !this._compareView( gameState.view, testInfo ) )
            {
                testInfo.status = "fail";
                tests.push( testInfo );
                didSucceed = false;
            }

            return didSucceed;
        },

        /**
         * Compares the current state of the view with a recorded state.
         *
         * @param {object} view The recorded view state.
         * @param {object} testInfo Information about the test to be filled in.
         *
         * @return {boolean} Whether or not the comparison failed or succeeded. 
         */
        _compareModel: function( modelData, testInfo )
        {
            var didSucceed = true;
            var model = null;

            model = this._game.model;

            if( ( modelData.score !== model.score ) )
            {
                testInfo.assertMessage += " Values of property: score are not equal. Recorded data: " + modelData.score + " Game data: " + model.score;
                didSucceed = false; 
            }

            if( ( modelData.totalResponses !== model.totalResponses ) )
            {
                testInfo.assertMessage += " Values of property: totalResponses are not equal. Recorded data: " + modelData.totalResponses + " Game data: " + model.totalResponses;
                didSucceed = false; 
            }

            if( ( modelData.totalResponsesCorrect !== model.totalResponsesCorrect ) )
            {
                testInfo.assertMessage += " Values of property: totalResponsesCorrect are not equal. Recorded data: " + modelData.totalResponsesCorrect + " Game data: " + model.totalResponsesCorrect;
                didSucceed = false; 
            }

            if( ( modelData.multiplier !== model.multiplier ) )
            {
                testInfo.assertMessage += " Values of property: multiplier are not equal. Recorded data: " + modelData.multiplier + " Game data: " + model.multiplier;
                didSucceed = false; 
            }

            if( ( modelData.meterValue !== model.meterValue ) )
            {
                testInfo.assertMessage += " Values of property: meterValue are not equal. Recorded data: " + modelData.meterValue + " Game data: " + model.meterValue;
                didSucceed = false; 
            }

            if( ( modelData.prevMultiplier !== model.prevMultiplier ) )
            {
                testInfo.assertMessage += " Values of property: prevMultiplier are not equal. Recorded data: " + modelData.prevMultiplier + " Game data: " + model.prevMultiplier;
                didSucceed = false; 
            }

            if( ( modelData.isOnBeginCalled !== model.isOnBeginCalled ) )
            {
                testInfo.assertMessage += " Values of property: isOnBeginCalled are not equal. Recorded data: " + modelData.isOnBeginCalled + " Game data: " + model.isOnBeginCalled;
                didSucceed = false; 
            }

            if( ( modelData.isGamePaused !== model.isGamePaused ) )
            {
                testInfo.assertMessage += " Values of property: isGamePaused are not equal. Recorded data: " + modelData.isGamePaused + " Game data: " + model.isGamePaused;
                didSucceed = false; 
            }

            if( ( modelData.isGameRunning !== model.isGameRunning ) )
            {
                testInfo.assertMessage += " Values of property: isGameRunning are not equal. Recorded data: " + modelData.isGameRunning + " Game data: " + model.isGameRunning;
                didSucceed = false; 
            }

            if( ( modelData.isGameOver !== model.isGameOver ) )
            {
                testInfo.assertMessage += " Values of property: isGameOver are not equal. Recorded data: " + modelData.isGameOver + " Game data: " + model.isGameOver;
                didSucceed = false; 
            }
    
            return didSucceed;
        },

        /**
         * Compares the current state of the view with a recorded state.
         *
         * @param {object} view The recorded view state.
         * @param {object} testInfo Information about the test to be filled in.
         *
         * @return {boolean} Whether or not the comparison failed or succeeded. 
         */
        _compareView: function( view, testInfo )
        {
            var didSucceed = true;

            // Compare display state.
            if( !this._compareDisplayState( view.displayState, testInfo ) )
            {
                didSucceed = false;
            }

            // Compare how to play view.
            if( !this._compareHowToPlay( view.howToPlay, testInfo ) )
            {
                didSucceed = false;
            }

            return didSucceed;
        },

        /**
         * Compares the current state of the display state with a recorded state.
         *
         * @param {object} view The recorded display state.
         * @param {object} testInfo Information about the test to be filled in.
         *
         * @return {boolean} Whether or not the comparison failed or succeeded. 
         */
        _compareDisplayState: function( displayState, testInfo )
        {
            var didSucceed = true;
            var view = null;

            view = this._game.view;

            if( ( displayState.prevDisplayState !== view._prevDisplayState ) )
            {
                testInfo.assertMessage += " Values of property: prevDisplayState are not equal. Recorded data: " + displayState.prevDisplayState + " Game data: " + view._prevDisplayState;
                didSucceed = false; 
            }

            if( ( displayState.curDisplayState !== view._curDisplayState ) )
            {
                testInfo.assertMessage += " Values of property: curDisplayState are not equal. Recorded data: " + displayState.curDisplayState + " Game data: " + view._curDisplayState;
                didSucceed = false; 
            }
    
            return didSucceed;
        },

        /**
         * Compares the current state of the how to play view with a recorded state.
         *
         * @param {object} view The recorded how to play view state.
         * @param {object} testInfo Information about the test to be filled in.
         *
         * @return {boolean} Whether or not the comparison failed or succeeded. 
         */
        _compareHowToPlay: function( howToPlay, testInfo )
        {
            var didSucceed = true;
            var view = null;
            var howToPlayWidget = null;

            view = this._game.view;
            howToPlayWidget = view.getHowToPlayWidget();

            if( ( !howToPlayWidget && howToPlay ) )
            {
                // Doesn't use how to play widget.
                return true;
            }

            if( ( !howToPlayWidget && howToPlay ) )
            {
                testInfo.assertMessage += " Recorded data contains info but widget doesn't exist.";
                return false; 
            }

            if( howToPlayWidget.getPageCount() !== howToPlay.pageCount )
            {
                testInfo.assertMessage += " Values of property: pageCount are not equal. Recorded data: " + howToPlay.pageCount + " Game data: " + howToPlayWidget.getPageCount();
                didSucceed = false;
            }
            
            if( ( howToPlayWidget.getCurrentPageNumber() !== howToPlay.currentPageNumber ) )
            {
                testInfo.assertMessage += " Values of property: currentPageNumber are not equal. Recorded data: " + howToPlay.currentPageNumber + " Game data: " + howToPlayWidget.getCurrentPageNumber();
                didSucceed = false;
            }
            
            if( ( howToPlayWidget.getPageText() !== howToPlay.pageText ) )
            {
                testInfo.assertMessage += " Values of property: pageText are not equal. Recorded data: " + howToPlay.pageText + " Game data: " + howToPlayWidget.getPageText();
                didSucceed = false;
            }

            return didSucceed;
        },
        
        /**
         * Begins playing back a script of interactions. In order for these to work, update() must
         * be called on every frame update. A script is simply an ordered array of objects, with an "interaction"
         * key and a "time" key. As soon as the time is exceeded, that interaction will be fired.
         *
         * @param {Array} interactionScript An array of objects, with "interaction" and "time" keys.
         * @param {Object} listener An object with an "interaction_received( interaction )" function. This will be fired as per the script.
         */
        replayScript: function( interactionScript, listener )
        {
            this._script = interactionScript;
            // NAF: Script start time can later be used to play data back at a given time in the data.
            // lumosity.game.frametime;
            this._scriptStartTime = 0;
            this._scriptIndex = 0;
            this._scriptListener = listener;
        },
        
        /**
         * Must be called every frame if you want interaction scripts to work.
         * Can fire any number of interactions.
         *
         * @param {number} delta The delta since last frame.
         */
        onUpdate: function(  delta )
        {
            this.timer += delta;

            if ( !this._script || !this._scriptListener )
            {
                return;
            }

            //the time relative to the running script
            var scriptTime = this.timer - this._scriptStartTime;
            
            // lumosity.game.frametime can't be used here. If the game hitches frametime might still
            // be updated which will cause the logic below to spam input. Input may be blocked by the game
            // after the first button press causeing the rest of the input that was spammed to not occur.
            // This will throw the data out of sync.
            //the time relative to the running script
            //var scriptTime = lumosity.game.frametime - this._scriptStartTime;
            
            //go through all the interactions until one is beyond the current time, and fire them all
            var interactionIndex;
            var interactionData;
            var interaction;
            var tests = null;
            var testInfo = null;
            var i;
            for ( interactionIndex = this._scriptIndex; interactionIndex < this._script.length; interactionIndex++ )
            {
                interactionData = this._script[ interactionIndex ];
                
                //if this is before the current time, fire it
                if ( interactionData.time <= scriptTime )
                {

                    tests = [];

                    //create an actual Interaction object
                    interaction = this.createInteraction( interactionData.interaction );
                    
                    // Interaction was recorded before being performed, so we need to compare before we
                    // fire off the interaction.
                    if( !this._compareGameState( interactionData.gameState, tests ) )
                    {
                        // Evaluate the state of the game based on recorded data.
                        testInfo = null;
                        for( i = 0; i < tests.length; i++ )
                        {
                            // Set interaction type and time for each test recorded.
                            testInfo = tests[ i ];
                            testInfo.interactionType = this._typeNameForInteractionType( interaction.type );
                            testInfo.time = interactionData.time;
                            testInfo.locale = lumosity.game.locale;
                        }
                    }

                    //fire the Interaction
                    this._scriptListener.interaction_received( interaction );
                    
                    //advance the index
                    this._scriptIndex = interactionIndex + 1;
                }
                //otherwise, since this is an ordered list, break out
                else
                {
                    break;
                }

                if( tests.length !== 0 )
                {
                    this.report.tests = this.report.tests.concat( tests );
                }
            }
            
            //clear out the script if we've finished it
            if ( this._scriptIndex >= this._script.length )
            {
                this._script = null;
            }
        },
        
        /**
         * Will begin recording interactions by initializing a new script. You must
         * still tell the automator when an individual interaction happens.
         *
         * @param {boolean} [useRelativeTime=false] If true, then the script will be started from whenever this function was called. If false, then the script time is the same as the frame time.
         */
        beginRecordingScript: function( useRelativeTime )
        {
            // Recorded start time will always be 0 for now.
            // This is what it looked like = this._recordingStartTime = useRelativeTime ? lumosity.game.frametime : 0;
            this._recordingScript = [];
            this._recordingStartTime = 0;
        },
        
        /**
         * Finishes recording a script by returning the final script and clearing the recording
         * variables.
         *
         * @returns {Object} A valid interaction script that can be replayed.
         */
        finishRecordingScript: function()
        {
            var script = this._recordingScript;
            
            if ( script )
            {
                this._lastRecordedScript = script;
            }
            this._recordingScript = null;
            
            return script;
        },
        
        /**
         * Records an interaction into the recorded interaction script.
         * This can be used to later replay a script.
         *
         * @param {Interaction} interaction The interaction that should be recorded.
         */
        recordInteraction: function( interaction )
        {
            if ( !this._recordingScript )
            {
                return;
            }
            
            var scriptTime = lumosity.game.frametime - this._recordingStartTime;
            
            this._recordingScript.push(
            {
                "interaction" : this._getApiDataFromInteraction( interaction ),
                "gameState" : this._getApiDataFromGame(),
                "time" : scriptTime
            } );
        },
        
        /**
         * Gives the last Interaction script that was recorded, if any.
         *
         * @returns {Object} A valid interaction script that can be replayed, or null.
         */
        getLastRecordedScript: function()
        {
            return this._lastRecordedScript || this._recordingScript || null;
        },
        
        _getApiDataFromInteraction: function( interaction )
        {
            var interactionData = lu.duplicateObject( interaction.getData(), true );
            if ( interactionData.swipeDirection !== undefined )
            {
                interactionData.swipeDirection = this._reverseSwipeDirectionMap[ interactionData.swipeDirection ];
            }
            
            var apiData =
            {
                type : this._reverseTypeMap[ interaction.getType() ],
                phase : this._reversePhaseMap[ interaction.getPhase() ],
                data : interactionData
            };
            
            return apiData;
        },

        _getApiDataFromGame: function()
        {
            var apiData = null;
      
            apiData = 
            {
                model: this._getModelDataFromGame(),
                view: this._getViewDataFromGame()
            };

            return lu.duplicateObject( apiData, true );
        },

        _getModelDataFromGame: function()
        {
            var model = this._game.model;
            var modelData = null;

            modelData =
            {
                score: model.score,
                totalResponses: model.totalResponses,
                totalResponsesCorrect: model.totalResponsesCorrect,
                consecutiveResponsesCorrect: model.consecutiveResponsesCorrect,
                multiplier: model.multiplier,
                meterValue: model.meterValue,
                prevMultiplier: model.prevMultiplier,
                isOnBeginCalled: model.isOnBeginCalled,
                isGamePaused: model.isGamePaused,
                isGameRunning: model.isGameRunning,
                isGameOver: model.isGameOver
            };

            return modelData;
        },

        _getViewDataFromGame: function()
        {
            var viewData = null;

            viewData =
            {
                howToPlay: this._getHowToPlayDataFromGame(),
                displayState: this._getDisplayStateFromGame()
            };

            return viewData;
        },

        _getHowToPlayDataFromGame: function()
        {
            var view = this._game.view;
            var howToPlayWidget = view.getHowToPlayWidget();
            var howToPlayData = null;

            if( !howToPlayWidget )
            {
                return null;
            }

            howToPlayData =
            {
                pageCount: howToPlayWidget.getPageCount(),
                currentPageNumber: howToPlayWidget.getCurrentPageNumber(),
                pageText: howToPlayWidget.getPageText()
            };

            return howToPlayData;
        },

        _getDisplayStateFromGame: function()
        {
            var view = this._game.view;
            var displayStateData = null;

            // Data to be stored about the display state of the game.
            displayStateData = 
            {
                prevDisplayState: view._prevDisplayState,
                curDisplayState: view._curDisplayState
            };

            return displayStateData;
        },

        /**
         *  Compares the recorded data and the live data.
         *
         *  @param {Object} x The recorded data.
         *  @param {Object} y The live data.
         *  @param {object} testInfo Information about the test to be filled in.
         *
         *  @returns {Boolean} Whether or not the two models are equal.
         **/
        // _compare: function( x, y, testInfo )
        // {
        //     var p;

        //     // Both are equal.
        //     if( x === y ) 
        //     {
        //         return true;
        //     }

        //     // If they are not strictly equal, they both need to be Objects
        //     if( !( x instanceof Object ) || !( y instanceof Object ) ) 
        //     {
        //         testInfo.assertMessage += "One of the compared types is not an object. Data structure missmatch.";
        //         return false;
        //     }

        //     for( p in x ) 
        //     {
        //         if( !x.hasOwnProperty( p ) ) 
        //         {
        //             continue;
        //         }
               
        //         if( !y.hasOwnProperty( p ) ) 
        //         {
        //             testInfo.assertMessage += "Recorded data contains a property that is not found in game data: " + p;
        //             return false;
        //         }
               
        //         if( x[ p ] === y[ p ] ) 
        //         {
        //             //assertMessage += "Values of property: " + p + " are equal. Recorded data: " + x[ p ] + " Game data: " + y[ p ] + "\n";
        //             continue;
        //         }
               
        //        // Numbers, Strings, Functions, Booleans must be strictly equal
        //         if( typeof( x[ p ] ) !== "object" )
        //         {
        //             testInfo.assertMessage += "Values of property: " + p + " are not equal. Recorded data: " + x[ p ] + " Game data: " + y[ p ];
        //             return false;
        //         }

        //         this._compare( x[ p ],  y[ p ] );
        //     }

        //     return true;
        // },

        _typeNameForInteractionType: function( interactionType )
        {
            if( interactionType === InteractionType.Point )
            {
                return "point";
            }
            
            if( interactionType === InteractionType.MultiplePoints )
            {
                return "multiPoint";
            }
            
            if( interactionType === InteractionType.Button )
            {
                return "button";
            }
            
            if( interactionType === InteractionType.Swipe )
            {
                return "swipe";
            }

            return "unknown";
        },

        /**
         *  Filters tests based on key so that multiple entries are not listed.
         *
         *  @param {string} key The key to filter the data by.
         *
         *  @returns {Array} The filtered array.
         **/
        filterTests: function( key ) 
        {
            var object = {};
            var tests = this.report.tests;
            var filteredTests = null;
            var i;
            var data;

            for( i = 0; i < tests.length; i++ )
            {
                if( object[ tests[ i ][ key ] ] )
                {
                    // Continue if object data already exists.
                    continue;
                }

                // Store the test information.
                object[ tests[ i ][ key ] ] = tests[ i ];
            }

            filteredTests = [];
            for( data in object )
            {
                filteredTests.push( object[ data ] );
            }

            return filteredTests;
        },
        
        _populateReverseMaps:function()
        {
            var typeKey;
            for ( typeKey in this._typeMap )
            {
                this._reverseTypeMap[ this._typeMap[ typeKey ] ] = typeKey;
            }
            
            var phaseKey;
            for ( phaseKey in this._phaseMap )
            {
                this._reversePhaseMap[ this._phaseMap[ phaseKey ] ] = phaseKey;
            }
            
            var swipeKey;
            for ( swipeKey in this._swipeDirectionMap )
            {
                this._reverseSwipeDirectionMap[ this._swipeDirectionMap[ swipeKey ] ] = swipeKey;
            }
        }
     });
}() );

InteractionAutomator.screenshotCount = 1;
