/*global require
global cc
global lumosity*/

/**
 * LumosityGlobals is loaded first thing in main.js so that everyone everywhere can use it.
 * You can either use LumosityGlobals.function(), or use its alias lg.function(), or many functions
 * have individual aliases to make them easier to use.
 *
 * @namespace LumosityGlobals
 */
var LumosityGlobals = 
{
    /** @lends LumosityGlobals */
    
    /** Stores all the strings that we found, a concatenation of sharedStrings and strings jsons. */
    localizedStrings: {},
    
    /** Stores the current build number for Cocos2dGames. This is read from a file on disk each time. */
    buildNumber: 0,

    /** Stores the current asset number for Cocos2dGames. This is read from a file on disk each time. */
    assetNumber: 0,  

    /** Files that have already been required have their filename as a key, then true is set. */
    requiredFiles: {},
    
    /**
     * Initializes the LumosityGlobals. This involves loading the localization
     * files and storing them locally.
     */
    initialize:function()
    {
        this.localizedStrings = {};
        this.requiredFiles = {};
        
        //for html5
        if ( !cc.FileUtils.getInstance().isFileExist && lumosity._fileExists )
        {
            cc.FileUtils.getInstance().isFileExist = lumosity._fileExists;
        }
                
        //initialize the localized strings
        var stringKey;
        
        var sharedStrings = lumosity.io_read( this.getLocalizedPath( "sharedStrings.json" ) );
        if ( sharedStrings )
        {
            sharedStrings = JSON.parse( sharedStrings );
            
            for ( stringKey in sharedStrings )
            {
                this.localizedStrings[ stringKey ] = sharedStrings[ stringKey ];
            }
        }
        
        var gameStrings = lumosity.io_read( this.getLocalizedPath( "strings.json" ) );
        if ( gameStrings )
        {
            gameStrings = JSON.parse( gameStrings );
            
            for ( stringKey in gameStrings )
            {
                this.localizedStrings[ stringKey ] = gameStrings[ stringKey ];
            }
        }
        
        //first, use the npm version, if it exists, for the build number
        var packageString = lumosity.io_read( "package.json" );
        if ( packageString )
        {
            try
            {
                var packageJSON = JSON.parse( packageString );
                this.buildNumber = packageJSON.version;
            }
            catch ( err )
            {
                cc.log( "ERROR loading package.json file, it's not valid JSON." );
            }
        }

        var assetString = lumosity.io_read( this.getLocalizedPath( "versionInfo.json" ) );
        if ( assetString )
        {
            try
            {
                var assetJSON = JSON.parse( assetString );
                this.assetNumber = assetJSON.version
            }
            catch ( err )
            {
                cc.log( "ERROR loading versionInfo.json file, it's not valid JSON." );
            }
        }
        
        //if there was no package.json to get a build number from, use the buildData.json
        //see if there's a build data json. if there is, get data from it
        if ( !this.buildNumber )
        {
            var buildDataString = lumosity.io_read( "buildData.json" );
            if ( buildDataString )
            {
                //since this JSON is auto-generated, it's important that we don't throw for exceptions parsing it
                try
                {
                    var buildData = JSON.parse( buildDataString );
            
                    this.buildNumber = buildData.version;
                }
                catch ( err2 )
                {
                    cc.log( "ERROR loading buildData.json file, it's not valid JSON." );
                }
            }
        }

        // sometimes we get game results as a string, but we need it to be an object
        if ( lumosity.game.params.last_results && typeof( lumosity.game.params.last_results ) === "string" )
        {
            lumosity.game.params.last_results = JSON.parse ( lumosity.game.params.last_results );
        }
    },
    
    /**
     * Localizes a string by taking a key. You may also provide a default in case you aren't sure
     * the key exists. Will use the strings lookup dictionary to decide what to give. If you do
     * not provide a default and the key also does not exist in the dictionary, the key is returned.
     *
     * @param {string} key The key to look up.
     * @param {string} [defaultValue=key] The default to return in case the key does not exist.
     * @returns {string} A localized string.
     */
    localize:function( key, defaultValue )
    {
        var string = this.localizedStrings[ key ];
        if ( string )
        {
            return string;
        }
        
        return ( defaultValue || key );
    },
   
    /**
     * @param {array} keyArray If the key is not found, or is empty or an empty string, then omit it.  
     *                         This allows optional lines.  Example:  For all games: ["skipTutorial_line1", "skipTutorial_line2"].
     *                         Starting with Highway Hazards, and hopefully future games, only "skipTutorial_line1" is optimal, 
     *                         because it saves time for each translator to translate a full sentence string rather than 
     *                         two separate half sentence strings which may be grammatically unrelated in their language.
     *                         Example:  Pretend the right side is a foreign language with a different grammatical order.
     *                              "Watch signs\nto dodge collisions." --> "To dodge, signs\nshould be observed." 
     *                         is faster to translate than:
     *                              "Watch signs"          --> "To dodge, signs"
     *                              "to dodge collisions." --> "should be observed."
     *                         In the latter case, the individual string translations are nonsense, because 
     *                         "dodge" does not appear in "Watch signs" and "observed" does not appear in "to dodge collisions".
     * @returns {array} Each localized string is split into lines by Unix line-ending "\n", becomming a separate array element.
     *                         Example: ["Watch signs\nto dodge collisions.", undefined] --> ["Watch signs", "to dodge collisions"]
     *                         Example @see ../Game/Base/GameView.js
     */
    localizeLines:function( keyArray )
    {
        var lines = [];
        var key;
        var text;
        for (var k = 0; k < keyArray.length; k++) {
            key = keyArray[k];
            text = this.localizedStrings[ key ];
            if (text) {
                lines = lines.concat( text.split("\n") );
            }
        }
        return lines;
    },

    /**
     * This is a smart require. It stores off any required files, and will only
     * require them if they haven't already been required.
     *
     * @param {string} file The file to require. Uses typical search paths.
     */
    require:function( file )
    {
        if ( this.requiredFiles[ file ] )
        {
            return;
        }
        
        this.requiredFiles[ file ] = true;
        this.cocosRequire( file );
    },
    
    /**
     * The given path is localized using the locale folder and the passed in locale.
     * Use this for any assets that you have localized versions of. Note that this
     * function will check to see if multiple files exist. It can potentially do this
     * twice.
     *
     * For example, if you passed in "strings.json", the order of preference is:
     * 1) "locale/ja/strings.json" (ja being whatever the current locale is)
     * 2) "locale/en/strings.json"
     * 3) "strings.json"
     *
     * @param {string} path The unlocalized path to this resource.
     * @returns {string} The localized path to this resource.
     */
    getLocalizedPath:function( path )
    {
        var pathTest = "locale/" + lumosity.game.locale + "/" + path;
                
        if ( this.fileExists( pathTest ) )
        {
            return pathTest;
        }
        
        pathTest = "locale/en/" + path;
        
        if ( this.fileExists( pathTest ) )
        {
            return pathTest;
        }
        
        return path;
    },
    
    fileExists:function( path )
    {
        return cc.FileUtils.getInstance().isFileExist( cc.FileUtils.getInstance().fullPathForFilename( path ) );
    }
};

var lg = LumosityGlobals;
var localize = LumosityGlobals.localize.bind( LumosityGlobals );
var localizedPath = LumosityGlobals.getLocalizedPath.bind( LumosityGlobals );

LumosityGlobals.cocosRequire = require;
this[ "require" ] = LumosityGlobals.require.bind( LumosityGlobals );
