var cc = cc;

var MultiDimensionArray = cc.Class.extend(
{
    /** @lends MultiDimensionArray.prototype */
    
    array:[],
    dimensionMultiples:[],
    dimensionLengths:[],
    totalLength:0,
    
    /**
     * A MultiDimensionArray is a simple wrapper of a 1D array which allows you to treat it
     * like a multi-dimension array of any size. This is mainly because JS multi-dimension arrays
     * can be annoying in that they can be staggered and inconsistent. This class enforces typical
     * behavior you would expect from an n*m... array. You can also pass in each dimension as a
     * separate argument if you prefer that.
     *
     * @param {Array} dimensionsArrayOrFirstArgument An array where each element specifies how many items in a certain dimension to create. The length of this array determines how many dimensions you have.
     * @constructs
     */
    ctor:function( dimensionsArrayOrFirstArgument )
    {
        if ( dimensionsArrayOrFirstArgument instanceof Array )
        {
            this.dimensionLengths = dimensionsArrayOrFirstArgument;
        }
        else
        {
            this.dimensionLengths = Array.prototype.slice.call( arguments );
        }
        
        this.totalLength = 1;
        var dimensionIndex;
        for ( dimensionIndex = 0; dimensionIndex < this.dimensionLengths.length; dimensionIndex++ )
        {
            this.totalLength *= this.dimensionLengths[ dimensionIndex ];
        }
        this.array = new Array( this.totalLength );
        
        this.dimensionMultiples = new Array( this.dimensionLengths.length );
        this._findDimensionMultiples();
    },
    
    /**
     * Returns the array element at the passed multi-dimensional indices.
     * This function can take any number of parameters, but will return undefined
     * if you provide it with a number that doesn't match the dimension size
     * of this array.
     *
     * @param index0 The index of the 0th dimension.
     * @param index1 The index of the 1st dimension.
     * @param ... The index of the nth dimension.
     * @returns The value in the array at the expected indices, or undefined if there is an improper index count provided.
     */
    get:function()
    {
        //cc.log("in me get");
        var actualIndex = this._getActualIndex.apply( this, arguments );
        return this.array[ actualIndex ];
    },
    
    /**
     * Sets a value on this array. The value is expected as the first parameter, and
     * the indices are expected as parameters following that. The number of indices
     * must match the dimension size of this array, or nothing will be set.
     *
     * @param value Any value that you want to store in this array.
     * @param index0 The index of the 0th dimension.
     * @param index1 The index of the 1st dimension.
     * @param ... The index of the nth dimension.
     */
    set:function( value )
    {
        //do not include the first parameter, since that's the value and not an index
        //since arguments isn't actually an array, we much do this hoodoo magic to slice it like it is one
        var actualIndex = this._getActualIndex.apply( this, Array.prototype.slice.call( arguments, 1 ) );
        if ( actualIndex >= 0 )
        {
            this.array[ actualIndex ] = value;
        }
    },
    
    /**
     * Returns whether or not the element at the passed multi-dimensional indices exists.
     * It is considered to not exist if it is either undefined or null. A value of 0 DOES
     * exist, however. This function can take any number of parameters, but will always
     * return false if you provide it with a number that doesn't match the dimension size.
     *
     * @returns {boolean} Whether or not the obejct at the passed indices exists.
     */
    exists:function()
    {
        var actualIndex = this._getActualIndex.apply( this, arguments );
        return ( this.array[ actualIndex ] !== undefined && this.array[ actualIndex ] !== null );
    },
    
    /**
     * Fills the entire array with the passed value.
     *
     * @param value Any value you want to fill the entire array with.
     */
    fill:function( value )
    {
        var index;
        for ( index = 0; index < this.array.length; index++ )
        {
            this.array[ index ] = value;
        }
    },
    
    /**
     * Returns the actual length of the array that backs the multi-dimension array.
     *
     * @returns {number} The total number of elements in this multi-dimension array.
     */
    getActualLength:function()
    {
        return this.array.length;
    },
    
    /**
     * Returns the number of dimensions that this multi-dimension array has.
     *
     * @returns {number} The dimension count.
     */
    getDimensionCount:function()
    {
        return this.dimensionLengths.length;
    },
    
    /**
     * Returns the length of a given dimension in this array.
     *
     * @param {number} dimension The dimension to get the length of.
     * @returns {number} The length of the passed dimension. Returns -1 if that dimension is too large or small.
     */
    getLength:function( dimension )
    {
        if ( dimension < 0 || dimension >= this.dimensionLengths.length )
        {
            return -1;
        }
        
        return this.dimensionLengths[ dimension ];
    },
    
    /**
     *  Convenience function to get the "width" of the array. That means the length of the first dimension.
     *
     *  @returns {number} The length of the first (index 0) dimension.
     */
    getWidth:function()
    {
        return this.getLength( 0 );
    },
    
    /**
     *  Convenience function to get the "height" of the array. That means the length of the second dimension.
     *
     *  @returns {number} The length of the second (index 1) dimension.
     */
    getHeight:function()
    {
        return this.getLength( 1 );
    },
    
    /**
     * Retroactively sets the length of a given dimension to a new size.
     * Note that this is a slow function since it will have to create a brand
     * new array and copy everything over.
     *
     * @param {number} dimension The dimension index to change the length of.
     * @param {number} newSize The new length of that dimension.
     */
    setLength:function( dimension, newSize )
    {
        var targetLength = 1;
        var dimensionIndex;
        for ( dimensionIndex = 0; dimensionIndex < this.dimensionLengths.length; dimensionIndex++ )
        {
            if ( dimension === dimensionIndex )
            {
                targetLength *= newSize;
            }
            else
            {
                targetLength *= this.dimensionLengths[ dimensionIndex ];
            }
        }
        
        var newArray = new Array( targetLength );
        var minLength = targetLength < this.array.length ? targetLength : this.array.length;
        var arrayIndex;
        for ( arrayIndex = 0; arrayIndex < minLength; arrayIndex++ )
        {
            newArray[ arrayIndex ] = this.array[ arrayIndex ];
        }
        this.array = newArray;
        this.dimensionMultiples[ dimension ] = newSize;
        this.dimensionLengths[ dimension ] = newSize;
        
        this._findDimensionMultiples();
    },
    
    /**
     * Returns the actual index for the passed parameter indices. This function accepts
     * any number of parameters, but will return -1 if the passed number of parameters does
     * not match the number of dimensions in this array.
     *
     * @returns {number} The real index in our array, or -1 if the parameter count is invalid.
     */
    _getActualIndex:function()
    {
        if ( arguments.length !== this.dimensionLengths.length )
        {
            cc.log( "[MultiDimensionArray] WARNING: Improper number of index params passed to _getActualIndex(). Expected " + this.dimensionLengths.length + ", but got " + arguments.length + ". Returning -1 for index." );
            return -1;
        }
        
        //cc.log("index index index");
        var actualIndex = 0;
        
        var dimensionIndex;
        for ( dimensionIndex = 0; dimensionIndex < this.dimensionLengths.length; dimensionIndex++ )
        {
            actualIndex += this.dimensionMultiples[ dimensionIndex ] * arguments[ dimensionIndex ];
        }
        
        return actualIndex;
    },
    
    _findDimensionMultiples:function()
    {
        this.dimensionMultiples[ this.dimensionLengths.length - 1 ] = 1;
        
        var dimensionIndex;
        for ( dimensionIndex = this.dimensionLengths.length - 2; dimensionIndex >= 0; dimensionIndex-- )
        {
            this.dimensionMultiples[ dimensionIndex ] = this.dimensionLengths[ dimensionIndex + 1 ] * this.dimensionMultiples[ dimensionIndex + 1 ];
        }
    }
});
