/*global require*/
/*global cc*/
/*global LumosityButton*/
/*global LumosityButtonState*/
/*global ll*/
/*global LumosityUtils*/

require( "src/View/LumosityButton.js" );
require( "src/Utility/LumosityLogger.js");
require( "src/Utility/LumosityUtils.js" );

var LumosityButtonLayer;
( function()
{
    "use strict";
    LumosityButtonLayer = cc.Layer.extend(
    {
        /** All the buttons that exist in this layer. They can be anywhere in the child hierarchy. */
        _buttons: null,

        /** All buttons that are currently being touched, by touch ID. */
        _touchedButtons: null,

        /**
         *  The LumosityButtonLayer can be used in conjunction with one or more LumosityButton's
         *  in order to register button clicks and whatever. This is used to replace CCMenu since
         *  that only supports buttonReleased events. To use this, create this layer, then add it
         *  as a child to a view, and finally create one or more buttons and add them as a child to this.
         *
         *  @constructs
         */
        ctor: function()
        {
            this._super();
            this.init();
            
            this._buttons = [];
            this._touchedButtons = [];

            this.setTouchEnabled( true );
        },

        /**
         * Adds a button to be tested for touches.
         * ( The button must be in a child hierarchy inorder to receive touches. )
         * 
         * @param {LumosityButton} button
         */
        addButton: function( button )
        {
            if( this._buttons.contains( button ) )
            {
                ll.warning("LumosityButtonLayer", "addButton", "button already added. It can't be added again.");
                return;
            }

            this._buttons.push( button );
        },

        /**
         * Adds collection of buttons to the layer.
         *
         * @param buttons
         */
        addButtons: function( buttons )
        {
            if ( buttons && buttons.length > 0 )
            {
                var i;
                for ( i = 0; i < buttons.length; i++ )
                {
                    this.addButton( buttons[ i ] );
                }
            }
        },

        /**
         * Retrieves a button by button name.
         *
         * @param name
         *
         * @return {LumosityButton} button
         */
        getButton: function( name )
        {
            var i;
            var button = null;

            for( i = 0; i < this._buttons.length; i++ )
            {
                button = this._buttons[ i ];
                if( button.nodeName === name )
                {
                    return button;
                }
            }

            return null;
        },

        /**
         * Removes a button from touch testing.
         * 
         * @param  {LumosityButton} button
         */
        removeButton: function( button )
        {
            if( !this._buttons.contains( button ) )
            {
                //ll.warning("LumosityButtonLayer", "removeButton", "button not found.");
                return;
            }

            this._buttons.remove( button );
        },

        /**
         * Removes collection of buttons from layer.
         * 
         */
        removeButtons: function( buttons )
        {
            if ( buttons && buttons.length > 0 )
            {
                var i;
                for ( i = 0; i < buttons.length; i++ )
                {
                    this.removeButton( buttons[ i ] );
                }
            }
        },

        /**
         * Removes all the buttons saved to this object.
         * 
         */
        removeAllButtons: function()
        {
            this._buttons = [];
        },

        /**
         * Resets all radio buttons to unpressed state.
         */
        clearRadioButtonPresses:function()
        {
            var i = 0;
            for( i = 0; i < this._buttons.length; i++ )
            {
                if( this._buttons[i].isRadioButton )
                {
                    this._buttons[i].handleIsPressed( false );
                }
            }
        },

        /**
         * Cocos2dX onTouchesBegan overridden implementation.
         * 
         * @param  {array} touches
         */
        onTouchesBegan: function( touches )
        {
            var i;
            var touch = null;
            var touchedButton = false;
            var alreadyHandledRadioButton = false;
            var radioButtonIndex = 0;

            // Sort touches numerically ascending based on touch id because touches
            // don't always come in in order.
            touches.sort( this._numericallyAscendingSort );

            for( i = 0; i < touches.length; i++ )
            {
                touch = touches[ i ];
                if( this._isExclusiveTouchBlocked( touch ) )
                {
                    break;
                }

                touchedButton = this._handleTouchBegan( touch, !alreadyHandledRadioButton );

                if( touchedButton !== null && touchedButton.isRadioButton )
                {
                    // only handle first radio button press
                    alreadyHandledRadioButton = true;

                    // clear presses for all other radio buttons
                    for ( radioButtonIndex = 0; radioButtonIndex < this._buttons.length; radioButtonIndex++ )
                    {
                        if( this._buttons[radioButtonIndex] !== touchedButton && this._buttons[radioButtonIndex].isRadioButton )
                        {
                            this._buttons[radioButtonIndex].handleIsPressed( false );
                        }
                    }
                    break;
                }
            }
        },

        /**
         * Cocos2dX onTouchBegan overridden implementation.
         * 
         * @param  {cc.Touch} touch
         */
        onTouchBegan: function( touch )
        {
            if( this._isExclusiveTouchBlocked( touch ) )
            {
                return;
            }

            this._handleTouchBegan( touch );
        },

        /**
         * Handles touch began for a single touch.
         * TODO - this should be able to handle multiple button touches with a single touch ID.
         * but that would require a big refactor which is not worth it.
         * 
         * @param  {cc.Touch} touch
         */
        _handleTouchBegan: function( touch, handleRadioButtons )
        {
            var touchedButton = this.getTouchBeganButton( touch, handleRadioButtons );
            var touchID = -1;
            if( touchedButton === null )
            {
                return null;
            }

            touchID = touch.getId();
            this._touchedButtons[ touchID ] = touchedButton;
            touchedButton.handleIsPressed( true );

            return touchedButton;
        },

        /**
         * Returns the button that was touched.
         * 
         * @param  {cc.Touch} touch
         */
        getTouchBeganButton: function( touch, handleRadioButtons )
        {
            var touchedButtons = this._findTouchedButtons( touch );
            var touchID = -1;

            if( touchedButtons.length <= 0 )
            {
                return null;
            }
            
            var touchedButton = touchedButtons[ 0 ];

            if( !handleRadioButtons && touchedButton.isRadioButton )
            {
                return null;
            }

            touchID = touch.getId();
            if( touchID < 0 )
            {
                return null;
            }

            return touchedButton;
        },

        /**
         * Cocos2dX onTouchesMoved overridden implementation.
         * 
         * @param  {array} touches
         */
        onTouchesMoved: function( touches )
        {
            var i;
            var touch = null;

            for( i = 0; i < touches.length; i++ )
            {
                touch = touches[ i ];
                this._handleTouchMoved( touch );
            }
        },

        /**
         * Cocos2dX onTouchMoved overridden implementation.
         * 
         * @param  {cc.Touch} touch
         */
        onTouchMoved: function( touch )
        {
            this._handleTouchMoved( touch );
        },

        /**
         * Handles touch moved for a single touch.
         * TODO, this should handle multiple touched buttons
         * 
         * @param  {cc.Touch} touch
         */
        _handleTouchMoved: function( touch )
        {
            var touchedButton = this.getTouchMovedButton( touch );
            if( touchedButton === null )
            {
                return null;
            }

            // cancel touches if we've moved off a normal button or if we've moved onto a different drag-begins-touch button
            if( !touchedButton || !touchedButton.dragBeginsTouch || ( touchedButton !== this._touchedButtons[touch.getId()] ) )
            {
                this._handleTouchCancelled( touch );
            }

            // handle a new touch if we've moved onto a new unpressed drag-begins-touch button
            var dragBeginningTouch = touchedButton && touchedButton.dragBeginsTouch && ( touchedButton.currentState() === LumosityButtonState.ButtonStateNormal );
            if( dragBeginningTouch )
            {
                this._handleTouchBegan( touch );
            }
        },

        /**
         * Returns the button that was touched.
         * 
         * @param  {cc.Touch} touch
         */
        getTouchMovedButton: function( touch )
        {
            var touchedButton = null;

            // Check if the button is still being touched.
            var touchedButtons = this._findTouchedButtons( touch );
            touchedButton = touchedButtons.length > 0 ? touchedButtons[ 0 ] : null;

            if( touchedButton && !touchedButton.dragBeginsTouch )
            {
                return null;
            }

            return touchedButton;
        },

        /**
         * Cocos2dX onTouchesEnded overridden implementation.
         * 
         * @param  {array} touches
         */
        onTouchesEnded: function( touches )
        {
            var i;
            var touch = null;

            for( i = 0; i < touches.length; i++ )
            {
                touch = touches[ i ];
                this._handleTouchEnded( touch );
            }
        },

        /**
         * Cocos2dX onTouchMoved overridden implementation.
         * 
         * @param  {cc.Touch} touch
         */
        onTouchEnded: function( touch )
        {
            this._handleTouchEnded( touch );
        },

        /**
         * Handles touch ended for a single touch.
         * 
         * @param  {cc.Touch} touch
         */
        _handleTouchEnded: function( touch )
        {
            var touchedButton = this.getTouchEndedButton( touch );
            var touchID = -1;
            if( touchedButton === null )
            {
                return null;
            }

            touchID = touch.getId();
            this._touchedButtons[ touchID ] = null;
            touchedButton.handleIsPressed( false );

            return touchedButton;
        },

        /**
         * Returns the button that was touched.
         * 
         * @param  {cc.Touch} touch
         */
        getTouchEndedButton: function( touch )
        {
            var touchedButton = null;
            var touchID = -1;

            touchID = touch.getId();
            if( touchID < 0 )
            {
                return null;
            }

            touchedButton = this._touchedButtons[ touchID ];
            if( !touchedButton )
            {
                return null;
            }

            if( touchedButton.isRadioButton )
            {
                // radio buttons are only deselected when another is selected
                return null;
            }

            return touchedButton;
        },

        /**
         * Cocos2dX onTouchesCancelled overridden implementation.
         * 
         * @param  {array} touches
         */
        onTouchesCancelled: function( touches )
        {
            var i;
            var touch = null;

            for( i = 0; i < touches.length; i++ )
            {
                touch = touches[ i ];
                this._handleTouchCancelled( touch );
            }
        },

        /**
         * Cocos2dX onTouchCancelled overridden implementation.
         * 
         * @param  {cc.Touch} touch
         */
        onTouchCancelled: function( touch )
        {
            this._handleTouchCancelled( touch );
        },

        /**
         * Handles touch cancelled for a single touch.
         * 
         * @param  {cc.Touch} touch
         */
        _handleTouchCancelled: function( touch )
        {
            var storedTouchedButton = this.getTouchCancelledButton( touch );
            var touchID = -1;
            var willHandleCallbacks;
            if( storedTouchedButton === null )
            {
                return;
            }

            touchID = touch.getId();
            this._touchedButtons[ touchID ] = null;

            // Do not brodcast callbacks when touch is cancelled.
            willHandleCallbacks = storedTouchedButton.willHandleCallbacks;
            storedTouchedButton.willHandleCallbacks = false;
            storedTouchedButton.handleIsPressed( false, true );
            storedTouchedButton.willHandleCallbacks = willHandleCallbacks;

            return storedTouchedButton;
        },

        /**
         * Returns the button that was touched.
         * 
         * @param  {cc.Touch} touch
         */
        getTouchCancelledButton: function( touch )
        {
            var storedTouchedButton = null;
            var touchID = -1;

            touchID = touch.getId();
            if( touchID < 0 )
            {
                return null;
            }

            storedTouchedButton = this._touchedButtons[ touchID ];
            if( !storedTouchedButton )
            {
                return null;
            }

            if( storedTouchedButton.isRadioButton )
            {
                // radio buttons are only deselected when another is selected
                return null;
            }

            return storedTouchedButton;
        },

        /**
         * Returns the button being touched.
         * 
         * @param  {cc.Touch} touch
         * @return {LumosityButton}
         */
        _findTouchedButtons: function( touch )
        {
            var i;
            var location = touch.getLocation();
            var button = null;
            var touchedButtons = [];

            for( i = 0; i < this._buttons.length; i++ )
            {
                button = this._buttons[ i ];
                if( button.currentState() !== LumosityButtonState.ButtonStateDisabled && button.containsWorldPoint( location ) )
                {
                    this._addButtonSortedByZ( touchedButtons, button );
                    //ll.verbose( "Touched button: " + button.nodeName + " and its Z is " + this._findButtonZ( button ) );
                }
            }

            //ll.verbose( "Touched " + touchedButtons.length + " buttons." );

            return touchedButtons;
        },
        
        //add a button to a list of other buttons, where the list is sorted by z, and greater z is first
        _addButtonSortedByZ: function( buttons, button )
        {
            var buttonZ = this._findButtonZ( button );
            
            var buttonIndex;
            for ( buttonIndex = 0; buttonIndex < buttons.length; buttonIndex++ )
            {
                var z = this._findButtonZ( buttons[ buttonIndex ] );
                
                if ( buttonZ > z )
                {
                    buttons.splice( buttonIndex, 0, button );
                    return;
                }
            }
            
            buttons.push( button );
        },
        
        _findButtonZ: function( button )
        {
            var z = button.getZOrder();
            
            var parent = button.getParent();
            while ( parent )
            {
                z += parent.getZOrder();
                //ll.verbose( "button parent " + parent.nodeName + " z is " + parent.getZOrder() );
                parent = parent.getParent();
            }
            
            return z;
        },

        /**
         * Returns whether or not the touch will be blocked by exclusivity.
         * 
         * @param  {cc.Touch}  touch
         * @return {Boolean}
         */
        _isExclusiveTouchBlocked: function( touch )
        {
            var i;
            var isExclusiveTouchBlocked = false;
            var touchID = -1;
            var touchedButton = null;


            touchID = touch.getId();
            for( i = 0; i < ( touchID + 1 ); i++ )
            {
                touchedButton = this._touchedButtons[ i ];
                if( touchedButton && ( touchedButton.isExclusiveTouch === true ) )
                {
                    // radio buttons are unaffected by exclusivity, since one can be already be selected when another is pressed
                    if( !touchedButton.isRadioButton )
                    {
                        isExclusiveTouchBlocked = true;
                        break;
                    }
                }
            }

            return isExclusiveTouchBlocked;
        },

        /**
         * Sorts an array of touches numerically ascending by touchId.
         * 
         * @param  {cc.Touch} touch1
         * @param  {cc.Touch} touch2
         * 
         * @return {int}
         */
        _numericallyAscendingSort: function( touch1, touch2 )
        {
            var touchID1 = touch1.getId();
            var touchID2 = touch2.getId();

            if( touchID1 < touchID2 )
            {
                return -1;
            }

            if( touchID1 > touchID2 )
            {
                return 1;
            }

            return 0;
        }

    } );
}() );
