/*global require*/
/*global cc*/
/*global LumosityUtils*/

require( "src/Utility/LumosityUtils.js" );

var LumosityInverseNode;
(function()
{
    "use strict";

    LumosityInverseNode = cc.Node.extend(
    {
        /** @lends LumosityInverseNode.prototype */

        /**
         * A LumosityInverseNode is a node that calculates and applies the inverse scale
         * of all of its parents to itself. This creates a node who is not affected by its
         * parents scale but can be scaled directly.
         *
         * Disclaimer: Scaling the parent after the this node has been added will affect the node.
         *
         * For crisp text, divide compensating scale by global content scale.
         * Cocos2D engine compensates text and textures already by a global content scale.
         * GAME-3050 Test case:  2015-03 Trouble Brewing CocosBuilder 320x480.  iPad Air 1536x2048.  Jon Edwards expects text fits screen.  Got text x4 at over 4096 pixels.
         * GAME-3518 Test case:  2015-05-26 Splitting Seeds CocosBuilder 800x1200.  Optimus Fuel phone 320x480.  Breanna expects crisp text.
         */
        onEnter: function()
        {
            var hierarchyScale = cc.PointZero();
            var newScale = cc.p( 1, 1 );
            this._super();

            hierarchyScale = LumosityUtils.getNodeHierarchicalScale( this );
            newScale = cc.p( this.getScaleX() * hierarchyScale.x, this.getScaleY() * hierarchyScale.y );
            
            var contentScale = cc.Director.getInstance().getContentScaleFactor();
            newScale.x /= contentScale;
            newScale.y /= contentScale;
            // XXX TEST blurry
            // newScale.x *= 0.125, newScale.y *= 0.125;
            // ll.log("newScale " + newScale.x + ", " + newScale.y + " contentscale " + contentScale);
            
            this.setScaleX( newScale.x );
            this.setScaleY( newScale.y );
        },

        onExit: function()
        {
            this._super();

            this.setScaleX( 1 );
            this.setScaleY( 1 );
        },

        setScale: function( scale, scaleY )
        {
            var inverseScale = cc.PointZero();
            var tempScaleX = scale;
            var tempScaleY = (scaleY || scaleY === 0) ? scaleY : scale;

            inverseScale = LumosityUtils.getInverseScale( cc.p( tempScaleX, tempScaleY ) );
            this._super( inverseScale );
        },

        setScaleX: function( newScaleX )
        {
            var inverseScale = cc.PointZero();

            inverseScale = LumosityUtils.getInverseScale( cc.p( newScaleX, 1 ) );
            this._super( inverseScale.x );
        },

        setScaleY: function( newScaleY )
        {
            var inverseScale = cc.PointZero();

            inverseScale = LumosityUtils.getInverseScale( cc.p( 1, newScaleY ) );
            this._super( inverseScale.y );

            //cc.log( this.nodeName + " LumosityLabel::setScaleY inverseScale: " + JSON.stringify( inverseScale ) );
        }
    } );
}() );
