/*global require*/
/*global cc*/
/*global LumosityLabel*/
/*global LumosityUtils*/

require( "src/View/LumosityButton.js" );
require( "src/Utility/LumosityUtils.js" );

var LumosityJoystick;
( function()
{
    "use strict";
    LumosityJoystick = cc.Node.extend(
        {
            _knob:null,

            _radius:45,

            _inputAxis:cc.POINT_ZERO,

            _isPressed:false,

            /**
             * LumosityJoystick constructor.
             *
             * @param {cc.Node} knob    Node manipulated by the player (usually sprite or button)
             * @param {number}  radius  Radius of knob's contstrained movement
             */
            ctor:function( knob, radius )
            {
                this._super();
                this.init();

                this._knob = knob;
                this.addChild( this._knob );

                this._radius = radius || this._radius;
            },

            /**
             * Allows the joystick to update its state based on its pressed or unpressed status.
             *
             * @param  {boolean} pressed        Whether or not the joystick is currently being pressed.
             */
            handleIsPressed:function( pressed )
            {
                if ( pressed && !this._isPressed )
                {
                }
                else if ( !pressed && this._isPressed )
                {
                    // no longer pressed, so unschedule input axis callback and reset input position
                    this._setInputAxis( cc.POINT_ZERO );
                    this._sendInputAxisCallback();
                }
                this._isPressed = pressed;
            },

            /**
             * Allows the joystick to update its new position after the player has moved his touch.
             *
             * @param  {cc.p} deltaPosWorld     The player's change in touch position.
             */
            handleTouchMoved:function( deltaPosWorld )
            {
                var buttonPosWorld = cc.pAdd( this.convertToWorldSpace( cc.POINT_ZERO ), deltaPosWorld );
                var buttonPosNode = this.convertToNodeSpace( buttonPosWorld );
                var inputAxis = cc.pMult( buttonPosNode, 1 / this._radius );
                this._setInputAxis( inputAxis );
                this._sendInputAxisCallback();
            },

            /**
             * Determines whether the worldPoint is within the bounds of the joystick's knob.
             *
             * @param  {cc.p} worldPoint
             * @return {Boolean} is worldPoint within the bounds of the joystick's knob.
             */
            containsWorldPoint:function( worldPoint )
            {
                if( this._knob )
                {
                    var localPoint = this._knob.convertToNodeSpace( worldPoint );
                    var contentSize = this._knob.getContentSize();

                    localPoint = cc.p( Math.floor( localPoint.x ), Math.floor( localPoint.y ) );

                    return ( ( ( localPoint.x >= 0 ) && ( localPoint.x <= contentSize.width ) ) &&
                        ( ( localPoint.y >= 0 ) && ( localPoint.y <= contentSize.height ) ) );
                }
            },

            /**
             * Sets the target for joystick callbacks.
             *
             * @param {object}      Object that will receive joystick callbacks.
             */
            setTarget: function( target )
            {
                this._target = target;
            },

            /**
             * Sets the callback for when the joystick's input axis changes.
             *
             * @param {function} inputAxisCallback      Callback function to be called on target when input axis changes.
             */
            setInputAxisCallback: function( inputAxisCallback )
            {
                this._inputAxisCallback = inputAxisCallback;
            },

            /**
             * Sets current input axis value.
             *
             * @param {cc.p} inputAxis      Current input values, from [-1, 1] in x- and y-dimensions.
             * @private
             */
            _setInputAxis:function( inputAxis )
            {
                // normalize input axes that are too long
                var inputAxisLen = cc.pLength( inputAxis );
                if( inputAxisLen > 1 )
                {
                    inputAxis = cc.pNormalize( inputAxis );
                }

                this._inputAxis = inputAxis;

                // update knob's position to match input access
                var knobPos = cc.pMult( inputAxis, this._radius );
                this._knob.setPosition( knobPos );
            },

            /**
             * Calls input axis callback when axis changes.
             *
             * @private
             */
            _sendInputAxisCallback: function()
            {
                if ( this._inputAxisCallback != null && this._target !== null )
                {
                    this._inputAxisCallback.call( this._target, this._inputAxis );
                }
            }

        } );
}() );