/*global cc*/
/*global require*/
/*global lumosity*/
/*global audio*/

require( "src/Utility/LumosityUtils.js" );
/*global LumosityUtils*/

require( "src/Animation/LayoutLoader.js" );
/*global LayoutLoader*/

require( "src/Animation/AnimationChanges.js" );
/*global AnimationChanges*/

require( "src/Animation/AlphaAnimator.js" );
/*global AlphaAnimator*/

require( "src/Animation/FramePositionAnimator.js" );
/*global FramePositionAnimator*/

require( "src/Animation/PositionAnimator.js" );
/*global PositionAnimator*/

require( "src/Animation/RotationAnimator.js" );
/*global RotationAnimator*/

require( "src/Animation/ScaleAnimator.js" );
/*global ScaleAnimator*/

require( "src/Animation/SkewAnimator.js" );
/*global SkewAnimator*/

require( "src/Animation/VisibleAnimator.js" );
/*global VisibleAnimator*/

//turn this on if you want a lot of verbose logging telling you what's going on
var __animationDebugLogging = false;
//if this is set, then we'll only print out info for a specific animation
var __animationDebugWatchAnimation = "";
//if this is set, then we'll only print out info for a specific node
var __animationDebugWatchNode = "";
//this is how quickly the animations play back, slow it down to find changes
//the higher the number, the SLOWER it is, since it multiplies the duration
var __animationDebugDurationMultiple = 1.0;

/**
 *  Uses the watch values to decide whether or not to log for a particular animation and node.
 *
 * @name AnimationController#__animationShouldDebugLog
 * @function
 * @param {string} string The message to log.
 * @returns {boolean} True if we should log this message.
 */
var __animationShouldDebugLog = function( animName, nodeName )
{
    return ( __animationDebugLogging && ( !__animationDebugWatchAnimation || __animationDebugWatchAnimation === animName ) &&
                                        ( !__animationDebugWatchNode || __animationDebugWatchNode === nodeName ) );
};

/**
 *  Log a message about animation, but only if the debug log flag is set to true.
 *
 * @name AnimationController#__animationDebugLog
 * @function
 * @param {string} string The message to log.
 */
var __animationDebugLog = function( string, animName, nodeName )
{
    if ( __animationShouldDebugLog( animName, nodeName ) )
    {
        cc.log( "[AnimationController] " + string );
    }
};

/**
 *  Gets a time value as a nice to read string in the 00:00.00 format.
 *
 * @name AnimationController#__animationDebugTime
 * @function
 * @param {number} time The time in seconds.
 * @returns {string} A string representation of the time.
 */
var __animationDebugTime = function( time )
{
    if ( __animationDebugLogging )
    {
        var minutes = 0;
        var seconds = time;
        if ( Math.abs( time ) >= 60 )
        {
            minutes = Math.floor( time / 60 );
            seconds = time - minutes * 60;
        }
    
        var str = "";
        if ( minutes < 10 )
        {
            str += "0";
        }
        str += minutes + ":";
        if ( seconds < 10 )
        {
            str += "0";
        }
        str += seconds;
    
        return str;
    }
    return "";
};

/**
 *  Logs a specific change in a node's property. Calls the same should log etc functions so behaves the same as regular debug logging.
 *
 * @name AnimationController#__animationLogNodeProperty
 * @function
 * @param {string} animationName The animation name.
 * @param {string} nodeName The node name.
 * @param {string} propertyKey The property that's being modified.
 * @param {number} actionTime The time at which this will happen.
 * @param {number} [actionValue] The value that's being set (there may be no value).
 * @param {number} [actionDuration] The duration to set the value (for animations).
 */
var __animationLogNodeProperty = function( animationName, nodeName, propertyKey, actionTime, actionValue, actionDuration )
{
    if ( __animationShouldDebugLog( animationName, nodeName ) )
    {
        var message = animationName + "/" + nodeName + " <" + propertyKey + ">: at " + __animationDebugTime( actionTime ) + ", will ";
        message += ( actionDuration > 0 ) ? "animate " : "set ";
        message += propertyKey + ( actionValue !== undefined ? " to " + JSON.stringify( actionValue ) : "" );
        message += ( actionDuration > 0 ) ? " for " + actionDuration + " seconds." : ".";
        
        __animationDebugLog( message, animationName, nodeName );
    }
};

var AnimationController = cc.Class.extend(
{
    /** @lends AnimationController.prototype */
    
    /** The data for this animation, will be loaded from one or more JSON files. */
    animationData: {},
    
    /** The object that we will look for nodes to animate. */
    view: null,
    
    /** The object that we will call Cocos runAction() calls on. */
    actionTarget: null,
    
    /** An array of targets that we will try to call cue functions on. */
    cueTargets: [],
    
    /** The global scale to apply to this animation, will automatically be chosen based upon the screenSize that was passed into the ctor. */
    globalScale: null,
    
    /** The global size change to apply to this animation, will automatically be chosen based upon the screenSize that we passed into the ctor. */
    globalSizeChange: null,
    
    /**
     *  An AnimationController is a class that can load up a JSON file of animation data and then animate multiple
     *  views in accordance with this data. Views can be cc.Node or any subclass of it. Its animations are done using
     *  cc.Action calls, which will be called on its actionTarget. Any cues can be set up so that they are automatically
     *  called on any number of objects.
     *
     * @constructs
     * @param {string} fileLocation The location on the disk for the animation data.
     * @param {Object} view The object that contains all views to be animated. Can be any kind of object.
     * @param {cc.Node} [actionTarget=view] The object that we will call runAction on. Must be a cc.Node or subclass. If undefined, will use the view.
     * @param {Array|Object} [cueTargetOrTargets=actionTarget] The object or objects to call cues functions on. If undefined, will use the actionTarget.
     * @param {cc.Point} [sourceSize=640x960] The size of the source stage that was used to create the animation data.
     * @param {cc.Point} [viewSize=view.getContentSize(), actionTarget.getContentSize()] The size of the root view that we animate on. Will be compared to sourceSize for auto-sizing.
     */
    ctor:function( fileLocation, view, actionTarget, cueTargetOrTargets, sourceSize, viewSize )
    {
        this.animationData = {};
        
        this.view = view;
        this.actionTarget = actionTarget || view;
        this.cueTargets = cueTargetOrTargets || actionTarget;
        sourceSize = sourceSize || cc.size( 640, 960 );
        
        if ( !viewSize )
        {
            if ( this.view.getContentSize )
            {
                viewSize = this.view.getContentSize();
            }
            else if ( this.actionTarget.getContentSize )
            {
                viewSize = this.actionTarget.getContentSize();
            }
        }
        
        //find our global scale
        if ( viewSize )
        {
            this.globalScale = cc.p( viewSize.width / sourceSize.width, viewSize.height / sourceSize.height );
            this.globalSizeChange = cc.p( viewSize.width - sourceSize.width, viewSize.height - sourceSize.height );
        }
        else
        {
            this.globalScale = cc.p( 1.0, 1.0 );
            this.globalSizeChange = cc.p( 0.0, 0.0 );
        }
        
        //make sure we have an array of cue targets
        if ( !( this.cueTargets instanceof Array ) )
        {
            this.cueTargets = [ this.cueTargets ];
        }
        
        //if we have an array of file locations, concat them all.
        if ( fileLocation instanceof Array )
        {
            var fileLocationIndex;
            for ( fileLocationIndex = 0; fileLocationIndex < fileLocation.length; fileLocationIndex++ )
            {
                this.concatAnimationFile( fileLocation[ fileLocationIndex ] );
            }
        }
        else
        {
            this.concatAnimationFile( fileLocation );
        }
        
        this._propertyAnimators =
        {
            "alpha" : new AlphaAnimator(),
            "frame" : new FramePositionAnimator( this.globalSizeChange, this.globalScale ),
            "origin" : new PositionAnimator( this.globalSizeChange, this.globalScale ),
            "rotation" : new RotationAnimator(),
            "size" : new ScaleAnimator( this.globalScale ),
            "skew" : new SkewAnimator(),
            "visible" : new VisibleAnimator()
        };
    },

    /**
     * Will return the nodes that are affected by a particular animation sequence.
     * 
     * @param  {string} animation_name Expected to be a well-formed string.
     * @return {array} Node objects that are used in animation.
     */
    getAffectedNodes:function( animation_name )
    {
        "use strict";
        var key = "",
            keys = [],
            nodes = this.animationData[ animation_name ].nodes;

        for ( key in nodes )
        {
            if( nodes.hasOwnProperty( key ) )
            {
                keys.push( key );
            }
        }
        return keys;
    },

    /**
     * Returns the duration of the passed animation.
     * 
     * @param  {string} animation_name Expected to be a well-formed string.
     * @return {number} A float representing how long this animation lasts. 
     */
    getDuration:function( animation_name )
    {
        "use strict";
        return this.animationData[ animation_name ].duration * __animationDebugDurationMultiple;
    },

    /**
     * Attempts to gather the keyframe info for a sequence of animation on a particular node.
     * 
     * @param  {string} animation_name Name of the animation sequenceAction.
     * @param  {cc.Node} node Should be a node in the hierarchy.
     * @return {array} Returns an array of keyframes.
     */
    getKeyframes:function( animation_name, node )
    {
        "use strict";
        return this.animationData[ animation_name ].nodes[ node ].keyframes;
    },
    
    /**
     *  Appends one animation file onto what is already loaded into this animation. If nothing
     *  is loaded, simply writes the entire file. This function will only copy over root keys,
     *  and will replace anything that already exists. For example, if the first file you
     *  load has an animation "beginGame" and the second one does as well, then the first file's
     *  animation will be replaced by the second's. However, if the first has an animation
     *  "popupShow" and the second does not have it, then you will still use the first.
     *
     *  @param {string} fileLocation The location of the animation file to load.
     */
    concatAnimationFile:function( fileLocation )
    {
        var jsonDataString = lumosity.io_read( fileLocation );
        var jsonData = JSON.parse( jsonDataString, null );
        
        var key;
        for ( key in jsonData )
        {
            this.animationData[ key ] = jsonData[ key ];
        }
    },
    
    /**
     *  Plays an animation. Multiple animations can play concurrently, but they will do wonky things
     *  if they both affect the same nodes. You can do what you want, though, at your own risk. :-)
     *  You may even play the same animation twice, and they will play concurrently. In the future,
     *  we will provide support to decide what you want to do in this case (replace the old anim,
     *  ignore the new anim, etc.)
     *
     *  @param {string} name The name of the animation to play. Should correspond with an animation in the file.
     *  @param {function} [onDoneFunction] The function to call when the animation is finished.
     *  @param {Object} [onDoneTarget] The object to call the function on upon finish. This is a required param if you include a function.
     */
    play:function( name, onDoneFunction, onDoneTarget )
    {
        __animationDebugLog( "*********Playing animation: \"" + name + "\".*********", name );
        
        var animation = this.animationData[ name ];
        
        if ( !animation )
        {
            __animationDebugLog( "!!! Unable to play animation: \"" + name + "\" Does not exist in animation data. !!!", name );
            
            if ( onDoneFunction )
            {
                var target = ( onDoneTarget ? onDoneTarget : this.actionTarget );
                if ( target && target[ onDoneFunction ] )
                {
                    onDoneFunction.call( target );
                }
            }
            
            //also call the automatically finished function if it exists
            var cueTargetIndex;
            for ( cueTargetIndex = 0; cueTargetIndex < this.cueTargets.length; cueTargetIndex++ )
            {
                var cueTarget = this.cueTargets[ cueTargetIndex ];
                if ( cueTarget[ "animationFinished_" + name ] )
                {
                    cueTarget[ "animationFinished_" + name  ].call( cueTarget );
                }
            }
            
            return;
        }
                
        var nodeName;
        for ( nodeName in animation.nodes )
        {
            var node = this.view[ nodeName ];
            if ( node )
            {
                //if the node is actually an array, then loop through it and apply the transformation to all objects within the array.
                if ( node instanceof Array )
                {
                    var nodeIndex;
                    for ( nodeIndex = 0; nodeIndex < node.length; nodeIndex++ )
                    {
                        this._animateNode( node[ nodeIndex ], animation.nodes[ nodeName ], name, nodeName );
                    }
                }
                //otherwise, just animate the one node
                else
                {
                    this._animateNode( node, animation.nodes[ nodeName ], name, nodeName );
                }
            }
        }

        var soundId;
        for ( soundId in animation.audio ) 
        {
            var intervals = animation.audio[ soundId ];
            if( intervals.length )
            {
                var i;
                for( i = 0; i < intervals.length; i++ )
                {
                    __animationDebugLog( name + " <sounds>: at " + __animationDebugTime( intervals[ i ] ) + ", will play sound \"" + soundId + "\".", name );
                    
                    this._playSound( soundId, intervals[ i ] );
                }
            }
            else
            {
                __animationDebugLog( name + " <sounds>: playing sound " + soundId + " now.", name );
                
                this._playSound( soundId, intervals );
            }
        }
        
        this._setUpCues( animation.cues, name );
        this._setUpCues( animation.otherAnimationPlays, name );
        
        //register the onDoneFunction if it was passed in
        if ( onDoneFunction )
        {
            __animationDebugLog( name + " <onFinish>: at " + ( animation.duration * __animationDebugDurationMultiple ) + ", will call the onFinish function.", name );
            
            var doneAction = cc.CallFunc.create( onDoneFunction, onDoneTarget ? onDoneTarget : this.actionTarget );
            var animDurationInterval = cc.DelayTime.create( animation.duration * __animationDebugDurationMultiple );
            this.actionTarget.runAction( cc.Sequence.create( animDurationInterval, doneAction ) );
        }
        
        //also register the automatically finished function if it exists
        var cueTargetIndex2;
        for ( cueTargetIndex2 = 0; cueTargetIndex2 < this.cueTargets.length; cueTargetIndex2++ )
        {
            var cueTarget2 = this.cueTargets[ cueTargetIndex2 ];
            if ( cueTarget2[ "animationFinished_" + name ] )
            {
                __animationDebugLog( name + " <onFinish>: at " + ( animation.duration * __animationDebugDurationMultiple ) + ", will call the animationFinished_" + name + " function.", name );
                
                var doneAction2 = cc.CallFunc.create( cueTarget2[ "animationFinished_" + name  ], cueTarget2 );
                var animDurationInterval2 = cc.DelayTime.create( animation.duration * __animationDebugDurationMultiple );
                this.actionTarget.runAction( cc.Sequence.create( animDurationInterval2, doneAction2 ) );
            }
        }
    },
    
    /**
     *  Stops an animation from playing. Warning: because actions are applied to individual nodes,
     *  if you call this function then it will remove all actions from a give node. That means if you
     *  are playing multiple animations simultaneously on one node, or if you have a custom action on
     *  it, that action will be stopped as well. Also note that this will still call all actions on
     *  its action target, which generally means sounds, cues, and animationFinished callbacks.
     *
     *  @param {string} animationName The animation to stop.
     */
    stop:function( animationName )
    {
        //loop through all nodes in this animation and remove all their actions.
        var animation = this.animationData[ animationName ];
        
        if ( animation )
        {
            var nodeName;
            for ( nodeName in animation.nodes )
            {
                var node = this.view[ nodeName ];
                if ( node )
                {
                    node.stopAllActions();
                }
            }
        }
    },
    
    /**
     *  Stops all animations on all nodes. Warning: this will stop every action on every
     *  single node this animation uses, not necessarily the ones that are playing or that
     *  the animation controller is responsible for. So if you add custom actions, beware
     *  this may remove them. You may also have this stop all actions on the actionTarget,
     *  which usually consists of sounds, cues, and animationFinished callbacks.
     *
     *  @param {boolean} [stopActionTarget=false] If true, we will stop actions on the actionTarget as well.
     */
    stopAll:function( stopActionTarget )
    {
        this._callFunctionOnAllNodes( function( node ) { node.stopAllActions(); } );
        if ( this.actionTarget && stopActionTarget )
        {
            this.actionTarget.resumeSchedulerAndActions();
            this.actionTarget.stopAllActions();
        }
    },
    
    /**
     *  Pauses all actions in progress on all nodes that are affected by this AnimationController.
     *  Note that this may also pause any other actions you may have added to each node.
     *  You may also have this pause all actions on the actionTarget,
     *  which usually consists of sounds, cues, and animationFinished callbacks.
     *
     *  @param {boolean} [pauseActionTarget=false] If true, we will pause actions on the actionTarget as well.
     */
    pause:function( pauseActionTarget )
    {
        this._callFunctionOnAllNodes( function( node ) { node.pauseSchedulerAndActions(); } );
        if ( this.actionTarget && pauseActionTarget )
        {
            this.actionTarget.pauseSchedulerAndActions();
        }
    },
    
    /**
     *  Unpauses all actions that were paused. This may resume any actions you added on a node that
     *  you manually paused.
     *  You may also have this unpause all actions on the actionTarget,
     *  which usually consists of sounds, cues, and animationFinished callbacks.
     *
     *  @param {boolean} [unpauseActionTarget=false] If true, we will unpause actions on the actionTarget as well.
     */
    resume:function( unpauseActionTarget )
    {
        this._callFunctionOnAllNodes( function( node ) { node.resumeSchedulerAndActions(); } );
        if ( this.actionTarget && unpauseActionTarget )
        {
            this.actionTarget.resumeSchedulerAndActions();
        }
    },
    
    /**
     *  Adds an object that will receive cue events. The object can implement any
     *  cue functions that it wants, anything it doesn't implement will simply be
     *  ignored. All cue functions start with "cue_" and end with the cue name.
     *
     *  @param {Object} target The target for cue events.  
     */
    addCueTarget:function( target )
    {
        this.removeCueTarget( target );
        this.cueTargets.push( target );
    },
    
    /**
     *  Removes an object from the cue list so that it no longer receives cue events.
     *
     *  @param {Object} target The object to remove from the cue list.
     */
    removeCueTarget:function( target )
    {
        var cueIndex;
        for ( cueIndex = this.cueTargets.length; cueIndex >= 0; cueIndex-- )
        {
            if ( this.cueTargets[ cueIndex ] === target )
            {
                this.splice( cueIndex, 1 );
                return;
            }
        }
    },
    
    /**
     *  You can call this on a given node in order to ignore changes from the animation. Use
     *  The AnimationChanges enum to choose which to ignore. You can pass in an array in order
     *  to ignore multiple changes per node.
     *
     *  You should probably use the AnimationChanges class instead if you're calling this externally,
     *  since you can pass a node into that rather than using the nodeName.
     *
     *  @param {string} nodeName The node name to ignore those changes on.
     *  @param {Array|AnimationChanges} changes The changes to ignore.
     */
    setNodeIgnoredChanges:function( nodeName, changes )
    {
        if ( !this.view[ nodeName ] )
        {
            return;
        }
        
        AnimationChanges.setNodeIgnoredChanges( this.view[ nodeName ], changes );
    },
    
    /**
     * Returns whether or not a given node ignores an animation change. Easier to probably just
     * use the AnimationChanges class if you're going to call this externally.
     *
     *  @param {string} nodeName The node name to check for the ignore on.
     *  @param {AnimationChanges} change The change to look for.
     */
    nodeIgnoresChange:function( nodeName, change )
    {
        if ( !this.view[ nodeName ] )
        {
            return false;
        }
        
        return AnimationChanges.nodeIgnoresChange( this.view[ nodeName ], change );
    },
    
    /**
     * Returns whether or not the specified animation exists.
     *
     * @param {string} name The animation to check for.
     * @returns {boolean} True if the animation exists, false otherwise.
     */
    hasAnimation:function( name )
    {
        return ( !!this.animationData[ name ] );
    },
    
    /**
     *  Applies a function to every single node that may be used by this Animation.
     *
     *  @param {function} func Calls this function, passing it the params ( node, nodeName ).
     *  @param {bool} [forceUniques=true] If true, will never call a function more than once per node. This is more costly to call.
     */
    _callFunctionOnAllNodes:function( func, forceUniques )
    {
        var appliedNodes = [];
        forceUniques = forceUniques || true;
        
        var animationName;
        for ( animationName in this.animationData )
        {
            var animation = this.animationData[ animationName ];
            
            if ( animation )
            {
                var nodeName;
                for ( nodeName in animation.nodes )
                {
                    var node = this.view[ nodeName ];
                    if ( node )
                    {
                        if ( node instanceof cc.Node )
                        {
                            if ( (!forceUniques || appliedNodes.indexOf( node ) < 0 ) )
                            {
                                func( node, nodeName );
                                appliedNodes.push( node );
                            }
                        }
                        else
                        {
                            cc.log( "[AnimationController] WARNING: " + nodeName + " is not a cc.Node! Why did you do this?" );
                        }

                    }
                }
            }
        }
    },
    
    _animateNode:function( node, animationData, animationName, nodeName )
    {
        var visibleSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "visible", AnimationChanges.Visibility );
        var alphaSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "alpha", AnimationChanges.Alpha );
        var rotSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "rotation", AnimationChanges.Rotation );
        var skewSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "skew", AnimationChanges.Skew );
        
        //these are special cased due to legacy animations
        var posSequence = null;
        var scaleSequence = null;
        
        //current support, animate the position and the scale separately, do not use a bounding box
        if ( animationData.origin )
        {
            posSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "origin", AnimationChanges.Position );
            scaleSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "size", AnimationChanges.Scale );
        }
        //legacy support, use the frame (bounding box) for both the position and for the scale
        else
        {
            posSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "frame", AnimationChanges.Position );
            scaleSequence = this._animateNodeProperty( node, animationData, animationName, nodeName, "frame", AnimationChanges.Scale, this._propertyAnimators.size );
        }

        if ( posSequence )
        {
            node.runAction( posSequence );
        }

        if ( scaleSequence )
        {
            node.runAction( scaleSequence );
        }

        if ( alphaSequence )
        {
            node.runAction( alphaSequence );
        }
        
        if ( rotSequence )
        {
            node.runAction( rotSequence );
        }
        
        if ( skewSequence )
        {
            node.runAction( skewSequence );
        }

        if ( visibleSequence )
        {
            node.runAction( visibleSequence );
        }

        return ( animationData.duration || 0 ) * __animationDebugDurationMultiple;
    },
    
    /**
     * Will animate a specific property on a node. Properties can be alpha, position, frame, etc etc
     *
     * @param {cc.Node} node The node to animate.
     * @param {Object} animationData The animation data for this particular node in this particular animation.
     * @param {string} animationName The name of the animation that we're playing.
     * @param {string} nodeName The name of the node that we're animating.
     * @param {string} dataKey The name of the data in the animationData that we'll be using for this particular property.
     * @param {AnimationChanges} [changes] The changes that we ignore. If we find them ignored, we'll do nothing.
     */
    _animateNodeProperty:function( node, animationData, animationName, nodeName, dataKey, changes, propertyAnimator )
    {
        propertyAnimator = propertyAnimator || this._propertyAnimators[ dataKey ];
        
        if ( changes !== undefined && this.nodeIgnoresChange( nodeName, changes ) )
        {
            return;
        }
        
        var sequenceString = "";
        var sequence = null;
        var lastActionTime = 0;
        var delay = 0;
        var capitalDataKey = dataKey.charAt( 0 ).toUpperCase() + dataKey.substring( 1 );
        var endDataKey = "end" + capitalDataKey; //endFrame, endAlpha, etc
        
        var keyframeIndex;
        for ( keyframeIndex = 0; keyframeIndex < animationData.keyframes.length; keyframeIndex++ )
        {
            var keyframe = animationData.keyframes[ keyframeIndex ];
            var actionTime = 0;

            //we can have either or both startData and endData.
            //we immediately set any startData, and we animate to any endData.
            var startData = keyframe[ dataKey ];
            var endData = keyframe[ endDataKey ];
            if ( startData !== undefined || endData !== undefined )
            {
                actionTime = ( keyframe.startTime || 0 ) * __animationDebugDurationMultiple;
                var actionDelay = actionTime - lastActionTime;
                lastActionTime = actionTime;
                
                //if there is a delay since the last time we added an action in this sequence, we must add an action for that
                if ( actionDelay > 0 )
                {
                    __animationLogNodeProperty( animationName, nodeName, "delay", actionTime );
                    
                    sequence = this._addWait( sequence, actionDelay );
                    
                    if ( __animationShouldDebugLog( animationName, nodeName ) )
                    {
                        sequenceString += "[Wait for " + actionDelay + "]";
                    }
                }
                
                //immediately set any startData that was defined
                if ( startData !== undefined )
                {
                    __animationLogNodeProperty( animationName, nodeName, dataKey, actionTime, startData );
                    
                    //sequence = this[ "_addSet" + capitalDataKey ]( sequence, node, startData, actionTime <= 0 );
                    sequence = propertyAnimator.addSetProperty( sequence, node, startData, actionTime <= 0 );
                    
                    if ( __animationShouldDebugLog( animationName, nodeName ) )
                    {
                        if ( actionTime <= 0 )
                        {
                            __animationDebugLog( animationName + "/" + nodeName + " <immediate>: " + dataKey + " being set to " + JSON.stringify( startData ) + " instantly." );
                        }
                        else
                        {
                            sequenceString += "[Set " + dataKey + " to " + JSON.stringify( startData ) + "]";
                        }
                    }
                }
                
                //animate towards any endData that was defined
                if ( endData !== undefined )
                {
                    var duration = ( keyframe.motionDuration || 0 ) * __animationDebugDurationMultiple;
                    var easeType = keyframe.easeType;
                    __animationLogNodeProperty( animationName, nodeName, dataKey, actionTime, endData, duration );
                    
                    //sequence = this[ "_addAnimate" + capitalDataKey ]( sequence, node, endData, duration, easeType, actionTime );
                    sequence = propertyAnimator.addAnimateProperty( sequence, node, endData, duration, easeType, actionTime );
                    
                    lastActionTime += duration;
                    
                    if ( __animationShouldDebugLog( animationName, nodeName ) )
                    {
                        sequenceString += "[Animate " + dataKey + " to " + JSON.stringify( endData ) + " for " + duration + "]";
                    }
                }
            }
        }
        
        if ( sequenceString && __animationShouldDebugLog( animationName, nodeName ) )
        {
            __animationDebugLog( animationName + "/" + nodeName + " <sequence>: " + sequenceString, animationName, nodeName );
        }
        
        return sequence;
    },

    _addWait:function( prevSequence, duration )
    {
        if ( duration <= 0.0 )
        {
            return prevSequence;
        }
        else
        {
            var interval = cc.DelayTime.create( duration );
            
            if ( !prevSequence )
            {
                return interval;
            }
            return cc.Sequence.create( prevSequence, interval );
        }
    },
    
    //TODO - move this to a utility function of some kind?
    _playSound:function ( soundName, delay )
    {
        if( soundName )
        {
            var soundLoc = "sounds/" + soundName;
            if( delay )
            {
                var playFunc = function( sound )
                {
                    audio.playEffect( sound );
                }.bind( undefined, soundLoc );

                var playAction = cc.CallFunc.create( playFunc, this );
                var intervalAction = cc.DelayTime.create( delay );
                this.actionTarget.runAction( cc.Sequence.create( intervalAction, playAction ) );
            }
            else
            {
                audio.playEffect( soundLoc );
            }
        }
    },
    
    _setUpCues:function( cues, animName )
    {
        if ( !cues )
        {
            return;
        }
        
        __animationDebugLog( animName + " <cues>: setting up cues.", animName );
        
        var maxCueTime = 0;
        
        //loop through all cues, and apply any that exist and have matching target functions
        var cueId;
        for ( cueId in cues )
        {
            __animationDebugLog( animName + " <cues>: trying to setup cue: " + cueId, animName );
            
            var cueTime = cues[ cueId ];
            var cueFunctionName = "cue_" + cueId;
                            
            //loop through all cue targets to apply cues to all of them
            var cueTargetIndex;
            for ( cueTargetIndex = 0; cueTargetIndex < this.cueTargets.length; cueTargetIndex++ )
            {
                var cueTarget = this.cueTargets[ cueTargetIndex ];
                if ( cueTarget[ cueFunctionName ] !== undefined )
                {
                    __animationDebugLog( animName + ": <cues>    cue target #" + cueTargetIndex + " can do cue, correctly linked", animName );
                    //if the cue isn't fired immediately, create a Cocos delayed action to fire it later
                    if( cueTime > 0 )
                    {
                        var runCue = cc.CallFunc.create( cueTarget[ cueFunctionName ], cueTarget );
                        var interval = cc.DelayTime.create( cueTime );
                        this.actionTarget.runAction( cc.Sequence.create( interval, runCue ) );
                    }
                    else
                    {
                        cueTarget[ cueFunctionName ]();
                    }
                }
                else
                {
                    __animationDebugLog( animName + ": <cues>    cue target #" + cueTargetIndex + " can't do cue, no link", animName );
                }
            }
            
            maxCueTime = Math.max( cueTime, maxCueTime );
        }
        return maxCueTime;
    }
});