/*global cc */
/*global require*/
/*global lumosity*/
/*global document*/

/*global DisplayState*/

require( "src/Utility/Random.js" );
/*global Random*/

require( "src/Input/Interaction.js" );
/*global InteractionType*/
/*global InteractionPhase*/

/*global GameComponentTimer*/
/*global GameComponentAudioMixer*/
/*global GameComponentScoreMultiplier*/
/*global GameComponentScore*/
/*global GameComponentMetadata*/
/*global metadataAggregateKeys*/
/*global metadataTrialKeys*/
/*global DisplayState*/
/*global Random*/
/*global AutomatedTester*/

require( "src/Input/InteractionAutomator.js" );
/*global InteractionAutomator*/

require( "src/Input/SwipeInteractionRecognizer.js" );
/*global SwipeGestureDirection*/

require( "src/Utility/LumosityUtils.js" );
/*global LumosityUtils*/
/*global lu*/

require( "src/Utility/LumosityGlobals.js" );
/*global LumosityGlobals*/
/*global lg*/

require( "src/Utility/LumosityLogger.js" );
/*global ll*/

require( "src/Game/Component/GameComponentTimer.js" );
/*global GameComponentTimer*/

require( "src/Game/Component/GameComponentTimeBonus.js" );
/*global GameComponentTimeBonus*/

require( "src/Game/Component/GameComponentMetadata.js" );
/*global GameComponentMetadata*/

require( "src/Game/Component/GameComponentAudioMixer.js" );
/*global GameComponentAudioMixer*/

require( "src/Game/Component/GameComponentScoreMultiplier.js" );
/*global GameComponentScoreMultiplier*/

require( "src/Game/Component/GameComponentScore.js" );

require( "src/Utility/AutomatedTester.js" );

var Game = cc.Class.extend(
{
    /** @lends Game.prototype */
    
    /** Get this game's GameSettings. */
    settings: null,
    
    /** Get this game's GameModel. */
    model: null,
    
    /** Get this game's GameView. */
    view: null,
    
    /** An array of components attached to this game. */
    _components: null,

    /** Automator used to record and playback animations. */
    _interactionAutomator: null,
    
    /**
     * A Game is the base class for all... games. It provides functionality for entering and exiting a game,
     * and it also constructs its GameModel, GameSettings, and GameView.
     *
     *   @constructs
     */
    ctor:function()
    {
        this._configureSearchPaths();

        // Initialize the automated test data.
        // This reads the file and assigns the Random.seed.
        // ( Must be done before game code because of Random. )
        Random.initialSeed = Random.seed;
        AutomatedTester.initialize();

        // Create the interaction automator.
        // Games can override this method to return their own interaction automator.
        this._interactionAutomator = this._createInteractionAutomator();

        //must be done after the search paths are set up
        LumosityGlobals.initialize();

        this.settings = this._createSettings();
        this.model = this._createModel();

        // Tells view if its practice mode.
        if ( LumosityUtils.isFitTestModeOn() )
        {
            if ( !this.model.persistentData.isPracticeMode && !this.model.isFitTestMode )
            {
                this.model.restartToPracticeMode = true;
            }
        }

        this.view = this._createView( cc.Director.getInstance().getWinSize() );
        
        this.view.model = this.model;
        this.view.settings = this.settings;
        this.view.setInteractionResponder( this );
        this.view.setAnimationResponder( this );
        
        if ( LumosityUtils.isFitTestModeOn() )
        {
            if ( !this.model.persistentData.isPracticeMode && !this.model.isFitTestMode )
            {
                this.model.shouldCreateTimer = false;
            }
        }

        // Keeps the state of metadata from Practice Mode
        this.tutorialMetadata = {
                                  _incongruentStimuliCorrect: 0,
                                  _trialCongruentTimes: [],
                                  _trialIncongruentTimes: [],
                                  _trialCorrectTimes: [],
                                  _trialIncorrectTimes: [],
                                  _trialResponseTimes: [],
                                  _trialNoSwitchTimes: [],
                                  _trialSwitchTimes: [],
                                  starts: 0, finishes: 0,
                                  correct: 0, incorrect: 0,
                                  time: 0,
                                  _trials: []
                                };   

        this._components = {};
        this._createComponents();

        this._initializeLumosityObject();

        this.configureMetaDataKeys();
        
        cc.log( "LUMOSITY PARAMS: " + JSON.stringify( lumosity.game.params ) );
    },
    
    _initializeLumosityObject:function()
    {
        lumosity.game.onPause = this.onBackgrounded.bind( this );
        lumosity.game.fireInteraction = function( apiData ) { this.interaction_received( this._interactionAutomator.createInteraction( apiData ) ); }.bind( this );
        lumosity.game.playInteractionScript = function( scriptData ) { this._interactionAutomator.replayScript( scriptData, this ); }.bind( this );
        lumosity.game.recordInteractionScript = function( useRelativeTime ) { this._interactionAutomator.beginRecordingScript( useRelativeTime ); }.bind( this );
        lumosity.game.finishInteractionScript = function() { return this._interactionAutomator.finishRecordingScript(); }.bind( this );
        lumosity.game.getLastRecordedInteractionScript = function() { return this._interactionAutomator.getLastRecordedScript(); }.bind( this );
        lumosity.game.setRandomSeed = function( seed ) { Random.seed = seed; }.bind( this );
        lumosity.game.getRandomSeed = function() { return Random.initialSeed; }.bind( this );
        lumosity.game.backdoor = function( data ) { this.interaction_alias( data.action, data.data ); }.bind( this );
    },

    /**
     * Whenever the game gets sent to the background as a result of
     * a phone call, multi-tasking, etc.
     * 
     */
    onBackgrounded: function()
    {
        if ( this.view.getDisplayState() !== DisplayState.HowToPlayMenu )
        {
            this.onPause(true);   // instant, no anim, to prevent cheating
        }
    },
    
    /**
    *  This is called every frame. Be very careful using Cocos functions in here, since that can tank the
    *  framerate if it gets called every frame.
    *  @param delta How long it was since the last update
    */
    onUpdate:function( delta )
    {
        LumosityUtils.callFunctionOnCollection( this._components, "onUpdate", [ delta ] );

        if( this._interactionAutomator )
        {
            this._interactionAutomator.onUpdate( delta );
        }
    },
    
    /**
    *  Gets called every second. Most useful for incrementing the timer and such. Be careful about calling
    *  too many Cocos functions from here, as it can be bad for framerate.
    */
    onUpdateEachSecond:function()
    {
        LumosityUtils.callFunctionOnCollection( this._components, "onUpdateEachSecond" );
    },
    
    /**
     *  Called when the Game has first opened. This is the entry point into the Game. It will kick off
     *  the gameBegin animation, as well as make sure anything that needs to be initialized is.
     *  It's important to reset any initialization here, rather than the constructor only, because
     *  if a user restarts a game or something similar it's possible (although not currently) that
     *  the state won't be properly reset.
     */
    onBegin:function()
    {
        LumosityUtils.callFunctionOnCollection( this._components, "onBegin" );

        ll.verbose( "Game", "onBegin()" );

        this.playStartingAnimation();

        this.model.isOnBeginCalled = true;
        this.model.isGamePaused = false;

        if ( LumosityUtils.isFitTestModeOn() )
        {
            if ( !this.model.persistentData.isPracticeMode && !this.model.isFitTestMode )
            {
                this.model.persistentData.isPracticeMode = true;
                this.model.shouldCreateTimer = false;
                this.model.restartToPracticeMode = true;
                this.view.getHudWidget().enableFitTestMode ( true );
                this.view.createDimLayer();
            }
            else
            {
                this.model.shouldCreateTimer = true;
                this.view.getHudWidget().enableFitTestMode ( false );
                this.view.removeDimLayer();
            }

            this.view.getPauseMenu().setFitTestModeEnabled( true );
        }

        //nab focus if this is HTML5
        if ( lumosity.device.id === "html5" && document && document.getElementById )
        {
            var gameCanvas = document.getElementById( "gameCanvas" );
            if ( gameCanvas )
            {
                gameCanvas.focus();
            }
        }
    },
    
    playStartingAnimation: function()
    {
        //if we're in practice mode and we have a practice begin animation, play that
        if ( ( this.model.restartToPracticeMode || this.model.persistentData.isPracticeMode ) && this.view.hasAnimation( "practiceBegin" ) )
        {
            this.view.playAnimation( "practiceBegin", this.onStartGameplay );
        }
        //otherwise, play the game begin animation
        else
        {
            this.view.playAnimation( "gameBegin", this.onStartGameplay );
        }
    },

    /**
     *  Called when the main gameplay has begun. This is generally after the gameBegin animation completes.
     *  At this point, the timer will actually be started, or the first trial will be shown, etc.
     */
    onStartGameplay:function()
    {
        LumosityUtils.callFunctionOnCollection( this._components, "onStartGameplay" );

        ll.verbose( "Game", "onStartGameplay()" );

        this.model.isGameRunning = true;
        this.model.isGameOver = false;

        this.model.startTime = lumosity.game.frametime; //new Date().getTime();

        if ( this.model.persistentData.isPracticeMode )
        {
            this.tutorialMetadata.starts += 1;
        }

        if ( this.settings.printNodeHierarchy === true )
        {
            ll.verbose( "Game", "onStartGameplay()", "HIERARCHY ");
            ll.verbose( "Game", "onStartGameplay()", "****************************************" );
            LumosityUtils.printNodeHierarchy( this.view );
            ll.verbose( "Game", "onStartGameplay()", "****************************************" );
        }

        // Make sure the script is not running before deciding to do a playback or record.
        if( !AutomatedTester.isRunning() )
        {
            if( !AutomatedTester.playback() )
            {
                // Start recording script if playback fails.
                if( !AutomatedTester.record() && lumosity.game.params.interaction_script )
                {
                    // AutomationTester isn't running but a playback script was provided but unable to play back the script and recording
                    // can't happen, then we need to quit the game.
                    this.onFinish( true );
                }
            }
        }
    },

    /**
     *  Called when the main gameplay in a game has ended. At this point, results will be sent to the C++ layer
     *  so that they can be sent to the server, and the gameEnd animation will begin playing.
     *  No user input will be accepted after this event.
     */
    onEndGameplay:function()
    {
        this.model.isGameRunning = false;
        this.model.isGameOver = true;

        this.finalizeMetadata();

        if ( this.model.persistentData.isPracticeMode )
        {
            this.tutorialMetadata.finishes += 1;
        }

        LumosityUtils.callFunctionOnCollection( this._components, "onEndGameplay" );

        this.playEndAnimation();

        //finish the interaction script in case there is one (if you want it absolutely tied to the session)
        //this is harmless to call if nothing is being recorded.
        if( this._interactionAutomator )
        {
            this._interactionAutomator.finishRecordingScript();
        }

        //for automation support, you can hook into this to end automation
        if ( lumosity.game.gameplayEnded )
        {
            lumosity.game.gameplayEnded();
        }
    },

    exportAutomatedTestReport: function()
    {
        var testInfo = null;
        if( !lumosity.game.params.interaction_script || !this._interactionAutomator || !this._interactionAutomator.report )
        {
            return null;
        }

        // Filter the tests based on assertMessage property to remove duplicate entries of the same assert.
        // This will only report first occurrence of the issue.
        this._interactionAutomator.report.tests = this._interactionAutomator.filterTests( "assertMessage" );
        this._interactionAutomator.report.filePath = AutomatedTester.filePath;

        // Automated test never ran but was provided an interaction_script then we report an error.
        if( !AutomatedTester.isRunning() )
        {
            testInfo = { test: "Playback Interaction Script", assertMessage: "", status: "fail" };
            testInfo.assertMessage += "Tried to run interaction script but " + lumosity.game.params.interaction_script + " does not exist.";
            this._interactionAutomator.report.tests.push( testInfo );
        }

        if( this._interactionAutomator.report.tests.length !== 0 )
        {
            this._interactionAutomator.report.status = "fail";
        }

        cc.log( "Automated Test Report: " + JSON.stringify( this._interactionAutomator.report ) );
        return this._interactionAutomator.report;
    },

    exportResults: function()
    {
        var results = JSON.parse( this.model.exportResults() );

        results.game_result_data.bad_play = 'F';
        var hasMetThreshold = this.model.hasMetValidResponseAccuracy( this.settings.responseAccuracyThreshold );
        if( hasMetThreshold === false )
        {
            // Threshold was not met, setting 0 for best stat and flagging game as bad play.
            results.gameStats[ results.bestStatKey ] = 0;
            results.stat = 0;
            results.game_result_data.bad_play = 'T';
        }
         
        cc.log( "Game Results: " + JSON.stringify( results ) );
        return results;
    },

    // OVERRIDE for 
    // playEndAnimation: function ()
    // {
    //     lumosity.game.report( this.model.exportResults() );
    //     this.view.playAnimation( "gameEnd", this.onFinish );
    // },

    playEndAnimation: function ()
    {
        if ( !this.model.persistentData.isPracticeMode )
        {
            var results = this.exportResults();
            var automatedTestReport = this.exportAutomatedTestReport();
            if( automatedTestReport )
            {
                // Append automated test report if we have one.
                results.automatedTestReport = automatedTestReport;
            }

            lumosity.game.report( JSON.stringify( results ) );
            this.view.playAnimation( "gameEnd", this.onFinish.bind( this, false ) );
        }
        else
        {
            this.view.playAnimation( "gameEnd", this.onStartFitTest );
        }
    },

    /** Called when a practice/tutorial is over, to start a fitTest game. */
    onStartFitTest: function ()
    {
        this.model.restartToPracticeMode = false;

        this.model.tutorialEnded = true;

        this.model.shouldCreateTimer = true;

        this.onRestart();
    },

    /** Called when a game should begin its tutorial sequence. **/
    onStartTutorial: function( fromApp )
    {
        this.model.persistentData.isPracticeMode = true;
        if( fromApp )
        {
            this.onBegin();
        }
        else
        {
            this.onRestart();
        }
        this.view.createDimLayer();
    },

    /** Called when a game should begin its tutorial completion sequence. **/
    onCompleteTutorial: function( skipped )
    {
        cc.log( "Game::onCompleteTutorial" );

        this.model.tutorialSkipped = skipped;

        this.view.playAnimation( "endTutorial" );
    },

    /** Called when a game has ended its tutorial sequence. **/
    onEndTutorial: function()
    {
        this.model.persistentData.isPracticeMode = false;
        this.onRestart();
        this.view.removeDimLayer();
    },
    
    /**
     *  Called when a Game has completely finished. This could be because the gameplay has naturally ended
     *  and the gameEnd animation has finished, or it could be because the user prematurely exited the game.
     *  Regardless, a Game will always go through this function when it's being exited.
     *
     *  @param {boolean} wasQuit If true, the game was prematurely ended.
     */
    onFinish:function( wasQuit )
    {
        LumosityUtils.callFunctionOnCollection( this._components, "onFinish");

        // Fix Jira bug GAME-3053 06/03/2015
        // [QA][iOS] Quitting a game can cause the music to continue playing in the app, or during other game's HTPs
        cc.AudioEngine.getInstance().stopMusic( true );
        cc.AudioEngine.getInstance().stopAllEffects();
        
        // Added for automated testing.
        // Need to report to the app the autmoated test report on quit.
        if( wasQuit )
        {
            var results = {};
            var automatedTestReport = this.exportAutomatedTestReport();
            if( automatedTestReport )
            {
                // Append automated test report if we have one.
                results.automatedTestReport = automatedTestReport;
                lumosity.game.report( JSON.stringify( results ) );
            }

            if( lumosity.device.id === "html5" && document && document.getElementById )
            {
                if( this._interactionAutomator )
                {
                    this._interactionAutomator.finishRecordingScript();
                }

                if ( lumosity.game.gameplayEnded )
                {
                    lumosity.game.gameplayEnded();
                }
            }
        }
        
        lumosity.game.quit( wasQuit ? "completed" : "quit" );
    },

    /**
     * Called when a Game has been paused.
     */
    onPause: function ( instant )
    {
        ll.verbose( "Game", "onPause" );


        this.model.isGamePaused = true;
        this.view.onPause();
        this.view.centerOnGameMenu(instant);

        this.view.pauseWidgets();

        this.model.addPause( lumosity.game.frametime, instant );

        LumosityUtils.callFunctionOnCollection( this._components, "onPause" );
        if ( instant === true )
        {
            cc.Director.getInstance().drawScene();
        }

        //this.view.getPauseMenu().setTutorialModeEnabled( this.model.persistentData.isPracticeMode );
    },

    /**
     * Called when a Game has been unpaused.
     */
    onResume: function()
    {
        ll.verbose( "Game", "onResume" );

        this.model.isGamePaused = false;

        this.view.onResume();
        this.view.centerOnGameStage();

        this.model.addResume( lumosity.game.frametime );
        
        this.view.resumeWidgets();
        LumosityUtils.callFunctionOnCollection( this._components, "onResume" );
    },
    
    /**
     * Called when the Game has been restarted within JS.
     * By default, this will re-create the model, stop all
     * animations, then set everything as in the layout and
     * call gameBegin. Not at all guaranteed to work out of the box.
     */
    onRestart: function()
    {
        this.model.persistentData.num_restart++;
        var restartToFitTestMode = false;
        // Undo this comment and comment out the line below it to restart to fitTest mode.
        // rather than restarting to the tutorial.
        // if ( LumosityUtils.isFitTestModeOn() && !this.model.restartToPracticeMode )
        if ( LumosityUtils.isFitTestModeOn() && this.model.tutorialEnded )
        {
            restartToFitTestMode = true;
            this.model.tutorialEnded = false;
        }

        var tutorialPopUpPlaying = this.view.guidedEndPopUpPlaying ? true : false;

        // store off Persistent data
        var persistentData = this.model.persistentData;
        //create a brand new model
        this.model = this._createModel();
        this.model.persistentData = persistentData;
        this.view.model = this.model;

        if ( LumosityUtils.isFitTestModeOn() )
        {
            if ( restartToFitTestMode )
            {
                this.model.isFitTestMode = true;
                this.model.shouldCreateTimer = true;
                this.model.restartToPracticeMode = false;
                this.model.persistentData.isPracticeMode = false;
            }
            else
            {  
                // Restarts tutorial metaData
                this.tutorialMetadata = {
                                          _incongruentStimuliCorrect: 0,
                                          _trialCongruentTimes: [],
                                          _trialIncongruentTimes: [],
                                          _trialIncorrectTimes: [],
                                          _trialCorrectTimes: [],
                                          _trialResponseTimes: [],
                                          _trialSwitchTimes: [],
                                          _trialNoSwitchTimes: [],
                                          starts: this.tutorialMetadata.starts,
                                          finishes: this.tutorialMetadata.finishes,
                                          correct: 0, incorrect: 0,
                                          time: 0,
                                          _trials: []
                                        };
                this.model.shouldCreateTimer = false;
                this.view.removeDimLayer();
                if ( tutorialPopUpPlaying )
                {
                    this.view.removeFrontDimLayer();
                }
                this.model.restartToPracticeMode = true;
            }
        }
    
        //re-create all our components
        this._components = {};
        this._createComponents();
    
        //have the view restart itself
        this.view.onRestart();

        //and transition back from the pause menu
        this.view.centerOnGameStage();
    
        //now begin the game again
        this.onBegin();
    },

    /**
     * Cue for showing the tutorial popup.
     * Place this cue in your flash file.
     */
    cue_showEndTutorialPopup:function()
    {
        this.view.showEndTutorialPopup( this.model.tutorialSkipped );
    },

    /**
     * Cue for ending the tutorial mode.
     * Place this cue in your flash file.
     */
    cue_endTutorialMode:function()
    {
        this.onEndTutorial();
    },

    /*
     *  Called each time the component timer ticks.
     *  Override this to be notified when the component has ticked.
     */
    onComponentTimerTick:function( timeRemaining )
    {
        
    },

     /**
     *  Called when the component timer has reached 0.
     *  Override this to be notified when the component timer has ended.
     */
    onComponentTimerEnd:function()
    {
        ll.verbose( "Game", "onComponentTimerEnd" );
        this.onEndGameplay();
    },
    
    /**
     *  When the user responds, you probably want to call this. It's a shorthand function
     *  to cover most cases. All it does is it increases / decreases the bonus multiplier
     *  and then tells the view to show its feedback.
     *
     *  @param {boolean} correct If the user responded correctly, this is true. False if incorrectly.
     */
    evaluateUserResponse:function( correct )
    {
        var scoreMultiplierComponent = this._components[ GameComponentScoreMultiplier.prototype.getType() ];
        
        if( correct )
        {
            scoreMultiplierComponent.increaseMultiplierProgress();
            this.view.showFeedback( true );
        }
        else
        {
            scoreMultiplierComponent.decreaseMultiplierProgress();
            this.view.showFeedback( false );
        }
    },

    /**
     *  Used to record interactions for interaction automator playback.
     *
     *  @param {Interaction} interaction The interaction that happened.
     */
    interaction_record: function( interaction )
    {
        if( this._interactionAutomator )
        {
            this._interactionAutomator.recordInteraction( interaction );
        }
    },
    
    /**
     *  Override this to respond to different sorts of interactions. By default,
     *  this handles button release events.
     *
     *  @param {Interaction} interaction The interaction that happened.
     */
    interaction_received:function( interaction, isLegacy )
    {
        if( isLegacy === undefined )
        {
            isLegacy = false;
        }

        if( this.model.isGameOver === true ){ return; }
        
        if ( interaction.getType() === InteractionType.Swipe )
        {
            if( this.view.getDisplayState() === DisplayState.HowToPlayMenu )
            {
                if ( interaction.getSwipeDirection() === SwipeGestureDirection.Left )
                {
                    this.view.slideHowToPlayCards( true );
                }
                else if ( interaction.getSwipeDirection() === SwipeGestureDirection.Right )
                {
                    this.view.slideHowToPlayCards( false );
                }
            }
            else if ( this.view.getDisplayState() === DisplayState.PauseMenu )
            {
                if ( interaction.getSwipeDirection() === SwipeGestureDirection.Up )
                {
                    this._pauseSwiped( true );
                }
                else if ( interaction.getSwipeDirection() === SwipeGestureDirection.Down )
                {
                    this._pauseSwiped( false );
                }
            }
        }

        if( isLegacy === false )
        {
            if ( interaction.getType() === InteractionType.Button )
            {
                if ( interaction.getPhase() === InteractionPhase.Ended )
                {
                    if( !this.simulateButtonRelease( interaction.getButtonName() ) )
                    {
                        if( !this.simulateSystemButtonRelease( interaction.getButtonName() ) )
                        {
                            this.simulateKeyboardRelease( interaction.getRawButtonName() );
                        }
                    }
                }
                else if ( interaction.getPhase() === InteractionPhase.Began )
                {
                    if( !this.simulateButtonPressed( interaction.getButtonName() ) )
                    {
                        this.simulateKeyboardPressed( interaction.getRawButtonName() );
                    }
                }
                
                //any interaction with a button is considered to be held, but only bother if we care
                if ( this.usesInteractionType( InteractionType.Button, InteractionPhase.Changed ) )
                {
                    this._buttonHeld( interaction.getButtonName(), interaction );
                }
            }
        }
        else if( isLegacy === true )
        {
            // Legacy
            if ( interaction.getType() === InteractionType.Button )
            {
                if ( interaction.getPhase() === InteractionPhase.Ended )
                {
                    this._buttonReleased( interaction.getButtonName(), interaction );
                }
                else if ( interaction.getPhase() === InteractionPhase.Began )
                {
                    this._buttonPressed( interaction.getButtonName(), interaction );
                }
                
                //any interaction with a button is considered to be held, but only bother if we care
                if ( this.usesInteractionType( InteractionType.Button, InteractionPhase.Changed ) )
                {
                    this._buttonHeld( interaction.getButtonName(), interaction );
                }
            }
        }
    },

    /**
     * Simulates the press of a button by retrieving the button from the button layer and
     * calling it's button handler directly.
     */
    simulateButtonPressed: function( name )
    {
        var button = null;

        button = this.view.getButton( name );
        if( button === null )
        {
            return false;
        }

        button.handleIsPressed( true );

        return true;
    },

    /**
     * Simulates the release of a button by retrieving the button from the button layer and
     * calling it's button handler directly.
     */
    simulateButtonRelease: function( name )
    {
        var button = null;

        button = this.view.getButton( name );
        if( button === null )
        {
            return false;
        }

        button.handleIsPressed( false );

        return true;
    },

    /**
     * Simulates the release of a system button.
     * specifically used by Android devices. 
     */
    simulateSystemButtonRelease: function( name )
    {
        cc.log( "SimulateSystemButtonRelease buttonName: " + name );
        if( name === "system_menu" )
        {
            this.onFinish( true );
            return true;
        }

        if( name === "system_back" )
        {
            this._backButtonTapped();
            return true;
        }

        return false;
    },

    /**
     * Responds to the press of a button from the keyboard.
     */
    simulateKeyboardPressed: function( name )
    {
        // ` Key on the keyboard.
        if( name === "keyboard.192" )
        {
            if( lumosity.game.takeScreenshot )
            {
                // Take a screenshot
                var screenshotName = "Screenshot" + InteractionAutomator.screenshotCount;
                lumosity.game.takeScreenshot( screenshotName );
                InteractionAutomator.screenshotCount += 1;
                cc.log( "Game::simulateKeyboardPressed taking screenshot: " + screenshotName );
            }
        }
    },

    /**
     * Responds to the release of a button from the keyboard.
     */
    simulateKeyboardRelease: function( name )
    {
       
    },
    
    /**
     * Fired by backdoor calls or playbacks of interaction scripts. You can create aliases to actual
     * interaction objects in your settings, or you can override this function and do whatever you
     * want with a particular alias.
     *
     * @param {string} alias The alias. If this exists as a key in interactionAliases in your settings, this fire that interaction.
     * @param {Object} [data] Any data that gets sent through with the alias, may or may not exist.
     */
    interaction_alias:function( alias, data )
    {
        if ( this._interactionAutomator && this.settings.interactionAliases && this.settings.interactionAliases[ alias ] )
        {
            var interaction = this._interactionAutomator.createInteraction( this.settings.interactionAliases[ alias ] );
            if ( interaction )
            {
                this.interaction_received( interaction );
            }
        }
        
        //only do how to play functions if it's open
        if ( this.view.getDisplayState() === DisplayState.HowToPlayMenu )
        {
            if ( alias === "close_how_to_play" )
            {
                this.howToPlay_playNowTapped();
            }
            else if ( alias === "cancel_how_to_play" )
            {
                this.howToPlay_backNowButtonTapped();
            }
            else if ( alias === "go_to_next_how_to_play" )
            {
                this.view.goToNextHowToPlaySlide();
            }
        }
        
        if ( alias === "pause" && this.view.getDisplayState() === DisplayState.Game )
        {
            this.onPause();
        }
        else if ( alias === "unpause" && this.view.getDisplayState() === DisplayState.PauseMenu )
        {
            this.onResume();
        }
        
        if ( alias === "finish_game" )
        {
            this.onEndGameplay();
        }
        
        if ( alias === "select_how_to_play" && this.view.getDisplayState() === DisplayState.PauseMenu )
        {
            this.gamePauseMenu_howToPlay();
        }
    },
    
    /**
    *  Because it's expensive to report every single bit of information from
    *  C++ to JS, the view will check with the game using this function to see if
    *  it should bother sending any Interactions of certain types.
    *  
    *  This is especially relevant for touch and mouse events, because their move
    *  events will likely be triggered every single frame. By default, Button
    *  events are accepted, since they are very low-cost and won't get triggered
    *  unless you have a button anyway. The exception here is button held events,
    *  which are by default ignored since they fire every frame.
    *  
    *  @param type The InteractionType that the view wants to send events for.
    *  @return Whether or not the passed type is supported by this game.
    */
    usesInteractionType:function( type, phase )
    {
        //by default, allow pressing and releasing of buttons, but don't allow held since it's costly
        if ( type === InteractionType.Button && phase !== InteractionPhase.Changed )
        {
            return true;
        }
        else if ( type === InteractionType.Swipe )
        {
            return true;
        }

        return false;
    },
    
    /**
     *  You must override this to create your GameSettings subclass.
     *  If you do not, it will be set to null, and stuff will break.
     */
    _createSettings:function()
    {
        return null;
    },
    
    /**
     *  You must override this to create your GameModel subclass.
     *  If you do not, it will be set to null, and stuff will break.
     */
    _createModel:function()
    {
        return null;
    },
    
    /**
     *  You must override this to create your GameView subclass.
     *  If you do not, it will be set to null, and stuff will break.
     *  
     *  @param {cc.Size} screenSize The size of the screen. The view may construct itself differently based on this.
     */
    _createView:function( /*screenSize*/ )
    {
        return null;
    },

    /**
     *  You can override this to create a Game specific interaction automator.
     *  If you do not, it will be set to the base class InteractionAutomator that handles base level automation checks.
     */
    _createInteractionAutomator: function()
    {
        return new InteractionAutomator( this );
    },
    
    /**
     *  Override this to respond to button interactions. By default, this will handle
     *  the pause button.
     *  
     *  @param {string} buttonName The name of the button that was released.
     *  @param {Interaction} interaction The interaction that triggered this function, may be needed to determine source (keyboard etc)
     */
    _buttonReleased:function( buttonName, interaction )
    {
        if( buttonName === "pause" )
        {
            this.onPause();
        }
        else if( buttonName === "system_menu" )
        {
            this.onFinish( true );
        }
        else if( buttonName === "system_back" )
        {
            this._backButtonTapped();
        }
    },

    _backButtonTapped: function()
    {
        if( this.view.getDisplayState() === DisplayState.Transition ){ return; }

        if( this.model.isGamePaused === true && this.model.isGameRunning === true )
        {
            if( this.view.getDisplayState() === DisplayState.PauseMenu )
            {
                this.onResume();
            }
            else if( this.view.getDisplayState() === DisplayState.HowToPlayMenu )
            {
                this.view.centerOnGameMenu();
            }
        }
        else if( this.view.getDisplayState() === DisplayState.HowToPlayMenu && this.model.isGameRunning === false )
        {
            this.onFinish( true );
        }
        else
        {
            this.onPause();
        }
    },
    
    /**
     *  Override this to respond to button interactions. By default, this will handle
     *  the pause button.
     *  
     *  @param {string} buttonName The name of the button that was held.
     *  @param {Interaction} interaction The interaction that triggered this function, may be needed to determine source (keyboard etc)
     */
    _buttonHeld:function( buttonName, interaction )
    {
        //
    },
    
    /**
     *  Override this to respond to button interactions. By default, this will handle
     *  the pause button.
     *  
     *  @param {string} buttonName The name of the button that was pressed.
     *  @param {Interaction} interaction The interaction that triggered this function, may be needed to determine source (keyboard etc)
     */
    _buttonPressed:function( buttonName, interaction )
    {
        //
    },
    
    /**
     *  Override this to create whatever components your game may need. Make sure to
     *  add them to this._components object so that they get updated properly.
     *
     * GAME-3437 Upon creating the metadata component, assign the game's tutorial metadata to the metadata component.  Test case:  2015-07-27 Editor's Choice.  Boiler.  Play.  Pause.  How to Play.  Complete tutorial.  Aaron expects end tutorial popup.  Got error that tutorialMetadata was undefined (with error message that it was accessing a read-only property).  No symptom in the mobile app.  Test case: 2015-05 Splitting Seeds.  Complete a session.  Finalize metadata.  Tyler expects the metadata component had already assigned the tutorial metadata to the game's tutorial metadata.
     */
    _createComponents:function()
    {
        if( this.settings.showTime === true )
        {
            if ( LumosityUtils.isFitTestModeOn() )
            {
                if( this.model.shouldCreateTimer )
                {
                    this._addComponent( new GameComponentTimer( this ) );
                }
            }
            else
            {
                this._addComponent( new GameComponentTimer( this ) );
            }
        }

        if( this.settings.showScore === true )
        {
            // GameComponent Score Multiplier
            if( this.settings.scoreType === 1 )
            {
                this._addComponent( new GameComponentScoreMultiplier( this ) );
            }
            // GameComponent Score
            else if( this.settings.scoreType === 2 )
            {
                this._addComponent( new GameComponentScore( this ) );
            }
        }

        if( this.settings.showTimeBonus === true )
        {
            this._addComponent( new GameComponentTimeBonus( this ) );
        }

        this._addComponent( new GameComponentAudioMixer( this ) );
        this._createMetadataComponent();
    },

    /**
     * Creates a metadata component.
     * Should be overriden for custom component.
     *
     * @private
     */
    _createMetadataComponent: function ()
    {
        var metadataComponent = new GameComponentMetadata( this );
        metadataComponent.tutorialMetadata = this.tutorialMetadata;
        this._addComponent( metadataComponent );
    },

    /**
     * Adds the passed component to the array of components.
     * @param {GameComponent} component Component to add to our array (must extend from GameComponent).
     */
    _addComponent: function( component )
    {
        var componentType = null;
        if( component )
        {
            componentType = component.getType();
            this._components[ componentType ] = component;
        }
    },

    _configureSearchPaths:function( )
    {
        var defaultSearchPaths = cc.FileUtils.getInstance().getSearchPaths();
                
        //for HTML5
        if ( !defaultSearchPaths )
        {
            return;
        }
        
        // Find the device folder (and bigger) that we'll be using.
        var deviceFolderNames = this._findDeviceFolderNames();
        
        // Default behavior is to add device paths to the default search paths.
        var configuredSearchPaths = [];
        var basePath = lumosity.game.path;
        var addBasePath = ( basePath && basePath.length > 0 );
        var assetPath = lumosity.game.params.asset_path || lumosity.game.params.assetPath;      // Path where the game assets are located.
        var addCustomGameAssetPaths = ( assetPath && assetPath.length > 0 );
        var sharedGlobalAssets = lumosity.game.params.shared_global_assets;       // Path where we keep the shared art between games.
        var addCustomSharedGlobalAssets = ( sharedGlobalAssets && sharedGlobalAssets.length > 0 );
        var sharedGameAssets = lumosity.game.params.shared_game_assets;           // Path for game-specific shared assets.
        var addCustomSharedGameAssets = ( sharedGameAssets && sharedGameAssets.length > 0 );
        var soundsPath = lumosity.game.params.sounds_path;
        var addCustomSoundsPath = ( soundsPath && soundsPath.length > 0 );
        var sharedSoundsPath = lumosity.game.params.shared_sounds_path;
        var addCustomSharedSoundsPath = ( sharedSoundsPath && sharedSoundsPath.length > 0 );
        
        if( addBasePath )
        {
            basePath = lu.addTrailingSlashToPath( basePath );
            configuredSearchPaths.push( basePath );
        }

        if( addCustomGameAssetPaths )
        {
            assetPath = lu.addTrailingSlashToPath( assetPath );
            this._addDeviceFolders( configuredSearchPaths, assetPath, deviceFolderNames );
            //this._addLegacyDeviceFolders( configuredSearchPaths, assetPath );
            configuredSearchPaths.push( assetPath );
        }

        if( addCustomSoundsPath )
        {
            soundsPath = lu.addTrailingSlashToPath( soundsPath );
            configuredSearchPaths.push( soundsPath );
        }

        if( addCustomSharedGlobalAssets )
        {
            sharedGlobalAssets = lu.addTrailingSlashToPath( sharedGlobalAssets );
            configuredSearchPaths.push( sharedGlobalAssets );
        }

        if( addCustomSharedGameAssets )
        {
            sharedGameAssets = lu.addTrailingSlashToPath( sharedGameAssets );
            configuredSearchPaths.push( sharedGameAssets );
        }

        if( addCustomSharedSoundsPath )
        {
            sharedSoundsPath = lu.addTrailingSlashToPath( sharedSoundsPath );
            configuredSearchPaths.push( sharedSoundsPath );
        }
        
        var pathIndex;
        for( pathIndex = 0; pathIndex < defaultSearchPaths.length; pathIndex++ )
        {
            var path = defaultSearchPaths[pathIndex];
            if( path )
            {
                var sharedPath = path.substring( 0, path.lastIndexOf( "shared" ) );
                if( sharedPath && sharedPath.length > 0 )
                {
                    // Add the resolution-based device folders.
                    this._addDeviceFolders( configuredSearchPaths, sharedPath, deviceFolderNames );

                    // Add the legacy device folders, which will be removed soon.
                    //this._addLegacyDeviceFolders( configuredSearchPaths, path );
            
                    // push back the original path.
                    configuredSearchPaths.push( path );
                }
            }
        }

        // Depricated: See LumosityGlobals initialize method on searching for localized strings.
        //add the localization lookup last
        // for( pathIndex = 0; pathIndex < defaultSearchPaths.length; pathIndex++ )
        // {
        //     if ( !defaultSearchPaths[ pathIndex ] )
        //     {
        //         break;
        //     }
            
        //     this._addLocalizationFolders( configuredSearchPaths, defaultSearchPaths[ pathIndex ] );
        // }

        cc.FileUtils.getInstance().setSearchPaths( configuredSearchPaths );

        cc.log( "default paths: " + JSON.stringify( defaultSearchPaths ) );
        cc.log( "search paths: " + JSON.stringify( configuredSearchPaths ) );
    },

    _findDeviceFolderNames:function()
    {
        var names = [];
        var deviceWidth = cc.Director.getInstance().getWinSize().width;
        var searchPathWidthIndex;
        names.push( LumosityUtils.searchPathWidths[ 0 ] );
        
        for ( searchPathWidthIndex = 1; searchPathWidthIndex < LumosityUtils.searchPathWidths.length; searchPathWidthIndex++ )
        {
            var searchPathWidth = LumosityUtils.searchPathWidths[ searchPathWidthIndex ];
                        
            if ( deviceWidth > searchPathWidth )
            {
                break;
            }
            
            names.push( searchPathWidth );
        }
        
        //we need to reverse the names so that we have them in the right order (the smallest found should be loaded first)
        names.reverse();
        
        return names;
    },

    _addDeviceFolders:function( configuredSearchPaths, path, deviceFolderNames )
    {
        //do not expand Cocos2dShared, it contains only code
        if ( path.indexOf( "Cocos2dShared" ) >= 0 )
        {
            return;
        }
        
        var folderNameIndex;
        for ( folderNameIndex = 0; folderNameIndex < deviceFolderNames.length; folderNameIndex++ )
        {
            var devicePath = path + deviceFolderNames[ folderNameIndex ] + "/";
            configuredSearchPaths.push( devicePath );
        }
    },

    _addLocalizationFolders:function( configuredSearchPaths, path )
    {
        var localizationPath = path.substring( 0, path.lastIndexOf( "shared" ) ) + lumosity.game.locale + "/";
        configuredSearchPaths.push( localizationPath );
        
        if ( lumosity.game.locale !== "en" )
        {
            var enLocalizationPath = path.substring( 0, path.lastIndexOf( "shared" ) ) + "en" + "/";
            configuredSearchPaths.push( enLocalizationPath );
        }
    },

    // _getLegacyDeviceFolders:function(device)
    // {
    //     if(device === 'iphone5_retina')
    //     {
    //         return [ 'iphone5_retina', 'iphone_retina' ];
    //     }
    //     else
    //     {
    //         return [device];
    //     }
    // },
    //
    // _addLegacyDeviceFolders:function( configuredSearchPaths, path )
    // {
    //     var deviceFolders = this._getLegacyDeviceFolders( lumosity.device.id );
    //     var deviceIndex;
    //     for( deviceIndex = 0; deviceIndex < deviceFolders.length; ++deviceIndex)
    //     {
    //         ll.verbose( "Game", "_addLegacyDeviceFolders", "Device Folder: " + deviceFolders[ deviceIndex ] );
    //         var devicePath = path;
    //         var position = devicePath.lastIndexOf( "shared" );
    //         devicePath = [ devicePath.slice( 0, position ), deviceFolders[ deviceIndex ] + '/' ].join( '' );
    //         configuredSearchPaths.push( devicePath );
    //     }
    // },

    gamePauseMenu_stateChanged: function( paused )
    {
        if ( paused )
        {
            this.onPause();
        }
        else
        {
            this.onResume();
        }
    },

    gamePauseMenu_closeMenu: function()
    {
        this.onResume();
    },

    gamePauseMenu_howToPlay: function()
    {
        this.showHowToPlay ( false );
    },

    gamePauseMenu_restartGame: function()
    {
        // Fix Jira Bug [Android] - Fit Test game tutorials show timer/trial and score instead of Tutorial
        // https://lumoslabs.atlassian.net/browse/GAME-3154
        if( LumosityUtils.isFitTestModeOn() )
        {
            this.model.persistentData.isPracticeMode = false;
        }
        this.onRestart();
    },

    gamePauseMenu_quitGame: function()
    {
        this.onFinish( true );
    },

    skipTutorialButton_Tapped: function()
    {
        this.onCompleteTutorial( true );
    },

    howToPlay_playNowTapped: function()
    {
        if( this.settings.shouldShowHowToPlayOnStart() && this.model.isOnBeginCalled === false )
        {
            this.view.centerOnGameStage();
            this.onBegin();
        }
        else
        {
            this.onRestart();
        }
    },

    howToPlay_backNowButtonTapped: function()
    {
        if( this.settings.shouldShowHowToPlayOnStart() && this.model.isOnBeginCalled === false )
        {
            this.onFinish( true );
        }
        else
        {
            this.view.centerOnGameMenu();
        }
    },

    showHowToPlay: function( instant )
    {
        if ( this.view.getDisplayState() === DisplayState.PauseMenu || this.view.getPrevDisplayState() === DisplayState.App )
        {
            if( this.settings.hasTutorial )
            {
               this.onStartTutorial( this.view.getPrevDisplayState() === DisplayState.App );
            }
            else
            {
                this.view.centerOnSideStage( instant );
            }
        }
    },

    /**
     * Convenience for getting the active scoring component.
     * 
     * @return {object} Will return a GameComponentScoreMultiplier for speed scoring types or
     *                              a GameComponentScore for manual score type.
     *                  or [AnotherClassName] for others.
     */
    getScoringComponent: function()
    {
        "use strict";

        // GameComponent Score Multiplier
        if( this.settings.scoreType === 1 )
        {
            return this._components[ GameComponentScoreMultiplier.prototype.getType() ];
        }
        // GameComponent Score
        else if( this.settings.scoreType === 2 )
        {
            return this._components[ GameComponentScore.prototype.getType() ];
        }
        return null;
    },
    
    _pauseSwiped:function( up )
    {
        if ( up )
        {
            //correct order for initial swipe, advance the swipe state
            if ( !this.model.pauseDebugSwipeStep || this.model.pauseDebugSwipeStep <= 0 )
            {
                this.model.pauseDebugSwipeStep = 1;
            }
            //correct order for last swipe, tell the view to show the debug info
            else if ( this.model.pauseDebugSwipeStep >= 2 )
            {
                this.view.showDebugInfo();
            }
            //incorrect order, reset the swipe state
            else
            {
                this.model.pauseDebugSwipeStep = 0;
            }
        }
        else
        {
            //correct order for second swipe, advance the state
            if ( this.model.pauseDebugSwipeStep === 1 )
            {
                this.model.pauseDebugSwipeStep = 2;
            }
            //incorrect order, reset the swipe state
            else
            {
                this.model.pauseDebugSwipeStep = 0;
            }
        }
    },

    getMetadataComponent: function()
    {
        return this._components[ GameComponentMetadata.prototype.getType() ];
    },

    getAudioMixerComponent: function()
    {
        return this._components[ GameComponentAudioMixer.prototype.getType() ];
    },

    beginTrial: function()
    {
        this.getMetadataComponent().beginMetadataTrial();
    },

    endTrial: function( trialObject, trialCorrect, switchedTrial, congruentStimuli, switching, inhibition )
    {
        this.getMetadataComponent().endMetadataTrial( trialObject, trialCorrect, switchedTrial, congruentStimuli, switching, inhibition );
    },

    configureMetaDataKeys: function()
    {
        //
        // "OVERRIDE" ME BRO
        //
        return undefined;
    },

    finalizeMetadata: function()
    {
        //
        // "OVERRIDE" ME BRO
        //
    }
});
