/*global require*/
/*global cc*/
/*global audio*/
/*global lumosity*/
/*global GameComponent*/
/*global GameWidgetHud*/
/*global LumosityUtils*/

require( "src/Game/Component/GameComponent.js" );
require( "src/Game/Widget/GameWidgetHud.js" );
require( "src/Utility/LumosityUtils.js" );
//
// These are variable keys for setting the metadata with an explanation of what
// each key is supposed to represent. We may have a need to add keys to this list
// specific to every game. In that case, I would suggest appending keys at runtime
// in the ctor() of your GameModel.
// 
// NOTE: Some of the keys like "tutorial"* don't actually exist on mobile yet,
// thought there might be a need for these keys later.
//
var metadataAggregateKeys = 
{
    numTries:                 "num_tries",                    // Number of trials.
    numCorrect:               "num_correct",                  // Number of correct trials.
    numTriesInhibition:       "num_tries_inhibition",         // Number of trials with incongruent stimuli.
    numCorrectInhibition:     "num_correct_inhibition",       // Number of correct trials with incongruent stimuli.
    numTriesSwitch:           "num_tries_switch",             // Number of trials where the attended attribute was switched.
    numCorrectSwitch:         "num_correct_switch",           // Number of correct trials where the attended attribute was switched.
    responseTime:             "response_time",                // Average response time for all trials in milliseconds.
    responseTimeCorrect:      "response_time_correct",        // Average response time on correct trials in milliseconds.
    responseTimeIncorrect:    "response_time_incorrect",      // Average response time on incorrect trials in milliseconds.
    responseTimeInhibition:   "response_time_inhibition",     // Average response time on incongruent stimuli in milliseconds.
    responseTimeNoInhibition: "response_time_no_inhibition",  // Average response time on switched trials in milliseconds.
    responseTimeSwitch:       "response_time_switch",         // Average response time on incongruent stimuli in milliseconds.
    responseTimeNoSwitch:     "response_time_no_switch",      // Average response time on switched trials in milliseconds time.
    endBonus:                 "end_bonus",                    // Number of bonus points awarded at the end of the game.
    tutorialStart:            "tutorial_start",               // The number of times someone clicked "yes" on the tutorial dialog.
    tutorialDecline:          "tutorial_decline",             // The number of times someone clicked "no" on the tutorial dialog.
    tutorialFinish:           "tutorial_finish",              // The number of times the tutorial was completed.
    tutorialErrors:           "tutorial_errors",              // The number of errors made during the tutorial.
    boardLayout:              "board_layout",                 // Initial layout of the board (dependent on game).
    userLevel:                "user_level",                   // Max level reached.
    sessionLevel:             "session_level",                // Max level reached in previous session.
    startLevel:               "start_level",                  // First level played for the current session.
    endLevel:                 "end_level",                    // Final level reached.
    thresholdLevel:           "threshold",                    // Staircasing threshold level.
    maxLevel:                 "max_level",                    // Max level reached.
    aidedNumCorrect:          "aided_num_correct",            // Number of correct trials that included hints
    aidedNumTries:            "aided_num_tries",              // Number of trials that included hints
    aidedResponseTime:        "aided_response_time",          // Average response time for hinted trials in milliseconds
    time:                     "time",                         // Time in milliseconds from start of gameplay (end of countdown) to end of gameplay, subtracting pauses
    numPauses:                "num_pauses",                   // Number of times paused
    pauses:                   "pauses",                       // Time offsets in milliseconds when game was paused, semicolon delimited
    num_defocus:              "num_defocus",                  // Number of times the app lost device's focus
    num_restart:              "num_restart",                  // Number of times the game was restarted
    screenWidth:              "width",                        // Screen width in pixels
    screenHeight:             "height",                       // Screen height in pixels
    avgLevel:                 "avg_level",                    // Average level
    gameMode:                 "game_mode",                    // What type of game this is (Standard, Fit Test, or Practice Mode)
    version:                  "version",                      // Git commit ID (TODO)
    localeId:                 "locale_id",                    // ID of locale game played (TODO)
    defaultLocaleId:          "default_locale_id",            // ID of default locale of user (TODO)
    tutorialCorrect:          "num_correct_tutorial",         // Number of correct trials in Practice Mode.
    tutorialIncorrect:        "num_incorrect_tutorial",       // Number of incorrect trials in Practice Mode.
    accuracy:                 "accuracy",                     // Accuracy of trial in FitTest.
    fitTest:                  "fit_test",                     // "T" iff its on fitTest mode.
    tutorial:                 "tutorial",                     // Metadata object reported iff its on fitTest mode.
    numMatch:                 "num_match",                    // Number of consecutively matching stimuli
    responseTimeMatch:        "response_time_match"           // Average response time between consecutively matching stimuli.

};
//
// Just like metadataAggregateKeys above,   but for more game-specific per trial
// information. Should probably extend this list of keys in GameModel.
//
var metadataTrialKeys = 
{
    trialCSV:     "trial_csv",     // Key for trial data (exported in csv format).
    correct:      "correct",       // Was the trial correctly answered? (boolean).
    responseTime: "response_time", // Response time in milliseconds.
    timeOffset:   "time_offset",   // Number of milliseconds at start of trial since beginning of game.
    multiplier:   "multiplier"     // Multiplier applied to the current score increment.
};
//
// 
//
var GameComponentMetadata;
( function()
{
    "use strict";
    GameComponentMetadata = GameComponent.extend(
    {
        /** The average time it takes the player to respond in a trial (in milliseconds). */
        _averageResponseTime: 0,

        /** The average time it takes the player to respond in a trial (in milliseconds)
          * for trials that have a hint up. */
        _aidedResponseTimes: null,

        /** Collects the response times of correct answers with hints showing. */
        _aidedCorrectTimes: null,

        /** The average response time on correct trials (in milliseconds). */
        _averageResponseTimeCorrect: 0,

        /** Total number of correct trials featuring incongruent stimuli. */
        _incongruentStimuliCorrect: 0,

        /** Total number of trials in a game (should always be > 1). */
        _trialCount: 0,

        /** An array of trial dictionary objects that contain relevant data about each trial. */
        _trials: null,

        /** Response times in a given trial. */
        _trialResponseTimes: null,

        /** Collects the response times of correct answers. */
        _trialCorrectTimes: null,

        /** Collects the response times to incorrect answers. */
        _trialIncorrectTimes: null,

        /** Times where responses were recorded in the presence of congruent stimuli. */
        _trialCongruentTimes: null,

        /** Records time where responses recorded are from incongruent stimuli. */
        _trialIncongruentTimes: null,

        /** Recorded times where the trial stimuli has switched. */
        _trialSwitchTimes: null,

        /** Recorded times where the trial stimuli doesn't change. */
        _trialNoSwitchTimes: null,

        /** The time between beginTrial and endTrial. (in milliseconds) */
        _responseTime: 0,

        /** An array of times that records the number of milliseconds at start of trial since beginning of game.  */
        _trialTimeOffsets: null,

        /** Total running time of all trials. */
        _runningTrialTime: 0,

        /** Tally of correct trials when the stimuli switched. */
        _numberCorrectSwitches: 0,

        /** Object defined in controller that saves the state of metadata occuring in practice mode. */
        tutorialMetadata: 0,

        /**
         * How about setting _switching and _inhibition to false to avoid garbage metadata?
         */
        _switching: null,

        _inhibition: null,

        _usingLevels: false,

        _trialLevels: null,

        /* Records the time at which a new trial begins. Updated in beginMetadataTrial(). */
        _trialTimeOffset: 0,

        /** @lends GameComponentTimer.prototype */
        
        ctor:function( controller )
        {
             this._super( controller );

            var settings = this._controller.settings;
            this.time = settings.gameLengthInSeconds * 1000;

            this._aidedResponseTimes = [];
            this._aidedCorrectTimes = [];
            this._trials = [];
            this._trialResponseTimes = [];
            this._trialCorrectTimes = [];
            this._trialIncorrectTimes = [];
            this._trialCongruentTimes = [];
            this._trialIncongruentTimes = [];
            this._trialSwitchTimes = [];
            this._trialNoSwitchTimes = [];
            this._trialTimeOffsets = [];
            this._trialLevels = [];

        },

        /**
         * This wraps endMetadataTrial
         * This function simply stomps the internal timing variables with game reported times and proceeds as usual
         * Useful for parallel/multiple concurrent trials (like Train of Thought)
         * @param {Object} trialObject A dictionary containing all the different trial details for this particular trial
         * @param {boolean} trialCorrect Whether or not this trial was a correct one.
         * @param {int} startTime for this trial should be stored away by the game and sent back here to override time_offset
         */
        endTrialWithTimes: function( trialObject, trialCorrect, startTime)
        {
            // easier to store start tiem wih trialData and stomp it here
            this._trialTimeOffset =  this.getRunningTime();
            this._responseTime =  (this.getRunningTime() - startTime) / 1000;
            this.endMetadataTrial(trialObject, trialCorrect, false, false, false, false);
        },

        /**
         * Begins a new trial. Make sure to call this every time before calling endMetadataTrial().
         */
        beginMetadataTrial: function()
        {
            this._responseTime = 0;
            this._trialCount += 1;
            this._trialTimeOffset = ( this._runningTrialTime > 0 ) ? Math.round( LumosityUtils.convertSecondsToMilliseconds( this._runningTrialTime ) ) : 0;
            this._trialTimeOffsets.push( this._runningTrialTime );
        },

        /**
         * Ends a trial, and stores off its metadata.
         *
         * @param {Object} trialObject A dictionary containing all the different trial details for this particular trial. See the Games wiki for details.
         * @param {boolean} trialCorrect Whether or not this trial was a correct one.
         *
         * @param {boolean} switchedTrial Whether or not the sort of stimulus "switched", maybe? For example, in Ebb and Flow, when the leaves change color...
         * @param {boolean} congruentStimuli WTF I dunno. the opposite of switchedTrial?  
         * @param {boolean} switching Uh, is this the same as switchedTrial?  Set exactly to false to avoid irrelevant data from undefined triggering "flagUndefined".
         * @param {boolean} inhibition The user is not doing the wrong thing?  Set exactly to false to avoid irrelevant data from undefined triggering "flagUndefined".
         *
         * The last four parameters are highly game-specific, so could be cleaner to refactor into those specific games' trialObjects.  
         */
        endMetadataTrial: function( trialObject, trialCorrect, switchedTrial, congruentStimuli, switching, inhibition )
        {

            this._switching = switching;
            this._inhibition = inhibition;

            if ( !this._controller.model.persistentData.isPracticeMode )
            {
                this._trialResponseTimes.push( this._responseTime );

                if ( trialObject.aided === 'T' )
                {
                    this._aidedResponseTimes.push( this._responseTime );
                }
            }
            else
            {
                this._controller.tutorialMetadata._trialResponseTimes.push( this._responseTime ); 
            }
            
            if ( trialCorrect === true )
            {
                if ( this._controller.model.persistentData.isPracticeMode )
                {
                    this._controller.tutorialMetadata.correct += 1;

                    this._controller.tutorialMetadata._trialCorrectTimes.push( this._responseTime );
                }
                else
                {
                    this._trialCorrectTimes.push( this._responseTime );

                    if ( trialObject.aided === 'T' )
                    {
                        this._aidedCorrectTimes.push( this._responseTime );
                    }
                }
            }
            else
            {
                if ( this._controller.model.persistentData.isPracticeMode )
                {
                    this._controller.tutorialMetadata.incorrect += 1;
                    this._controller.tutorialMetadata._trialIncorrectTimes.push( this._responseTime );
                }
                else
                {
                    this._trialIncorrectTimes.push( this._responseTime );
                }
            }
            
            var metadataObject = this._controller.model.persistentData.isPracticeMode ? this._controller.tutorialMetadata : this;

            if ( congruentStimuli === true )
            {
                metadataObject._trialCongruentTimes.push( this._responseTime );
            }
            else if ( congruentStimuli === false )
            {
                if ( trialCorrect === true )
                {
                    metadataObject._incongruentStimuliCorrect += 1;
                }
                metadataObject._trialIncongruentTimes.push( this._responseTime );
            }

            if ( switchedTrial === true )
            {
                if ( trialCorrect === true )
                {
                    metadataObject._numberCorrectSwitches += 1;
                }
                metadataObject._trialSwitchTimes.push( this._responseTime );
            }
            else
            {
                metadataObject._trialNoSwitchTimes.push( this._responseTime );
            }

            if ( trialObject.level !== undefined && !this._controller.model.persistentData.isPracticeMode )
            {
                this._usingLevels = true;
                metadataObject._trialLevels.push( trialObject.level );
            }

            trialObject[ metadataTrialKeys.responseTime ] = Math.round( LumosityUtils.convertSecondsToMilliseconds( this._responseTime ) );
            trialObject[ metadataTrialKeys.timeOffset ] = this._trialTimeOffset;

            metadataObject._trials.push( trialObject );

        },

        getType: function()
        {
            return "GameComponentMetadata";
        },

        onUpdate: function( delta )
        {
            var model = this._controller.model;
            if ( model.isGameRunning )
            {   
                if ( ( model.isGamePaused || model.isGameOver ) === true ){ return; }

                this._responseTime += delta;
                this._runningTrialTime += delta;
            }
        },

        /**
         * Gets called when the game is over. This should be called just before we export results.
         * This function makes metadata rigid.  It bloats many special cases by default that need to be disabled instead of being disabled by default.
         *
         * One improvement would be for each key to ignore writing a value if they already contains a defined value.  Test case:  2014-12-13 Trouble Brewing has a clean and completely portable Metadata class that manages all the game-specific values, including "num_correct", "num_trials".  This generic metadata stomps those values with a complicated, rigid expectation to use the beginMetadataTrial and endMetadataTrial functions.  This component's onEndGameplay function is called after "finalizeMetadata" function, so there is no forward-compatible place to overwrite the metadata's bloated "onEndGameplay" values.
         *
         * GAME-4074 If trial_csv is already defined, do not overwrite it.  Test case:  2015-07-27 Editor's Choice. End. Aaron Kaluszka expects sub-arrays are semicolon-delimited.  Got comma-delimited.
         *
         * GAME-4074 Always record tutorial starts and finishes if tutorial metadata is defined.  Test case:  2015-07-27 Editor's Choice. End. Aaron Kaluszka expects to find tutorial_starts and tutorial_finishes.
         */
        onEndGameplay: function()
        {
            var model = this._controller.model;
            var windowSize = cc.Director.getInstance().getWinSize();
            //
            // Aggregate data.
            //
            if ( model.persistentData.isPracticeMode )
            {
                this.tutorialMetadata.time                                    = Math.round( this._runningTrialTime * 1000 );
                if ( model.metadata.num_match )
                {
                    this.tutorialMetadata.num_match                           = model.metadata.num_match;
                }
                if ( model.metadata.response_time_match )
                {
                    this.tutorialMetadata.response_time_match                 = model.metadata.response_time_match;
                }
            }

            model.metadata[ metadataAggregateKeys.time ]                      = Math.round( this._runningTrialTime * 1000 );
            model.metadata[ metadataAggregateKeys.numPauses ]                 = model.pauseTimes.length;
            model.metadata[ metadataAggregateKeys.pauses ]                    = this.getPauseList( model ).join();
            model.metadata[ metadataAggregateKeys.num_defocus ]               = model.num_defocus;
            model.metadata[ metadataAggregateKeys.num_restart ]               = model.persistentData.num_restart;

            model.metadata[ metadataAggregateKeys.screenWidth ]               = windowSize.width;
            model.metadata[ metadataAggregateKeys.screenHeight ]              = windowSize.height;

            model.metadata[ metadataAggregateKeys.numTries ]                  = this._trialCount;

            if (this.tutorialMetadata && this.tutorialMetadata.starts) {
                model.metadata[ metadataAggregateKeys.tutorialStart ]         = this.tutorialMetadata.starts;
                model.metadata[ metadataAggregateKeys.tutorialFinish ]        = this.tutorialMetadata.finishes;
            }
            if ( LumosityUtils.isFitTestModeOn() )
            {
                model.metadata[ metadataAggregateKeys.fitTest ]               = "T";
            }
            else if ( model.persistentData.isPracticeMode )
            {
                model.metadata[ metadataAggregateKeys.tutorialStart ]         = this.tutorialMetadata.starts;
                model.metadata[ metadataAggregateKeys.tutorialFinish ]        = this.tutorialMetadata.finishes;
                model.metadata[ metadataAggregateKeys.fitTest ]               = "F";
            }
            else
            {
                model.metadata[ metadataAggregateKeys.fitTest ]               = "F";
            }

            model.metadata[ metadataAggregateKeys.numCorrect ]                = this._trialCorrectTimes.length;

            if (this._aidedResponseTimes.length > 0)
            {
                model.metadata[ metadataAggregateKeys.aidedNumCorrect ]       = this._aidedCorrectTimes.length;
                model.metadata[ metadataAggregateKeys.aidedNumTries ]         = this._aidedResponseTimes.length;
                model.metadata[ metadataAggregateKeys.aidedResponseTime ]     = this._calculateAverageTimes( this._aidedResponseTimes );  
            }

            if ( this._inhibition )
            {
                model.metadata[ metadataAggregateKeys.numTriesInhibition ]    = this._trialIncongruentTimes.length;
                model.metadata[ metadataAggregateKeys.numCorrectInhibition ]  = this._incongruentStimuliCorrect;
            }

            if ( this._switching )
            {
                model.metadata[ metadataAggregateKeys.numTriesSwitch ]        = this._trialSwitchTimes.length;
                model.metadata[ metadataAggregateKeys.numCorrectSwitch ]      = this._numberCorrectSwitches;
            }

            if ( this._inhibition )
            {
                model.metadata[ metadataAggregateKeys.responseTimeIncorrect ]    = this._calculateAverageTimes( this._trialIncorrectTimes );
                model.metadata[ metadataAggregateKeys.responseTimeInhibition ]   = this._calculateAverageTimes( this._trialIncongruentTimes );
                model.metadata[ metadataAggregateKeys.responseTimeNoInhibition ] = this._calculateAverageTimes( this._trialCongruentTimes );
            }

            if ( this._switching )
            {
                model.metadata[ metadataAggregateKeys.responseTimeSwitch ]    = this._calculateAverageTimes( this._trialSwitchTimes );
                model.metadata[ metadataAggregateKeys.responseTimeNoSwitch ]  = this._calculateAverageTimes( this._trialNoSwitchTimes );
            }

            if ( this._usingLevels )
            {
                model.metadata[ metadataAggregateKeys.sessionLevel ]          = this._trialLevels[0];
                model.metadata[ metadataAggregateKeys.userLevel ]             = this._trialLevels[this._trialLevels.length - 1];
                model.metadata[ metadataAggregateKeys.avgLevel ]              = this._calculateAverageLevel( this._trialLevels );
            }
            else
            {
                // Don't report times in aggregate metadata for Level-Based games
                if ( !model.persistentData.isPracticeMode )
                {
                    model.metadata[ metadataAggregateKeys.responseTime ]          = this._calculateAverageTimes( this._trialResponseTimes );
                    model.metadata[ metadataAggregateKeys.responseTimeCorrect ]   = this._calculateAverageTimes( this._trialCorrectTimes );
                    model.metadata[ metadataAggregateKeys.responseTimeIncorrect ]   = this._calculateAverageTimes( this._trialIncorrectTimes );
                }
            }

            if ( LumosityUtils.isFitTestModeOn() || ( !LumosityUtils.isFitTestModeOn() && model.persistentData.isPracticeMode ) )
            {
                model.metadata[ metadataAggregateKeys.tutorial ]              = {
                                                                                  num_correct: this.tutorialMetadata.correct,
                                                                                  num_tries: this.tutorialMetadata.correct
                                                                                              + this.tutorialMetadata.incorrect,
                                                                                  num_errors: this.tutorialMetadata.incorrect,
                                                                                  time: this.tutorialMetadata.time
                                                                                };

                // Don't report times in aggregate metadata for Level-Based games
                if ( !this._usingLevels && !model.persistentData.isPracticeMode )
                {
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.responseTime ]        = this._calculateAverageTimes( this.tutorialMetadata._trialResponseTimes );
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.responseTimeCorrect ] = this._calculateAverageTimes( this.tutorialMetadata._trialCorrectTimes );
                }

                if ( this._inhibition )
                {
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.numTriesInhibition ]    = this.tutorialMetadata._trialIncongruentTimes.length;

                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.numCorrectInhibition ]  = this.tutorialMetadata._incongruentStimuliCorrect;
                    
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.responseTimeIncorrect ] = this._calculateAverageTimes( this.tutorialMetadata._trialIncorrectTimes );
                }

                if ( model.metadata.num_match !== undefined )
                {
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.numMatch ]              = this.tutorialMetadata.num_match;
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataAggregateKeys.responseTimeMatch ]     = this.tutorialMetadata.response_time_match; 
                }

                if (!(metadataTrialKeys.trialCSV in model.metadata[ metadataAggregateKeys.tutorial ] )) {
                    model.metadata[ metadataAggregateKeys.tutorial ][ metadataTrialKeys.trialCSV ] = this._formatTrialsData( true );
                }
            }

            if ( model.bonus > 0 )
            {
                model.metadata[ metadataAggregateKeys.endBonus ]                  = model.bonus;
            }

            if ( model.maxLevel !== undefined )
            {
                model.metadata[ metadataAggregateKeys.maxLevel ]                  = model.maxLevel;
            }
            
            //
            // Trials data.
            //
            if (!(metadataTrialKeys.trialCSV in model.metadata )) {
                model.metadata[ metadataTrialKeys.trialCSV ] = this._formatTrialsData();
            }
        },

        getPauseList: function( model )
        {
            var i;
            var pauseList = [];
            for ( i = 0; i < model.pauseTimes.length; i++ )
            {
                pauseList.push( model.pauseTimes[i] + ":" + model.resumeTimes[i] + ":" + model.pauseTriggers[i] );
            }
            return pauseList;
        },

        getResponseTime: function( milliseconds )
        {
            if ( milliseconds === undefined )
            {
                milliseconds = true;
            }

            if ( milliseconds )
            {
                return LumosityUtils.convertSecondsToMilliseconds( this._responseTime );
            }
            else
            {
                return this._responseTime;
            }
        },

        getRunningTime: function ( milliseconds )
        {
            if ( milliseconds === undefined )
            {
                milliseconds = true;
            }

            if ( milliseconds )
            {
                return LumosityUtils.convertSecondsToMilliseconds( this._runningTrialTime );
            }
            else
            {
                return this._runningTrialTime;
            }
        },

        /**
         * This function is redundant to LumosityUtils.jsToCsv 
         * except for the tutorial and header special cases.
         *
         * GAME-4074 Sub-arrays are converted to semicolon-separated values.  Test case:  2015-07-28 Editor's Choice metadata.  Aaron Kaluszka expects semicolon-separated nested values.  Got comma separated values, which get confused with the 
         */
        _formatTrialsData: function( isFormattingTutorial )
        {
            var i = 0;
            var j = 0;
            var trial = null;
            var trialString = "";
            var header = "";
            var trialHeader = "";
            var headers = this._getTrialHeaders( isFormattingTutorial ); 
            if( headers.length <= 0 ){ return trialString; }
            //
            // Add the headers to our string first.
            //
            for( i = 0; i < headers.length; i += 1 )
            {
                header = headers[ i ];
                if( i !== ( headers.length - 1 ) )
                {
                    header += ",";
                }

                trialString = trialString.concat( header );
            }

            var trials = isFormattingTutorial ? this.tutorialMetadata._trials : this._trials;

            //
            // Add the formatted trial data.
            //
            for( i = 0; i < trials.length; i += 1 )
            {
                trial = trials[ i ];
                trialString = trialString.concat( "/" );

                for( j = 0; j < headers.length; j += 1 )
                {
                    trialHeader = trial[ headers[ j ] ];
                    if (this._isArray(trialHeader)) {
                        trialHeader = trialHeader.join(";");
                    }
                    // Only add a comma if we are not at the end of the trial data.
                    if( j !== ( headers.length - 1 ) )
                    {
                        trialHeader += ",";
                    }
                    trialString = trialString.concat( trialHeader );
                }
            }

            return trialString;
        },

        // http://blog.niftysnippets.org/2010/09/say-what.html
        _isArray: function(value)
        {
            return "[object Array]" == Object.prototype.toString.call(value);
        },

        _getTrialHeaders: function( isFormattingTutorial )
        {
            var trials = isFormattingTutorial ? this.tutorialMetadata._trials : this._trials;
            var trial = trials[ 0 ];
            var key = "";
            var headers = [];
            for( key in trial )
            {
                if( trial.hasOwnProperty( key ) )
                {
                    if ( isFormattingTutorial )
                    {
                        if ( key !== "multiplier" && key !== "progress" )
                        {
                            headers.push( key );
                        }
                    }
                    else
                    {
                        headers.push( key );
                    }
                }
            }
            return headers;
        },

        _calculateAverageLevel: function( times )
        {
            var index = 0;
            var runningTotal = 0;
            var arrLength = times.length;

            if( arrLength <= 0 ) { return 0; }

            for( index = 0; index < arrLength; index += 1 )
            {
                runningTotal += times[ index ];
            }

            return Math.round(runningTotal / arrLength);

        },

        _calculateAverageTimes: function( times )
        {
            var index = 0;
            var runningTotal = 0;
            var arrLength = times.length;
            var result = 0;

            if( arrLength <= 0 ) { return 0; }

            for( index = 0; index < arrLength; index += 1 )
            {
                runningTotal += times[ index ];
            }

            result = LumosityUtils.convertSecondsToMilliseconds( runningTotal );
            result = result / arrLength;
            result = Math.round( result );

            return result;
        }
    });
}());
