/*global require*/
/*global cc*/
/*global audio*/
/*global GameComponent*/
/*global GameWidgetHud*/

require( "src/Game/Component/GameComponent.js" );
require( "src/Game/Widget/GameWidgetHud.js" );

var GameComponentTimer;
( function()
{
	"use strict";

	GameComponentTimer = GameComponent.extend(
	{
        _timeRemaining: 0,
        _timeTotal: 0,
        _currDelta: 0,
		_hasTimerEnded: false,
        _isComponentPaused: false,

	    /** @lends GameComponentTimer.prototype */
	    
	    /**
	     * A GameComponentTimer is a subclass of GameComponent that can be used to set up games with
	     * limited time. Provides functionality to be fired every frame or every second.
	     * @constructs
	     */
	    ctor:function( controller )
	    {
	        this._super( controller );

            var model = this._controller.model;
            var settings = this._controller.settings;
            var view = this._controller.view;
            var hudWidget = view._widgets[ GameWidgetHud.prototype.getType() ];

            // Initialize the model and hud widget timer value
            this._timeTotal = settings.getGameLengthInSeconds();
            this._timeRemaining = this._timeTotal;
            this._suppressTimerAudio = settings.suppressTimerAudio;
            model.timeRemaining = this._timeTotal;
            hudWidget.setTimeRemaining( this._timeTotal );
	    },

        /**
         * @returns {[float]} Seconds consumed since start, including fraction of a second.
         *
         * Example @see Disillusion.
         */
        getTimeConsumed:function()
        {
            return this._timeTotal + this._currDelta - this._timeRemaining;
        },

        /**
         * @param   seconds {float} Set total and remaining time.
         *
         * Example @see Disillusion.
         */
        reset:function(seconds)
        {
            seconds = (seconds != undefined || seconds === 0) ? seconds : this._timeTotal;
            this._timeRemaining = seconds;
            this._timeTotal = seconds;
            this._currDelta = 0;
        },

        /**
         * Returns Component type as a string.
         *
         * @returns {string}
         */
        getType:function()
        {
            return "GameComponentTimer";
        },

        /**
         * Sets the paused/unpaused state of the component.
         * 
         * @param {Boolean} isPaused
         */
        setIsComponentPaused: function( isComponentPaused )
        {
            this._isComponentPaused = isComponentPaused;
        },

        /**
         * Returns the paused/unpaused state of the component.
         * 
         * @returns {Boolean}
         */
        getIsComponentPaused: function()
        {
            return this._isComponentPaused;
        },

        /**
         * Updates the game each frame.
         * 
         * @param  {[float]} delta
         */
        onUpdate:function( delta )
        {
            var model = this._controller.model;

            if( this._hasTimerEnded === true )
            {
                return;
            }

            if( model.isGameRunning === false )
            {
                return;
            }

            if( model.isGamePaused === true )
            {
                return;
            }

            if( this._isComponentPaused === true )
            {
                return;
            }

            this._currDelta += delta;

            if( this._currDelta >= 1.0 )
            {
                this._currDelta = this._currDelta - 1.0;
                this._updateTimer();
            }
        },

        /**
        *  Updates the visual timer.
        */
        _updateTimer: function()
        {
            var view = this._controller.view;
            var hudWidget = view._widgets[ GameWidgetHud.prototype.getType() ];
            var model = this._controller.model;

            this._timeRemaining -= 1;
            model.timeRemaining = this._timeRemaining;

            hudWidget.setTimeRemaining( model.timeRemaining );
            this._controller.onComponentTimerTick( model.timeRemaining );

            if( model.timeRemaining <= 0 )
            {
                this._hasTimerEnded = true;
                this._controller.onComponentTimerEnd();
            }
            else if ( model.timeRemaining <= 5 && model.timeRemaining > 0 && !this._suppressTimerAudio )
            {
                audio.playEffect( "sounds/timer.mp3" );
            }
        }
	} );
}() );
