/*global cc*/
/*global require*/
var audio = cc.AudioEngine.getInstance();

require( "src/Game/Widget/GameWidget.js" );
/*global GameWidget*/

require( "src/View/LumosityLabel.js" );
/*global LumosityLabel*/

require( "src/Utility/LumosityUtils.js" );
/*global LumosityUtils*/

require( "src/View/LumositySprite.js" );
/*global LumositySprite*/

var GameWidgetLevelCard = GameWidget.extend(
{
    /** @lends GameWidgetLevelCard.prototype */

    /** Background sprite. */
    _backgroundSprite:null,

    /** Text object for level label. */
    _levelLabel:null,

    /** Text object for level value. */
    _levelValue:null,

    /** Default desired content size. */
    _desiredContentSize:cc.size( 390, 390 ),

    /**
     * Returns Widget type as a string.
     *
     * @returns {string}
     */
    getType: function()
    {
        return "GameWidgetLevelCard";
    },
    
    /**
     *  A GameWidgetLevelCard is shown to display the game's current level.
     *
     *  @constructs
     *  @param {Object} settings An object that contains popup-related settings.
     *  @param {Object} [delegate] An object to call popup-related functions on.
     *  @param {boolean} [preload=true] Whether or not to preload all popup images upon creation.
     */
    ctor:function( settings, delegate, preload )
    {
        this._super();
        
        this._delegate = delegate;

        //by default, preload all our images on creation
        if ( ( preload === undefined || preload ) && settings !== undefined )
        {
            this.preload( settings );
        }

        // Create sub views
        this._createViews();

        // Make sure it is positioned dead center
        this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
    },

    /**
     * Sets text for level label.
     * @param text  Valid string.
     */
    setLevelLabel:function( text )
    {
        if( text && this._levelLabel )
        {
            this._levelLabel.setString( text );
        }
    },

    /**
     * Sets text for level value.
     * @param text  Valid string.
     */
    setLevelValue:function( text )
    {
        if( text && this._levelValue )
        {
            this._levelValue.setString( text );
        }
    },

    /**
     * Sets the color of the widget's text labels.
     *
     * @param colorOrR  Either a color object or the red element of four channel arguments (0-255).
     * @param g         If four channel arguments supplied, the value in the green channel (0-255).
     * @param b         If four channel arguments supplied, the value in the blue channel (0-255).
     * @param a         If four channel arguments supplied, the value in the alpha channel (0-255).
     */
    _setTextColor:function( textElement, colorOrR, g, b, a )
    {
        var color;

        if ( ( colorOrR.r !== undefined ) || ( colorOrR.r !== null ) )
        {
            color = colorOrR;
        }
        else if ( colorOrR.Red )
        {
            color = cc.c4b( colorOrR.Red, colorOrR.Green, colorOrR.Blue, colorOrR.Alpha );
        }
        else
        {
            color = cc.c4b( colorOrR, g, b, a );
        }

        if ( textElement )
        {
            textElement.setColor( color );
        }
    },

    setTextColor: function( color )
    {
        if ( this._levelLabel )
        {
            this._setTextColor( this._levelLabel, color );
        }

        if ( this._levelValue )
        {
            this._setTextColor( this._levelValue, color );
        }
    },

    _setTextStyle: function ( textElement, font, fontSize, fontColor )
    {
        if ( textElement )
        {
            textElement.setFontName( LumosityUtils.getFontLocation( font ) );
            textElement.setFontSize( fontSize );

            this._setTextColor( textElement, fontColor );
        }
    },

    setLabelTextStyle: function ( font, fontSize, fontColor )
    {
        if ( this._levelLabel )
        {
            this._setTextStyle( this._levelLabel, font, fontSize, fontColor );
        }
    },

    setValueTextStyle: function ( font, fontSize, fontColor )
    {
        if ( this._levelValue )
        {
            this._setTextStyle( this._levelValue, font, fontSize, fontColor );
        }
    },

    /**
     * Preloads images
     *
     * @param settings Game settings that provides
     */
    preload:function( settings )
    {
    },

    /**
     * Because the countdown is always full screen and can merely be positioned
     * (something that should be changed in the future probably), we normally
     * don't want to allow anyone to change the content size. If someone does,
     * then we need to stop it. TODO - just fix the whole widget.
     *
     * @param {cc.Size} size The content size to set.
     * @param {bool} [force=false] If true, will actually set the size. Otherwise it won't. :-P
     */
     setContentSize:function( size, force )
     {
         if ( force )
         {
             this._super( size );
         }
     },

    setDesiredContentSize: function( size )
    {
        this._desiredContentSize = cc.size( size.width, size.height );
        this._updateContentSize();
    },

    /**
     * Creates views.
     *
     * @private
     */
    _createViews: function()
    {
        // Background image
        this._backgroundSprite = new LumositySprite( "levelCardBackground.png" );

        this._backgroundSprite.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        this.addChild( this._backgroundSprite );

        this._levelLabel = LumosityLabel.create( "LEVEL", "fonts/MuseoSans-700.ttf", 40 );
        this._levelLabel.setColor( cc.c4b( 88, 57, 25, 255 ) );
        this._levelLabel.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        this._levelLabel.setPosition( cc.p( 0, 110 ) );
        this._levelLabel.setHorizontalAlignment( cc.TEXT_ALIGNMENT_CENTER );
        this._levelLabel.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
        this.addChild( this._levelLabel );

        this._levelValue = LumosityLabel.create( "22", "fonts/MuseoSans-500.ttf", 220 );
        this._levelValue.setColor( cc.c4b( 88, 57, 25, 255 ) );
        this._levelValue.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        this._levelValue.setPosition( cc.p( 0, -10 ) );
        this._levelValue.setHorizontalAlignment( cc.TEXT_ALIGNMENT_CENTER );
        this._levelValue.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
        this.addChild( this._levelValue );

        this._updateContentSize();
    },

    /**
     * Updates size to match desired content size.
     * @private
     */
    _updateContentSize:function()
    {
        if( this._backgroundSprite )
        {
            this._backgroundSprite.setContentSize( this._desiredContentSize, true, true );
        }
    }
});
