/*global cc*/
/*global require*/
/*global GameWidget*/
/*global GameWidgetFlipbook*/
/*global AnimationController*/
/*global LumosityButtonState*/
/*global LumosityButton*/
/*global LumosityUtils*/
/*global LayoutLoader*/
/*global AnimationChanges*/
/*global DisplayState*/

require( "src/Game/Widget/GameWidget.js" );
require( "src/Game/Widget/GameWidgetFlipbook.js" );
require( "src/Animation/AnimationController.js" );
require( "src/View/LumosityButton.js" );
require( "src/Utility/LumosityUtils.js" );
require( "src/Animation/LayoutLoader.js" );

var GameWidgetPauseButton;

( function()
{
    "use strict";

    GameWidgetPauseButton = GameWidget.extend(
    {
        /** @lends GameWidgetPauseButton.prototype */

        /** Game Model */
        _model: null,

        _gameplayPauseButton: null,
        _gameplayPauseButtonMask: null,
        _gameplayPauseButtonClipper: null,

        _animationInProgress: false,
        _isPaused: false,
        _isShown: undefined,

        /**
         * Returns Widget type as a string.
         *
         * @returns {string}
         */
        getType: function()
        {
            return "GameWidgetPauseButton";
        },

        /**
         *  A GameWidgetPauseButton shows in game pause menu.
         *
         *  @constructs
         *  @param {Object} settings An object that contains popup-related settings.
         *  @param {Object} [delegate] An object to call popup-related functions on.
         */
        ctor:function( model, settings, delegate )
        {
            this._super();

            this._model = model;

            this._delegate = delegate;

            this._buttons = [];

            this._animationInProgress = false;
            this._isPaused = false;
            this._isShown = undefined;

            // Get screensize
            var screenSize = cc.Director.getInstance().getWinSize();

            // Content size needs to match the size of the screen
            this.setContentSize( screenSize );

            // Make sure it is positioned dead center
            this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this.setPosition( cc.p( screenSize.width / 2, screenSize.height / 2 ) );

            // Load the layout.
            this._loadLayout();

            // Create the pause button to be clipped
            this._createViews();

            // Setup animation controller
            this._createAnimationController();
        },

        setEnabled: function( isEnabled )
        {
            this._gameplayPauseButton.setEnabled( isEnabled );
        },
        
        /**
         *  Set whether or not we're paused. May cause an animation to play.
         *
         *  @param {boolean} paused If true, we'll go to the pause menu, if false, we'll remain.
         *
         *  GAME-1885 Multiply pause mask by content scale.  Test case:  2015-01-22 Trouble Brewing.  iOS.  Hardware button.  [~rivey] expects full button visible.  Got clipped.  The clipping size varies by device resolution.
         */
        setPaused: function( paused, playAnimation, notifyDelegate )
        {
            if ( playAnimation === undefined )
            {
                playAnimation = true;
            }

            if ( notifyDelegate === undefined )
            {
                notifyDelegate = false;
            }

            if ( !this._animationInProgress && paused !== this._isPaused )
            {
                this._isPaused = paused;
                var contentScale = cc.Director.getInstance().getContentScaleFactor();

                if ( this._isPaused )
                {
                    if ( this.animationController )
                    {
                        if ( playAnimation )
                        {
                            this._animationInProgress = true;
                            this.animationController.play( "extendPauseButton", this._onPauseAnimationFinished, this );
                        }
                        else
                        {
                            this._animationInProgress = true;

                            // Cancel all animations
                            this.animationController.stopAll ();

                            // Hack, instant animation isn't instant enough
                            //this.animationController.play( "extendPauseButtonInstant", this._onPauseAnimationFinished, this );
                            if ( this.pauseMask )
                            {
                                this.pauseMask.setScaleX( 5 * contentScale );
                                this.pauseMask.setScaleY( contentScale );
                            }
                            if ( this.pauseClipper )
                            {
                                this.pauseClipper.setPosition( cc.p ( 0, 0 ) );
                            }
                            this._onPauseAnimationFinished();
                        }
                    }

                    this._gameplayPauseButton.setVisible( false );
                    this._menuPauseButton.setVisible( true );
                    this.pausedLabel.setVisible( true );
                }
                else
                {
                    if ( this.animationController )
                    {

                        if ( playAnimation )
                        {
                            this._animationInProgress = true;
                            this.animationController.play( "shrinkPauseButton", this._onUnpauseAnimationFinished, this );
                        }
                        else
                        {
                            this._animationInProgress = true;

                            // Hack, instant animation isn't instant enough
                            //this.animationController.play( "shrinkPauseButtonInstant", this._onUnpauseAnimationFinished, this );
                            if ( this.pauseMask )
                            {
                                this.pauseMask.setScaleX( contentScale );
                                this.pauseMask.setScaleY( contentScale );
                            }
                            this._onUnpauseAnimationFinished();
                        }
                    }
                }

                if ( notifyDelegate && this._delegate && this._delegate.gamePauseMenu_stateChanged )
                {
                    this._delegate.gamePauseMenu_stateChanged( this._isPaused );
                }
            }
        },

        showPauseButton: function( show, playAnimation )
        {
            if ( playAnimation === undefined )
            {
                playAnimation = true;
            }

            if ( this._isShown !== show )
            {
                this._isShown = show;

                if ( this._isShown )
                {
                    if ( playAnimation )
                    {
                        if ( this.animationController )
                        {
                            this.animationController.play( "pauseButtonIn" );
                        }
                    }
                    else
                    {
                        if ( this.pauseClipper )
                        {
                            this.pauseClipper.setPosition( cc.p ( 0, 0 ) );
                        }
                    }
                }
                else
                {
                    if ( playAnimation )
                    {
                        if ( this.animationController )
                        {
                            this.animationController.play( "pauseButtonOut" );
                        }
                    }
                    else
                    {
                        if ( this.pauseClipper )
                        {
                            this.pauseClipper.setPosition( cc.p ( 0, 91 ) );
                        }
                    }
                }
            }
        },

        _loadLayout: function()
        {
            LayoutLoader.loadLayoutFile( "pauseButton_layout.json", this, this , this );
        },
        
        layout_creatingNode: function( nodeName, nodeData )
        {
             switch ( nodeName )
             {
                 case "pauseMask":
                     this.pauseMask = cc.Sprite.create( "pauseMenuMask.png" );
                     this.pauseMask.doNotAdd = true;
                     this._buttonGameSize = nodeData.size;
                     return this.pauseMask;
                 case "pauseClipper":
                      this.pauseClipper = cc.ClippingNode.create( this.pauseMask );
                      this.pauseClipper.doNotAdd = true;
                      return this.pauseClipper;
                 case "gameContainer":
                     return false;
                 case "settings":
                     return false;
                 case "pauseFullSize":
                     this._buttonMenuSize = nodeData.size;
                     return false;
                 case "howToPlayContainer":
                     return false;
                 case "howToPlayPlayButton":
                     return false;
             }
            
            return null;
        },
        
        layout_settingNodeResizingInfo: function( nodeName, nodeData, node, resizingData )
        {
            //the pauseMask exists in a separate scenegraph because it's the mask for a clipping node, never getting added to the hierarchy.
            //that means it should exist in an untransformed space, therefore we do not want it to even be allowed to have resizing info
            switch ( nodeName )
            {
                case "pauseMask":
                    return false;
            }
            return true;
        },

        /**
         * Creates Pause Button and all of its elements.
         *
         * @private
         *
         * GAME-1885 Multiply pause button by content scale.  Test case:  2015-01-22 Trouble Brewing.  iOS.  Hardware button.  [~rivey] expects full button visible.  Got clipped.  The clipping size varies by device resolution.
         */
        _createViews: function()
        {
            // Animated clipper view.
            this.pauseClipperContainer.addChild( this.pauseClipper );

            // Gameplay pause button.

            var contentScale = cc.Director.getInstance().getContentScaleFactor();
            this.pauseMask.setScaleX( contentScale );
            this.pauseMask.setScaleY( contentScale );
            this._gameplayPauseButton = this._createPauseButton( "buttons/game_buttonPauseUp.png", "buttons/game_buttonPauseDown.png", cc.PointZero(), this.pauseMask.getContentSize(), this, this._pauseButtonTapped );
            this._gameplayPauseButton.nodeName = "pauseButtonGameplay";
            this.pauseClipper.addChild( this._gameplayPauseButton );
            this._gameplayPauseButton.setScaleX( contentScale );
            this._gameplayPauseButton.setScaleY( contentScale );

            // Menu pause button.
            this._menuPauseButton = this._createPauseButton( "buttons/game_buttonPauseActiveUp.png", "buttons/game_buttonPauseActiveDown.png", cc.PointZero(), this._buttonMenuSize, this, this._pauseButtonTapped );
            this._menuPauseButton.nodeName = "pauseButtonMenu";
            this._menuPauseButton.setVisible( false );
            this._menuPauseButton.setEnabled( false );
            this.pauseClipper.addChild( this._menuPauseButton );

            // Add buttons to view.
            this._buttons.push( this._gameplayPauseButton );
            this._buttons.push( this._menuPauseButton );

            // Move paused label to clipping node, so it clips like the button.
            this.pausedLabel.removeFromParent( false );
            this.pausedLabel.setVisible( false );
            this.pausedLabel.setDimensions( cc.size( 0 ,0 ) );
            this.pausedLabel.setAnchorPoint( 0, 0.5 );
            this.pausedLabel.setString( localize( "PausedButtonText" ) );
            this.pauseClipper.addChild( this.pausedLabel );

            // Add and scale text background to match length of label.
            this._menuPauseTextSprite = cc.Sprite.create( "buttons/game_buttonPauseActiveUpText.png" );
            this._menuPauseTextSprite.setAnchorPoint( 0, 0 );
            this._menuPauseTextSprite.setPosition( cc.p( this._menuPauseButton.getContentSize().width, 0 ) );
            var textScaleX = ( this.pausedLabel.getContentSize().width + 25 ) / this._menuPauseTextSprite.getContentSize().width;
            var textScaleY = this._buttonMenuSize.width / this._menuPauseTextSprite.getContentSize().width;
            this._menuPauseTextSprite.setScaleX( textScaleX );
            this._menuPauseTextSprite.setScaleY( textScaleY );
            this.pauseClipper.addChild( this._menuPauseTextSprite );
        },

        _createButton: function( bgUpImage, bgDownImage, bgDisabledImage )
        {   
            var button = null;

            button = new LumosityButton();
            button.setBackgroundImageForState( bgUpImage, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( bgDownImage, LumosityButtonState.ButtonStateSelected );
            button.setBackgroundImageForState( bgDisabledImage, LumosityButtonState.ButtonStateDisabled );
          
            return button;
        },

        _createPauseButton: function( bgUpImageName, bgDownImageName, position, size, target, callback )
        {
            var button = null;
            var bgUpImage = null;
            var bgDownImage = null;
            var bgDisabledImage = null;

            bgUpImage = cc.Sprite.create( bgUpImageName );
            bgDownImage = cc.Sprite.create( bgDownImageName );
            bgDisabledImage = cc.Sprite.create( bgUpImageName );

            button = this._createButton( bgUpImage, bgDownImage, bgDisabledImage );

            button.setContentSize( size );
            button.setAnchorPoint( cc.PointZero() );
            button.setPosition( position );
            button.setTarget( target );
            button.setReleasedCallback( callback );

            return button;
        },

        /**
         *  Creates this view's AnimationController from the game's settings file. That way of doing it
         *  is temporary, later we will want to pass the animation file name into this function. The
         *  controller is also automatically created with the shared game animation data.
         */
        _createAnimationController:function()
        {
            this.animationController = new AnimationController( "pauseButton_animations.json", this, this, [ this ] );
            this.animationController.play( "layout" );
        },

        /**
         * Returns true if there is an animation in progress.
         *
         * @returns {boolean}
         */
        isAnimationInProgress: function()
        {
            return this._animationInProgress;
        },

        /**
         * Internal callback for the pause button tap.
         *
         * @private
         */
        _pauseButtonTapped: function()
        {
            if( this._delegate.view.getDisplayState() === DisplayState.Transition ){ return; }
            this.setPaused( !this._isPaused, true, true );
        },

        /**
         * Called whenever PAUSE animation is finished.
         *
         * @private
         */
        _onPauseAnimationFinished: function()
        {
            this._animationInProgress = false;
            this._menuPauseButton.setEnabled( true );
        },

        /**
         * Called whenever UNPAUSE animation is finished.
         * @private
         */
        _onUnpauseAnimationFinished: function()
        {
            this._animationInProgress = false;
            this._gameplayPauseButton.setVisible( true );
            this._menuPauseButton.setVisible( false );
            this._menuPauseButton.setEnabled( false );
            this.pausedLabel.setVisible( false );

            // don't reveal gameplay pause button if we're in tutorial mode
            var tutorialMode = ( !LumosityUtils.isFitTestModeOn() && this._model.persistentData.isPracticeMode );
            if( tutorialMode )
            {
                this.showPauseButton( false, false );
            }
        }
      } );
}() );
