"use strict";

/**
 * This class can be automatically converted into JavaScript.
 *
 * Minimally parse Excel or Google formatted string of a CSV to an array of objects.
 * Supported data types:  String, Boolean, float, int.
 * Does not support double-quotes, quoted commas ",", newlines in fields, nonsquare CSVs, blank lines.
 * If quoting "," becomes required, then we could pass in field delimiter and use tab "\t".
 * Example @see TestCsv.as
 */
var Csv = cc.Class.extend(
{



});

Csv.depthSeparators = [",", ";", "%", ":"];

Csv.trim = function(text)
{
    return text.replace(/^\s+|\s+$/g, "");
};

/**
 * One element per line, converting and removing line separator characters.
 * 14/3/21 Nick expects to fix normalize consecutively empty Mac or Unix line endings.
 * Example @see TestCsv
 */
Csv.normalizeLines = function(csv)
{
    var windows= /\r\n/g;
    var mac= /\r/g;
    var unix = "\n";
    var normal = csv.replace(windows, unix);
    normal = normal.replace(mac, unix);
    normal = Csv.trim(normal);
    var lines = normal.split(unix);
    return lines;
};

/**
 * Minimally parse CSV string to 2D array of rows with columns.
 * Infers data types:  String, Boolean, float, int.
 */
Csv.array = function(csv)
{
    var lines = Csv.normalizeLines(csv);
    var rows= [];
    for (var i=0; i < lines.length; i++) {
        var columns = lines[i].split(",");
        for (var c=0; c < columns.length; c++) {
            if ("TRUE" == columns[c]) {
                columns[c] = true;
            }
            else if ("FALSE" == columns[c]) {
                columns[c] = false;
            }
            else {
                var integer= parseInt(columns[c]);
                var float= parseFloat(columns[c]);
                if (!isNaN(integer)) {
                    if (integer == float) {
                        columns[c] = integer;
                    }
                    else {
                        columns[c] = float;
                    }
                }
            }
        }
        rows.push(columns);
    }
    return rows;
};

/**
 * First row is header row, to create fieldnames.
 * Example @see TestCsv.as
 */
Csv.arrayOfObjects = function(csv)
{
    var rows = Csv.array(csv);
    var objects = new Array();
    for (var r=1; r < rows.length; r++) {
        var object = new Object();
        for (var c=0; c < rows[r].length; c++) {
            object[rows[0][c]] = rows[r][c];
        }
        objects.push(object);
    }
    return objects;
};

/**
 * Convert embedded CSV text file to an array of objects.
 * String casting a ByteArray is not automatically convertible to JavaScript.
 * Example @see TestCsv.as
 */
Csv.read = function(CsvClass)
{
    var csv = TextFile.read(CsvClass);
    return Csv.arrayOfObjects(csv);
};

/**
 * Convert array of objects to CSV string.  Unquoted, unescaped.
 * Does not support quoting "," in a field.
 * In MySQL YAML a "\n" is converted to "\n\n    ", so consider line delimiting by "/".
 */
Csv.tabulate = function(rows, headers, lineDelimiter)
{
    var table = "";
    table += headers.join(",");
    for (var r = 0; r < rows.length; r++) {
        var row = rows[r];
        var tableRow = [];
        for (var h = 0; h < headers.length; h++) {
            var header = headers[h];
            tableRow.push(row[header]);
        }
        table += lineDelimiter + tableRow.join(",");
    }
    return table;
};

/**
 * Sorted array of strings.
 */
Csv.head = function(row)
{
    var headers = [];
    for (var h in row) {
        headers.push(h);
    }
    headers.sort();
    return headers;
};

/**
 * Conveniently tabulate with default line delimiter that is convenient for Daniel.
 */
Csv.tab = function(rows, lineDelimiter)
{
    if (undefined === lineDelimiter) {
        lineDelimiter="/";
    }
    return Csv.tabulate(rows, Csv.head(rows[0]), lineDelimiter);
};

/**
 * Converts an array to string separated by the specified separator character, used for nested arrays
 * e.g. input=[1,2,3] returns="1;2;3"
 */
Csv.arrayToString = function(arr, separator)
{
    if (undefined === separator) {
        separator=";";
    }
    var output = "";
    for (var i = 0; i < arr.length-1; i++) {
        output = output + String(arr[i]) + separator;
    }
    output = output + arr[i];

    return output;
};

/**
 * GAME-3304 Use only to nest a variable list (such as a list of words), not a sub-table.  Test case:  2015-05-01 Redshift expects metadata tables are not nested.
 *
 * Like tabulate, but supports limited nesting of sub-array and sub-hashes. 
 * @param   depth   Pick a separator character based on depth.  Be sure not to use those characters in the data.
 * Example @see TestCsv
 * Example @see ../games/tugOfWords
 */
Csv.tabulateNested = function(rows, headers, lineDelimiter, depth)
{
    if (undefined === depth) {
        depth=0;
    }
    function getSeparator(depth)
    {
        if (Csv.depthSeparators.length <= depth) {
            throw new Error("Csv does not support depth " + depth);
        }
        var separator = Csv.depthSeparators[depth];
        return separator;
    }
    function isFlat(value)
    {
        return (typeof value == "string") || (typeof value == "number") || null == value;
    }
    var separator = getSeparator(depth);
    var table = "";
    table += headers.join(separator);
    for (var r = 0; r < rows.length; r++) {
        var row = rows[r];
        var tableRow = [];
        for (var h = 0; h < headers.length; h++) {
            var header = headers[h];
            var value= row[header];
            if (isFlat(value)) {
            }
            else if (value instanceof Array) {
                if (1 <= value.length && !isFlat(value[0])) {
                    value = Csv.tabNested(value, getSeparator(depth + 1), depth + 2);
                }
                else {
                    value = Csv.arrayToString(value, getSeparator(depth + 1));
                }
            }
            else {
                if (!(value instanceof Array)) {
                    value = [value];
                }
                value = Csv.tabNested(value, getSeparator(depth + 1), depth + 2);
            }
            tableRow.push(value);
        }
        table += lineDelimiter + tableRow.join(separator);
    }
    return table;
};

/**
 * Conveniently nest sub-arrays and sub-hashes.
 */
Csv.tabNested = function(rows, lineDelimiter, depth)
{
    if (undefined === lineDelimiter) {
        lineDelimiter="/";
    }
    if (undefined === depth) {
        depth=0;
    }
    var headers = Csv.head(rows[0]);
    return Csv.tabulateNested(rows, headers, lineDelimiter, depth);
};
