/*global cc*/
/*global require*/

require( "src/View/LumosityButton.js" );
/*global LumosityButton*/
/*global LumosityButtonState*/

require( "src/Utility/LumosityUtils.js" );
/*global lu*/

var ButtonStateData;
( function()
{
    ButtonStateData = cc.Class.extend(
    {
        /** @lends ButtonStateData.prototype */
        
        /**
         * ButtonStateData is just a struct that holds info about multiple button states.
         *
         * @param {*} normal The button data for the normal (up) state.
         * @param {*} [selected=normal] The button data for the selected (down) state.
         * @param {*} [disabled=normal] The button data for the disabled state.
         * @param {*} [highlighted=normal] The button data for the highlighted state.
         */
        ctor: function( normal, selected, disabled, highlighted )
        {
            if ( highlighted === undefined )
            {
                highlighted = normal;
            }
            
            if ( disabled === undefined )
            {
                disabled = normal;
            }
            
            if ( selected === undefined )
            {
                selected = normal;
            }
            
            this[ LumosityButtonState.ButtonStateNormal ] = normal;
            this[ LumosityButtonState.ButtonStateHighlighted ] = highlighted;
            this[ LumosityButtonState.ButtonStateDisabled ] = disabled;
            this[ LumosityButtonState.ButtonStateSelected ] = selected;
        }
    });
}());

var ButtonFactory;
( function()
{
    "use strict";

    /**
     * The ButtonFactory is a static class that can be used to create buttons of various types.
     * You can call these functions directly, or you can call them from ViewFactory.
     *
     * @class
     * @name ButtonFactory
     */
    ButtonFactory =
    {
        /** @lends ButtonFactory */
        
        //this gets populated below
        _buttonStates: [],
        
        /**
         * Creates a button with text in it, and no images.
         *
         * @param {string|ButtonStateData} title The text for the button
         * @param {cc.Color|ButtonStateData} [titleColor] The text color for the button
         * @param {string} [fontName] The font name
         * @param {number} [fontSize] The size for the font - note this may not be actual size, as the size will change based on parent scale.
         * @param {cc.size} [contentSize] The content size for the button. If not provided, will use the label's size.
         * @returns {LumosityButton} A created button.
         */
        createTextButton: function( title, titleColor, fontName, fontSize, contentSize )
        {
            var button = new LumosityButton();
            
            this.applyTextDataToButton( button, title, titleColor, fontName, fontSize );
            this.applyCommonDataToButton( button, contentSize );
            
            return button;
        },
        
        /**
         * Creates a button with an image. You can provide a string for the image ref, or a cc.Sprite, or
         * a button state data with either strings or cc.Sprites in it.
         *
         * @param {string|cc.Sprite|ButtonStateData} [backgroundImage] The background image for the button
         * @param {string|cc.Sprite|ButtonStateData} [iconImage] The icon (top) image for the button
         * @param {cc.size} [contentSize] The content size for the button. If not provided, will use the background image's size.
         * @returns {LumosityButton} A created button.
         */
        createImageButton: function( backgroundImage, iconImage, contentSize )
        {
            var button = new LumosityButton();
            
            this.applyImageDataToButton( button, backgroundImage, iconImage );
            this.applyCommonDataToButton( button, contentSize );
            
            return button;
        },
        
        /**
         * Creates a button with an image and text on it. This is a "full" button with all options.
         * For the image, you can provide a string for the image ref, or a cc.Sprite, or
         * a button state data with either strings or cc.Sprites in it.
         *
         * @param {string|cc.Sprite|ButtonStateData} [backgroundImage] The background image for the button
         * @param {string|cc.Sprite|ButtonStateData} [iconImage] The icon (top) image for the button
         * @param {string|ButtonStateData} [title] The text for the button
         * @param {cc.Color|ButtonStateData} [titleColor] The text color for the button
         * @param {string} [fontName] The font name
         * @param {number} [fontSize] The size for the font - note this may not be actual size, as the size will change based on parent scale.
         * @param {cc.size} [contentSize] The content size for the button. If not provided, will use the background image's size.
         * @returns {LumosityButton} A created button.
         */
        createImageTextButton: function( backgroundImage, iconImage, title, titleColor, fontName, fontSize, contentSize )
        {
            var button = new LumosityButton();
            
            this.applyTextDataToButton( button, title, titleColor, fontName, fontSize );
            this.applyImageDataToButton( button, backgroundImage, iconImage );
            this.applyCommonDataToButton( button, contentSize );
            
            return button;
        },
        
        /**
         * Creates a simple button - has a background and text. Uses Museo-500 as the font, size 18, white color.
         *
         * @param {string|cc.Sprite|ButtonStateData} [backgroundImage] The background image for the button
         * @param {string|ButtonStateData} [title] The text for the button
         * @returns {LumosityButton} A created button.
         */
        createSimpleButton: function( backgroundImage, title, contentSize )
        {
            var button = new LumosityButton();
            
            this.applyTextDataToButton( button, title, cc.c3b(255,255,255), lu.getFontLocation( "MuseoSans-300" ), 32 );
            this.applyImageDataToButton( button, backgroundImage );
            this.applyCommonDataToButton( button, contentSize );
            
            return button;
        },
        
        /**
         * Convenience function, simply constructs a ButtonStateData.
         *
         * @param {*} normal The button data for the normal (up) state.
         * @param {*} [selected=normal] The button data for the selected (down) state.
         * @param {*} [disabled=normal] The button data for the disabled state.
         * @param {*} [highlighted=normal] The button data for the highlighted state.
         * @returns {Object} Button state data.
         */
        createButtonStateData: function( normal, selected, disabled, highlighted )
        {
            return new ButtonStateData( normal, selected, disabled, highlighted );
        },
        
        /**
         * Applies data common across all button types to a button.
         *
         * @param {LumosityButton} button The button to apply the data to.
         * @param {cc.size} [contentSize] The content size to apply to the button.
         */
        applyCommonDataToButton: function( button, contentSize )
        {
            if ( contentSize )
            {
                button.setContentSize( contentSize );
                
                if ( button.titleLabel )
                {
                    button.titleLabel.setPosition( cc.p( contentSize.width / 2, contentSize.height / 2 ) );
                }
            }
        },
        
        /**
         * Applies text data to a button.
         *
         * @param {LumosityButton} button The button to apply the data to.
         * @param {string|ButtonStateData} [title] The text for the button.
         * @param {cc.Color|ButtonStateData} [titleColor] The text color for the button
         * @param {string} [fontName] The font name
         * @param {number} [fontSize] The size for the font - note this may not be actual size, as the size will change based on parent scale.
         */
        applyTextDataToButton: function( button, title, titleColor, fontName, fontSize )
        {
            var stateIndex;
            
            if ( title !== undefined )
            {
                if ( typeof( title ) === "string" )
                {
                    title = this.createButtonStateData( title );
                }
            
                for ( stateIndex = 0; stateIndex < this._buttonStates.length; stateIndex++ )
                {
                    button.setTitleForState( title[ this._buttonStates[ stateIndex ] ], this._buttonStates[ stateIndex ] );
                }
            }
            
            if ( button.titleLabel )
            {
                if ( titleColor )
                {
                    if ( !titleColor[ LumosityButtonState.ButtonStateNormal ] )
                    {
                        titleColor = this.createButtonStateData( titleColor );
                    }
                
                    for ( stateIndex = 0; stateIndex < this._buttonStates.length; stateIndex++ )
                    {
                        button.setTitleColorForState( titleColor[ this._buttonStates[ stateIndex ] ], this._buttonStates[ stateIndex ] );
                    }
                }
            
                if ( fontName )
                {
                    button.titleLabel.setFontName( fontName );
                }
            
                if ( fontSize !== undefined )
                {
                    button.titleLabel.setInitialFontSize( fontSize );
                }
            }
        },
        
        /**
         * Applies image data to a button. You can provide a string for the image ref, or a cc.Sprite, or
         * a button state data with either strings or cc.Sprites in it.
         *
         * @param {LumosityButton} button The button to apply the data to.
         * @param {string|cc.Sprite|ButtonStateData} [backgroundImage] The background image for the button
         * @param {string|cc.Sprite|ButtonStateData} [iconImage] The icon (top) image for the button
         */
        applyImageDataToButton: function( button, backgroundImage, iconImage )
        {
            var stateIndex;
            var backgroundImageData = this._getButtonStateDataForImage( backgroundImage );
            if ( backgroundImageData )
            {
                for ( stateIndex = 0; stateIndex < this._buttonStates.length; stateIndex++ )
                {
                    button.setBackgroundImageForState( backgroundImageData[ this._buttonStates[ stateIndex ] ], this._buttonStates[ stateIndex ] );
                }
            }
            
            var iconImageData = this._getButtonStateDataForImage( iconImage );
            if ( iconImageData )
            {
                for ( stateIndex = 0; stateIndex < this._buttonStates.length; stateIndex++ )
                {
                    button.setImageForState( iconImageData[ this._buttonStates[ stateIndex ] ], this._buttonStates[ stateIndex ] );
                }
            }
        },
        
        _getButtonStateDataForImage: function( image )
        {
            var imageData = null;
            var buttonStateIndex;
            
            if ( image )
            {
                if ( typeof( image ) === "string" )
                {
                    imageData = this.createButtonStateData( cc.Sprite.create( image ) );
                }
                else if ( image instanceof cc.Sprite )
                {
                    imageData = this.createButtonStateData( image );
                }
                else
                {
                    if ( typeof( image[ LumosityButtonState.ButtonStateNormal ] ) === "string" )
                    {
                        imageData = new ButtonStateData();
                        for ( buttonStateIndex = 0; buttonStateIndex < this._buttonStates.length; buttonStateIndex++ )
                        {
                            imageData[ this._buttonStates[ buttonStateIndex ] ] = cc.Sprite.create( image[ this._buttonStates[ buttonStateIndex ] ] );
                        }
                    }
                    else
                    {
                        imageData = image;
                    }
                }
            }
            
            return imageData;
        }
    };
    
    //populate an array with all the button states, for convenience
    var buttonStateKey;
    for ( buttonStateKey in LumosityButtonState )
    {
        ButtonFactory._buttonStates.push( LumosityButtonState[ buttonStateKey ] );
    }
}());