/*global cc*/
/*global require*/
/*global lu*/
/*global audio*/
/*global LumosityUtils*/
/*global LumosityButton*/
/*global LumosityButtonLayer*/
/*global LumositySprite*/

require( "src/Utility/LumosityUtils.js" );
require( "src/View/LumosityButtonLayer.js" );
require( "src/View/LumosityButton.js" );
require( "src/View/LumositySprite.js" );

var LumositySwipeView = cc.Node.extend(
{
    /** @lends LumositySwipeView.prototype */

    /** Design size, matches flash */
    _designSize: cc.size( 640, 840 ),

    /** Name of the view */
    nodeName: "LumositySwipeView",

    /** Delegate for interactions and state changes. */
    _delegate: null,

    _dotsContainer: null,
    _pagesContainer: null,
    /** Array of buttons in this widget. */
    _buttons: [],
    _rightButton: null,
    _leftButton: null,
    _showButtons: false,

    _pages: [],
    _dots: [],
    _currentPage: undefined,

    // Settings
    _dotSize: 14,
    _dotSpacing: 30,
    _dotsPositionY: 80,
    _pageSlideTime: 0.3,
    _buttonShowHideTime: 0.3,

    /** Controls when the view is transitioning between pages. */
    _transitioning: false,

    /**
     * Returns Widget type as a string.
     *
     * @returns {string}
     */
    getType: function()
    {
        return "LumositySwipeView";
    },

    /**
     * Sets new delegate for the widget.
     *
     * @param delegate
     */
    setDelegate: function( delegate )
    {
        this._delegate = delegate;
    },

    /**
     * Returns array of buttons in this widget.
     *
     * @returns {*}
     */
    getButtons: function()
    {
        return this._buttons;
    },


    /**
     *  A GameWidgetPopup shows in game popup window that can have up to 3 slides and up to 3 lines of text
     *  per slide.
     *
     *  @constructs
     *  @param {Object} settings An object that contains swipe-related settings.
     *                           XXX Unused: The settings parameter is not used by this or any extended class.
     *  @param {Object} [delegate] An object to call swipe-related functions on.
     *  @param {boolean} [showButtons] Whether or not to preload all swipe button images upon creation.
     *                           XXX Unused: The showButtons parameter is not used by this or any extended class.
     */
    ctor:function( settings, delegate, showButtons )
    {
        this._super();

        this._delegate = delegate;

        this._designSize = cc.size( 640, 840 );
        this._buttons = [];
        this._showButtons = showButtons;

        this._pages = [];
        this._dots = [];
        this._currentPage = 0;

        this.setContentSize( this._designSize );
        this._createDotsContainer();
        this._createPagesContainer();
    },


    /**
     * Creates container for dots.
     *
     * @private
     */
    _createDotsContainer: function()
    {
        this._dotsContainer = cc.Node.create();
        this._dotsContainer.setAnchorPoint( cc.p( 0.5, 0 ) );
        this._dotsContainer.setPosition( cc.p( this._designSize.width / 2, this._dotsPositionY ) );
        this._dotsContainer.nodeName = "LumositySwipeView::DotsContainer";
        this.addChild( this._dotsContainer, 100 );
    },

    /**
     * Creates container for pages.
     *
     * @private
     */
    _createPagesContainer: function()
    {
        this._pagesContainer = cc.Node.create();
        this._pagesContainer.setContentSize( this._designSize );
        this._pagesContainer.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        this._pagesContainer.setPosition( cc.p( this._designSize.width / 2, this._designSize.height / 2 ) );
        this._pagesContainer.nodeName = "LumositySwipeView::PagesContainer";
        this.addChild( this._pagesContainer );
    },

    /**
     * Updates container for dots.
     * First, all old dots are deleted.
     * Then, new dots are added to the container and the array.
     *
     * @private
     */
    _updateDots: function()
    {
        if ( this._pages && this._dots && this._dotsContainer )
        {
            this._dotsContainer.removeAllChildren( false );
            this._dots.length = 0;

            if ( this._pages && this._pages.length > 0 )
            {
                // content size
                var size = cc.size( this._pages.length * this._dotSize + (this._pages.length - 1) * this._dotSpacing, this._dotSize );
                this._dotsContainer.setContentSize( size );
                this._dotsContainer.setPosition( cc.p( this._designSize.width / 2, this._dotsPositionY ) );

                // Add dots
                var i;
                var x = this._dotSize / 2;
                var dot;
                for ( i = 0; i < this._pages.length; i++ )
                {
                    dot = new LumositySprite ( "howToPlays/howToPlay_dotOff.png" );
                    dot.setAnchorPoint( cc.p( 0.5, 0 ) );
                    dot.setPosition( cc.p( x, 0 ) );
                    dot.setContentSize ( cc.size ( this._dotSize, this._dotSize ), true, true );
                    this._dotsContainer.addChild( dot );

                    this._dots.push( dot );

                    x += this._dotSize + this._dotSpacing;
                }

                // Light up current dot
                this._lightUpDot( this._currentPage );
            }
        }
    },


    _dimDot: function( dotIndex )
    {
        if ( dotIndex >= 0 && dotIndex < this._dots.length )
        {
            this._dots[ dotIndex ].setTexture( "howToPlays/howToPlay_dotOff.png" );
            this._dots[ dotIndex ].setContentSize ( cc.size ( this._dotSize, this._dotSize ), true, true );
        }
    },

    _lightUpDot: function( dotIndex )
    {
        if ( dotIndex >= 0 && dotIndex < this._dots.length )
        {
            this._dots[ dotIndex ].setTexture( "howToPlays/howToPlay_dotOn.png" );
            this._dots[ dotIndex ].setContentSize ( cc.size ( this._dotSize, this._dotSize ), true, true );
        }
    },

    /**
     * Adds a new page to the view.
     *
     * @param node
     */
    addPage: function( node )
    {
        // Add to array
        this._pages.push( node );

        // Add child
        var viewerSize = this._designSize;
        node.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        node.setPosition( cc.p( viewerSize.width / 2 + ((this._pages.length - 1) * viewerSize.width), viewerSize.height / 2 ) );
        this._pagesContainer.addChild( node );

        this._updateDots();
        this._updatePageVisibility();
    },

    /**
     * Moves to the next page.
     */
    nextPage: function()
    {
        "use strict";
        this.setPage( this._currentPage + 1 );
    },

    /**
     * Moves to the previous page.
     */
    previousPage: function( instant )
    {
        "use strict";
        this.setPage( this._currentPage - 1 );
    },

    getCurrentPageNumber: function()
    {
        return this._currentPage;
    },

    getCurrentPage: function()
    {
        return this._pages[ this._currentPage ];
    },

    setPage: function( pageNumber, instant )
    {
        "use strict";

        if ( instant === undefined )
        {
            instant = false;
        }

        var stepDirection = Math.abs( pageNumber - this._currentPage );
        if ( pageNumber > this._currentPage )
        {
            stepDirection = stepDirection * -1;
        }

        if( pageNumber < 0 || pageNumber > this._pages.length - 1 ){ return; }
        if( pageNumber === this._currentPage ){ return; }
        if( this._transitioning === true ){ return; }
        //
        //NOTE: We should remove this hack once the sound effect
        //      has been removed from the flash file. (Otherwise
        //      it will sound like we're doubling up on this sound)
        //
        if( this._currentPage > 0 )
        {
            audio.playEffect( "sounds/cardSlide.mp3" );
        }

        this._transitioning = true;

        // Dim current dot
        this._dimDot( this._currentPage );

        // Advance the page
        this._currentPage = pageNumber;

        // Light up current dot
        this._lightUpDot( this._currentPage );

        // Make sure the page is visible
        this._pages[ this._currentPage ].setVisible( true );

        // Slide the container, so we can see the page
        if ( this._pagesContainer )
        {
            if ( !instant )
            {
                var moveAction = cc.MoveBy.create( this._pageSlideTime, cc.p( this._designSize.width * stepDirection, 0 ) );
                var moveActionEase = cc.EaseInOut.create( moveAction, 2 );
                var visibilityAction = cc.CallFunc.create( this._updatePageVisibility, this );
                var finishCallbackAction = cc.CallFunc.create( this._onCenterPage, this );
                var sequence = cc.Sequence.create( moveActionEase, visibilityAction, finishCallbackAction );

                this._pagesContainer.runAction( sequence );
            }
            else
            {
                var newPosition = this._pagesContainer.getPosition();
                newPosition.x += this._designSize.width * stepDirection;
                this._pagesContainer.setPosition( newPosition );
                this._updatePageVisibility();
                this._onCenterPage();
            }
        }

        // Determine which buttons should be visible
        if( this._delegate )
        {
            // Left Button
            if ( this._delegate.swipeView_showLeftButton )
            {
                if ( this._currentPage === 0 )
                {
                    this._delegate.swipeView_showLeftButton ( false );
                }
                if ( this._currentPage > 0 )
                {
                    this._delegate.swipeView_showLeftButton ( true );
                }
            }

            if ( this._delegate.swipeView_showRightButton )
            {
                if( this._currentPage === this._dots.length - 1 )
                {
                    this._delegate.swipeView_showRightButton ( false );
                }
                if( this._currentPage < this._dots.length - 2 )
                {
                    this._delegate.swipeView_showRightButton ( true );
                }
            }
        }
    },

    _onCenterPage: function()
    {
        this._transitioning = false;

        if ( this._delegate && this._delegate.swipeView_onCenterPage )
        {
            this._delegate.swipeView_onCenterPage();
        }
    },

    getPageCount: function()
    {
        if ( this._pages )
        {
            return this._pages.length;
        }
        else
        {
            return 0;
        }
    },


    /**
     * Makes all but current page invisible.
     *
     * @private
     */
    _updatePageVisibility: function()
    {
        var page;
        var i;
        for ( i = 0; i < this._pages.length; i++ )
        {
            page = this._pages[ i ];
            page.setVisible( i === this._currentPage );
        }
    },


    /**
     * Returns true if need to show the left button to go to the previous slide.
     *
     * @returns {boolean}
     */
    showLeftButton: function()
    {
        if ( this._pages && this._currentPage > 0 )
        {
            return true;
        }

        return false;
    },


    /**
     * Returns true if need to show the right button to go to the next slide.
     *
     * @returns {boolean}
     */
    showRightButton: function()
    {
        if ( this._pages && this._currentPage < this._pages.length - 1 )
        {
            return true;
        }

        return false;
    },

    /**
     * Returns the current transition status as a bool.
     * 
     * @return {Boolean} Returns true for when the view is currently transitioning between pages.
     */
    isTransitioning: function()
    {
        return this._transitioning;
    },



    /**
     * Sets spacing for dots
     *
     * @param x
     * @param y
     */
    setDotSpacing: function( x, y )
    {
        this._dotSpacing = x;
    }
});
