/*global cc*/
/*global ll*/
/*global Game*/
/*global require*/
/*global audio*/
/*global LumosityUtils*/
/*global MemoryMatrixView*/
/*global MemoryMatrixModel*/
/*global MemoryMatrixSettings*/
/*global InteractionType*/
/*global MemoryMatrixBoardModel*/
/*global MemoryMatrixBoardTileModel*/
/*global metadataAggregateKeys*/
/*global metadataTrialKeys*/


var MemoryMatrix;
(function()
{
    "use strict";

    require( "src/Game/Base/Game.js" );
    require( "src/Game/MemoryMatrixView.js" );
    require( "src/Game/MemoryMatrixModel.js" );
    require( "src/Game/MemoryMatrixSettings.js" );
    require( "src/Utility/LumosityLogger.js" );
    require( "src/Utility/LumosityUtils.js" );
    require( "src/Game/MemoryMatrixBoardModel.js" );
    require( "src/Game/MemoryMatrixBoardTileModel.js" );
    require( "src/Game/DebugMenu.js" );

    MemoryMatrix = Game.extend(
    {
        /** @lends MemoryMatrix.prototype */

        /**
         *  MemoryMatrix is the main class for the game, MemoryMatrix!
         *
         *  @constructs
         */
        ctor:function()
        {
            this._super();
            this.debugMenu = new DebugMenu(this);
        },

        _createSettings:function()
        {
            return new MemoryMatrixSettings( "settings.json" );
        },

        _createModel:function()
        {
            return new MemoryMatrixModel( this.settings );
        },

        _createView:function( screenSize )
        {
            return new MemoryMatrixView( screenSize, this.settings, this.model, this );
        },

        /**
         * GAME-3659 On each begin (which includes each restart), resave settings to the model.  The model is destroyed and recreated each restart or after the tutorial and beginning of the fit test game, so any data on the old model is lost.  Test case:  2015-07-06 Reset.  End.  Expect level saved.  Got crash.  Test case:  2015-07-06 Short game.  Fit test.  Tutorial.  Finish.  Restart occurs automatically.  Finish fit test.  Zassmin Montes expects user level is an integer.  Got null.
         */
        onBegin: function()
        {
            this._super();
            this.model.settings = this.settings;
            // LumosityUtils.printNodeHierarchy( this.view );

            ll.verbose( "MemoryMatrix", "onBegin()" );

            this.setInitialHudState();

            // Figure out starting difficulty
            this.calculateStartingDifficulty();
        },

        onStartGameplay:function()
        {
            this._super();
            // LumosityUtils.printNodeHierarchy( this.view );

            var screenSize = cc.Director.getInstance().getWinSize();
            // if ( this.view.tray.getPosition().x > screenSize.width || this.view.tray.getPosition().x < 0 )
            // {
            //     // this.view.boardSlideInAnim();
            //     this.view.tray.setPosition( cc.p( screenSize.width / 2, this.view.tray.getPosition().y ) );
            // }

            ll.verbose( "MemoryMatrix", "onStartGameplay()" );

            this.startNextTrial();

            // Set initial meta data
            this.model.maxTileCount = this.model.currentMarkedTileCount;
        },

        skipTutorialButton_Tapped: function()
        {
            // LumosityUtils.printNodeHierarchy( this.view );
            if ( !this.model.tutorialSkipped )
            {
                this.tutorialMetadata.finishes += 1;
                this.getMetadataComponent().tutorialMetadata = this.tutorialMetadata; 
                LumosityUtils.callFunctionOnCollection( this._components, "onEndGameplay" );

                this.onCompleteTutorial( true );
            }
        },

        onEndGameplay:function()
        {
            if ( !this.model.isGameOver )
            {
                this.model.isGameRunning = false;
                this.model.isGameOver = true;

                this.finalizeMetadata();

                if ( this.model.persistentData.isPracticeMode )
                {
                    this.tutorialMetadata.finishes += 1;

                    if ( !LumosityUtils.isFitTestModeOn() )
                    {
                        this.view.showTutorialPrompt ( false, "FitTest_RememberTiles" );
                        this.view.showTutorialPrompt ( false, "FitTest_TapTiles" );
                        this.onCompleteTutorial();
                    }
                }

                if ( LumosityUtils.isFitTestModeOn() || ( !LumosityUtils.isFitTestModeOn() && this.model.persistentData.isPracticeMode ) )
                {
                    this.getMetadataComponent().tutorialMetadata = this.tutorialMetadata; 
                }

                LumosityUtils.callFunctionOnCollection( this._components, "onEndGameplay" );

                this.playEndAnimation();

                //finish the interaction script in case there is one (if you want it absolutely tied to the session)
                //this is harmless to call if nothing is being recorded.
                if( this._interactionAutomator )
                {
                    this._interactionAutomator.finishRecordingScript();
                }

                //for automation support, you can hook into this to end automation
                if ( lumosity.game.gameplayEnded )
                {
                    lumosity.game.gameplayEnded();
                }
            }
        },

        onCompleteTutorial: function( skipped )
        {
            cc.log( "Game::onCompleteTutorial" );

            this.model.tutorialSkipped = skipped;

            this.view.showTutorialPrompt ( false, "FitTest_RememberTiles" );
            this.view.showTutorialPrompt ( false, "FitTest_TapTiles" );

            this.view.playAnimation( "endTutorial" );
            // LumosityUtils.printNodeHierarchy( this.view );
        },

        playStartingAnimation: function()
        {
            if ( this.model.restartToPracticeMode || ( this.model.persistentData.isPracticeMode && LumosityUtils.isFitTestModeOn() ) )
            {
                this.view.playAnimation( "practiceBegin", this.onStartGameplay );
            }
            else if ( this.model.persistentData.isPracticeMode && !this.model.isFitTestMode ) 
            {
                // this.view.stopAnimation( "gameBegin" );
                // this.view.stopAllAnimations( this.view.gameContent );
                this.view.playAnimation( "startTutorial", this.onStartGameplay );
            }
            else
            {
                this.view.playAnimation( "gameBegin", this.onStartGameplay );
            }
        },

        playEndAnimation: function ()
        {
            if ( !this.model.persistentData.isPracticeMode )
            {
                var results = this.exportResults();
                var automatedTestReport = this.exportAutomatedTestReport();
                if( automatedTestReport )
                {
                    // Append automated test report if we have one.
                    results.automatedTestReport = automatedTestReport;
                }

                lumosity.game.report( JSON.stringify( results ) );
                this.view.playAnimation( "gameEnd", this.onFinish.bind( this, false ) );
            }
            else if ( LumosityUtils.isFitTestModeOn() )
            {
                this.view.playAnimation( "gameEnd", this.onStartFitTest );
            }
            else
            {

            }
        },

        onUpdateEachSecond:function()
        {
            this._super();

            this._updateIdlePrompt();
        },

        _updateIdlePrompt: function()
        { 
            if ( this.model.isGameRunning && !this.model.isGamePaused && !this.model.persistentData.isPracticeMode )
            {
                if ( this.model.idleTimeTilShowPrompt > 0 )
                {
                    this.model.idleTimeTilShowPrompt--;
                    if ( this.model.idleTimeTilShowPrompt === 0 )
                    {
                        this.view.showPrompt( true );
                    }
                }
            }
        },

        setInitialHudState: function ()
        {
            if ( this.model.persistentData.isPracticeMode && !LumosityUtils.isFitTestModeOn() )
            {
                this.view._widgets[ GameWidgetHud.prototype.getType() ]._hideAll();
            }
            else
            {
                if ( !this.model.persistentData.isPracticeMode )
                {
                    this.view._widgets[ GameWidgetHud.prototype.getType() ].showScore();
                    this.view._widgets[ GameWidgetHud.prototype.getType() ].showTrials();
                }

                // Set the trial, need to call with "1" as parameter because model is not updated till the actual trial starts
                this.view.setTrial( 1, this.settings.getGameLengthInTrials ( this.model.isFitTestMode ) );

                // Set the score
                this.view.setScore( 0 );
            }
        },

        updateHud: function()
        {
            // Set the trial
            this.view.setTrial( this.model.trialCount, this.settings.getGameLengthInTrials ( this.model.isFitTestMode ) );

            // Set the score
            this.view.setScore( this.model.score );
        },
        
        /**
         * Determine starting difficulty.
         * Set session level in metadata.
         *
         * GAME-3659 Save and load user_level.  Test case:  2015-06-02 Read user_level.  Zassmin Montes expects next session level (last trial level - 3).  Got value undefined.  Was saving start session level into user_level.
         *
         * GAME-3659 If last session end_level equals next session user_level, then subtract 3.  Otherwise, user_level is correct.
         */
        calculateStartingDifficulty: function()
        {
            var startingMarkedTileCount = 0;

            if ( !this.model.isFitTestMode && !this.model.persistentData.isPracticeMode )
            {
                // get level info from last results
                var lastUserLevel = this.model.loadFromLastResults(this.settings.startingMarkedTileCount, 
                    this.model.LastTileCountStat, this.settings.getMaxTileCount());
                var lastEndLevel = this.model.loadFromLastResults(this.settings.startingMarkedTileCount, 
                    this.model.LastTileCountStat, this.settings.getMaxTileCount(), 
                    -this.settings.startingMarkedTileOffsetFromPreviousLast, "end_level");
                startingMarkedTileCount = lastUserLevel;
                ll.log("MemoryMatrix.calculateStartingDifficulty: user_level " + lastUserLevel + " end_level " + lastEndLevel);
                if (lastUserLevel === lastEndLevel) {
                    startingMarkedTileCount = this.model.getNextSessionLevel(lastUserLevel);
                }
                var metadata = this.model.metadata;
                if ( metadata )
                {
                    metadata.start_level = startingMarkedTileCount;
                    metadata.session_level = lastUserLevel;
                }
            }
            else if ( this.model.persistentData.isPracticeMode )
            {
                startingMarkedTileCount = this.settings.practiceMode.startingMarkedTileCount;
            }
            else if ( this.model.isFitTestMode )
            {
                startingMarkedTileCount = this.settings.fitTestMode.startingMarkedTileCount;
            }

            this.model.currentMarkedTileCount = startingMarkedTileCount;
            this.model.maxTileCount = this.model.currentMarkedTileCount;

            ll.verbose( "MemoryMatrix", "calculateStartingDifficulty()", "Starting the game with " + this.model.currentMarkedTileCount + " marked tiles." );
        },

        /**
         * Prepares for the new trial (new piece)
         *
         */
        prepareNextTrial: function()
        {
            ll.verbose( "MemoryMatrix", "prepareNextTrial()" );

            // Save current marked tile count for threshold level calculations later
            this.model.markedTileCountHistoryArray.push ( this.model.currentMarkedTileCount );

            this._generateNewBoard();
            this.view.generateTileBoard();
        },

        /**
         *  Starts new trial (new piece)
         *
         */
        startNextTrial: function()
        {
            ll.verbose( "MemoryMatrix", "startNextTrial()" );

            // Reset model
            this.model.idleTimeTilShowPrompt = this.settings.idlePromptTime;
            this.model.resetForTrial();

            // Update trials
            this.model.trialCount++;
            this.updateHud();

            // Set remaining tile flips
            this.model.remainingTileFlips = this.model.currentMarkedTileCount;

            // Preview marked tiles
            this.view.previewMarkedTiles();

            //
            if ( this.model.persistentData.isPracticeMode )
            {
                if ( !this.model.tutorialSkipped )
                {
                    this.view.showTutorialPrompt ( true, "FitTest_RememberTiles" );
                }
            }

            // Meta data
            this.beginTrial();
        },

        _trialFinishedWithSuccess: function()
        {
            this.model.trialsCorrect++;
            this.model.trialsCorrectStreak++;

            // Calculate bonus points
            this.model.bonusPoints = this.settings.successTrialBonusMultiplier * this.model.currentMarkedTileCount;
            this.model.score += this.model.bonusPoints;

            if ( this.model.currentMarkedTileCount > this.model.maxTileCount )
            {
                this.model.maxTileCount = this.model.currentMarkedTileCount;
            }
            if ( this.model.currentMarkedTileCount < this.settings.getMaxTileCount() )
            {

                this.model.currentMarkedTileCount++;
            }

            if ( this.view )
            {
                // hide the help prompt
                this.view.showPrompt( false );

                this.view.updateBonusPoints();

                if ( !this.model.persistentData.isPracticeMode )
                {
                    this.view.playAnimation( "trialEnd", this._trialEndAnimationFinished, this );
                }
                else
                {
                    this.view.playAnimation( "trialEndIncorrect", this._trialEndAnimationFinished, this );
                }
            }

            this.endTrial( this.model.currentTrial, true );
        },

        _trialFinishedWithFailure: function()
        {
            // trial was failed
            this.model.currentTrialFailed = true;

            // Lock all tiles
            this.view.setTilesEnabled( false );

            this.model.trialsCorrectStreak = 0;

            // If two or more are missed, we will reduce difficulty
            if ( this.model.currentBoard.remainingMarkedTilesCount >= this.settings.levelDownTileMissCountThreshold )
            {
                // Reduce difficulty
                if ( this.model.currentMarkedTileCount > 1 )
                {
                    this.model.currentMarkedTileCount--;
                }

            }

            // reveal remaining marked tiles that were missed
            this._revealRemainingMarkedTiles();

            if ( this.view )
            {
                // hide the help prompt
                this.view.showPrompt( false );

                // Play end of trial animation
                this.view.playAnimation( "trialEndIncorrect", this._trialEndAnimationFinished, this );
            }

            this.endTrial( this.model.currentTrial, false );
        },

        _trialEndAnimationFinished: function()
        {
            if ( this.settings.enableRipple )
            {
                var rippleOriginTile;

                // On small board ripple from last tile looks kinda crappy, so it will happen from the top left corner
                var boardSize = this.model.currentBoard.getBoardSize();
                if ( boardSize.width >=  this.settings.rippleOriginSwitchBoardSize.width && boardSize.height >= this.settings.rippleOriginSwitchBoardSize.height )
                {
                    rippleOriginTile = this.model.lastTileTapped;
                }
                else
                {
                    rippleOriginTile = this.model.currentBoard.getTopLeftTile();
                }

                if ( rippleOriginTile )
                {
                    rippleOriginTile.isRippled = true;
                    rippleOriginTile.tileView.rippleTile();
                    this.model.activeRipplesCount++;

                    if ( this.model.currentBoard.remainingMarkedTilesCount === 0 )
                    {
                        // Play flip sound
                        audio.playEffect( "sounds/tileCascade.mp3" );
                    }
                }
            }
            else
            {
                this._moveToNextTrial();
            }
        },



        _gameEndConditionSatisfied: function()
        {
            if ( this.model.trialCount >= this.settings.getGameLengthInTrials ( this.model.isFitTestMode ) )
            {
                return true;
            }
            else if ( this.model.persistentData.isPracticeMode && this.model.trialsCorrect >= this.settings.practiceMode.correctTrialsRequiredToFinish )
            {
                return true;
            }

            return false;
        },


        _moveToNextTrial: function()
        {
            if ( !this.model.tutorialSkipped )
            {
                if ( this._gameEndConditionSatisfied() )
                {
                    this.onEndGameplay();
                }
                else if ( this.model.restartToPracticeMode || this.model.persistentData.isPracticeMode )
                {
                    this.view.playAnimation( "slideLeft", this.startNextTrial, this );
                }
                else
                {
                    if ( this.model.currentTrialSucceeded )
                    {
                        if ( this.view )
                        {
                            this.view.playAnimation( "levelUp", this.startNextTrial, this );
                        }
                    }
                    else if ( this.model.currentTrialFailed )
                    {
                        if ( this.view )
                        {
                            this.view.playAnimation( "levelDown", this.startNextTrial, this );
                        }
                    }
                    else
                    {
                        ll.error( "MemoryMatrix", "_moveToNextTrial", "Trial was finished but not won or lost." );
                    }
                }
            }
        },


        /**
         * This cue sets tile count.
         */
        cue_setTileCount: function()
        {
            this.view.setMarkedTileCount( this.model.currentMarkedTileCount );
        },


        /**
         * This cue triggers board generation.
         */
        cue_generateBoardView: function()
        {
            //cc.log( "Cue called: generateBoardView" );
            this.prepareNextTrial();
        },


        _generateNewBoard: function()
        {
            var boardSize = this.settings.getBoardSize( this.model.currentMarkedTileCount );
            this.model.currentBoard = new MemoryMatrixBoardModel( boardSize );
            this.model.currentBoard.generateNewBoard();
            this.model.currentBoard.setRandomLocationsMarked( this.model.currentMarkedTileCount );
        },


        /**
         * Reveals any remaining marked tiles. Called after trial has finished.
         *
         * @private
         */
        _revealRemainingMarkedTiles: function()
        {
            if ( this.model.currentBoard )
            {
                var tileModels = this.model.currentBoard.getReaminingMarkedTiles();
                if ( tileModels && tileModels.length > 0 )
                {
                    this.model.revealingRemainingMarkedTiles = true;

                    var tileModel;
                    var i;
                    for ( i = 0; i < tileModels.length; i++ )
                    {
                        tileModel = tileModels[ i ];
                        if ( tileModel )
                        {
                            this.model.revealingRemainingMarkedTilesCount++;
                            tileModel.tileView.revealTile();
                        }
                    }
                }
            }
        },

        tileView_onTileTapped: function( tileModel )
        {
            //reset the time we wait to show the prompt
            this.model.idleTimeTilShowPrompt = this.settings.idlePromptTime;

            if ( this.model.isGameRunning && this.model.remainingTileFlips > 0 )
            {
                if ( tileModel && tileModel.tileView )
                {
                    if ( !tileModel.isFlipped )
                    {
                        if ( !this.model.persistentData.isPracticeMode )
                        {
                            // Decrement how many tile flips user has left
                            this.model.remainingTileFlips--;

                            // Set tile as flipped
                            tileModel.isFlipped = true;

                            // Keep a handle to the last tile tapped
                            this.model.lastTileTapped = tileModel;

                            if ( tileModel.isMarked )
                            {
                                this.model.currentBoard.remainingMarkedTilesCount--;

                                // Play flip sound
                                audio.playEffect( "sounds/tileFlip1.mp3" );

                                tileModel.tileView.flipTile();

                                this.model.score += this.settings.tilePoints;
                                this.updateHud();

                                this.model.trialTaps++;
                            }
                            else
                            {
                                // show incorrect feedback
                                this.view.showFeedbackOverTile( false, tileModel );

                                // show prompt if more flips
                                if ( this.model.remainingTileFlips > 0 )
                                {
                                    this.view.showPrompt( true );
                                }

                                this.model.trialTaps++;
                            }

                            this.updateTrialMetadataLayout ( tileModel.x, tileModel.y, tileModel.isMarked );
                            this.updateTrialMetadataResponseTime();
                        }
                        else
                        {
                            // Keep a handle to the last tile tapped
                            this.model.lastTileTapped = tileModel;

                            if ( tileModel.isMarked )
                            {
                                // Decrement how many tile flips user has left
                                this.model.remainingTileFlips--;

                                // Set tile as flipped
                                tileModel.isFlipped = true;

                                // Decrement how many marked tiles are left
                                this.model.currentBoard.remainingMarkedTilesCount--;

                                // Play flip sound
                                audio.playEffect( "sounds/tileFlip1.mp3" );

                                tileModel.tileView.flipTile();

                                this.model.score += this.settings.tilePoints;
                                this.updateHud();

                                this.model.trialTaps++;
                            }
                            else
                            {
                                // show incorrect feedback
                                this.view.showFeedbackOverTile( false, tileModel );

                                // show prompt if more flips
                                if ( this.model.remainingTileFlips > 0 )
                                {
                                    //this.view.showPrompt( true );
                                }

                                this.model.trialTaps++;
                            }

                            this.updateTrialMetadataLayout ( tileModel.x, tileModel.y, tileModel.isMarked );
                            this.updateTrialMetadataResponseTime();
                        }
                    }
                    else
                    {
                        // Play flip sound
                        audio.playEffect( "sounds/tileLock.mp3" );
                    }
                }
            }
        },

        tileView_onFlipFinished: function( tileModel )
        {
            // hide any prompts after a tile is flipped
            if ( !this.model.persistentData.isPracticeMode )
            {
                this.view.showPrompt( false );
            }

            // Make sure trial is finished
            if ( !this.model.currentTrialSucceeded && !this.model.currentTrialFailed )
            {
                // we might have multiple tiles flipping, make sure this is the last one user tapped
                if ( this.model.lastTileTapped === tileModel )
                {
                    // if all flipped, success!!
                    if ( this.model.currentBoard && this.model.currentBoard.remainingMarkedTilesCount === 0 )
                    {
                        // trial was won
                        this.model.currentTrialSucceeded = true;

                        // Lock all tiles
                        this.view.setTilesEnabled( false );

                        // show correct feedback
                        this.view.showFeedbackOverTile( true, tileModel );
                    }
                    // we ran out of flips for this trial, so we fail
                    else if ( this.model.remainingTileFlips === 0 )
                    {
                        this._trialFinishedWithFailure();
                    }
                }
            }
            // this case handles the reveal of the left over tiles
            else if ( this.model.revealingRemainingMarkedTiles && this.model.revealingRemainingMarkedTilesCount > 0 )
            {
                this.model.revealingRemainingMarkedTilesCount--;

                // If there are no more, we are done with this trial
                if ( this.model.revealingRemainingMarkedTilesCount <= 0 )
                {
                    this.model.revealingRemainingMarkedTiles = false;
                    this._trialFinishedWithFailure();
                }
            }
        },

        tileView_onPreviewFinished: function( tileModel )
        {
            if ( this.model.currentTrialTilePreviewComplete === false )
            {
                this.model.currentTrialTilePreviewComplete = true;
                this.view.setTilesEnabled( true );

                if ( this.model.persistentData.isPracticeMode )
                {
                    if ( !this.model.tutorialSkipped ) 
                    {
                        this.view.showTutorialPrompt ( true, "FitTest_TapTiles" );
                    }
                }
            }
        },


        feedback_isLargest: function( feedback )
        {
            if ( feedback && feedback.tileModel && feedback.tileModel.tileView )
            {
                if ( feedback.tileModel.isMarked )
                {
                    feedback.tileModel.tileView.setCorrectTexture();
                }
                else
                {
                    if ( !this.model.persistentData.isPracticeMode )
                    {
                        feedback.tileModel.tileView.setWrongTexture();
                    }
                }
            }
        },

        feedback_onDisappearing: function( feedback )
        {
            this.view.hideFeedback( feedback );
        },

        feedback_onTiny: function( feedback )
        {
            this.view.hideFeedback( feedback );
        },

        feedback_isFinished: function( feedback )
        {
            if ( feedback )
            {
                if ( feedback.tileModel === this.model.lastTileTapped )
                {
                    // trial failed
                    if ( this.model.remainingTileFlips === 0 && !this.model.currentBoardWin )
                    {
                        if ( LumosityUtils.isFitTestModeOn() && this.model.persistentData.isPracticeMode )
                        {
                            this.model.currentBoardWin = true;
                        }
                        // If trial was marked as success, finish the trial
                        if ( this.model.currentTrialSucceeded )
                        {
                            this._trialFinishedWithSuccess();
                        }
                        // If trial was marked as failed or we ran out of flips, we fail
                        else if ( this.model.currentTrialFailed || this.model.remainingTileFlips === 0 )
                        {
                            this._trialFinishedWithFailure();
                        }
                    }
                }
            }
        },



        tileView_onMidRipple: function( tileModel )
        {
            var neighborTiles = this.model.currentBoard.getTileNeighbors( tileModel );
            if ( neighborTiles )
            {
                if ( neighborTiles.length > 0 )
                {
                    var i;
                    for ( i = 0; i < neighborTiles.length; i++ )
                    {
                        if ( !neighborTiles[ i ].isRippled )
                        {
                            neighborTiles[ i ].isRippled = true;
                            neighborTiles[ i ].tileView.rippleTile();

                            this.model.activeRipplesCount++;
                        }
                    }
                }
            }
        },

        tileView_onRippleFinished: function( tileModel )
        {
            this.model.activeRipplesCount--;

            if ( this.model.activeRipplesCount === 0 )
            {
                this._moveToNextTrial();
            }
        },


        beginTrial: function()
        {
            this._super();
            var metadata = this.model.metadata;
            if ( metadata )
            {
            }

            this.model.currentTrial = {};
            var trialMetadata = this.model.currentTrial;
            if ( trialMetadata )
            {
                trialMetadata [ metadataTrialKeys.level ] = this.model.currentMarkedTileCount;

                var boardSize =  this.model.currentBoard.getBoardSize();
                var layoutLength = boardSize.width * boardSize.height;
                trialMetadata [ metadataTrialKeys.tileLayout ] = Array.apply( null, new Array ( layoutLength ) ).map( Number.prototype.valueOf, this.settings.metadataTrialLayoutDefaultCharacter );
            }
        },

        endTrial: function( trialObject, trialCorrect )
        {
            var metadata = this.model.metadata;
            if ( metadata )
            {
                metadata[ metadataAggregateKeys.endLevel ] = this.model.currentMarkedTileCount;
                metadata[ metadataAggregateKeys.maxLevel ] = this.model.maxTileCount;
            }

            var trialMetadata = this.model.currentTrial;
            if ( trialMetadata )
            {
                trialMetadata [ metadataTrialKeys.correct ] = trialCorrect ? "T" : "F";
                trialMetadata [ metadataTrialKeys.correctClicks ] = this.model.currentBoard.markedTilesCount - this.model.currentBoard.remainingMarkedTilesCount;
                trialMetadata [ metadataTrialKeys.tileLayout ] = LumosityUtils.stringifyArray( trialMetadata [ metadataTrialKeys.tileLayout ], ";", this.model.currentBoard.getBoardSize().width );
            }
            this._super( trialObject, trialCorrect, false, false, false, false );
        },

        updateTrialMetadataLayout: function ( x, y, correct )
        {
            var trialMetadata = this.model.currentTrial;
            if ( trialMetadata )
            {
                var boardSize =  this.model.currentBoard.getBoardSize();
                var index = y * boardSize.width + x;

                if ( correct )
                {
                    trialMetadata [ metadataTrialKeys.tileLayout ] [ index ] = this.settings.metadataTrialLayoutCorrectCharacterSequence [ this.model.trialTaps - 1 ];
                }
                else
                {
                    trialMetadata [ metadataTrialKeys.tileLayout ] [ index ] = this.settings.metadataTrialLayoutIncorrectCharacterSequence [ this.model.trialTaps - 1 ];
                }
            }
        },

        /**
         * This method updates current trial metadata with:
         *      first_response_time - Response time of first click in milliseconds
         *      response_time - Response time (of last click) in milliseconds
         */
        updateTrialMetadataResponseTime: function()
        {
            var responseTime = this.getMetadataComponent().getResponseTime();

            var trialMetadata = this.model.currentTrial;
            if ( trialMetadata )
            {
                if ( trialMetadata [ metadataTrialKeys.firstResponseTime ] === undefined )
                {
                    trialMetadata [ metadataTrialKeys.firstResponseTime ] = Math.round( responseTime );
                }

                trialMetadata [ metadataTrialKeys.responseTime ] = Math.round( responseTime );
            }
        },

        configureMetaDataKeys: function()
        {
            this._super();

            metadataTrialKeys.level = "level";                              // Level
            metadataTrialKeys.tileLayout = "tile_layout";                   // Board configuration where period (.) represents empty tiles, zero (0) are unclicked tiles, click sequence is represented following the following order: abcdefghijklmnopqrstuvwxyz!@#$ for correct clicks and 123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ for incorrect clicks. Each row is separated by a semicolon (;)
            metadataTrialKeys.correct = "correct";                          // True if correct
            metadataTrialKeys.correctClicks = "correct_clicks";             // Number of correct clicks made
            metadataTrialKeys.firstResponseTime = "first_response_time";    // Response time of first click in milliseconds
        },

        finalizeMetadata: function()
        {
            this._super();

            var metadata = this.model.metadata;
            if ( metadata )
            {
                metadata [ metadataAggregateKeys.thresholdLevel ] = this.model.getThresholdLevel();
            }
        }
    });

}());
