/*global require*/
/*global cc*/
/*global LumosityButton*/
/*global LumosityButtonState*/
/*global LumosityUtils*/
/*global lu*/

require( "src/View/LumosityButton.js" );
require( "src/Utility/LumosityUtils.js" );


MemoryMatrixBoardTileView = LumosityButton.extend(
{
    /** @lends MemoryMatrixBoardTileView.prototype */

    // Settings : Images
    DefaultTileTexture: "tiles/tileBrown.png",
    MarkedTileTextureNormal: "tiles/tileTeal.png",
    MarkedTileTextureSelected: "tiles/tileDark.png",
    CorrectTileTexture: "tiles/tileCorrect.png",
    WrongTileTextureUp: "tiles/tileWrong.png",
    WrongTileTextureDown: "tiles/tileWrongDark.png",

    // Settings: Size
    tileSize: cc.size ( 100, 100 ),
    imageSize: cc.size ( 90, 90 ),

    // Settings: Timing
    tileFlippingTime: 1,
    tilePreviewTime: 1,
    tileRippleTime: 0.15,

    tileModel: null,

    _delegate: null,
    _settings: null,

    _currentNormalTexture: null,
    _currentSelectedTexture: null,
    _currentDisabledTexture: null,

    /**
    *  MemoryMatrixBoardTileView is the view class for a button Memory Matrix tiles.
    *
    *  @constructs
    */
    ctor: function( tileModel, settings, delegate )
    {
        this._super();

        this._delegate = delegate;
        this._settings = settings;

        this.tileSize = this._settings.tileSize;
        this.imageSize = this._settings.tileImageSize;

        this.nodeName = "MemoryMatrixTile";

        // Setup links
        this.tileModel = tileModel;
        this.tileModel.tileView = this;

        this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        this.setContentSize ( this.tileSize );

        this.setTarget( this );
        this.setPressedCallback( this._onTileTapped );

        if ( settings )
        {
            this.tileFlippingTime = settings.tileFlippingTime;
            this.tilePreviewTime = settings.tilePreviewTime;
        }

        this.setDefaultTexture();
    },

    _onTileTapped: function()
    {
        "use strict";

        if ( this.tileModel )
        {
            if ( this._delegate && this._delegate.tileView_onTileTapped )
            {
                this._delegate.tileView_onTileTapped( this.tileModel );
            }
        }
    },

    setTileSize: function ( tileSize, tileImageSize )
    {
        this.tileSize = tileSize;
        this.imageSize = tileImageSize;

        this._updateTextures();
    },

    _updateTextures: function ()
    {
        var normalTileSprite = cc.Sprite.create( this._currentNormalTexture );
        var selectedTileSprite = cc.Sprite.create( this._currentSelectedTexture );
        var disabledTileSprite = cc.Sprite.create( this._currentDisabledTexture );

        this.setBackgroundImageForState( normalTileSprite, LumosityButtonState.ButtonStateNormal );
        this.setBackgroundImageForState( disabledTileSprite, LumosityButtonState.ButtonStateDisabled );
        this.setBackgroundImageForState( selectedTileSprite, LumosityButtonState.ButtonStateSelected );

        var anchorPoint = cc.p ( 0.5, 0.5 );
        var position = cc.p ( this.tileSize.width / 2, this.tileSize.height / 2 );

        normalTileSprite.setAnchorPoint( anchorPoint );
        normalTileSprite.setPosition ( position );
        lu.scaleForContentSize ( normalTileSprite, this.imageSize );

        selectedTileSprite.setAnchorPoint( anchorPoint );
        selectedTileSprite.setPosition ( position );
        lu.scaleForContentSize ( selectedTileSprite, this.imageSize );

        disabledTileSprite.setAnchorPoint( anchorPoint );
        disabledTileSprite.setPosition ( position );
        lu.scaleForContentSize ( disabledTileSprite, this.imageSize );

        var previousWillResizeBackgroundImage = this.willResizeBackgroundImage;
        this.willResizeBackgroundImage = false;
        this.setContentSize ( this.tileSize );
        this.willResizeBackgroundImage = previousWillResizeBackgroundImage;
    },

    setTileTexture: function( textureNormal, textureSelected )
    {
        this._currentNormalTexture = textureNormal;
        this._currentSelectedTexture = textureSelected;
        this._currentDisabledTexture = textureNormal;

        this._updateTextures();
    },

    setDefaultTexture: function()
    {
        "use strict";

        if ( this.tileModel && this.tileModel.isMarked && lu.isDebugModeOn() )
        {
            this.setTileTexture( this.MarkedTileTextureSelected, this.MarkedTileTextureSelected );
        }
        else
        {
            this.setTileTexture( this.DefaultTileTexture, this.DefaultTileTexture );
        }
    },

    setMarkedTexture: function()
    {
        this.setTileTexture( this.MarkedTileTextureNormal, this.MarkedTileTextureSelected );
    },

    setCorrectTexture: function()
    {
        this.setTileTexture( this.CorrectTileTexture, this.CorrectTileTexture );
    },

    setWrongTexture: function()
    {
        this.setTileTexture( this.WrongTileTextureUp, this.WrongTileTextureDown );
    },

    setRippleTexture: function()
    {
        this.setTileTexture( this.DefaultTileTexture, this.DefaultTileTexture );
    },


    setDefaultState: function()
    {
        this.setScaleX ( 1 );
        this.setScaleY ( 1 );
    },


    flipTile: function()
    {
        var totalFlippingDuration = this.tileFlippingTime;

        var scaleActionPart1 = cc.ScaleTo.create( totalFlippingDuration/2, 0, 1 );
        var scaleActionPart1Ease = cc.EaseOut.create ( scaleActionPart1, 2 );
        var textureAction = cc.CallFunc.create( this.setMarkedTexture, this );
        var scaleActionPart2 = cc.ScaleTo.create( totalFlippingDuration/2, 1, 1 );
        var scaleActionPart2Ease = cc.EaseIn.create ( scaleActionPart2, 2 );

        var callbackAction = cc.CallFunc.create( this._onFlipFinished, this );

        var sequence = cc.Sequence.create( scaleActionPart1Ease, textureAction, scaleActionPart2Ease, callbackAction );

        this.runAction( sequence );
    },

    _onFlipFinished: function()
    {
        "use strict";

        if ( this.tileModel )
        {
            if ( this._delegate && this._delegate.tileView_onFlipFinished )
            {
                this._delegate.tileView_onFlipFinished( this.tileModel );
            }
        }
    },

    previewTile: function()
    {
        "use strict";

        // if tile is marked, we should display proper texture during preview.
        var firstImageFunc = this.setDefaultTexture;
        if ( this.tileModel && this.tileModel.isMarked )
        {
            firstImageFunc = this.setMarkedTexture;
        }

        var totalFlippingDuration = this.tileFlippingTime;

        // First flip
        var scaleActionPart1 = cc.ScaleTo.create( totalFlippingDuration/2, 0, 1 );
        var scaleActionPart1Ease = cc.EaseOut.create ( scaleActionPart1, 2 );
        var textureAction = cc.CallFunc.create( firstImageFunc, this );
        var scaleActionPart2 = cc.ScaleTo.create( totalFlippingDuration/2, -1, 1 );
        var scaleActionPart2Ease = cc.EaseIn.create ( scaleActionPart2, 2 );

        // FIX THIS SHIT
        var delay = cc.DelayTime.create( this.tilePreviewTime );

        // Second flip
        var scaleBackActionPart1 = cc.ScaleTo.create( totalFlippingDuration/2, 0, 1 );
        var scaleBackActionPart1Ease = cc.EaseOut.create ( scaleBackActionPart1, 2 );
        var textureBackAction = cc.CallFunc.create( this.setDefaultTexture, this );
        var scaleBackActionPart2 = cc.ScaleTo.create( totalFlippingDuration/2, 1, 1 );
        var scaleBackActionPart2Ease = cc.EaseIn.create ( scaleBackActionPart2, 2 );

        // Callback
        var callbackAction = cc.CallFunc.create( this._onPreviewFinished, this );

        var uberSequence = cc.Sequence.create( scaleActionPart1Ease, textureAction, scaleActionPart2Ease, delay, scaleBackActionPart1Ease, textureBackAction, scaleBackActionPart2Ease, callbackAction );
        this.runAction( uberSequence );
    },

    _onPreviewFinished: function()
    {
        "use strict";

        if ( this.tileModel )
        {
            if ( this._delegate && this._delegate.tileView_onPreviewFinished )
            {
                this._delegate.tileView_onPreviewFinished( this.tileModel );
            }
        }
    },

    revealTile: function()
    {
        var totalFlippingDuration = this.tileFlippingTime;

        var scaleActionPart1 = cc.ScaleTo.create( totalFlippingDuration/2, 0, 1 );
        var scaleActionPart1Ease = cc.EaseOut.create ( scaleActionPart1, 2 );
        var textureAction = cc.CallFunc.create( this.setMarkedTexture, this );
        var scaleActionPart2 = cc.ScaleTo.create( totalFlippingDuration/2, 1, 1 );
        var scaleActionPart2Ease = cc.EaseIn.create ( scaleActionPart2, 2 );

        var callbackAction = cc.CallFunc.create( this._onRevealFinished, this );

        var sequence = cc.Sequence.create( scaleActionPart1Ease, textureAction, scaleActionPart2Ease, callbackAction );

        this.runAction( sequence );
    },

    _onRevealFinished: function()
    {
        "use strict";

        if ( this.tileModel )
        {
            if ( this._delegate && this._delegate.tileView_onRevealFinished )
            {
                this._delegate.tileView_onRevealFinished( this.tileModel );
            }
        }
    },

    rippleTile: function()
    {
        "use strict";

        var totalDuration = this.tileRippleTime;

        var scaleActionPart1 = cc.ScaleTo.create( totalDuration/2, 0, 1 );
        var scaleActionPart1Ease = cc.EaseOut.create ( scaleActionPart1, 2 );
        var textureAction = cc.CallFunc.create( this.setRippleTexture, this );
        var midCallbackAction = cc.CallFunc.create( this._onMidRipple, this );
        var scaleActionPart2 = cc.ScaleTo.create( totalDuration/2, 1, 1 );
        var scaleActionPart2Ease = cc.EaseIn.create ( scaleActionPart2, 2 );

        var callbackAction = cc.CallFunc.create( this._onRippleFinished, this );

        var sequence = cc.Sequence.create( scaleActionPart1Ease, textureAction, midCallbackAction, scaleActionPart2Ease, callbackAction );

        this.runAction( sequence );
    },

    _onMidRipple: function()
    {
        "use strict";

        if ( this.tileModel )
        {
            if ( this._delegate && this._delegate.tileView_onMidRipple )
            {
                this._delegate.tileView_onMidRipple( this.tileModel );
            }
        }
    },

    _onRippleFinished: function()
    {
        "use strict";

        if ( this.tileModel )
        {
            if ( this._delegate && this._delegate.tileView_onRippleFinished )
            {
                this._delegate.tileView_onRippleFinished( this.tileModel );
            }
        }
    }
});
