/*global cc*/
/*global require*/
/*global PlannerNode*/

require( "src/AI/PlannerNode.js" );

/***************************** BEST FIRST *********************************************************/
// Best First uses problem specific knowledge to speed up the search process.
// Best First is a heuristic search (as opposed to exhaustive) and tries to head right for the goal.
// Instead of keeping the open list sorted by time or cost, it uses the distance of each node to the goal.
// The distance-to-goal serves as a heuristic that promotes the regions of the map that are more likely to lead to the goal.
// This causes the algorithm to head directly to the goal. It is important to note that heading towards
// the goal is not always the right thing to do. Just because you are getting closer to the goal does not
// mean that you will not go down a dead end track. Yet generally speaking, it is a good rule of thumb
// (or estimate) to head towards the goal first.

// Best First is extremely faster than Breadth First and Dijkstra. It typically creates very few nodes and
// finds a rather straight path to the goal. Best First suffers from a few disadvantages. First, it does not
// understand terrain cost. It also heads to the goal and can find a not so optimal path because it is only
// concerned with getting closer to the goal, and not how far a path might have traveled before getting close
// to the goal.

var COMPUTE_HASH_CODE = function( x, y )
{
    "use strict";

    return ( 0xFFFF0000 & ( x << 16 ) | ( 0xFFFF & y ) );
};

var MAX_GRID_SIZE = 35;

var PathPlannerBestFirst;
( function()
{
    "use strict";
    
    PathPlannerBestFirst = cc.Class.extend(
    {   
        _target: null,

        // Used as reference to the current node
        _current: null,

        _open: null,
        _solution: null,

        _startX: 0,
        _startY: 0,
        _endX: 0,
        _endY: 0,
        _columnSize: 0,
        _rowSize: 0,

        _isDone: false,
        _foundGoal: false,

        _closedHashTable: null,
        _closedHashTableSize: 0,

        /** @lends PathPlannerBestFirst.prototype */
        
        /**
         *  PathPlannerBestFirst is the main class for path planning for best first!
         *
         *  @constructs
         */
        ctor: function( target )
        {
            var i;
            this._target = target;
            this._open = [];
            this._solution = [];

            this._closedHashTableSize = MAX_GRID_SIZE;
            this._closedHashTable = [];
            for( i = 0; i < this._closedHashTableSize; i++ )
            {
                this._closedHashTable[ i ] = null;
            }
        },

        planPath: function( startX, startY, endX, endY, columnSize, rowSize )
        {
            this._startX = startX;
            this._startY = startY;
            this._endX = endX;
            this._endY = endY;
            this._columnSize = columnSize;
            this._rowSize = rowSize;

            this._foundGoal = false;
            this._isDone = false;

            this._current = new PlannerNode( this._startX, this._startY );
            this._current.computeHeuristicCost( this._endX, this._endY );
            this.smartInsert( this._open, this._current );
        },

        smartInsert: function( list, node )
        {
            var insertPosition = 0;
            var i;
            var k;
            var l;
            var i1;

            for( k = list.length - 1; i <= k; 0)
            {
                l = ( i + k ) / 2;
                i1 = node.compareBestFirst( list[ l ] );

                // Less
                if( i1 < 0 )
                {
                    i = l + 1;
                    insertPosition = -( i + 1 );
                }
                else if( i1 > 0 ) // Greater
                {
                    k = l - 1;
                    insertPosition = -( i + 1 );
                }
                else // Same
                {
                    insertPosition = l;
                    break;
                }

                if( insertPosition < 0 )
                {
                    insertPosition = ( -insertPosition ) - 1;
                }
            }

            list.splice( insertPosition, 0, node );
        },

        run: function()
        {
            var tempNode = null;
            var finalNode = null;
            var iterator = 0;

            var x = 0;
            var y = 0;
            var tempCurrentX = 0;
            var tempCurrentY = 0;
            var tempX = 0;
            var shouldContinue = false;

            // Not including diagnals
            // Leaving here incase we ever do include diagnals
            var limit = 1;

            // yi represents initial position for y
            var yi;
            var r;

            while( this._open.length )
            {
                this._current = this._open.pop();

                // We found the final spot
                if( this._current.x === this._endX && this._current.y === this._endY )
                {
                    this._foundGoal = true;
                    finalNode = this._current;

                    // Back up, storing the solution
                    while( ( finalNode.x !== this._startX ) ||
                           ( finalNode.y !== this._startY ) )
                    {
                        this._solution.push( finalNode );
                        finalNode = finalNode.parent;
                    }

                    break;
                }

                for( yi = 0; yi < limit; yi++ )
                {
                    // create point
                    x = 1;
                    y = yi;
                    for( r = 0; r < 4; r++ )
                    {
                        shouldContinue = false;

                        // Rotate 90 degrees about ( 0, 0 )
                        tempX = x;
                        x = -y;
                        y = tempX;

                        // Get x, y, in world space
                        tempCurrentX = x + this._current.x;
                        tempCurrentY = y + this._current.y;

                        if( ( tempCurrentX < 0 ) || ( tempCurrentX >= this._columnSize ) )
                        {
                            continue;
                        }

                        if( ( tempCurrentY < 0 ) || ( tempCurrentY >= this._rowSize ) )
                        {
                            continue;
                        }

                        // Check to see if the node is already in the open list
                        for( iterator = 0; iterator < this._open.length; iterator++ )
                        {
                            tempNode = this._open[ iterator ];
                            if( ( tempNode.x === tempCurrentX ) && ( tempNode.y === tempCurrentY ) )
                            {
                                shouldContinue = true;
                            }
                        }

                        if( shouldContinue )
                        {
                            continue;
                        }

                        // Check to see if the node exists in the hash table.
                        if( this.hashTableFind( tempCurrentX, tempCurrentY ) )
                        {
                            continue;
                        }

                        if( this._target && this._target.isValidTile && !this._target.isValidTile( tempCurrentX, tempCurrentY ) )
                        {
                            continue;
                        }

                        // All pass, insert the node into the open lost.
                        tempNode = new PlannerNode( tempCurrentX, tempCurrentY, this._current );
                        tempNode.computeHeuristicCost( this._endX, this._endY );
                        this.smartInsert( this._open, tempNode );
                    }
                }

                this._current.computeHeuristicCost( this._endX, this._endY );
                this.hashTableInsert( this._current );
            }

            this._isDone = true;
        },

        solution: function()
        {
            return this._solution;
        },

        cleanup: function()
        {
            var i;
            this._current = null;
            this._open.length = 0;
            this._solution.length = 0;

            for( i = 0; i < this._closedHashTableSize; i++ )
            {
                this._closedHashTable[ i ] = null;
            }
        },

        // Hash Table Methods
        hashTableInsert: function( node )
        {
            if( this._closedHashTable[ COMPUTE_HASH_CODE( node.x, node.y ) % this._closedHashTableSize ] === null )
            {
                // Empty bucket, insert first node.
                this._closedHashTable[ COMPUTE_HASH_CODE( node.x, node.y ) % this._closedHashTableSize ] = node;
            }
            else if( this._closedHashTable[ COMPUTE_HASH_CODE( node.x, node.y ) % this._closedHashTableSize ] !== null )
            {
                // Found a node there, append to the front.
                node.hashTableNext = this._closedHashTable[ COMPUTE_HASH_CODE( node.x, node.y ) % this._closedHashTableSize  ];
                this._closedHashTable[ COMPUTE_HASH_CODE( node.x, node.y ) % this._closedHashTableSize ] = node;
            }
        },

        hasTableRemove: function( x, y )
        {
            var htTemp = this._closedHashTable[ COMPUTE_HASH_CODE( x, y ) % this._closedHashTableSize ];
            var htStorage = null;

            // Empty bucket
            if( htTemp === null )
            {
                return null;
            }

            // We found who we are looking for
            if( COMPUTE_HASH_CODE( htTemp.x, htTemp.y ) === COMPUTE_HASH_CODE( x, y ) )
            {
                // Remove
                this._closedHashTable[ COMPUTE_HASH_CODE( x, y ) % this._closedHashTableSize ] = htTemp.hashTableNext;
                return htTemp;
            }

            while( htTemp.hashTableNext !== null )
            {
                 if( COMPUTE_HASH_CODE( htTemp.x, htTemp.y ) === COMPUTE_HASH_CODE( x, y ) )
                 {
                    htStorage = htTemp.hashTableNext;
                    htTemp.hashTableNext = htTemp.hashTableNext.hashTableNext;
                    return htStorage;
                 }

                 htTemp = htTemp.hashTableNext;
            }

            return null;
        },

        hashTableFind: function( x, y )
        {
            var htTemp = this._closedHashTable[ COMPUTE_HASH_CODE( x, y ) % this._closedHashTableSize ];

            // Empty bucket
            if( htTemp === null )
            {
                return null;
            }

            while( htTemp !== null )
            {
                if( COMPUTE_HASH_CODE( htTemp.x, htTemp.y ) === COMPUTE_HASH_CODE( x, y ) )
                 {
                    return htTemp;
                 }

                 htTemp = htTemp.hashTableNext;
            }

            return null;
        }
    } );
}() );
