var cc = cc;
var lumosity = lumosity;

var GameSettings = cc.Class.extend(
{
    /** @lends GameSettings.prototype */

    stageSlideDuration: 0.3,

    showScore: true,
    showTime: true,
    showTrials: false,
    showTimeBonus: false,

    // 0 = No Score
    // 1 = Score Multiplier
    // 2 = Score
    scoreType: 0,

    gameLengthInSeconds: 60,
    gameLengthInTrials: 15,

    timeBonus: 0,
    timeBonusDecreasePerSecond: 10,

    // Score Multiplier settings
    minScore: 0,
    pointsPerCorrect: 50,
    pointsPerIncorrect: 0,
    minMultiplier: 1,
    minMeterValue: 1,
    maxMultiplier: 10,
    maxMeterValue: 4,
    meterIncreasePerCorrect: 1,
    meterDecreasePerWrong: 1,
    lastTrialBonus: 0,

    responseAccuracyThreshold: undefined,
    
    // Post-game popup text
    popupTopTextKeys:
    [
        "PopupTopText0",
        "PopupTopText1",
        "PopupTopText2",
        "PopupTopText3",
        "PopupTopText4",
        "PopupTopText5"
    ],
    
    //default values for the feedback widget
    correctAnswerImages :
    [
      "game_feedbackCorrect.png",
      "game_feedbackCorrectX2.png",
      "game_feedbackCorrectX3.png",
      "game_feedbackCorrectX4.png",
      "game_feedbackCorrectX5.png",
      "game_feedbackCorrectX6.png",
      "game_feedbackCorrectX7.png",
      "game_feedbackCorrectX8.png",
      "game_feedbackCorrectX9.png",
      "game_feedbackCorrectX10.png"
    ],
    
    incorrectAnswerImage : "game_feedbackIncorrect.png",
    
    feedbackPulseImages :
    [
      "game_feedbackPulse2Anim000.png",
      "game_feedbackPulse2Anim001.png",
      "game_feedbackPulse2Anim002.png",
      "game_feedbackPulse2Anim003.png",
      "game_feedbackPulse2Anim004.png",
      "game_feedbackPulse2Anim005.png",
      "game_feedbackPulse2Anim006.png",
      "game_feedbackPulse2Anim007.png",
      "game_feedbackPulse2Anim008.png",
      "game_feedbackPulse2Anim009.png",
      "game_feedbackPulse2Anim010.png",
      "game_feedbackPulse2Anim011.png",
      "game_feedbackPulse2Anim012.png",
      "game_feedbackPulse2Anim013.png",
      "game_feedbackPulse2Anim014.png"
    ],
    
    /**
     *  GameSettings represents the different constants that a game may need in order to function.
     *  Most of a settings's contents are read from a json file. However, each individual game can
     *  subclass this in order to add convenience functions or more data.
     *
     *  @constructs
     *  @param {string} fileLocation The location of this settings file on the disk.
     *  @param {boolean} [convertKeysToLowercase=true] If true, then all root keys from the JSON file will be converted to lowerCamelCase
     */
    ctor:function( fileLocation, convertKeysToLowercase )
    {
        //by default, we want all keys to be in lower camel case
        if ( convertKeysToLowercase === undefined )
        {
            convertKeysToLowercase = true;
        }
        
        if ( fileLocation === undefined )
        {
            cc.log( "No file location specified for settings file. It will be empty." );
        }
        else
        {
            //read the settings file and parse it
            var settingsString = lumosity.io_read( fileLocation );
            
            if ( settingsString )
            {
                var settingsFileData = JSON.parse( settingsString, null );
            
                //go through all the values in the settings data and dump it into this object
                var settingKey;
                for ( settingKey in settingsFileData )
                {
                    //since all keys in the settings files are uppercase, we probably want to convert to lower camel case
                    var localKey = settingKey;
                    if ( convertKeysToLowercase )
                    {
                        var firstChar = settingKey.charAt( 0 ).toLowerCase();
                        localKey = firstChar + settingKey.substring( 1 );
                    }
                    this[ localKey ] = settingsFileData[ settingKey ];
                }
            }
        }
    },


    getGameLengthInSeconds: function()
    {
        var retval = this.gameLengthInSeconds;

        if ( LumosityUtils.isFitTestModeOn() )
        {
            retval = this.fitTestLengthInSeconds;
        }

        if ( LumosityUtils.isShortGamesModeOn() )
        {
            retval = 5;
        }

        return retval;
    },


    getGameLengthInTrials: function()
    {
        var retval = this.gameLengthInTrials;

        if ( LumosityUtils.isShortGamesModeOn() )
        {
            retval = this.shortGameLengthInTrials || 3;
        }

        return retval;
    },


    shouldShowHowToPlayOnStart: function()
    {
        var retval = false;
        if ( lumosity.game.params && ( ( lumosity.game.params.how_to_play === "true" ) ||
                                       ( lumosity.game.params.how_to_play === true ) ) )
        {
            if( lumosity.game.params.fit_test_mode === "true" || lumosity.game.params.fit_test_mode === true )
            {
              retval = false;
            }
            else
            {
              retval = true;
            }
        }

        return retval;
    },


    /**
     * Returns the score multiplier at a given index, as determined by the settings file.
     * TODO - move this to a speed scoring component
     *
     * @param {number} index The multiplier index in the multipler array.
     * @returns {number} The multiplier at that index.
     */
    getMultiplier: function( index )
    {
        if ( index >= this.speedScoring.Multipliers.length )
        {
            index = this.speedScoring.Multipliers.length - 1;
        }

        //TODO - clean up this structure when we standardize this
        return this.speedScoring.Multipliers[ index ];
    },

    /**
     * Returns how many scoring multipliers exist in the settings file.
     * TODO - move this to a speed scoring component
     *
     * @returns {number} How many multipliers there are.
     */
    getMultiplierCount: function()
    {
        var retval = 0;

        if ( this.speedScoring && this.speedScoring.Multipliers )
        {
            retval = this.speedScoring.Multipliers.length;
        }

        return retval;
    },

    /**
     * Returns the score multiplier required streak amount at a given index, as determined by the settings file.
     * TODO - move this to a speed scoring component
     *
     * @param {number} index The required multiplier streak index in the multipler array.
     * @returns {number} The streak at that index.
     */
    getMultiplierStreak: function( index )
    {
        if ( this.speedScoring && this.speedScoring.MultiplierStreaks )
        {
            if ( index >= this.speedScoring.MultiplierStreaks.length )
            {
                index = this.speedScoring.MultiplierStreaks.length - 1;
            }

            if ( this.speedScoring.MultiplierStreaks[ index ] < 0 )
            {
                return 0;
            }

            return this.speedScoring.MultiplierStreaks[ index ];
        }
        else
        {
            return 0;
        }
    },

    /**
     * Returns the base score a player gets, according to the settings file.
     * TODO - move this to a speed scoring component
     *
     * @returns {number} The base score.
     */
    getBaseScore: function()
    {
        return this.speedScoring.BaseScore;
    }
});