/*global cc*/
/*global require*/
/*global lu*/

require( "src/Game/Widget/GameWidget.js" );
/*global GameWidget*/

require( "src/Animation/LayoutLoader.js" );
/*global LayoutLoader*/

require( "src/Animation/AnimationController.js" );
/*global AnimationController*/
/*global AnimationChanges*/

var HintOrientation =
{
	Down: 0,
	Up: 1
};

var GameWidgetHint;

( function()
{
    "use strict";

    GameWidgetHint = GameWidget.extend(
    {
    	/** @lends GameWidgetHint.prototype */

    	/** the core square that gets drawn. */
    	boxSprite: null,

    	/** the core triangle that gets drawn. */
    	arrowSprite: null,

    	/** A container for the feeback sprites. */
    	clipper: null,

    	/** Text object for hint text */
    	hintText: null,

    	/** Orientation of the arrow */
    	orientation: null,

    	/* If the hint is currently shown on screen. */
    	showing: null,

        /**
         * Returns Widget type as a string.
         *
         * @returns {string}
         */
        getType: function()
        {
            return "GameWidgetHint";
        },

	        /**
	     *  A GameWidgetHint is shown to display certain hint text
	     *
	     *  @constructs
	     *  @param {HintOrientation} orientation UP or DOWN: Tells which direction the arrow will face
	     *  @param {string} hintText The text to be written on the text box
	     *  @param {colorObject} textColor The color of the Text
	     *  @param {colorObject} hintFont The color of the Text
	     *  @param {integer} optionalWidth The width of the box. If you don't enter a width, it will stetch around the string entered
	     */
	    ctor:function( orientation, hintText, textColor, hintFont, optionalWidth )
	    {
	        this._super();

        	this.orientation = orientation;
        	this.showing = false;

        	this.hintRoot = cc.Node.create();
        	this.hintRoot.nodeName = "hintRoot";
        	this.addChild( this.hintRoot );

        	LayoutLoader.loadLayoutFile( "hint_layout.json", this.hintRoot, this, this );

        	this.animationController = new AnimationController( "hint_animations.json", this, this, [this]);

        	//create a buffer from the flash file
			var buffer = (this.originalBoxWidth - this.originalLabelWidth) / this.originalBoxWidth;

        	//CHANGE THE LABEL
        	if ( hintText )
        	{
        		this.label.setString(hintText);
        	}

        	if ( textColor )
        	{
        		this.setTextColor( textColor );
        	}

        	if ( hintFont )
        	{
        		this.label.setFontName( hintFont );
        	}


        	//Scale the text box to the text, unless somebody passed in an optional width
        	var tempScale;
			
        	if ( optionalWidth )
	        {
	        	tempScale = (optionalWidth / this.boxParent.getContentSize().width) + buffer;
	        }
	        else
	        {
	        	tempScale = (lu.measureText( this.label.getString(), this.label.getFontName(), this.label.getFontSize()) / this.boxParent.getContentSize().width) + buffer;
	       	}

	       	//scale according to the tempscale we figured out
	        this.boxParent.setScaleX( tempScale );

	        this.label.setDimensions( cc.size(0, 0) );

        	this.animationController.play( this._getDirectedAnimationName( "hidden" ) , this._onAnimationFinished, this );//this._getDirectedAnimationName( "hidden" )
	    },

	    setContentSize: function( size, height )
	    {
	    	if ( height === undefined )
	    	{
	    		this._super( size );
	    	}
	    	else
	    	{
	    		this._super( size, height );
	    	}

	    	var cs = this.getContentSize();

	    	this.hintRoot.setPosition( cc.p( 0, cs.height ) );
	    },

	    /**
	     * Creates views.
	     * 
	     * @private
	     */
	    _createViews: function( )
	    {
	    	return;

	    },

	    /**
	     * shows the hint
	     * 
	     * @param {boolean} animated Whether or not you want animation
	     *
	     */
	    show:function ( animated )
	    {
	    	this._setIsShowing(true, animated);
	    },

	    /**
	     * Hides the hint
	     * 
	     * @param {boolean} animated Whether or not you want animation
	     *
	     */
	    hide:function ( animated )
	    {
	    	this._setIsShowing(false, animated);
	    },

	    _setIsShowing:function (needToShow, animated)
	    {
	    	//this.animationController.stopAll( this );

	    	if ( needToShow )
	    	{
	    		if (!this.showing) 
	    		{
		    		if ( animated )
		    		{
		    			this.animationController.play( this._getDirectedAnimationName( "show" ) );
		    		}
		    		else
		    		{
		    			this.animationController.play( this._getDirectedAnimationName( "idle" ) );
		    		}
		    		this.showing = true;
	    		}
	    	}
	    	else
	    	{
	    		if (this.showing) 
	    		{
		    		if ( animated )
		    		{
		    			this.animationController.play( this._getDirectedAnimationName( "hide" ), this._callHiddenAnim );
		    		}
		    		else
		    		{
		    			this.animationController.play( this._getDirectedAnimationName( "hidden" )  );
		    		}
		    		this.showing = false;
	    		}
	    	}
	    },

	    _getDirectedAnimationName:function (animation) {
	    	var str = '';
	    	if ( this.orientation === HintOrientation.Up )
        	{
        		str = animation + "Up";
        	}
        	else if ( this.orientation === HintOrientation.Down )
        	{
				str = animation + "Down";
        	}
        	return str;
	    },

	    /**
	     * Sets the color of the widget's text labels.
	     *
	     * @param colorOrR  Either a color object or the red element of four channel arguments (0-255).
	     * @param g         If four channel arguments supplied, the value in the green channel (0-255).
	     * @param b         If four channel arguments supplied, the value in the blue channel (0-255).
	     * @param a         If four channel arguments supplied, the value in the alpha channel (0-255).
	     */
	    setTextColor:function( colorOrR, g, b, a )
	    {
	        var color;

	        if ( colorOrR.Red )
	        {
	            color = cc.c4b( colorOrR.Red, colorOrR.Green, colorOrR.Blue, colorOrR.Alpha );
	        }
	        else
	        {
	            color = cc.c4b( colorOrR, g, b, a );
	        }

	        this.label.setColor( color );
	    },

	    /**
	     * Sets the text that is currently in the hint
	     *
	     * @param {String} Hint Text The text that is to be displayed
	     * @param {integer} optionalWidth The width of the box. If you don't enter a width, it will stetch around the string entered 
	     */
	    setText:function( hintText, optionalWidth )
	    {	
	    	this.label.setString( hintText );

			var buffer = (this.originalBoxWidth - this.originalLabelWidth) / this.originalBoxWidth;

	    	
        	//Scale the text box to the text, unless somebody passed in an optional width
        	var tempScale;
			
        	if ( optionalWidth )
	        {
	        	tempScale = (optionalWidth / this.boxParent.getContentSize().width) + buffer;
	        }
	        else
	        {
	        	tempScale = (lu.measureText( this.label.getString(), this.label.getFontName(), this.label.getFontSize()) / this.boxParent.getContentSize().width) + buffer;
	       	}

	        this.boxParent.setScaleX( tempScale );
	    },

	    layout_readingData:function( nodeName, nodeData )
	    {
	    	if ( nodeName === "label" )
	    	{
	    		this.originalLabelWidth = nodeData[ LayoutLoader.sizeKey ].width;
	    	}
	    	else if ( nodeName === "hintBoxUp" )
	    	{
	    		this.originalBoxWidth = nodeData[ LayoutLoader.sizeKey ].width;
	    	}
	    },

	    layout_settingNodeResizingInfo:function( nodeName, nodeData, node, resizingData )
	    {
	    	return false;
	    },

	    _callHiddenAnim: function ()
	    {
	    	this.animationController.play( this._getDirectedAnimationName( "hidden" )  );
	    }

    } );
}() );