var Random =
{
    seed: 1,
    
    /**
     * Initializes the random number generator with a seed. If none is passed in, uses the current time.
     *
     * @param {number} [seed=new Date().getTime()] The seed to initialize with.
     */
    initialize: function( seed )
    {
        if ( seed !== undefined )
        {
            this.seed = seed;
        }
        else
        {
            this.seed = new Date().getTime();
        }
    },
    
    /**
     * Generate a random number, similar to Math.random() but it's seedable.
     *
     * It would be safest to use this PRNG or Math.random but not swap one for the other.
     * http://stackoverflow.com/questions/424292/seedable-javascript-random-number-generator
     * This distribution is not uniform.  It is slightly sloped toward extremes 0 and 1.
     * http://jsfiddle.net/bhrLT/17/
     * Resampling with Math.random a few times shows that this seeded PRNG is a little less random.
     *
     * @returns {number} A random floating point value between 0 (inclusive) and 1 (exclusive).
     */
    value:function()
    {
        var x = Math.sin( this.seed++ ) * 10000;
        return x - Math.floor( x );
    },
    
    /**
     * Generate a random number in a range. It's inclusive / exclusive if integers are
     * passed in, or inclusive / inclusive if floats are passed in. If integers are
     * passed in, will return an integer, if floats are passed in, will return a float.
     *
     * @param {number} min The minimum value to generate.
     * @param {number} max The maximum value to generate.
     * @returns {number} A random value between min and max.
     */
    range:function( min, max )
    {
        //integers
        if ( Math.floor( min ) === min && Math.floor( max ) === max )
        {
            return Math.floor( this.integerExclusive( min, max ) );
        }
        
        //floats
        return this.float( min, max );
    },

    /**
     *  Returns a random integer between min and max, both inclusive.
     *
     * @param {number} min The minimum value to generate.
     * @param {number} min The maximum value to generate.
     * @returns {number} A random value between min and max.
     */
    integerInclusive: function( min, max )
    {
        return Random.integer( min, max, false, false );
    },
    
    /**
     *  Returns a random integer between min and max, where max is exclusive.
     *
     * @param {number} min The minimum value to generate.
     * @param {number} min The maximum value to generate.
     * @returns {number} A random value between min and max.
     */
    integerExclusive: function( min, max )
    {
        return Random.integer( min, max, false, true );
    },
    
    /**
     *  Returns a random integer between min and max. This is a flexible function, you can get either
     *  inclusive or exclusive on either min or max.
     *
     * @param {number} min The minimum value to generate.
     * @param {number} min The maximum value to generate.
     * @param {boolean} [exclusiveMin=false] If true, min is exclusive, otherwise it is inclusive (default).
     * @param {boolean} [exclusiveMax=false] If true, max is exclusive, otherwise it is inclusive (default).
     * @returns {number} A random value between min and max.
     */
    integer:function( min, max, exclusiveMin, exclusiveMax )
    {
        var range = ( max - min + ( exclusiveMin ? ( exclusiveMax ? -1 : 0 ) : ( exclusiveMax ? 0 : 1 ) ) );
        var add = exclusiveMin ? min + 1 : min;
        return Math.floor( Random.value() * range + add );
    },
    
    /**
     *  Returns a random float between min and max. This is a inflexible function,
     *  min is always inclusive and max is always exclusive.
     *
     * @param {number} min The minimum value to generate.
     * @param {number} min The maximum value to generate.
     * @returns {number} A random value between min and max.
     */
    float:function( min, max )
    {
        var range = ( max - min );
        return Random.value() * range + min;
    }
};

//set up the random number seed
Random.initialize();

//also overrides Math.random() with our function
Math.random = Random.value.bind( Random );