/*global cc
global require
global LumosityUtils*/

require( "src/Utility/LumosityUtils.js" );

//TODO - this really handles multi-touch poorly...

var LumosityJoystickLayer = cc.Layer.extend(
{
    /** All the joysticks that exist in this layer. They can be anywhere in the child hierarchy. */
    _joysticks: null,

    /** All joysticks that are currently being touched, by touch ID. */
    _touchedJoysticks: null,

    /**
     *  The LumosityJoystickLayer can be used in conjunction with one or more LumosityJoystick's
     *  in order to register joystick movement. To use this, create this layer, then add it
     *  as a child to a view, and finally create one or more joysticks and add them to this with addJoystick().
     *
     *  @constructs
     */
    ctor: function()
    {
        this._super();
        this.init();

        this._joysticks = [];
        this._touchedJoysticks = [];

        this.setTouchEnabled( true );
    },

  /**
     * Adds a joystick to be tested for touches.
     * ( The joystick must be in a child hierarchy inorder to receive touches. )
     *
     * @param {LumosityJoystick} joystick
     */
    addJoystick: function( joystick )
    {
        if( this._joysticks.contains( joystick ) )
        {
            ll.warning("LumosityJoystickLayer", "addJoystick", "joystick already added. It can't be added again.");
            return;
        }

        this._joysticks.push( joystick );
    },

    /**
     * Removes a joystick from touch testing.
     *
     * @param  {LumosityJoystick} joystick
     */
    removeJoystick: function( joystick )
    {
        if( !this._joysticks.contains( joystick ) )
        {
            ll.warning("LumosityJoystickLayer", "removeJoystick", "joystick not found.");
            return;
        }

        this._joysticks.remove( joystick );
    },

    /**
     * Cocos2dX onTouchesBegan overridden implementation.
     *
     * @param  {array} touches
     */
    onTouchesBegan: function( touches )
    {
        var i;
        var touch = null;

        // Sort touches numerically ascending based on touch id because touches
        // don't always come in in order.
        touches.sort( this._numericallyAscendingSort );

        for( i = 0; i < touches.length; i++ )
        {
            touch = touches[ i ];
            this._handleTouchBegan( touch );
        }
    },

    /**
     * Cocos2dX onTouchBegan overridden implementation.
     *
     * @param  {cc.Touch} touch
     */
    onTouchBegan: function( touch )
    {
        this._handleTouchBegan( touch );
    },

    /**
     * Handles touch began for a single touch.
     *
     * @param  {cc.Touch} touch
     */
    _handleTouchBegan: function( touch )
    {
        var touchedJoystick = this._touchedJoystick( touch );
        var touchID = -1;

        if( !touchedJoystick )
        {
            return null;
        }

        touchID = touch.getId();
        if( touchID < 0 )
        {
            return null;
        }

        this._touchedJoysticks[ touchID ] = touchedJoystick;
        touchedJoystick.handleIsPressed( true );

        return touchedJoystick;
    },

    /**
     * Cocos2dX onTouchesMoved overridden implementation.
     *
     * @param  {array} touches
     */
    onTouchesMoved: function( touches )
    {
        var i;
        var touch = null;

        for( i = 0; i < touches.length; i++ )
        {
            touch = touches[ i ];
            this._handleTouchMoved( touch );
        }
    },

    /**
     * Cocos2dX onTouchMoved overridden implementation.
     *
     * @param  {cc.Touch} touch
     */
    onTouchMoved: function( touch )
    {
        this._handleTouchMoved( touch );
    },

    /**
     * Handles touch moved for a single touch.
     *
     * @param  {cc.Touch} touch
     */
    _handleTouchMoved: function( touch )
    {
        var touchId = touch.getId();
        if ( touchId !== undefined )
        {
            var joystick = this._touchedJoysticks[ touchId ];
            if ( !!joystick )
            {
                var deltaPosNode = cc.pSub( touch.getLocation(), touch.getStartLocation() );
                var deltaPosWorld = this.convertToNodeSpace( deltaPosNode );
                joystick.handleTouchMoved( deltaPosWorld );
            }
        }
    },

    /**
     * Cocos2dX onTouchesEnded overridden implementation.
     *
     * @param  {array} touches
     */
    onTouchesEnded: function( touches )
    {
        var i;
        for( i = 0; i < touches.length; i++ )
        {
            this._handleTouchEnded( touches[ i ] );
        }
    },

    /**
     * Cocos2dX onTouchMoved overridden implementation.
     *
     * @param  {cc.Touch} touch
     */
    onTouchEnded: function( touch )
    {
        this._handleTouchEnded( touch );
    },

    /**
     * Handles touch ended for a single touch.
     *
     * @param  {cc.Touch} touch
     */
    _handleTouchEnded: function( touch )
    {
        var touchedJoystick = null;
        var touchID = -1;

        touchID = touch.getId();
        if( touchID < 0 )
        {
            return;
        }

        touchedJoystick = this._touchedJoysticks[ touchID ];
        if( !touchedJoystick )
        {
            return;
        }

        this._touchedJoysticks[ touchID ] = null;
        touchedJoystick.handleIsPressed( false );
    },

    /**
     * Cocos2dX onTouchesCancelled overridden implementation.
     *
     * @param  {array} touches
     */
    onTouchesCancelled: function( touches )
    {
        this.onTouchesEnded( touches );
    },

    /**
     * Cocos2dX onTouchCancelled overridden implementation.
     *
     * @param  {cc.Touch} touch
     */
    onTouchCancelled: function( touch )
    {
        this.onTouchEnded( touch );
    },

    /**
     * Returns the joystick being touched.
     *
     * @param  {cc.Touch} touch
     * @return {LumosityJoystick}
     */
    _touchedJoystick: function( touch )
    {
        var i;
        var location = touch.getLocation();
        var worldPoint = this.convertToWorldSpace( location );
        var joystick = null;
        var touchedJoystick = null;

        for( i = 0; i < this._joysticks.length; i++ )
        {
            var joystick = this._joysticks[ i ];
            if( joystick.containsWorldPoint( worldPoint ) )
            {
                touchedJoystick = this._joysticks[ i ];
                break;
            }
        }

        return touchedJoystick;
    },

    /**
     * Sorts an array of touches numerically ascending by touchId.
     *
     * @param  {cc.Touch} touch1
     * @param  {cc.Touch} touch2
     *
     * @return {int}
     */
    _numericallyAscendingSort: function( touch1, touch2 )
    {
        var touchID1 = touch1.getId();
        var touchID2 = touch2.getId();

        if( touchID1 < touchID2 )
        {
            return -1;
        }

        if( touchID1 > touchID2 )
        {
            return 1;
        }

        return 0;
    }
});