/*global require*/
/*global cc*/
/*global LumosityInverseNode*/
/*global LumosityUtils*/
/*global lumosity*/

require( "src/View/LumosityInverseNode.js" );
require( "src/Utility/LumosityUtils.js" );

/**
 * When you apply a scale to a LumosityLabel, it will inversely scale itself and then
 * increase its font size. This is to keep text crisp.
 *
 * HEY! This is totally weird because you cannot override LabelTTF, because Cocos.
 * _super will actually call LumosityInverseNode functions, but "this" will call
 * LabelTTF functions. Fun! To get at a default LabelTTF function, you must
 * store that function off in the create call below. See setDimensions for an example.
 *
 * @name LumosityLabel
 * @class
 */

var LumosityLabel;
(function()
{
    /** @lends LumosityLabel.prototype */
    "use strict";

    LumosityLabel = LumosityInverseNode.extend(
    {
        // A Boolean value indicating whether the font size should be reduced in order to fit the title string into the label’s content size.
        adjustsFontSizeToFitWidth: false,
        // The size of the smallest permissible font with which to draw the label’s text.
        minimumFontSize: 10,

        /** @lends LumosityLabel.prototype */

        onEnter: function()
        {
            this._super();

            // Comment in to see bounding boxes for all Labels.
            // var points = [ cc.p( 0, 0), 
            //                cc.p( this.getContentSize().width, 0 ),
            //                cc.p( this.getContentSize().width, this.getContentSize().height ),
            //                cc.p( 0, this.getContentSize().height ) ];

            // var boundingBox = LumosityUtils.createDrawNodePoly( points, 2, cc.c4f( 0, 0, 1, 1 ) );
            // this.addChild( boundingBox );
        },

        onExit: function()
        {
            this._super();

            this.setFontSize( this._initialFontSize );
        },

        setInitialFontSize: function( initialFontSize )
        {
            if( !initialFontSize )
            {
                return;
            }
            
            this._initialFontSize = initialFontSize;
        },
        
        getInitialFontSize: function()
        {
            return this._initialFontSize;
        },

        setString: function( string )
        {
            var scaleX = 1;
            var defaultFontSize = 0;
            var stringWidth = 0;
            var contentSizeWidth = 0;

            this._setStringSuper( string );

            // If enabled, adjust the font size to fit contentSize width.
            if( this.adjustsFontSizeToFitWidth )
            {
                this._adjustFontSizeToFitWidth();
            }

            scaleX = 1 / this.getScaleX();
            defaultFontSize = Math.ceil( this._initialFontSize * scaleX );

            //Check wether we are already at the default font size.
            if( defaultFontSize !== this.getFontSize() )
            {
                contentSizeWidth = this.getContentSize().width;
                stringWidth = LumosityUtils.measureText( string, this.getFontName(), defaultFontSize );
                // Compare contentSizeWidth with the stringWidth of the defaultFontSize.
                if( stringWidth < contentSizeWidth )
                {
                    // Set font back to default font size.
                    // ( Was previously shrunk to fit container )
                    this.setFontSize( defaultFontSize );
                }
            }
        },

        setScale: function( scale, scaleY )
        {
            var newFontSize = cc.SizeZero();

            this._super( scale, scaleY );

            newFontSize = this._initialFontSize * scale;
            this.setFontSize( Math.ceil( newFontSize ) );            
        },

        setScaleX: function( newScaleX )
        {
            var newFontSize = cc.SizeZero();

            this._super( newScaleX );

            // Scale font size
            newFontSize = this._initialFontSize * newScaleX;
            this.setFontSize( Math.ceil( newFontSize ) );
        },
        
        setDimensions: function( dims )
        {
            var hierarchyScale = LumosityUtils.getNodeHierarchicalScale( this );
            dims.width *= hierarchyScale.x || 0;
            dims.height *= hierarchyScale.y || 0;
            
            dims.width = Math.floor( dims.width );
            dims.height = Math.floor( dims.height );

            //this is basically the only way to call _super
            this._setDimensionsSuper( dims );
        },
        
        setFontName: function( fontName )
        {
            this._setFontNameSuper( LumosityLabel.getCorrectFontName( fontName ) );
        },

        _adjustFontSizeToFitWidth: function()
        {
            var stringWidth = 0;
            var scaleFactor = 0;
            var contentSizeWidth = 0;
            var scaleX = 1;
            var newFontSize = 0;

            contentSizeWidth = this.getContentSize().width;
            scaleX = 1 / this.getScaleX();
            stringWidth = LumosityUtils.measureText( this.getString(), this.getFontName(), this.getFontSize() );
            newFontSize = Math.ceil( this._initialFontSize * scaleX );

            // Loop until the stringWidth is less than or equal to the content size of the label.
            while( stringWidth > contentSizeWidth )
            {
                scaleFactor = ( contentSizeWidth / stringWidth );

                // Calculate new font size.                
                newFontSize = Math.floor( newFontSize * scaleFactor );
                if( newFontSize < this.minimumFontSize )
                {
                    break;
                }

                // Adjust font size to fit width.
                this.setFontSize( newFontSize );

                contentSizeWidth = this.getContentSize().width;
                stringWidth = LumosityUtils.measureText( this.getString(), this.getFontName(), this.getFontSize() );
            }

            // When stringWidth and contentSizeWidth are equal, it will still try
            // to line wrap, adjust the font size to be smaller to be on the safe side.
            if( stringWidth === contentSizeWidth )
            {
                newFontSize = this.getFontSize() - 1;
                if( newFontSize >= this.minimumFontSize )
                {
                    this.setFontSize( newFontSize );
                }
            }
        }
    } );
}() );

LumosityLabel.getCorrectFontName = function( fontName )
{
    "use strict";
    //for HTML5 fonts, we need to use a font family name, not a location
    if ( lumosity.device.id === "html5" )
    {
        if ( fontName.indexOf( "fonts/" ) === 0 )
        {
            fontName = fontName.substring( "fonts/".length );
        }
        
        //stupid edge case...
        if ( fontName === "Report School" )
        {
            fontName = "ReportSchoolRg-Bold";
        }
        else
        {
            fontName = fontName.replace( /_Italic/g, "" ).replace( /_/g, "-" ).replace( /-/g, "-" ).replace( /.ttf/g, "" ).replace( / /g, "" );
        }
    }
    return fontName;
};

/**
 * To deal with not being able to subclass LabelTTF, you must use this to create
 * a LumosityLabel. It's kind of wonky. Treat it just like a LabelTTF.
 */
LumosityLabel.create = function( label, fontName, fontSize ) 
{
    "use strict";
    var lumosityLabel = new LumosityLabel();
    
    var newLabel = cc.LabelTTF.create( label, LumosityLabel.getCorrectFontName( fontName ), fontSize );
    
    newLabel._initialFontSize = fontSize;
    newLabel.adjustsFontSizeToFitWidth = lumosityLabel.adjustsFontSizeToFitWidth;
    newLabel.minimumFontSize = lumosityLabel.minimumFontSize;
    newLabel.onEnter = lumosityLabel.onEnter;
    newLabel.onExit = lumosityLabel.onExit;
    newLabel.setInitialFontSize = lumosityLabel.setInitialFontSize;
    newLabel.getInitialFontSize = lumosityLabel.getInitialFontSize;
    newLabel.setScale = lumosityLabel.setScale;
    newLabel.setScaleY = lumosityLabel.setScaleY;
    newLabel.setScaleX = lumosityLabel.setScaleX;
    newLabel._getNodeHierarchyScale = lumosityLabel._getNodeHierarchyScale;
    newLabel._adjustFontSizeToFitWidth = lumosityLabel._adjustFontSizeToFitWidth;
    
    //store off the old function so we can call it (like _super)
    newLabel._setDimensionsSuper = newLabel.setDimensions;
    newLabel.setDimensions = lumosityLabel.setDimensions;

    newLabel._setStringSuper = newLabel.setString;
    newLabel.setString = lumosityLabel.setString;
    
    newLabel._setFontNameSuper = newLabel.setFontName;
    newLabel.setFontName = lumosityLabel.setFontName;

    return newLabel;
};
