/*global cc*/
/*global require*/
/*global lumosity*/
/*global Random*/
/*global lg*/
/*global metadataAggregateKeys*/
/*global lu*/
/*global ll*/

var GameModel;
( function()
{
    "use strict";
    
    GameModel = cc.Class.extend(
    {
        /** @lends GameModel.prototype */
        
        /** Whether or not this game should block input. */
        isBlockingInput: true,

        /** Whether or not this game should block input. */
        isGameRunning: false,

        /** A flag that tells us whether the game is over or not. */
        isGameOver: false,

        /** Whether or not this game is paused. */
        isGamePaused: false,

        /** True iff tutorial mode has ended.
         *  Used to distinguish between using onRestart to start the fitTest,
         *  and actually restarting the game. */
        tutorialEnded: false,

        /** This is confusing state-wrangling stuff. It is a janky bool that's inserted to let us
         know when onBegin was called. It's important for edge cases with the How To Play menu
         and where that navigates users depending on global state. Shitty. */
        isOnBeginCalled: false,

        /** True if this is a practice/tutorial game */
        isPracticeMode: false,

        /** True if the tutorial is skipped */
        tutorialSkipped: false,

        /** True if this is a practice/tutorial game */
        restartToPracticeMode: false,

        /** True if this is a fit test game */
        isFitTestMode: false,

        /** True if you should show the timer during a fit test game */
        shouldCreateTimer: null,
        
        /** Timer component */
        timeRemaining: 0,

        /** Time Bonus Component */
        timeBonus: 0,

        /** Score */
        score: 0,
        bonus: 0,
        prevScore: 0,

        /** Min/Max Levels */
        minGameLevel: 0,
        maxGameLevel: 0,
        
        /** Score multiplier component */
        multiplier: 0,
        prevMultiplier: 0,
        scoreMultiplierIndex: 0,

        meterValue: 0,
        prevMeterValue:0,

        pauseTimes:     [],
        resumeTimes:    [],
        pauseTriggers:  [],
        num_defocus:    null,
        
        //COMPTODO some other scoring component?
        totalResponseCount: 0,
        totalCorrectResponseCount: 0,
        consecutiveCorrectCount: 0,
        correctAnswerStreak: 0,
        
        //COMPTODO speed scoring component?
        lastAnswerTime: lumosity.game.frametime,
        averageReactionTime: 0.0,

        // Debug End User Level Variables
        setUserLevel: false,
        debugUserLevel: 1,

        /** Dictionary object that will be sent back to the app with the results. */
        metadata: {},

        /** A dictionary object that is meant to hold keys for metadata. Use in derived classes. */
        currentTrial: {},

        /** Describes keys that are to be used for the stats object when exporting results. */
        StatsKeys:
        {
            CorrectStat: "correct", // Should represent the number of correct trials.
            TrialsStat:  "trials",  // The total number of trials a player reached in their playtime.
            TilesStat:   "tiles",   // The maximum number of tiles a player uncovered in a given trial (board).
            LevelStat:   "level",   // The maximum level a player reached.
            CardStat:    "cards",   // The total correct cards.
            BallStat:    "balls"    // The maximux number of balls player managed in the game.
        },

        persistentData:
        {
            isPracticeMode: false,
            num_restart: 0
        },
        
        /**
         *  A GameModel is a dumb object that contains all members a game may use.
         *  Each time it is run, a new model will be created. Once the game is complete,
         *  its model will be sent to the C++ layer with game completion information.
         *
         *  @constructs
         */
        ctor:function()
        {
            this.score = 0;
            this.pauseTimes = [];
            this.resumeTimes = [];
            this.pauseTriggers = [];
            this.num_defocus = 0;
        },
        
        isPlaying:function()
        {
            return this.isGameRunning && !this.isGamePaused && !this.isGameOver;
        },

        addPause: function( pause, instant )
        {
            this.pauseTimes.push( Math.round( pause * 1000 ) );
            this.pauseTriggers.push( instant ? "F" : "T" );
            if (instant)
            {
                this.num_defocus++;
            }
        },

        addResume: function ( resume )
        {
            this.resumeTimes.push( Math.round( resume * 1000 ) );
        },

        /**
         * Creates a results JSON string to send back to the application layer.
         * 
         * @param  {Object} stats       An object that contains keys describing particular stats.
         * @param  {String} bestStatKey A key with which to grab the "best stat" from.
         * @return {Object}             Results object.
         *
         * GAME-2504 Move session_level and user_level from metadata to game results.  Therefore, analytics can scrape session_level and user_level into game results database and gamesave events, which conventionally reserved a column for 'session_level' and 'user_level' for every game.  Aaron advises that games with levels save the metadata key 'session_level' (for the level being played) and 'user_level' for the level unlocked.  That has been the convention on web games since at least 2012; maybe since 2009.  Test case:  2015-01-23 Download gamesave of Disillusion or Trouble Brewing from s3.  Ethan's gamesave script expects a number in user_level and session_level column.  Got the string 'null'.  Analogous to the Flash bridge, [~aaron] expects Cocos2d base game to move keys to top level of game results.  Code review.  [~medwards] expects to user level appears once in metadata.  2015-03-23 [~aaron] expects to delete session_level.
         *
         * XXX 2015-03-31 Until iOS version reports top-level session_level and user_level, Aaron Kaluszka expects to not delete session_level and user_level.
         */
        createGameResultsData: function ( stats, bestStatKey )
        {
            var results = {};

            results.score = this.score;
            results.bestStatKey = bestStatKey;
            results.gameStats = stats;
            if (undefined !== this.metadata.session_level)
            {
                results.session_level = this.metadata.session_level;
                // XXX delete this.metadata.session_level;
            }
            if (undefined !== this.metadata.user_level)
            {
                results.user_level = this.metadata.user_level;
                // XXX delete this.metadata.user_level;
            }
            results.game_result_data = this.metadata;
            this._populateDefaultGameResultData( results.game_result_data );

            if( stats && bestStatKey )
            {
                results.stat = stats[ bestStatKey ];  //NOTE: I stepped on Mark's toes with this one. We may adopt this change later, but it requires a good amount of app-side work.
            }
            else
            {
                results.stat = 0;
            }

            return results;
        },

        /**
         * Retrieves "user_level" from last game results (passed in from app).
         *
         * @param   firstLevel      If level not found, default to first level.
         * @param   oldStatsKey     Optionally migrate a user who had saved user level to a non-standard key in stats.
         * @param   maxLevel        Clamp maximum level to load.
         * @param   oldStatsOffset  Optionally migrate a user who had saved user level to a non-standard key in stats with non-standard offset to load, such as Memory Matrix.
         *
         * @return  Highest level found, in case the level had been saved in both places, such as if the old key were also saved as "user_level" with a different meaning (as was the case for Memory Matrix).
         *
         * Adapted from Penguin Pursuit, except using metadataAggregateKeys.userLevel as key, which the server expects is "user_level".
         *
         * GAME-2504 Attempts to find user level in last result.  If not found, then attempts to find in game stats.  Test case:  2015-01-27 Code review.  [~medwards] expects to user level appears once in metadata.  A user plays in previous metadata format.  The user upgrades and has a result in previous format.  The program will expect to find the previously formatted user level.
         *
         * GAME-3659 Clamp level.  Make sure it is not below first level or not too high (in case we have users who cheated or achieved higher max tile count on another device or older version).  Test case:  2015-06-15 Memory Matrix saves end level without the next session level offset.  Expect to migrate that user.  Example:  Last session end trial level 10.  Expect next session start level 7.  Inspect metadata log of last session level and next session load.
         */
        loadFromLastResults: function(firstLevel, oldStatsKey, maxLevel, oldStatsOffset)
        {
            if (undefined === oldStatsKey) {
                oldStatsKey = metadataAggregateKeys.userLevel;
            }
            if (undefined === oldStatsOffset) {
                oldStatsOffset = 0;
            }
            var foundLevel;
            var lastLevel = firstLevel;
            var isFound = false;
            var isStat = false;
            if (lumosity && lumosity.game && lumosity.game.params && lumosity.game.params.last_results) {
                var last_results = lumosity.game.params.last_results;
                // ll.verbose("loadFromLastResults: " + JSON.stringify(last_results));
                if ( last_results.hasOwnProperty( metadataAggregateKeys.userLevel ) ) {
                    foundLevel = last_results[metadataAggregateKeys.userLevel];
                    isFound = lu.isNumeric( foundLevel );
                    if (isFound) {
                        lastLevel = foundLevel;
                    }
                }
                var gameStats = last_results.gameStats;
                if (gameStats) {
                    if ( gameStats.hasOwnProperty(oldStatsKey) ) {
                        foundLevel = gameStats[oldStatsKey];
                        isStat = lu.isNumeric( foundLevel );
                        if (isStat) {
                            lastLevel = Math.max(foundLevel + oldStatsOffset, lastLevel);
                        }
                    }
                }
            }
            lastLevel = Math.max(firstLevel, lastLevel);
            if (undefined !== maxLevel) {
                lastLevel = Math.min(lastLevel, maxLevel);
            }
            return lastLevel;
        },

        /**
         * Creates game results using 'createGameResultsData' call  and stringifies them.
         *
         * @param  {Object} stats       An object that contains keys describing particular stats.
         * @param  {String} bestStatKey A key with which to grab the "best stat" from.
         * @return {String}             String version of results object.
         */
        exportResults:function( stats, bestStatKey )
        {
            if (this.setUserLevel) {
                this.metadata.user_level = debugUserLevel;   
            }
            var results = this.createGameResultsData ( stats, bestStatKey );

            cc.log( "results: " + JSON.stringify(results) );

            return JSON.stringify(results);
        },
        
        _populateDefaultGameResultData:function( data )
        {
            data.locale_id = lumosity.game.locale;
            data.default_locale_id = "en";
            data.version = this._findMetadataVersionString();
            data.seed = Random.seed;
        },

        /**
         * Increase/decrease the debug option for setting the user_level at end of game play
         *
         * @param {int} value  Increase/Decrease amount for the debug used_level value
         */
        _incrementDebugUserLevel: function(value)
        {
            if (!(this.debugUserLevel + value > this.maxGameLevel) && !(this.debugUserLevel + value < this.minGameLevel))
            {
                this.debugUserLevel += value;            
            }
        },         

        responseAccuracy: function()
        {
            var accuracy = 0;
            if( this.totalResponseCount === 0 ||  this.totalCorrectResponseCount === 0 )
            {
                return 0;
            }

            accuracy = this.totalCorrectResponseCount / this.totalResponseCount;
            return accuracy;
        },

        hasMetValidResponseAccuracy: function( threshold )
        {
            cc.log( "Response Accuracy: " + this.responseAccuracy() );
            cc.log( "Response Accuracy Threshold: " + threshold ); 

            if( threshold === undefined )
            {
                ll.warning( "GameModel", "hasMetValidResponseAccuracy", "threshold is undefined." );
                return true;
            }

            // validResponseAccuracy has been defined, need to check if we have met response accuracy threshold.
            if( this.responseAccuracy() < threshold )
            {
                // Didn't meet valid response accuracy.
                return false;
            }

            return true;
        },
        
        _findMetadataVersionString: function()
        {
            //so - due to the old way versions worked where they were based on git commits, we were at something like
            //version 6000 for most games. That means we can't simply use our nice actualy version number from npm.
            //SO, we're multiplying each digit by 10000, 100, and 1, respectively.
            var mdVersion = String( lg.buildNumber );
            if ( !mdVersion )
            {
                return "0";
            }
            
            var versionParts = mdVersion.split( "." );
            if ( versionParts.length < 3 )
            {
                return mdVersion;
            }
            
            return String( 10000 * Math.floor( versionParts[0] ) + 100 * Math.floor( versionParts[1] ) + Math.floor( versionParts[2] ) );
        }
    });

}());
