/*global cc*/
/*global require*/
/*global GameWidget*/
/*global GameWidgetFlipbook*/
/*global AnimationController*/
/*global AnimationChanges*/
/*global LumosityLabel*/
/*global LumosityUtils*/
/*global LayoutLoader*/
/*global LayoutNodeFlags*/

require( "src/Game/Widget/GameWidget.js" );
require( "src/Game/Widget/GameWidgetFlipbook.js" );
require( "src/Animation/AnimationController.js" );
require( "src/View/LumosityLabel.js" );
require( "src/Utility/LumosityUtils.js" );
require( "src/Animation/LayoutLoader.js" );

var GameWidgetCountdown;
( function()
{
     "use strict";

    GameWidgetCountdown = GameWidget.extend(
    {
        _designSize: cc.size( 200, 200 ),
        _numberPos: cc.PointZero(),

        _countDownLabel: null,
        _containerNode: null,
        
        _numberTextData: null,

        /**
         * Returns Widget type as a string.
         *
         * @returns {string}
         */
        getType: function()
        {
            return "GameWidgetCountdown";
        },

        
        /**
         *  A GameWidgetCountdown shows during game begin animation that signifies start of the actual gameplay.
         *
         *  @constructs
         *  @param {Object} settings An object that contains popup-related settings.
         *  @param {Object} [delegate] An object to call popup-related functions on.
         *  @param {boolean} [preload=true] Whether or not to preload all popup images upon creation.
         */
        ctor:function( settings, delegate, preload )
        {
            var screenSize = cc.Director.getInstance().getWinSize();

            this._super();
            
            this._delegate = delegate;
            
            this.setContentSize( screenSize );
            this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this.setPosition( cc.p( screenSize.width >> 1, screenSize.height >> 1 ) );

            //by default, preload all our images on creation
            if( ( preload === undefined || preload ) && settings !== undefined )
            {
                this.preload( settings );
            }

            // Load the layout.
            this._loadLayout();

            // Setup animation controller
            this._createAnimationController();

            // Create the views.
            this._createViews();

            // Hide on creation.
            this.setVisible( false );
        },

        /**
         * Preloads images
         *
         * @param settings Game settings that provides
         */
        preload:function( settings )
        {
        },

        _loadLayout: function()
        {
            LayoutLoader.loadLayoutFile( "countdown_layout.json", this, this, this );
        },

        _createAnimationController:function()
        {
            this.animationController = new AnimationController( "countdown_animations.json", this, this, [ this ] );
            AnimationChanges.addNodeIgnoredChanges( this.countDown, [ AnimationChanges.Position ] );
        },

        _createViews: function()
        {
            var countDownPos = cc.PointZero();
            
            // Create containerNode and add it to countDown.
            this._containerNode = cc.Node.create();
            this._containerNode.setContentSize( this._designSize );
            this._containerNode.setPosition( cc.PointZero() );
            this._containerNode.setAnchorPoint( cc.PointZero() );
            this._containerNode.nodeName = "containerNode";
            this.countDown.addChild( this._containerNode );
            
            // Remove scale applied by countDown.
            this._containerNode.setScaleX( this._containerNode.getScaleX() / this.countDown.getScaleX() );
            this._containerNode.setScaleY( this._containerNode.getScaleY() / this.countDown.getScaleY() );
            
            // Calculate countDown position.
            this.countDownContainer.removeFromParent( false );
            this.countDownContainer.setContentSize( this.getContentSize() );
            this.countDownContainer.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this.countDownContainer.setPosition( cc.p( this.getContentSize().width / 2, this.getContentSize().height / 2 ) );
            this.addChild( this.countDownContainer );
            countDownPos = cc.p( this.countDownContainer.getContentSize().width / 2 , this.countDownContainer.getContentSize().height / 2 );
            
            // Add countDown to countDownContainer.
            this.countDown.removeFromParent( false );
            this.countDownContainer.addChild( this.countDown );
            this.countDown.setPosition( countDownPos );
            
            this._numberPos = cc.p( this._containerNode.getContentSize().width / 2, this._containerNode.getContentSize().height / 2 );
                
            // Create the count down label.
            this._countDownLabel = this._createCountDownLabel( this._numberTextData );
            this._countDownLabel.setPosition( this._numberPos );
            this._containerNode.addChild( this._countDownLabel );
        },

        _createCountDownLabel: function( textData )
        {
            var text = "4";
            var font = "fonts/MuseoSans-700.ttf";
            var fontSize = 100;
            var colorData = null;
            var color = cc.c3b( 255, 255, 255 );
            var countDownLabel = null;

            text = textData[ LayoutLoader.textStringKey ];
            font = LumosityUtils.getFontLocation( textData[ LayoutLoader.fontKey ] );
            fontSize = textData[ LayoutLoader.fontSizeKey ];
            colorData = textData[ LayoutLoader.fontColorKey ];
            color = LayoutLoader.getColor( colorData );
            
            countDownLabel = this._createLabel( text, font, fontSize, color );
            countDownLabel.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            countDownLabel.setHorizontalAlignment( cc.TEXT_ALIGNMENT_CENTER );
            
            return countDownLabel;
        },

        _createLabel: function( text, font, fontSize, color )
        {
            var label = null;

            label = LumosityLabel.create( text, font, fontSize );
            label.nodeName = "countDownLabel";
            label.setColor( color );

            return label;
        },

        layout_readingData: function( nodeName, nodeData )
        {
            if( nodeName === "numberText" )
            {
                this._numberTextData = nodeData[ LayoutLoader.textDataKey ];
            }
        },

        layout_creatingNode: function( nodeName, nodeData )
        {
            var result = null;

            if( nodeName === "numberText" )
            {
                //this._numberTextData = nodeData[ LayoutLoader.textDataKey ];
                result = false;
            }

            return result;
        },

        /**
         * Plays animation that shows countdown.
         */
        play: function()
        { 
            // Stop previous animation
            this.animationController.stopAll( true );

            // Show the countdown.
            this.setVisible( true );

            // Update text.
            this._setCountdownString( 4 );

            // Play countdown animation.
            this.animationController.play( "countdown", this._onFinished, this );
        },

        /**
         * Gets called whenever animation is finished.
         *
         * @private
         */
        _onFinished: function()
        {
            this.setVisible( false );

            if( this._delegate && this._delegate.countdown_onFinished !== undefined )
            {
                this._delegate.countdown_onFinished();
            }
        },

        /**
         *  Set the string within the gameBegin countdown. Currently expects the label name
         *  "countdown numbers.text" but this may change.
         *
         *  @param {string} value The countdown string to show. i.e. "3", "2", "1".
         */
        _setCountdownString:function( value )
        {
            if( !this._countDownLabel )
            {
                return;
            }

            this._countDownLabel.setString( value );
        },

        /**
         *  Cue function that sets the countdown string to 3.
         */
        cue_countdown3:function()
        {
            this._countDownLabel.setPosition( this._numberPos );
            this._setCountdownString( "3" );

            if( this._delegate && this._delegate.countdown_onThree !== undefined )
            {
                this._delegate.countdown_onThree();
            }

            //LumosityUtils.printNodeHierarchy( this.getParent() );
        },

        /**
         *  Cue function that sets the countdown string to 2.
         */
        cue_countdown2:function()
        {
            this._countDownLabel.setPosition( this._numberPos );
            this._setCountdownString( "2" );

            if( this._delegate && this._delegate.countdown_onTwo !== undefined )
            {
                this._delegate.countdown_onTwo();
            }
        },

        /**
         *  Cue function that sets the countdown string to 1.
         */
        cue_countdown1:function()
        {
            this._countDownLabel.setPosition( this._numberPos );
            this._setCountdownString( "1" );

            if( this._delegate && this._delegate.countdown_onOne !== undefined )
            {
                this._delegate.countdown_onOne();
            }
        },


        /**
         * Called whenever game is paused.
         */
        onPause: function()
        {
            this._super();

            if( this.animationController )
            {
                this.animationController.pause( true );
            }
        },

        /**
         * Called whenever game is unpaused (resumed).
         */
        onResume: function()
        {
            this._super();

            if( this.animationController )
            {
                this.animationController.resume( true );
            }
        },

        onRestart: function()
        {
            this._super();

            this.setVisible( false );

            // Update text
            this._setCountdownString( 4 );
        },

        layout_settingNodeFlags: function( nodeName, nodeData, node )
        {
            var nodeFlags = 0xFF;
            if( nodeName === "countDown" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }
            else if( nodeName === "countDownContainer" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }

            return nodeFlags;
        }
    } );
}() );
