/*global cc*/
/*global require*/
/*global lu*/
/*global ll*/
/*global localize*/
/*global GameWidget*/
/*global GameWidgetFlipbook*/
/*global LayoutLoader*/
/*global AnimationController*/
/*global AnimationChanges*/

require( "src/Game/Widget/GameWidget.js" );
require( "src/Game/Widget/GameWidgetFlipbook.js" );
require( "src/Animation/AnimationController.js" );
require( "src/Animation/LayoutLoader.js" );
require( "src/Utility/LumosityUtils.js" );


var GameWidgetPopup = GameWidget.extend(
{
    /** @lends GameWidgetPopup.prototype */

    /** A core object that gets drawn. */
    sprite: null,

    _backgroundSpriteStroke: null,
    _backgroundSpriteFill: null,
    _nodeTop: null,
    _labelTop: null,
    _nodeCenter: null,
    _labelCenter: null,
    _nodeButtom: null,
    _labelBottom: null,

    _textFirst: [],
    _textSecond: [],
    _textThird: [],
    _textFourth: [],
    
    /** The full scale of this popup widget. Can be increased to change the maximum size. */
    _fullScale: 1.0,


    /**
     * Returns Widget type as a string.
     *
     * @returns {string}
     */
    getType: function()
    {
        return "GameWidgetPopup";
    },

    
    /**
     *  A GameWidgetPopup shows in game popup window that can have up to 3 slides and up to 3 lines of text
     *  per slide.
     *
     *  @constructs
     *  @param {Object} settings An object that contains popup-related settings.
     *  @param {Object} [delegate] An object to call popup-related functions on.
     *  @param {boolean} [preload=true] Whether or not to preload all popup images upon creation.
     */
    ctor:function( settings, delegate, preload )
    {
        this._super();
        
        this._delegate = delegate;

        this._fullScale = this.getScale();

        this._buttons = [];

        this._textFirst = [];
        this._textSecond = [];
        this._textThird = [];
        this._textFourth = [];

        // Get screensize
        var screenSize = cc.Director.getInstance().getWinSize();

        // Content size needs to match the size of the screen in the flash file (ie 640x960 for iPhone 1-4)
        this.setContentSize( screenSize );

        // Make sure it is positioned dead center
        this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        this.setPosition( cc.p( screenSize.width / 2, screenSize.height / 2 ) );
        
        //by default, preload all our images on creation
        if ( ( preload === undefined || preload ) && settings !== undefined )
        {
            this.preload( settings );
        }

        // Load the layout
        this._loadLayout();

        // Setup animation controller
        this._createAnimationController();

        // HACK: ignore dimensions from the flash file
        this._setDimensions();

        // Hide all elements
        this._hideAll();
    },

    /**
     * Preloads images
     *
     * @param settings Game settings that provides
     */
    preload:function( settings )
    {
    },

    _loadLayout: function()
    {
        LayoutLoader.loadLayoutFile( "popup_layout.json", this, this,  this._layoutCreatingNode.bind( this ) );
    },

    _layoutCreatingNode:function( nodeName, nodeData )
    {
        switch( nodeName )
        {
            case "popUpWhite":
                this._backgroundSpriteFill = cc.Sprite.create( "whiteSquare.png" );
                this._backgroundSpriteFill.setAnchorPoint( cc.p( 0.5, 0.5 ) );
                this._backgroundSpriteFill.setColor( cc.WHITE );
                return this._backgroundSpriteFill;

            case "popUpStroke":
                this._backgroundSpriteStroke = cc.Sprite.create( "whiteSquare.png" );
                this._backgroundSpriteStroke.setAnchorPoint( cc.p( 0.5, 0.5 ) );
                this._backgroundSpriteStroke.setColor( cc.BLACK );
                return this._backgroundSpriteStroke;
        }
        return null;
    },


    /**
     *  Creates this view's AnimationController from the game's settings file. That way of doing it
     *  is temporary, later we will want to pass the animation file name into this function. The
     *  controller is also automatically created with the shared game animation data.
     */
    _createAnimationController:function()
    {
        this.animationController = new AnimationController( "popup_animations.json", this, this, [ this ] );
    },

    /**
     * Updates text arrays that will be used to set up labels during animation.
     *
     * @param slide
     * @private
     */
    _updateLabelText: function( slide )
    {
        var index = slide - 1;

        if ( this[ "popUpText1" ] && this._textFirst && this._textFirst.length > index )
        {
            this[ "popUpText1" ].setString( this._textFirst[index] );
        }

        if ( this[ "popUpText2" ] && this._textSecond && this._textSecond.length > index )
        {
            this[ "popUpText2" ].setString( this._textSecond[index] );
        }

        if ( this[ "popUpText3" ] && this._textThird && this._textThird.length > index )
        {
            this[ "popUpText3" ].setString( this._textThird[index] );
        }

        if ( this[ "popUpText4" ] && this._textFourth && this._textFourth.length > index )
        {
            this[ "popUpText4" ].setString( this._textFourth[index] );
        }
    },

    _setDimensions: function()
    {
        if ( this[ "popUpText1" ] )
        {
            this[ "popUpText1" ].setDimensions( cc.size( 0 ,0 ) );
        }

        if ( this[ "popUpText2" ] )
        {
            this[ "popUpText2" ].setDimensions( cc.size( 0 ,0 ) );
        }

        if ( this[ "popUpText3" ] )
        {
            this[ "popUpText3" ].setDimensions( cc.size( 0 ,0 ) );
        }

        if ( this[ "popUpText4" ] )
        {
            this[ "popUpText4" ].setDimensions( cc.size( 0 ,0 ) );
        }
    },

    /**
     * Hides all visible components.
     *
     * @private
     */
    _hideAll: function()
    {
        var childNodes = this.getChildren();
        var childIndex;
        var childNode;
        for( childIndex = 0; childIndex < childNodes.length; childIndex++ )
        {
            childNode = childNodes[ childIndex ];
            if ( childNode.nodeName !== "popUpNode" )
            {
                childNode.setVisible( false );
            }
        }
        this.popUpText1.setVisible( false );
        this.popUpText2.setVisible( false );
        this.popUpText3.setVisible( false );
        if( this.popUpText4 )
        {
            this.popUpText4.setVisible( false );
        }
    },

    /**
     * Updates text array with provided values.
     *
     * @param array
     * @param first
     * @param second
     * @param third
     * @private
     */
    _updateTextArray: function( array, first, second, third )
    {
        if ( array )
        {
            // Clear array
            array.length = 0;

            if ( first !== undefined )
            {
                array.push( first );

                if ( second !== undefined )
                {
                    array.push( second );

                    if ( third !== undefined )
                    {
                        array.push( third );
                    }
                }
            }
        }
    },

    setFirstText: function( first, second, third )
    {
        this._updateTextArray( this._textFirst, first, second, third );
    },

    setSecondText: function( first, second, third )
    {
        this._updateTextArray( this._textSecond, first, second, third );
    },

    setThirdText: function( first, second, third )
    {
        this._updateTextArray( this._textThird, first, second, third );
    },

    setFourthText: function( first, second, third )
    {
        this._updateTextArray( this._textFourth, first, second, third );
    },

    /**
     * Plays animation that corresponds to number of slides and lines.
     *
     * @param slides Number of slides to show. (3 is max)
     * @param lines Number if lines per slide. (3 is max)
     */
    play: function( slides, lines )
    {
        // Ensure the widget (node) is visible.
        this.setVisible( true );

        // Stop previous animation
        this.animationController.stopAll( true );

        // Hide all visible elements
        this._hideAll();

        // Update text
        this._updateLabelText( 1 );

        this.isPlaying = true;

        // Start animations
        if ( this.animationController )
        {
            if ( slides === 1 )
            {
                if ( lines === 1 )
                {
                    this.animationController.play( "popUp1Line1Slide", this._onFinished, this );
                }
                else if ( lines === 2 )
                {
                    this.animationController.play( "popUp2Lines1Slide", this._onFinished, this );
                }
                else if ( lines === 3 )
                {
                    this.animationController.play( "popUp3Lines1Slide", this._onFinished, this );
                }
                else if ( lines === 4 )
                {
                    this.animationController.play( "popUp4Lines1Slide", this._onFinished, this );
                }
            }
            else if ( slides === 2 )
            {
                if ( lines === 1 )
                {
                    this.animationController.play( "popUp1Line2Slides", this._onFinished, this );
                }
                else if ( lines === 2 )
                {
                    this.animationController.play( "popUp2Lines2Slides", this._onFinished, this );
                }
                else if ( lines === 3 )
                {
                    this.animationController.play( "popUp3Lines2Slides", this._onFinished, this );
                }
            }
            else if ( slides === 3 )
            {
                if ( lines === 2 )
                {
                    this.animationController.play( "popUp2Lines3Slides", this._onFinished, this );
                }
                else if ( lines === 3 )
                {
                    this.animationController.play( "popUp3Lines3Slides", this._onFinished, this );
                }
            }
        }
    },

    /**
     * Gets called whenever animation is finished.
     *
     * @private
     */
    _onFinished: function()
    {
        if ( this._delegate && this._delegate.popup_isFinished !== undefined )
        {
            this._delegate.popup_isFinished();
        }
    },

    /**
     * Cue called by animation, updates text for second slide.
     */
    cue_changePopupTextSlide2: function()
    {
        this._updateLabelText( 2 );
    },

    /**
     * Cue called by animation, updates text for third slide.
     */
    cue_changePopupTextSlide3: function()
    {
        this._updateLabelText( 3 );
    },

    onPause: function()
    {
        this._super();
        this.animationController.pause();
    },

    onRestart: function()
    {
        this._super();
        this.setVisible( false );

        // Stop previous animation
        this.animationController.stopAll( true );

        // Hide all visible elements
        this._hideAll();
        // this.animationController.stopAll( true );
    },

    onResume: function()
    {
        this._super();
        this.animationController.resume();
    }
});