/*global cc*/
/*global require*/
/*global GameWidget*/
/*global GameWidgetFlipbook*/
/*global AnimationController*/
/*global LumosityButtonState*/
/*global LumosityButton*/
/*global LumosityUtils*/
/*global LayoutLoader*/
/*global AnimationChanges*/

require( "src/Game/Widget/GameWidget.js" );
require( "src/Game/Widget/GameWidgetFlipbook.js" );
require( "src/Animation/AnimationController.js" );
require( "src/View/LumosityButton.js" );
require( "src/Utility/LumosityUtils.js" );
require( "src/Animation/LayoutLoader.js" );

var GameWidgetSkipTutorialButton;

( function()
{
    "use strict";

    GameWidgetSkipTutorialButton = GameWidget.extend(
    {
        /** @lends GameWidgetSkipTutorialButton.prototype */

        /** Game Model */
        _model: null,

        /** The button tapped to skip tutorials **/
        _skipButton: null,

        /** Whether the skip button is currently showing */
        _isShown: false,

        /**
         * Returns Widget type as a string.
         *
         * @returns {string}
         */
        getType: function()
        {
            return "GameWidgetSkipTutorialButton";
        },

        /**
         *  A GameWidgetSkipTutorialButton shows in game pause menu.
         *
         *  @constructs
         *  @param {Object} settings An object that contains popup-related settings.
         *  @param {Object} [delegate] An object to call popup-related functions on.
         */
        ctor:function( model, settings, delegate )
        {
            this._super();

            this._model = model;

            this._delegate = delegate;

            this._buttons = [];

            // Get screensize
            var screenSize = cc.Director.getInstance().getWinSize();

            // Content size needs to match the size of the screen
            this.setContentSize( screenSize );

            // Make sure it is positioned dead center
            this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this.setPosition( cc.p( screenSize.width / 2, screenSize.height / 2 ) );

            // Load the layout.
            this._loadLayout();

            // Create the pause button to be clipped
            this._createViews();

            // Setup animation controller
            this._createAnimationController();
        },

        /**
         * Shows or hides the skip tutorial button.
         *
         * @param show (Boolean) Whether to show or hide the button.
         */
        showButton: function( show )
        {
            if ( this._isShown !== show )
            {
                this._isShown = show;

                if ( this._isShown )
                {
                    this.setVisible( true );
                    this.animationController.play( "buttonIn" );
                }
                else
                {
                    this.animationController.play( "buttonOut" );
                }
            }
        },

        _loadLayout: function()
        {
            LayoutLoader.loadLayoutFile( "skipTutorialButton_layout.json", this, this , this );
        },
        
        layout_creatingNode: function( nodeName, nodeData )
        {
             switch ( nodeName )
             {
                 case "skipLabel":
                     this._skipLabelTextData = nodeData.textData;
                     break;
             }
            return null;
        },

        /**
         * Creates Pause Button and all of its elements.
         *
         * @private
         */
        _createViews: function()
        {
            this._createSkipButton();
        },

        /**
         * Creates the actual skip tutorial button.
         *
         * @returns {LumosityButton}
         * @private
         */
        _createSkipButton:function()
        {
            // get text data that was previously read in from the Flash file's skip label
            var textData = this._skipLabelTextData;
            var text = textData[ LayoutLoader.textStringKey ];
            var textKey = textData[ LayoutLoader.textKeyKey ];
            var localizedText = localize( textKey, text );
            var font = LumosityUtils.getFontLocation( textData[ LayoutLoader.fontKey ] );
            var fontSize = textData[ LayoutLoader.fontSizeKey ];
            var colorData = textData[ LayoutLoader.fontColorKey ];
            var fontUpColor = LayoutLoader.getColor( colorData );
            var fontDownColor = cc.WHITE;

            // localize skip label to get its final content size
            this.skipLabel.setDimensions( cc.size( 0 ,0 ) );
            this.skipLabel.setString( localizedText );

            // calculate final content size of new button
            var buttonSideMargin = this.skipLabel.getContentSize().height * 0.45;
            var buttonContentSize = cc.size( this.skipLabel.getContentSize().width + buttonSideMargin * 2,
                this.buttonContainer.getContentSize().height );

            // actually create LumosityButton based on skip label found in Flash file
            var button = new LumosityButton();

            button.nodeName = "skipTutorialButton";
            this.buttonContainer.addChild( button );
            this._buttons.push( button );
            this._skipButton = button;

            var bgUpImage = cc.Sprite.create( "buttons/pauseButtonBackgroundUp.png" );
            var bgDownImage = cc.Sprite.create( "buttons/pauseButtonBackgroundDown.png" );
            var bgDisabledImage = cc.Sprite.create( "buttons/pauseButtonBackgroundUp.png" );

            button.setBackgroundImageForState( bgUpImage, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( bgDownImage, LumosityButtonState.ButtonStateSelected );
            button.setBackgroundImageForState( bgDisabledImage, LumosityButtonState.ButtonStateDisabled );

            button.setContentSize( buttonContentSize );
            button.setAnchorPoint( cc.POINT_ZERO );
            button.setPosition( cc.POINT_ZERO );
            button.setTarget( this );
            button.setReleasedCallback( this._skipButtonTapped );

            button.setTitleColorForState( fontUpColor, LumosityButtonState.ButtonStateNormal );
            button.setTitleColorForState( fontDownColor, LumosityButtonState.ButtonStateSelected );
            button.setTitleColorForState( fontUpColor, LumosityButtonState.ButtonStateDisabled );

            button.setTitleForState( localizedText, LumosityButtonState.ButtonStateNormal );
            button.titleLabel.setFontName( font );
            button.titleLabel.setInitialFontSize( fontSize );
            button.titleLabel.setContentSize( this.skipLabel.getContentSize() );

            button.titleLabel.setAnchorPoint( this.skipLabel.getAnchorPoint() );
            button.titleLabel.setHorizontalAlignment( this.skipLabel.getHorizontalAlignment() );
            button.titleLabel.setPosition( buttonContentSize.width * 0.5, buttonContentSize.height * 0.5 );

            return button;
        },

        /**
         *  Creates this view's AnimationController from the game's settings file. That way of doing it
         *  is temporary, later we will want to pass the animation file name into this function. The
         *  controller is also automatically created with the shared game animation data.
         */
        _createAnimationController:function()
        {
            this.animationController = new AnimationController( "skipTutorialButton_animations.json", this, this, [ this ] );
            this.animationController.play( "layout" );
        },

        /**
         * Internal callback for the skip button tap.
         *
         * @private
         */
        _skipButtonTapped: function()
        {
            if( this._delegate && this._delegate.skipTutorialButton_Tapped )
            {
                this._delegate.skipTutorialButton_Tapped();
            }
        }
      } );
}() );