/*global require*/
/*global cc*/
/*global LumosityLabel*/
/*global ll*/
/*global LumosityUtils*/

require( "src/View/LumosityLabel.js" );
require( "src/Utility/LumosityLogger.js" );
require( "src/Utility/LumosityUtils.js" );

// make this true to see hit region outlines
var debugShowHitRegions = false;

var LumosityButtonState = 
{
    ButtonStateNormal:         "ButtonStateNormal",
    ButtonStateHighlighted:    "ButtonStateHighlighted",
    ButtonStateDisabled:       "ButtonStateDisabled",
    ButtonStateSelected:       "ButtonStateSelected"
};

var LumosityButton;
( function()
{
    "use strict";
    LumosityButton = cc.Node.extend(
    {
        titleLabel: null,
        willResizeBackgroundImage: true,
        willHandleCallbacks: true,
        isExclusiveTouch: true,
        isRadioButton: false,
        _isHighlighted: false,
        _isEnabled: true,
        dragBeginsTouch: false,

        _titles: null,
        _titleColors: null,
        _backgroundImages: null,
        _images: null,
        _currentState: LumosityButtonState.ButtonStateNormal,

        _currentBackgroundImage: null,
        _currentImage: null,

        _overridingHitRegion: false,
        _hitRegion: null,

        _target: null,
        _pressedCallback: null,
        _heldCallback: null,
        _releasedCallback: null,
        _cancelledCallback: null,

        /**
         * LumosityButton constructor.
         */
        ctor:function()
        {
            this._super();
            this.init();

            this._titles = {};
            this._titleColors = {};
            this._backgroundImages = {};
            this._images = {};
        },

        /**
         * Returns the title for the LumostiyButtonState.
         * 
         * @param  {LumosityButtonState} lumosityButtonState
         * @return {string} title
         */
        titleForState: function( lumosityButtonState )
        {
            var title = null;
            
            if( !Object.prototype.hasOwnProperty.call( this._titles, lumosityButtonState ) )
            {
                return title;
            }

            title = this._titles[ lumosityButtonState ];

            return title || "";
        },

        /**
         * Sets the title for the LumosityButtonState.
         * 
         * @param {string} title
         * @param {LumosityButtonState} lumosityButtonState
         */
        setTitleForState: function( title, lumosityButtonState )
        {
            this._titles[ lumosityButtonState ] = title;
            
            this._updateView();
        },

        /**
         * Returns the title color for the LumosityButtonState.
         * 
         * @param  {LumosityButtonState} lumosityButtonState
         * @return {cc.c3b} title color
         */
        titleColorForState: function( lumosityButtonState )
        {
            var titleColor = null;
            
            if( !Object.prototype.hasOwnProperty.call( this._titleColors, lumosityButtonState ) )
            {
                return titleColor;
            }

            titleColor = this._titleColors[ lumosityButtonState ];

            return titleColor || "";
        },

        /**
         * Sets the title color for the LumosityButtonState.
         * 
         * @param {cc.c3b} titleColor
         * @param {LumosityButtonState} lumosityButtonState
         */
        setTitleColorForState: function( titleColor, lumosityButtonState )
        {
            if( !titleColor )
            {
                ll.warning( "LumosityButton", "setTitleColorForState", "titleColor is null." );
                return;
            }

            this._titleColors[ lumosityButtonState ] = titleColor;

            this._updateView();
        },

        /**
         * Returns the background image for the LumosityButtonState.
         * 
         * @param  {LumosityButtonState} lumosityButtonState
         * @return {cc.Sprite} background image
         */
        backgroundImageForState: function( lumosityButtonState )
        {
            var backgroundImage = null;

            if( !Object.prototype.hasOwnProperty.call( this._backgroundImages, lumosityButtonState ) )
            {
                return null;
            }

            backgroundImage = this._backgroundImages[ lumosityButtonState ];

            return backgroundImage;
        },

        /**
         * Sets the background image for the LumosityButtonState.
         * 
         * @param {cc.Sprite} backgroundImage
         * @param {LumosityButtonState} lumosityButtonState
         */
        setBackgroundImageForState: function( backgroundImage, lumosityButtonState )
        {
            var willResizeBackgroundImage = this.willResizeBackgroundImage;
            var prevBackgroundImage = null;
            var bgImageContentSize = cc.SizeZero();
            var bgImageScaledContentSize = cc.SizeZero();
            var bgImageScale = cc.p( 1, 1 );

            if( backgroundImage )
            {
                backgroundImage.setAnchorPoint( cc.PointZero() );
                backgroundImage.nodeName = "backgroundImage";
                backgroundImage.nodeType = "buttonBackgroundImage";

                // No contentSize is defined or setting background image for ButtonStateNormal
                if( cc.sizeEqualToSize( this.getContentSize(), cc.SizeZero() ) ||
                    ( lumosityButtonState === LumosityButtonState.ButtonStateNormal ) )
                {
                    
                    bgImageContentSize = backgroundImage.getContentSize();
                    bgImageScale = cc.p( backgroundImage.getScaleX(), backgroundImage.getScaleY() );
                    bgImageScaledContentSize = cc.size( bgImageContentSize.width * bgImageScale.x, bgImageContentSize.height * bgImageScale.y );

                    // Disable willResizeBackgroundImage when setting content size with the images size.
                    this.willResizeBackgroundImage = false;
                    this.setContentSize( bgImageScaledContentSize );
                    // Restore willResizeBackgroundImage to it's previous value.
                    this.willResizeBackgroundImage = willResizeBackgroundImage;
                }

                // No longer using retain and release. Hiding the object instead.
                backgroundImage.setVisible( false );
                
                if ( backgroundImage.getParent() )
                {
                    backgroundImage.removeFromParent();
                }
                
                this.addChild( backgroundImage, 0 );
            }
            
            // Remove previous background image for state.
            prevBackgroundImage = this.backgroundImageForState( lumosityButtonState );
            if( prevBackgroundImage )
            {
                // Set current background image to null if we are removing "that" image.
                if( this._currentBackgroundImage === prevBackgroundImage )
                {
                    this._currentBackgroundImage = null;
                }

                prevBackgroundImage.removeFromParent( true );
                prevBackgroundImage = null;
            }

            this._backgroundImages[ lumosityButtonState ] = backgroundImage;

            this._updateView();

            this._updateHitRegion();
        },

        /**
         * Returns the image for the LumosityButtonState.
         * 
         * @param  {LumosityButtonState} lumosityButtonState
         * @return {cc.Sprite} image
         */
        imageForState: function( lumosityButtonState )
        {
            var image = null;

            if( !Object.prototype.hasOwnProperty.call( this._images, lumosityButtonState ) )
            {
                return null;
            }

            image = this._images[ lumosityButtonState ];

            return image;
        },

        /**
         * Sets the image for the LumosityButtonState.
         * 
         * @param {cc.Sprite} image
         * @param {LumosityButtonState} lumosityButtonState
         */
        setImageForState: function( image, lumosityButtonState )
        {
            var prevImage = null;

            if( image )
            {
                image.nodeName = "image";
                image.nodeType = "buttonImage";

                image.setVisible( false );
                
                if ( image.getParent() )
                {
                    image.removeFromParent();
                }
                
                this.addChild( image, 1 );
            }

            // Release previously stored image.
            prevImage = this.imageForState( lumosityButtonState );
            if( prevImage )
            {
                if( this._currentImage === prevImage )
                {
                    this._currentImage = null;
                }

                prevImage.removeFromParent( true );
                prevImage = null;
            }

            this._images[ lumosityButtonState ] = image;

            this._updateView();

            this._updateHitRegion();
        },

        /**
         * Determines whether the worldPoint is within the button's hit region.
         *
         * @param  {cc.p} worldPoint    The world-space point in question.
         * @return {Boolean}            Returns if worldPoint within the button's hit region.
         */
        containsWorldPoint:function( worldPoint )
        {
            var localPoint = this.convertToNodeSpace( worldPoint );
            if( this._hitRegion )
            {
                localPoint = cc.p( Math.floor( localPoint.x ), Math.floor( localPoint.y ) );
                return LumosityUtils.isPointInPolygon( localPoint, this._hitRegion );
            }
            return false;
        },

        /**
         * Sets button's hit region, allowing for override of generic content bounds.
         *
         * @param  {Array} polygon  Array of cc.p points that describe the verticies of a convex polygon
         */
        setHitRegion:function( polygon )
        {
            this._overridingHitRegion = !!polygon;
            this._hitRegion = polygon;
            this._updateHitRegion();
        },

        /**
         * Updates button's hit region to match its bounds, if necessary
         */
        _updateHitRegion:function()
        {
            var contentSize = this.getContentSize();
            if( !this._overridingHitRegion && contentSize )
            {
                // not overriding hit region, so recalculate default hit region using content size
                this._hitRegion =
                    [
                        cc.p( 0, 0 ),
                        cc.p( contentSize.width - 1, 0 ),
                        cc.p( contentSize.width - 1, contentSize.height - 1 ),
                        cc.p( 0, contentSize.height - 1 )
                    ];
            }

            if( debugShowHitRegions )
            {
                // debug option to see hit regions around buttons (set at top of file)
                this.__debugCreateHitRegionSprite();
            }
        },

        /**
         * Returns the current title.
         * 
         * @return {string} title
         */
        currentTitle: function()
        {
            var title = this.titleLabel.getString();
            return title;
        },

        /**
         * Sets the current title.
         * 
         * @param {string} title
         */
        _setCurrentTitle: function( title )
        {
            var willResizeBackgroundImage = this.willResizeBackgroundImage;

            if( title === null || title === undefined )
            {
                return;
            }

            // Create titleLabel if it doesn't exist.
            if( !this.titleLabel )
            {
                this.setTitleLabel( LumosityLabel.create( "", "fonts/MuseoSans-700.ttf", 10 ) );
            }

            this.titleLabel.setString( title );

            // No contentSize is defined
            if( cc.sizeEqualToSize( this.getContentSize(), cc.SizeZero() ) )
            {
                // Disable willResizeBackgroundImage when setting content size with the titles size.
                this.willResizeBackgroundImage = false;
                this.setContentSize( this.titleLabel.getContentSize() );
                // Restore willResizeBackgroundImage to it's previous value.
                this.willResizeBackgroundImage = willResizeBackgroundImage;
            }
        },
        
        /**
         * Allows you to manually set the label object that will be used for the button title.
         * If there is already a title label, it will be destroyed.
         *
         * @param {cc.LabelTTF} The title label to set. 
         */
        setTitleLabel: function( titleLabel )
        {
            if ( this.titleLabel )
            {
                this.titleLabel.removeFromParent();
                this.titleLabel = null;
            }
            
            this.titleLabel = titleLabel;
            this.titleLabel.nodeType = "buttonLabel";
            this.titleLabel.nodeName = "titleLabel";
            this.addChild( this.titleLabel, 1 );
        },

        /**
         * Returns the current title color.
         * 
         * @return {[type]} title color
         */
        currentTitleColor: function()
        {
            var titleColor = this.titleLabel.getColor();
            return titleColor;
        },

        /**
         * Sets the current title color.
         * 
         * @param {cc.c3b} titleColor
         */
        _setCurrentTitleColor: function( titleColor )
        {
            if( !titleColor )
            {
                return;
            }

            if( !this.titleLabel )
            {
                return;
            }

            this.titleLabel.setColor( titleColor );
            if( titleColor.a )
            {
                this.titleLabel.setOpacity( titleColor.a );
            }
        },

        /**
         * Returns the current background image.
         * 
         * @return {cc.Sprite} current background image
         */
        currentBackgroundImage: function()
        {
            return this._currentBackgroundImage;
        },

        /**
         * Sets the current background image.
         * 
         * @param {cc.Sprite} backgroundImage
         */
        _setCurrentBackgroundImage: function( backgroundImage )
        {
            // No background image to set.
            if( !backgroundImage )
            {
                if( this._currentBackgroundImage )
                {
                    this._currentBackgroundImage.setVisible( false );
                }

                return;
            }

            // Hide the current background image.
            if( this._currentBackgroundImage )
            {
                this._currentBackgroundImage.setVisible( false );
            }

            // Assign new current background image and show it.
            this._currentBackgroundImage = backgroundImage;
            this._currentBackgroundImage.setVisible( true );
        },

        /**
         * Returns the current image.
         * 
         * @return {cc.Sprite}
         */
        currentImage: function()
        {
            return this._currentImage;
        },

        /**
         * Sets the current image.
         * 
         * @param {cc.Sprite} image
         */
        _setCurrentImage: function( image )
        {
            // No image to set.
            if( !image )
            {
                if( this._currentImage )
                {
                    this._currentImage.setVisible( false );
                }

                return; 
            }

            if( this._currentImage )
            {
                this._currentImage.setVisible( false );
            }

            this._currentImage = image;
            this._currentImage.setVisible( true );
        },

        /**
         * Returns the current state.
         * 
         * @return {LumosityButtonState} current state
         */
        currentState: function()
        {
            return this._currentState;
        },

        /**
         * Sets the current state.
         * 
         * @param {LumosityButtonState} newState
         */
        _setCurrentState: function( newState )
        {
            this._currentState = newState;
        },

        /**
         * Sets the target.
         * 
         * @param {object} target
         */
        setTarget: function( target )
        {
            this._target = target;
        },

        /**
         * Sets the pressed callback for the target.
         * 
         * @param {Function} pressedCallback
         */
        setPressedCallback: function( pressedCallback )
        {
            this._pressedCallback = pressedCallback;
        },

        /**
         * Sets the held callback for the target.
         * 
         * @param {Function} heldCallback
         */
        setHeldCallback: function( heldCallback )
        {
            this._heldCallback = heldCallback;
        },

        /**
         * Sets the released callback for the target.
         * 
         * @param {Function} releasedCallback
         */
        setReleasedCallback: function( releasedCallback )
        {
            this._releasedCallback = releasedCallback;
        },

        /**
         * Sets the cancelled callback for the target.
         *
         * @param {Function} releasedCallback
         */
        setCancelledCallback: function( cancelledCallback )
        {
            this._cancelledCallback = cancelledCallback;
        },

        /**
         * Sends the callback.
         * 
         * @param  {Function} callback
         */
        _sendCallback: function( callback )
        {
            if ( callback )
            {
                if( this._target )
                {
                    callback.call( this._target, this );
                }
                else
                {
                    callback( this );
                }
            }
        },

        /**
         * Sends the held callback.
         */
        _sendHeldCallback: function()
        {
            this._sendCallback( this._heldCallback );
        },

        /**
         * Sends the cancelled callback.
         */
        _sendCancelledCallback: function()
        {
            this._sendCallback( this._cancelledCallback );
        },

        /**
         * Sets the enabled/disabled LumosityButtonState.
         * 
         * @param {Boolean} isEnabled
         */
        setEnabled: function( isEnabled )
        {
            this._isEnabled = isEnabled;

            if( this._isEnabled )
            {
                this._setCurrentState( LumosityButtonState.ButtonStateNormal );
            }
            else
            {
                this._setCurrentState( LumosityButtonState.ButtonStateDisabled );

                // Unschedule a held callback update.
                this.unschedule( this._sendHeldCallback );
            }

            this._updateView();
        },

        /**
         * Gets the enabled/disabled bool.
         * 
         * @return {Boolean} isEnabled
         */
        getIsEnabled: function()
        {
            return this._isEnabled;
        },

        /**
         * Sets highlighted/unhighlighted LumosityButtonState.
         * 
         * @param {Boolean} isHighlighted Whether to hightlight.
         */
        setHighlighted: function( isHighlighted )
        {
            this._isHighlighted = isHighlighted;

            if( this._isHighlighted )
            {
                this._setCurrentState( LumosityButtonState.ButtonStateHighlighted );
            }
            else
            {
                if( this._isEnabled )
                {
                    this._setCurrentState( LumosityButtonState.ButtonStateNormal );
                }
                else
                {
                    this._setCurrentState( LumosityButtonState.ButtonStateDisabled );
                }
            }

            this._updateView();
        },

        /**
         * Sets opacity on children as well as images for non-active states.
         * 
         * @param {int} opacity The opacity.
         */
        setOpacity: function( opacity )
        {
            var nodes = null;
            var node = null;
            var i = 0;

            nodes = this.getChildren();
            if( this.titleLabel )
            {
                nodes.addAll( [ this.titleLabel ] );
            }
            nodes.addAll( this._backgroundImages );
            nodes.addAll( this._images );

            for( i = 0; i < nodes.length; i++ )
            {
                node = nodes[ i ];
                LumosityUtils.setOpacityOnNode( node, opacity );
            }
        },

        /**
         * Handles pressed or not pressed states.
         * 
         * @param {Boolean} isPressed
         */
        handleIsPressed:function( isPressed, cancelled )
        {            
            var callback = null;
            var newState = null;

            if( this._isEnabled === false )
            {
                return;
            }

            if( isPressed )
            {
                callback = this._pressedCallback;
                newState = LumosityButtonState.ButtonStateSelected;
            }
            else
            {
                callback = this._releasedCallback;
                if( this._isHighlighted )
                {
                    newState = LumosityButtonState.ButtonStateHighlighted; 
                }
                else
                {
                    newState = LumosityButtonState.ButtonStateNormal;
                }
            }

            if( this._currentState === newState )
            {
                return;
            }
            this._setCurrentState( newState );
            
            // Schedule and unschedule a held callback update.
            if( isPressed )
            {
                if( this.willHandleCallbacks )
                {
                    this.schedule( this._sendHeldCallback, 0 );
                }
            }
            else
            {
                this.unschedule( this._sendHeldCallback );
            }

            if( this.willHandleCallbacks )
            {
                this._sendCallback( callback );
            }

            if( cancelled )
            {
                this._sendCancelledCallback();
            }

            this._updateView();
        },

        /**
         * Updates all visual components.
         */
        _updateView: function()
        {
            var backgroundImage = null;
            var image = null;
            var title = null;
            var titleColor = null;

            backgroundImage = this.backgroundImageForState( this._currentState );
            image = this.imageForState( this._currentState );
            title = this.titleForState( this._currentState );
            titleColor = this.titleColorForState( this._currentState );

            this._setCurrentBackgroundImage( backgroundImage );
            this._setCurrentImage( image );

            this._setCurrentTitle( title );
            this._setCurrentTitleColor( titleColor );
        },

        setContentSize:function( contentSize )
        {
            this._super( contentSize );
       
            if( this.willResizeBackgroundImage )
            {
                LumosityUtils.forEachIn( this._backgroundImages, function( lumosityButtonState, backgroundImage ) 
                {
                    //ll.verbose( "LumosityButton", "setContentSize", "lumosityButtonState: " + lumosityButtonState );

                    if( backgroundImage )
                    {
                        // Scale the background sprite to fill the content size.
                        LumosityUtils.scaleForContentSize( backgroundImage, contentSize, true );
                    }
                } );
            }

            this._updateHitRegion();
        },

        __debugCreateHitRegionSprite:function()
        {
            if( this.hitRegionSprite )
            {
                this.hitRegionSprite.removeFromParent( true );
                this.hitRegionSprite = null;
            }
            if( this._hitRegion )
            {
                this.hitRegionSprite = LumosityUtils.createDrawNodePoly( this._hitRegion, 2, cc.c4f( 1, 0, 0, 1 ) );
                this.addChild( this.hitRegionSprite );
            }
        }
    } );
}() );