/*jslint node: true */
"use strict";


var DebugMenu = cc.Class.extend(
{
    /**
     * @param   controller  Game that has model and view property.
     */
    ctor: function(controller)
    {
        this.controller = controller;
        // XXX Not for production:
        // lumosity.game.params.short_game = true;  // XXX Is there a way to set short game only in my LumosityMobile?
        // lumosity.game.params.debug_mode = true;  // XXX Is there a way to set debug only in my LumosityMobile?
        this._addDebugMenuButtons();
        DebugMenu.html5Accessor(controller);
    },

    isOpen: function()
    {
        return lu.isDebugModeOn() 
            && this.controller.view._debugMenu 
            && !this.controller.view._debugMenu.openButton.isVisible();
    },

    /**
     * @param   root    If no root, then ignore.  Test case:  2015-02-10 Android.  Select Trouble Brewing.  John Tighe expects to begin.  Got crash.  Log message: "JavaScript error: DebugMenu.js line 39.  TypeError: root is undefined"  (+Dom).  Xcode Locomotive iOS simulator:  No problem.
     * GAME-2177 Hide scene if debug menu is open.  Test case:  2014-12-16 Touch D.  Touch button.  Expect debug button activates.  Got level select button behind it activates.  
     *
     * GAME-2177 Unfortunately, only a few pixels on the outside edge of the debug button respond to a mouse click in Boiler, and probably none respond to a tap on a mobile device.  [~nick], On a mobile device, do you know how to select a debug button that is in front of a game button?  As [~shalin] mentioned in his Train of Thought post-mortem, unfortunately the level select LumosityButtons from LayoutLoader, although behind the debug buttons, block the click on the debug buttons.  Likewise the ingredient ControlButtons while visually behind the "D" debug button, block the debug button menu altogether.  When the debug buttons are drawn in front game content, I hope they could process a touch or a click.
     *
     * When not in level select mode, do not auto-show scene.  Test case:  2015-02-13 At level select.  Pause.  How to Play.  Mark Torii expects level select no longer visible.  Got appears in background.  Quit.  Select Debug mode.  Repeat same steps.
     */
    hideIfOpen: function(root, isEnabled)
    {
        if (!lu.isDebugModeOn()) {
            return;
        }
        if (!root) {
            return;
        }
        var isBlocked = isEnabled && this.isOpen();
        if (root.isVisible() == isBlocked) {
            ll.log("DebugMenu.hideIfDebugMenuOpen: " + isBlocked);
            root.setVisible(!isBlocked);
        }
    },

    _addDebugMenuButtons: function()
    {
        var view = this.controller.view;
        view.addDebugMenuButton( { labelText: "NEXT session: 13", 
            callback: function(){this._levelUp(130)}, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "NEXT session +1", 
            callback: this._levelUp, 
            target: this, closeMenuOnCallback: true } );
        view._createUserLevelDebugButtons();     

    },

    /**
     * GAME-3648 This affects next session only, because reloading the level select layout of this session appeared to be fragile.
     */
    _levelUp: function(level)
    {
        if (undefined === level) {
            level = DebugMenu.cycle(this.controller.model.persistentData.user_level, 240);
        }
        this.controller.model.persistentData.user_level = level;
        this.controller.onRestart();
    },

    /**
     * @return  GAME-2160 If short game, speed up game motion by delta times multiplier.  Otherwise return delta.  Test case:  2015-03-04 Short games mode.  Train of Thought.  Ethan expects train arrives in less than 3 seconds.  Got more than 10 seconds and felt impatient.
     */
    speedUp: function(delta, multiplier)
    {
        if (LumosityUtils.isShortGamesModeOn())
        {
            delta *= multiplier;
        }
        return delta;
    }
});


DebugMenu.cycle = function(previous, max)
{
    if (undefined === previous) {
        previous = 30;
    }
    var next = (previous + 10) % 250;
    return next;
};


/**
 * @param   controller  Assign to "lumosity.game.instance" for convenient debugging.
 */
DebugMenu.html5Accessor = function(controller)
{
    if ( lumosity.device.id === "html5" ) 
    {
        if (!lu.isDebugModeOn())
        {
            ll.CurrentLoggingLevel = LoggingLevel.Normal;
        }
        try
        {
            lumosity.game.instance = controller;
        }
        catch (err)
        {
        }
    }
};

/**
 * When pause menu reads layout it sets maximum width and height of build number text.  On high resolution, setContentScaleFactor multiplies content size.  Also in CCB, set text width and height, so that long text is wrapped.  Otherwise when setting long text, the texture may complain.  This is a workaround that prevents a warning and shows build number text.  A complete fix could be for LayoutLoader set dimensions for text, but that might not be backward compatible.  Test case:  2015-03-25 Start Trouble Brewing.  Jon Edwards expects no error message about image greater than 4096x4096.  Got warning:  "2015-03-26 09:40:45.827 Lumosity[44675:1555318] Cocos2d: cocos2d: WARNING: Image (5536 x 144) is bigger than the supported 4096 x 4096"
 */
DebugMenu.willClampBuildNumberText = function()
{
    var customReading = PauseMenu.prototype.layout_readingData;
    PauseMenu.prototype.layout_readingData = function(nodeName, nodeData)
    {
        if (customReading)
        {
            customReading.call(PauseMenu.prototype, nodeName, nodeData);
        }
        if ("buildNumberText" == nodeName)
        {
            nodeData.size.width = Math.min(220, nodeData.size.width);
        }
    }
};
