/*jslint node: true */
"use strict";

require("src/Game/TutorScene.js" );

var SimpleTutorScene = cc.Class.extend(
{
    ccbFileName: undefined,
    game: undefined,
    /**
     * GAME-2832 Specify game does not update while tutor scene is paused.  Test case:  2015-03-02 Level select screen.  Start tutorial.  Apple home button.  Lumosity App minimizes.  Refocus on Lumosity.  Resume.  [~rivey] expects red train still paused.  Got red train travels and switch is not toggled toward red station.  So tutorial now feels confusing.
     *
     * GAME-2935 Tutor pause also pauses game, but does not override the game's pause variable.  Test case:  2015-03-13 Android.  Tutor.  Train paused.  Tap Android back button.  [~rivey] expects to see pause menu.  Got train remains paused.  I simulated Android back button in Boiler:  Debug mode.  Console command:  lumosity.game.instance._backButtonTapped();
     */
    isGamePaused: undefined,
    isView: undefined,
    numTrials: undefined,
    owner: undefined,
    parent: undefined,
    root: undefined,

    ctor: function(game, parent)
    {
        this.game = game;
        this.parent = parent;
        this.owner = this;
        this.numTrials = 5;
        this.ccbFileName = "ccb/Tutor";
        CCBReaderAdapter.designWidth = 640;
        CCBReaderAdapter.designHeight = 960;
        DebugMenu.html5Accessor(this.game);
    },

    /**
     * GAME-2160 In FLA, Ethan moved label of "transition" and "gameplay" so that end of fit test tutorial stitches to beginning of "gameplay" label.  He added hud to tutorial, which only the fit test needs.  The program hides the HUD in default tutorial.
     */
    onBegin: function()
    {
        var isPracticeMode = this.game.model.persistentData.isPracticeMode;
        TutorScene.onBegin(this, isPracticeMode);
        var isFitTest = !isPracticeMode && LumosityUtils.isFitTestModeOn();
        if (isFitTest)
        {
            this.beginFitTest();
        }
    },

    beginFitTest: function()
    {
        this.game.selectLevel(7);
        this.game.view.levelSelectMenu.setVisible(false);
        this.game.animationFinished_transition();
    },

    /**
     * GAME-2159 Experiment with moving stations away from switch without making maximum dimension bigger. Keep room for text on mobile.  Test case: 2015-02-25 Read 100 recent customer complaints. Brittany counted about 30% were users who thought they were supposed to click the station not the switch.
     *
     * GAME-2159 Test case:  2015-02-27 How to Play.  Play.  How to Play.  Red train departs.  Rapidly tap switch.  Expect no callout.  Got callout appears.  Train arrives.  Expect black train departs.  Got tutorial ends.  Cannot reproduce.
     *
     * GAME-2159 Compensate callout size to match Lumosity custom scaling.  Setting content scale factor globally interferes with Train of Thought custom scaling of textures.  Test case:  2015-02-27 iPad.  Tutor.  Expect callout centered at switch.
     *
     * GAME-2159 Compensate text font size and invert font scale.  Test case:  2015-02-27 iPad Retina.  Tutor.  Expect text crisp.  Got slightly blurry.
     *
     * GAME-2835 Load tutorial board from first without rotating it.  Was loading diagram with Courier font embedded in Tutor.ccbi.  Test case:  2015-03-03 Android.  Select Train of Thought.  Rob Ivey expects to see train tracks enter.  Got application crashes.  iOS has no problem.
     *
     * GAME-2160 [~wouter] In Tutor.ccb, I shortened my "gameBegin" to 15 frames and "playBegin" to 5 frames.  Was 3 seconds total.  Test case:  2015-03-05 Begin fit test tutorial.  [~aaron] expects first train enters immediately.
     */
    onStartTutorial: function()
    {
        if (TutorScene.onStartTutorial(this)) 
        {
            CCBReaderAdapter.fitTextures(this.root);
        }
        this._androidCalloutImage(this.root.callout);
        var contentScale = 1.0 / CCBReaderAdapter.getScaleByAspect();
        this.root.callout.setScale(contentScale);
        var diagram = this.game.settings.boards[0][0];
        this.game.model.selectNumStations(3, diagram);
        this.game.model.isBlockingInput = true;
        this.game.model.isBlockingInputExceptAddress = undefined;
        this.game.view.showBoard();
        this.root.setZOrder(999);
        this.isGamePaused = false;
    },

    /**
     * GAME-2877 If callout image does not exist, then detect screen width and load the image "tutorCallout.png" from Android image folder.   Set transparent.  This is less work and less subscriber bandwidth than upgrading Train of Thought to v2 downloadable game.  Test case:  2015-03-06 Release version of Lumosity Android.  Download Train of Thought.  Expect to see callout.  Got invisible.
     */
    _androidCalloutImage: function(callout)
    {
        if (callout.getContentSize().width <= 0)
        {
            var resolutionDir = this._resolutionWidth().toString();
            var texturePath = "android/" + resolutionDir + "/tutorCallout.png";
            var texture = cc.TextureCache.getInstance().addImage(texturePath);
            var locContentSize = texture.getContentSize();
            var bounds = cc.rect(0, 0, locContentSize.width, locContentSize.height);
            var frame = cc.SpriteFrame.createWithTexture(texture, bounds);
            callout.setDisplayFrame(frame);
            callout.setOpacity(0);
        }
    },

    /**
     * Do you know of a shared function that returns the same integer?
     */
    _resolutionWidth: function()
    {
        var screenWidth = cc.Director.getInstance().getWinSize().width;
        var resolutions = [400, 640, 800, 1200, 1600];
        var width = resolutions[resolutions.length - 1];
        for (var r = 0; r < resolutions.length; r++)
        {
            var resolution = resolutions[r];
            if (screenWidth <= resolution)
            {
                width = resolution;
                break;
            }
        }
        return width;
    },

    /**
     * GAME-2159 After tutorial, reset level to session level.  Test case:  2015-02-27 Already at level 4 or higher.  Complete tutorial.  Level select screen.  Expect highest level has a highlight.  Got no highlight.
     */
    onCompleteTutorial: function(skipped)
    {
        var model = this.game.model;
        if (skipped) {
            this.isGamePaused = true;
        }
        model.isGameRunning = false;
        model.isBlockingInput = false;
        model.onToggleDelegate = undefined;
        model.onAnswerWrongDelegate = undefined;
        model.isBlockingInputExceptAddress = undefined;
        model.level = model.sessionLevel;
        this.game.stopMusic();
        TutorScene.onCompleteTutorial(skipped, this);
        if (!skipped) {
            cc.BuilderReaderHelper.setState(this.root, "trialEnd");
        }
    },

    /**
     * Ramp up density of trains with number correct.
     *
     * Test case:  2013-05-30 Tutorial.  Expect interested.  Density 0.125, I feel bored.
     * Test case:  2013-06-11 In tutorial, Miguel might expect to demo increasing difficulty and density.
     * Test case:  2013-06-17 Tutorial density 0.5.  Patty expects to not feel bored.
     * https://trello.com/card/railroad-switch-polish-program/506b18c9ad088b7d31d56a4f/650
     * Test case:  2013-07-08 Tutorial. Trains spawn, until last two. Erika expects to not feel overwhelmed.
     * Test case:  2013-07-12 Tutorial. Second train.  Pam expects to feel patient.
     * Test case:  2013-07-12 Tutorial. Pam expects few trials and trains start frequently.
     * Test case:  2013-07-16 Tutorial. Pam expects trains start infrequently.
     * https://trello.com/card/railroad-switch-polish-tutorial/506b18c9ad088b7d31d56a4f/660
     *
     * GAME-2160 If short game mode, tutorial game ends instantly.  Test case:  2015-03-02 Short game.  Fit test.  Expect bypass tutorial.
     *
     * GAME-2160 Always update tutor, even when game is not running, as happens before playBegin.  Test case:  2015-03-05 Begin fit test tutorial.  [~aaron] expects first train enters immediately.  
     */
    onUpdate: function()
    {
        var isUpdating = TutorScene.onUpdate(this);
        var board = this.game.model.board;
        if (isUpdating && this.game.model.isGameRunning && board) 
        {
            //- this.updatePause();
            if (LumosityUtils.isShortGamesModeOn() || (this.numTrials <= board.spawnCount && board.empty())) 
            {
                this.game.onCompleteTutorial(true);
            }
            else {
                board.setTravellerDistribution(Math.min(1.0, 
                    0.25 + 0.75 * this.game.model.numCorrect));
                board.setDestinationFraction(1.0);
            }
        }
    },
   
    /**
     * GAME-2935 At end of tutorial, unpause tutorial.  Test case:  2015-03-13 How to Play.  Play.  Expect train.  Got no train.  Tap switch.  Expect switch highlight disappears.  Got remains.
     *
     * If no root, then do not hide it.  Test case:  2015-07 Crittercism expects no handled JavaScript exception.  Got 144 occurences during 4 months with root undefined on line 187 of SimpleTutorScene.  https://app.crittercism.com/developers/exception-details/606b939441e8b080110653251aea39c2f65c20f8decbd21045f5d723#backtrace
     */
    onEndTutorial: function(skipped)
    {
        if (this.root)
        {
            this.root.setVisible(false);
        }
        this.isGamePaused = false;
    },

    /**
     * Test case:  2015-02-26 Tutor.  Red train.  Spam switch.  Expect only responds to first tap.  Black train.  Spam switch.  Expect only responds to first tap.  
     *
     * GAME-2160 In the tutorial, when toggling the callout switch during an example, stop all previous tutorial animations, then force a restart of the corresponding example end animation ("exampleEnd" or "example2End").  Test case:  2015-03-03 Fit test.  Train of Thought.  Tutor.  Red train travels.  Pause.  Restart.  Red train travels.  Pause.  Resume.  Expect callout disappears and black train appears.  Got callout appears and remains, then no update.  Inspect log.  See that gameBegin started, but playBegin never logs as completed.  Repeat all previous steps with black train.  
     */
    onToggle: function()
    {
        var model = this.game.model;
        cc.BuilderReaderHelper.visit(this.root, "stopAllActions");
        this.isGamePaused = false;
        this.game.playMusic();
        model.isBlockingInput = true;
        var nextState;
        if (model.board.spawnCount <= 1) {
            nextState = "exampleEnd";
        }
        else {
            nextState = "example2End";
            model.onToggleDelegate = undefined;
        }
        cc.BuilderReaderHelper.runAnimations(this.root, nextState);
    },

    pauseUntilToggle: function()
    {
        var model = this.game.model;
        model.onToggleDelegate = this;
        this.isGamePaused = true;
        this.game.stopMusic();
        model.isBlockingInput = false;
        model.isBlockingInputExceptAddress = 22;
    },

    onAnswerWrong: function()
    {
        var model = this.game.model;
        model.board.spawnMore(this.numTrials);
        model.numCorrect = 2;
    },

    /**
     * GAME-2159 Sequence "example2End" animation ends after 2 seconds, which schedules spawning 5 more trains.  Sequence "trial" ends after 8 seconds, which enables switches.  Test case:  2015-02-26 Black train.  Rapidly tap switch.  Expect switch does not toggle until first black train goes over.  Expect as soon as black train arrives another train departs.
     *
     * GAME-2159 Animate callout and text according to Trouble Brewing style:  Callout fades in and out 8 frames, linear ease.  Text fades in 10 frames, ease out; text fades out with an ease in.  Museo sans 500 font size 28 (at 640x960) or size 14 (at 320x480), horizontally and vertically centered, width 500.  Test case:  2015-02-27 Play Train of Thought tutorial.  Trouble Brewing tutorial.  Expect callout and text appear and animate in the same timing and size.  
     *
     * GAME-2159 Center and lift trial root board in xJSFL FLA.  Align callout and text in CocosBuilder.  Test case:  2015-02-27 iPhone 4, 5.  iPad 1, 3.  Tutorial.  Expect train tracks and text feels approximately vertically and horizontally centered.  Text paragraph height varies by language and device.  Expect callout precisely centers on switch.  
     */
    completedAnimationSequenceNamed: function(sequenceName)
    {
        TutorScene.completedAnimationSequenceNamed(sequenceName, this);
        var model = this.game.model;
        if ("gameBegin" == sequenceName) {
        }
        else if ("playBegin" == sequenceName) {
            this.game.animationFinished_gameplay();
            this.game.model.board.spawnTraveller(1);
            cc.BuilderReaderHelper.runAnimations(this.root, "example");
        }
        else if ("example" == sequenceName) {
            this.pauseUntilToggle();
        }
        else if ("exampleHint" == sequenceName) {
        }
        else if ("exampleEnd" == sequenceName) {
            this.game.model.board.spawnTraveller(0);
        }
        else if ("example2" == sequenceName) {
            this.pauseUntilToggle();
        }
        else if ("example2End" == sequenceName) {
            model.board.spawnMore(this.numTrials);
            model.onAnswerWrongDelegate = this;
        }
        else if ("trial" == sequenceName) {
            model.isBlockingInput = false;
            model.isBlockingInputExceptAddress = undefined;
        }
    }
});
