/*jslint node: true */
"use strict";

require("src/Game/Component/GameComponentAudioMixer.js");
require("src/Game/Component/GameComponentTimer.js");
require("src/View/CCBReaderAdapter.js");

/**
 * @return  If name ends with ending.  If name is shorter than ending, name cannot end with ending.
 */
String.endsWith = function(name, ending)
{
    return ending.length <= name.length 
            && name.indexOf(ending) 
            == name.length - ending.length;
};


var TutorScene = TutorScene || {};

/**
 * @param   GameClass       Call super on start, complete, and end tutorial.
 * @param   gameInstance    Override start, complete, end tutorial callbacks.  
 *                          After finish animation "endTutorial" call game.onEndTutorial.
 *
 * GAME-2160 At end of tutorial, if in fit test, start fit test.  Test case:  2015-03-02 Fit test mode.  Short games mode.  Train of Thought.  End tutorial popup disappears.  Expect to see board with 7 stations.  Got tutorial restarts.
 */
TutorScene.extend = function(GameClass, gameInstance, tutorScene)
{
    GameClass.prototype.onCompleteTutorial = function(skipped) 
    {
        tutorScene.onCompleteTutorial(skipped);
        GameClass.__super__.onCompleteTutorial.call(gameInstance, skipped);
    };

    GameClass.prototype.onEndTutorial = function(skipped) 
    {
        if (tutorScene.onEndTutorial)
        {
            tutorScene.onEndTutorial(skipped);
        }
        if (LumosityUtils.isFitTestModeOn()) 
        {
            tutorScene.game.onStartFitTest();
        }
        else 
        {
            var seconds = tutorScene.game.settings.getGameLengthInSeconds();
            tutorScene.game._components.GameComponentTimer.reset(seconds);
            GameClass.__super__.onEndTutorial.call(gameInstance, skipped);
        }
    };

    GameClass.prototype.animationFinished_endTutorial = function()
    {
        gameInstance.onEndTutorial();
    };
};


TutorScene.extendHud = function(hud)
{
    hud.show = function(isAllVisible, isTutorialVisible)
    {
        var i;
        for ( i = 0; i < this._components.length; i++ )
        {
            this._components[ i ].container.setVisible ( isAllVisible );
        }
        var component = this._getTutorialComponent(true);
        if (component)
        {
            component.container.setVisible(isTutorialVisible);
        }
        this._positionComponents();
    };

    /**
     * GAME-2160 If not in fit test yet in tutorial, hide HUD.  Test case: 2015-03-03 Fit test.  Short game.  Train of Thought.  Expect to see Tutorial in HUD.  Play.  Expect to see HUD.  Got no hud.  Not fit test, short game.  How to Play.  Train of Thought.  Expect no HUD.  Got HUD.  Play.  Expect HUD.
     *
     * GAME-2160 If not in fit test, hide HUD.  Test case: 2015-03-03 Fit test.  Expect to see Tutorial in HUD.  Not fit test.  How to Play.  Expect no HUD.  Got HUD.
     */
    hud.visibles = function(isPracticeMode)
    {
        var isFitTest = LumosityUtils.isFitTestModeOn();
        var isAllVisible = !isPracticeMode;
        var isTutorialVisible = isFitTest && isPracticeMode;
        return [isAllVisible, isTutorialVisible];
    };

    hud.onBeginMayShow = function(isPracticeMode)
    {
        var hudVisibles = this.visibles(isPracticeMode);
        this.show.apply(this, hudVisibles);
    };

    /**
     * GAME-2160 Before beginning, disable HUD hiding elements in fit test.  Test case:  2015-03-02 Fit test mode.  Short games mode.  Train of Thought.  Look at HUD.  Expect to read Time and Correct.  Got only Time.
     */
    hud.enableFitTestMode = function(isEnabled) { };
};


/**
 * GAME-1874 GAME-2159 Override creating dim layer.  Test case:  2015-01-05 Trouble Brewing, Train of Thought:  Tutor.  [~wouter] expects background is bright as normal.  Got dimmed, which interferes with foreground in Trouble Brewing and obscures tracks in Train of Thought.
 */
TutorScene.disableDimLayer = function(view)
{
    view.createDimLayer = function() {};
    view.removeDimLayer = function() {};
};


TutorScene.getCurrentState = function(root)
{
    var state;
    if (root) 
    {
        state = root.animationManager.getRunningSequenceName();
    }
    return state;
};


/**
 * If next state ends with "End" or "End2" then do not change string.
 * Test case:  2014-12-16 Tutor 2 ingredients.  Do not expect to flash Tutor 1 ingredient message.
 */
TutorScene.isNextMessage = function(nextState)
{
    if (!nextState) {
        return false;
    }
    else {
        var countless = nextState.split("2")[0];
        return !String.endsWith(countless, "End");
    }
};


TutorScene.maySetString = function(labelNode, key)
{
    var may = false;
    if (key in LumosityGlobals.localizedStrings) 
    {
        may = true;
        ll.log("TutorScene.maySetString: " + key);
        var string = localize(key);
        labelNode.setString(string);
    }
    return may;
};


/**
 * @param   labelNode   The CC label being set, by calling method labelNode.setString.
 * @param   nextState   Set to localized string with prefix "tutor_" If no string, do not set.  Test case:  2014-12-16 Tutor third trial.  Expect message.  Got overflow message.
 *
 * GAME-2160 Try to set message to state with suffix "_fitTest".  If does not exist, then set without suffix.  Test case:  2015-03-05 Train of Thought fit test.  Tutor.  Unlike the web fit test, [~aaron] expects a custom message for the mobile fit test.  Got the same text for default tutorial and fit test tutorial.
 */
TutorScene.updateMessage = function(labelNode, nextState)
{
    if (TutorScene.isNextMessage(nextState)) 
    {
        var key = "tutor_" + nextState;
        if (!LumosityUtils.isFitTestModeOn() 
         || !TutorScene.maySetString(labelNode, key + "_fitTest")) {
            TutorScene.maySetString(labelNode, key);
        }
    }
};


/**
 * GAME-2403 Instead of set animation manager delegate, poll each frame for the currently running animation.  Test case:  2015-01-17 Start game.  setDelegate.  Cocos2d-x js v2.2.2 crash.  Cocos2d-html5 ok.
 * @param   tutorScene   If no tutorScene.root, add tutorScene.ccbFileName to tutorScene.parent with tutorScene.owner.  GAME-2159 If already created, ignore.  Test case:  2015-02-26 Train of Thought.  Play.  Pause.  How to Play.  Expect tutor appears.  Got crash, since animation label "none" was not found.
 * @return  tutorScene.root
 */
TutorScene.addChild = function(tutorScene)
{
    if (!tutorScene.root)
    {
        tutorScene.root = CCBReaderAdapter.addChild(tutorScene.parent, tutorScene.ccbFileName, 
            tutorScene.root, tutorScene.owner);
        tutorScene.root.setZOrder(0);
    }
    return tutorScene.root;
};


/**
 * GAME-2160 When beginning practice, create a timer.  At least Train of Thought, Trouble Brewing, and Disillusion are exceptions to the base engine rule to not create a timer during fit test practice.  Test case:  2015-03-03 Fit test mode.  Train of Thought.  [~aaron] expects to see train tracks enter.  Got crash.
 *
 * GAME-2159 When starting tutorial, reset timer to 1000 years, which is plenty of time, while still being precise below one tenth of a millisecond as a IEEE floating point number.  When ending tutorial, reset time to time in settings.  This allows for tutorial timed and paused behavior to reuse the code of play behavior.  Test case:  2015-02-26 Tutor.  Wait 2 minutes.  Expect do not run out of time.
 */
TutorScene.setPlentyOfTime = function(game)
{
    if (!game._components.GameComponentTimer)
    {
        game._addComponent( new GameComponentTimer( game ) );
    }
    var MILLENIUM_IN_SECONDS = Math.pow(2, 35);
    game._components.GameComponentTimer.reset(MILLENIUM_IN_SECONDS);
};


/**
 * GAME-2834 Start tutorial in practice mode after game begins.  Test case:  2015-03-03 iOS.  How to Play mode.  Train of Thought.  iOS system button.  Lumosity app.  At pause screen, restart.  Rapidly tap.  [~rivey] expects see scene animate into view.  Got crash.
 *
 * GAME-2160 On begin fit test practice, start tutorial and resume animations.  Test case:  2015-03-03 Fit test.  Train of Thought.  Tutor.  Red train travels.  Pause.  Restart.  Red train travels.  Pause.  Resume.  Expect callout disappears and black train appears.  Got callout appears and remains, then no update.  Inspect log.  See that gameBegin started, but playBegin never logs as completed.  
 *
 * Test case:  2015-03-02 Fit test mode.  Train of Thought.  Expect start tutorial.  Got crash.  Board not constructed.  How to Play mode.  Train of Thought.  Tutorial does start.
 *
 * GAME-2160 If starting tutorial, tutorScene.isVisible otherwise false, so view knows to dispose of the tutor images.  Cocos2d-x v2.2.2 disposing is fragile and does not tolerate trying to dispose twice.  Test case:  2015-03-04 How to Play.  Train of Thought.  Skip tutorial.  Play.  Expect only one train set.  Got tutor train set in background.  Restart.  Play.  Restart.  At level select, restart.  Play.  Expect only one train set.  Got tutor train set in background.  Fit test.  Train of Thought.  Restart.  Expect only one train set in view.
 */
TutorScene.onBegin = function(tutorScene, isPracticeMode)
{
    tutorScene.isVisible = false;
    if (tutorScene.game.settings.hasTutorial) {
        if (isPracticeMode)
        {
            tutorScene.isVisible = true;
            TutorScene.setPlentyOfTime(tutorScene.game);
            tutorScene.onStartTutorial();
            cc.BuilderReaderHelper.resume(tutorScene.root);
        }
        tutorScene.game.view.hud.onBeginMayShow(isPracticeMode);
    }
};


/**
 * @return  isLoading   If constructing a new root.
 *
 * GAME-2403 On begin, hide scene except beginning.  On load, show root yet reset animation.  Test case:  2015-01-10 Play.  Restart.  Expect board disappeared.  Got board remained.  How to Play.  Expect level select disappeared.  Got level select remained.  Wait 5 seconds.  Expect tutorial continues.  Got level menu appears behind tutorial.
 *
 * GAME-2403 Preload animations.  2014-12-18 Tap level.  Matt Leung expects to animate immediately.  Got impatient.  (+Sonia, +Beverly, +Melissa Malski, 2015-01-09 +Emma).
 *
 * GAME-2159 Show tutor scene.  Test case:  2015-02-26 How to Play mode.  Skip.  Play.  How to Play.  Expect to read text and see callout.  Not shown.
 *
 * GAME-2159 Reset current sequence.  Test case: 2015-02-27 How to Play.  Skip.  Play.  How to Play.  Red train emerges.  Rapidly tap switch.  Expect no response.  Got can switch twice and then blocks input.
 *
 * GAME-2160 If in fit test, show pause button, otherwise show skip tutorial button.  Test case:  2015-03-02 Fit test.  Expect pause button.  Got skip tutorial button.
 */
TutorScene.onStartTutorial = function(tutorScene)
{
    var isLoading = TutorScene.load(tutorScene);
    var root = tutorScene.root;
    root.setVisible(true);
    root.currentSequence = undefined;
    cc.BuilderReaderHelper.runAnimations(root, "gameBegin");
    var nextState = TutorScene.getCurrentState(root);
    TutorScene.updateMessage(root.text, nextState);
    var view = tutorScene.game.view;
    if (LumosityUtils.isFitTestModeOn())
    {
        view.cue_showPauseButton();
    }
    else
    {
        view.cue_showSkipTutorialButton();
    }
    return isLoading;
};


/**
 * @return  isLoading   If constructing a new root.
 */
TutorScene.load = function(tutorScene)
{
    var isLoading = undefined == tutorScene.root;
    tutorScene.root = TutorScene.addChild(tutorScene);
    return isLoading;
};


/**
 * GAME-2159 End play, stop animations, hide skip tutorial button.
 * If skipping, pause animations, hide tutor scene, and stop music.
 *
 * GAME-2160 If in fit test, hide pause button, otherwise hide skip tutorial button.  Test case:  2015-03-02 Fit test.  End tutor.  Expect hide pause button.  Got skip tutorial button.
 *
 * Null check root of tutorial scene before setting invisible.  Did the user skip before the game had finished constructing the root?  In the last crash report it appears images were not loaded, so maybe the scene was also not loaded.  Perhaps the images were not fully downloaded.  Test case:  2015-03-13 Android.  Unknown step.  Skip tutorial.  Expect popup.  Got JavaScript error.  https://app.crittercism.com/developers/exception-details/4067b42fe3feaa8bc1461b64ac572e1dde0cefd52affeeeb26d3dfb5#backtrace
 * https://app.crittercism.com/developers/exception-details/41cdf7014f7ee1ee1257b9ec647c9536bb106c509df86bf9e4c59a14#backtrace
 * https://app.crittercism.com/developers/exception-details/207eaaa07a2ca9b1ae3f13695c055cc2461324b72e453277291dc1d5
 */
TutorScene.onCompleteTutorial = function(skipped, tutorScene)
{
    if (skipped)
    {
        cc.BuilderReaderHelper.pause(tutorScene.root);
        if (tutorScene.root)
        {
            tutorScene.root.setVisible(false);
        }
        var mixer = GameComponentAudioMixer.getInstance(tutorScene.game);
        mixer.stopSound();
    }
    var view = tutorScene.game.view;
    view.stopAllAnimations(skipped);
    if (LumosityUtils.isFitTestModeOn())
    {
        view.cue_hidePauseButton();
    }
    else
    {
        view.cue_hideSkipTutorialButton();
    }
};


/**
 * GAME-2159 If in practice mode and running, complete tutorial.  Test case:  2015-02-26 Skip tutorial.  Expect popup.  Got crash due to recursively completing.
 */
TutorScene.mayCompleteTutorial = function(game)
{
    if (game.model.persistentData.isPracticeMode && game.model.isGameRunning) 
    {
        game.onCompleteTutorial(false);
    }
};


/**
 * Update root animation and tutor state.
 * If not in practice mode, ignore.
 */
TutorScene.onUpdate = function(tutorScene)
{
    var isPracticeMode = tutorScene.game.model.persistentData.isPracticeMode;
    if (isPracticeMode) {
        cc.BuilderReaderHelper.update(tutorScene.root, tutorScene);
    }
    return isPracticeMode;
};


/**
 * After any game end animation, ask game to play its game animation.
 */
TutorScene.completedAnimationSequenceNamed = function(sequenceName, tutorScene)
{
    ll.log('TutorScene.completedAnimationSequenceNamed: "' + sequenceName + '"');
    if (sequenceName == "gameBegin")
    {
    }
    else if (0 == sequenceName.indexOf("playBegin")) 
    {
        tutorScene.game.onStartGameplay();
    }
    else if (0 == sequenceName.indexOf("trialEnd")) 
    {
        tutorScene.game.onCompleteTutorial(false);
    }
    else if (0 == sequenceName.indexOf("gameEnd")) 
    {
        tutorScene.game.playEndAnimation(true);
    }
    var nextState = TutorScene.getCurrentState(tutorScene.root);
    TutorScene.updateMessage(tutorScene.root.text, nextState);
};


GameComponentAudioMixer.getInstance = function(controller)
{
    if (!controller._components[ GameComponentAudioMixer.prototype.getType() ]) 
    {
        controller._addComponent( new GameComponentAudioMixer( controller ) );
    }
    var mixer = controller._components[ GameComponentAudioMixer.prototype.getType() ];
    return mixer;
};


/**
 * Scale Sprites and Scale9Sprites, but do not recurse their descendents.
 * Also scale the animated keyframes of scale.
 * Test case: Animate.  Expect texture still fits.
 * Test case: ControlButton, which has had its Scale9Sprite disabled.
 * Expect at same visual position as in SpriteBuilder.
 * Test case:  Tap button. Expect does image remains in place.
 *
 * Invert root scale on LabelTTF, to keep font crisp.
 *
 * Invert scale according to root scale on iPad in button.
 * Test case:  2014-12-17 iOS simulator.  Level Select button.  Compare to SpriteBuilder.
 * SpriteBuilder Helvetica font.
 * Expect at most 2% difference in margins due to Helvetica and Museo Sans fonts.
 * Observe log:
 * iPhone 4:    CCBReaderAdapter.fitTextures: textureScaleFactor 0.625 labelScaleInverse 2
 * iPad Retina: CCBReaderAdapter.fitTextures: textureScaleFactor 0.25 labelScaleInverse 4.266666666666667
 * 
 * Warning:  Boiler shows different scale than iOS simulator on iPad resolutions.
 *
 * GAME-1872 Multiply button font 1.5; was 2.5.  Test case:  2015-01-02 SpriteBuilder.  Xcode simulator.  [~wouter] expects font proportion to button size is approximate.  Got Xcode font was larger.
 * GAME-2403 Set variable "didScaleTextures" on each node scaled, and do not scale them again.  Test case: 2015-01-10 Play.  Restart.  How to Play.  Look at coffee machine.  I expect it is same size as when playing.  Got shrunken due to reusing the blender twice and the board once.  
 */
CCBReaderAdapter.scaleTextures = function(node, textureScaleFactor, labelScaleInverse)
{
    var isScaling = node instanceof cc.Sprite || node instanceof cc.Scale9Sprite;
    isScaling = isScaling && !node.didScaleTextures;
    if (node instanceof cc.LabelTTF)
    {
        var parent = node.getParent();
        var labelScale;
        var size = node.getFontSize();
        if (parent && parent instanceof cc.ControlButton) {
            var buttonFontMultiplier = 1;
                                       // 1.5; // ?
                                       // 2;
                                       // 2.2;
                                       // 2.4;
                                       // 2.5;
            labelScale = labelScaleInverse * buttonFontMultiplier;  
        }
        else {
            labelScale = labelScaleInverse;
        }
        node.setFontSize(Math.round(labelScale * size));
        var dimensions = node.getDimensions();
        var labelDimensions = new cc.size(Math.round(labelScaleInverse * dimensions.width), 
                                          Math.round(labelScaleInverse * dimensions.height));
        node.setDimensions(labelDimensions);
        isScaling = true;
        textureScaleFactor = 1.0 / labelScaleInverse;
    }
    if (isScaling)
    {
        node.didScaleTextures = true;
        var sX = (node.getScaleX() * textureScaleFactor );
        var sY = (node.getScaleY() * textureScaleFactor );
        node.setScaleX(sX);
        node.setScaleY(sY);
        var man = node.animationManager;
        if (man) {
            var base = man._getBaseValue(node, PROPERTY_SCALE);
            if (base) {
                base[0] *= textureScaleFactor;
                base[1] *= textureScaleFactor;
                man.adjustScale(node, textureScaleFactor);
            }
        }
    }
    else if (!node.didScaleTextures)
    {
        var children = node.getChildren();
        for(var i = 0; i < children.length; i++)
        {
            this.scaleTextures(children[i], textureScaleFactor, labelScaleInverse);
        }
    }
};


/**
 * Sharpen text on iPad Retina, by fitting textures and text font size to device resolution.  Expects no element's scale is animated.
 * Does not set global content scale factor, because rest of the game was created in xJSFL or code at a custom scale that is incompatible with global contentScaleFactor.  Test case:  2015-02-27 iPad.  How to Play mode.  Train of Thought.  Expect the trains, created at custom game content scale snap to each other and that text is crisp.
 */
CCBReaderAdapter.fitTextures = function(root, labelScaleInverse)
{
    if (undefined === labelScaleInverse) 
    {
        labelScaleInverse = root.getScale();
    }
    var frameWidth = cc.Director.getInstance().getWinSize().width;
    var textureScaleFactor = CCBReaderAdapter.calculateContentScale(frameWidth);
    // cc.log("CCBReaderAdapter.fitTextures: textureScaleFactor " + textureScaleFactor 
    //     + " labelScaleInverse " + labelScaleInverse );
    CCBReaderAdapter.scaleTextures(root, textureScaleFactor, labelScaleInverse);
};
